/*
 * linux/arch/arm/drivers/block/adfsimg.c
 *
 * Copyright (c) 1996 Russell King.
 *
 * Implement image files under Linux (allows Linux to access partition files
 *  on ADFS disks).
 *
 * This file is split into two halves: the first half is the actual image
 * mapper creation code, and the second half is the device driver.
 *
 * Changelog:
 *  28-03-1996	RMK	Created
 *  02-04-1996	RMK	Cleaned up per-device data into structure [should
 *			  improve code].
 *  24-06-1996	RMK	Marked 4GB problem areas.
 *			Altered blocksize checking code - now uses the larger
 *			  of the sector size and map bit size.
 *  04-06-1996	RMK	Re-wrote mapping function.  Now applies mapping as
 *			  blocks come to be put on the queues.
 *  06-06-1996	RMK	Pass through blocks that need to be split up.
 */

#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/malloc.h>
#include <linux/blkdev.h>
#include <linux/errno.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>

#include <asm/segment.h>
#include <asm/bitops.h>

#define MAJOR_NR ADFSIMG_MAJOR
#include "blk.h"
#include "adfs.h"

/*----------------------------------------------------------------------------------------
 * User configuration
 *
 * Number of minors per device, and number of devices.  Note that we have
 * a limit of 256 on the number of minors at the moment.
 */
#define ADFSIMG_SHIFT	5
#define ADFSIMG_NRDEV	(256 / (1 << ADFSIMG_SHIFT))
#define ADFSIMG_PARMASK	((1 << ADFSIMG_SHIFT) - 1)
/*
 * This maps a device onto an index into the adfsimage structure
 */
#define TARGETMAP(kdev) DEVICE_NR(kdev)
/*
 * Debugging contants...
 */
#define DEBUG_SWAP		32
#define DEBUG_CHANGED		16
#define DEBUG_MAPPING		8
#define DEBUG_BUFFERS		4
#define DEBUG_REQUESTS		2
#define DEBUG_ADDMAPPING	1
#define DEBUG			(DEBUG_CHANGED)

/*
 * Set this if you are developing anything in kernel space.
 * If any of the maps gets overwritten, then this code will not detect it
 * without this set!!!
 */
#define EXTRA_SAFE_BUT_SLOW
/*
 *--------------------------------------------------------------------------------------
 */

/*
 * This is left over from the old code - the old !Config filenames for the files
 * get copied into this array.
 */
static char arc_hd_files[8][128];

struct adfsimage {
    kdev_t		dev;		/* device on which to look for filename		*/
    char		*filename;	/* adfs filename without the 'fsname::drive.$.'	*/
    /*
     * Private data
     */
    kdev_t		mapdev;		/* Current device map array is associated with	*/
    struct map		*mapstrt;	/* Map array					*/
    unsigned int	mapsize;	/* number of map entries - 1			*/
    unsigned int	mapalloced;	/* Size of map array				*/
    unsigned int	blocksize;	/* blocksize this device supports		*/
    unsigned int	accesscount;	/* number of times this has been opened		*/
    unsigned int	capacity;	/* number of sectors in this image file		*/
    struct request	*req;		/* Currently executing request			*/
    struct file		filp;		/* fake filp for open device			*/
    struct inode	inode;		/* fake inode for open device			*/
    unsigned char	busy:1,		/* Device is currently busy			*/
			changed:1;	/* Mapped device has been changed		*/
#ifdef EXTRA_SAFE_BUT_SLOW
    unsigned int	map_checksum[3];
#endif
} adfsimg_images[ADFSIMG_NRDEV] = {
  { 0, arc_hd_files[0], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[1], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[2], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[3], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[4], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[5], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[6], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 },
  { 0, arc_hd_files[7], 0x000, NULL, 0, 0, 0, 0, 0, NULL, {0, }, {0, }, 0, 0 }
};

void adfsimg_add_image (int slot, kdev_t dev, char *path)
{
    adfsimg_images[slot].dev = dev;
    strcpy (adfsimg_images[slot].filename, path);
}

struct map {
    unsigned long strt_sblk;
    unsigned long strt_dblk;
    unsigned long length;
};

static struct wait_queue *busy_wait;

/*
 * These are private to the map generation code
 */
static void *_adfsimg_map;
static struct disc_record adfsimg_dr;
static int adfsimg_zonesize;
static int adfsimg_idsperzone;
static int adfsimg_maplen;

#ifdef EXTRA_SAFE_BUT_SLOW
static void adfsimg_generatechecksum (struct adfsimage *img)
{
    unsigned int c1, c2, c3;
    int i;

    c1 = c2 = c3 = 0;

    for (i = 0; i <= img->mapsize; i++) {
	c1 = ((c1 << 3) | (c1 >> 29)) ^ img->mapstrt[i].strt_sblk;
	c2 = ((c2 << 3) | (c2 >> 29)) ^ img->mapstrt[i].strt_dblk;
	c3 = ((c3 << 3) | (c3 >> 29)) ^ img->mapstrt[i].length;
    }
    img->map_checksum[0] = c1;
    img->map_checksum[1] = c2;
    img->map_checksum[2] = c3;
}

static int adfsimg_checkchecksum (struct adfsimage *img)
{
    unsigned int c1, c2, c3;
    int i;

    c1 = c2 = c3 = 0;

    for (i = 0; i <= img->mapsize; i++) {
	c1 = ((c1 << 3) | (c1 >> 29)) ^ img->mapstrt[i].strt_sblk;
	c2 = ((c2 << 3) | (c2 >> 29)) ^ img->mapstrt[i].strt_dblk;
	c3 = ((c3 << 3) | (c3 >> 29)) ^ img->mapstrt[i].length;
    }
    return (img->map_checksum[0] != c1 ||
	img->map_checksum[1] != c2 ||
	img->map_checksum[2] != c3);
}
#endif

/*
 * Lock a device - prevents opens/closes/revalidates all happening at the same time.
 */
static inline void
adfsimg_lockdev (struct adfsimage *img)
{
    while (img->busy) {
    	barrier ();
	sleep_on (&busy_wait);
    }
    img->busy = 1;
}    

static inline void
adfsimg_unlockdev (struct adfsimage *img)
{
    img->busy = 0;
    wake_up (&busy_wait);
}

/*
 * read blocks to an area of memory.  [eg map, boot block etc].
 */
static void *
adfsimg_read (void *data, struct adfsimage *img, unsigned long blkaddr, int length)
{
    struct buffer_head *bh;
    unsigned char *p;

    for (p = (unsigned char *)data; length > 0; blkaddr ++, length -= img->blocksize) {
	bh = breada (img->dev, blkaddr, img->blocksize, 0, PAGE_SIZE);

	if (!bh)
	    return NULL;
	memcpy (p, bh->b_data, img->blocksize);
	p += img->blocksize;
	brelse (bh);
    }

    return data;
}

/*
 * Calculate block number for zone, zone index.
 */
static inline unsigned long
adfsimg_calcblockno (struct adfsimage *img, unsigned int zone, unsigned int start)
{
    /* *** 4GB LIMITATION ALERT! *** */
    return (((zone ? zone * adfsimg_zonesize - RECSIZE * 8 : 0) + start) << adfsimg_dr.log2bpmb) /
    	   img->blocksize;
}

static inline void
adfsimg_addmapping (struct adfsimage *img, unsigned int zone, unsigned int start, unsigned int len)
{
    struct map newmap;

    /*
     * totaladdr  - starting block number for this segment
     * startaddr  - sector on media for start
     * lengthaddr - length in sectors of this block
     */
#if DEBUG & DEBUG_ADDMAPPING
    printk ("Z: %8d S: %8d E: %8d ", zone, start, start + len);
#endif
    newmap.strt_sblk = 0;
    newmap.strt_dblk = adfsimg_calcblockno (img, zone, start);
    /* *** 4GB LIMITATION ALERT! *** */
    newmap.length    = (len << adfsimg_dr.log2bpmb) / img->blocksize;

    if (!img->mapstrt) {
	img->mapstrt = kmalloc (5 * sizeof (newmap), GFP_KERNEL);
	img->mapsize = 0; /* Size - 1 */
	img->mapalloced = 5; /* allocated size */
	memcpy (img->mapstrt, &newmap, sizeof (newmap));
    } else {
	struct map *mp;
	mp = img->mapstrt + img->mapsize;
	if (mp->strt_dblk + mp->length == newmap.strt_dblk)
	    mp->length += newmap.length;
	else
	if (img->mapsize < img->mapalloced - 1) {
	    newmap.strt_sblk = mp->strt_sblk + mp->length;
	    img->mapsize ++;
	    memcpy (&img->mapstrt[img->mapsize], &newmap, sizeof (newmap));
	} else {
	    newmap.strt_sblk = mp->strt_sblk + mp->length;
	    mp = kmalloc ((img->mapalloced + 5) * sizeof (newmap), GFP_KERNEL);
	    img->mapalloced += 5;
	    img->mapsize ++;
	    memcpy (mp, img->mapstrt, img->mapsize * sizeof (newmap));
	    kfree (img->mapstrt);
	    img->mapstrt = mp;
	    memcpy (&img->mapstrt[img->mapsize], &newmap, sizeof (newmap));
	}
    }
#if DEBUG & DEBUG_ADDMAPPING
    printk ("SSB: %08lX STB: %08lX LEN: %08lX\n", img->mapstrt[img->mapsize].strt_sblk,
		img->mapstrt[img->mapsize].strt_dblk, img->mapstrt[img->mapsize].length);
#endif
}

static unsigned long
adfsimg_scanmap (struct adfsimage *img, unsigned long internaladdr, int objlength, int simple)
{
    unsigned int fragment, zone, startzone, sectoroff, mapoff, found;

    fragment  = (int)internaladdr >> 8;
    startzone = zone = fragment / adfsimg_idsperzone;
    sectoroff = (internaladdr & 0xff) ? (internaladdr & 0xff) - 1 : 0;
/*
 * Problem area: if log2secsize < log2bpmb, then the map offset calculation
 *    breaks horribly.  However, can an internal address refer to a non-allocation
 *    unit sector offset?
 *
 * Ideas:
 *  if (adfsimg_dr.log2secsize > adfsimg_dr.log2bpmb)
 *	mapoff = sectoroff << (adfsimg_dr.log2secsize - adfsimg_dr.log2bpmb);
 *  else
 *	mapoff = sectoroff >> (adfsimg_dr.log2bpmb - adfsimg_dr.log2secsize);
 *
 *  shift = adfsimg_dr.log2secsize - adfsimg_dr.log2bpmb;
 *  if (shift < 0)
 *	mapoff = sectoroff >> (- shift);
 *  else
 *	mapoff = sectoroff << shift;
 */
    mapoff = sectoroff << (adfsimg_dr.log2secsize - adfsimg_dr.log2bpmb);
    objlength >>= adfsimg_dr.log2bpmb;
    found = 0;

    do {
        unsigned char const *map;
        unsigned int bmoff, bmoff_max;

	map = ((unsigned char *)_adfsimg_map) + (zone << adfsimg_dr.log2secsize) + 4 +
			(zone == 0 ? RECSIZE : 0);
	bmoff_max = adfsimg_zonesize - ((zone == 0) ? RECSIZE << 3 : 0);
	bmoff = 0;

	while (bmoff < bmoff_max) {
	    unsigned long t;
	    unsigned int n, start;
	    int z;

	    start = bmoff;
	    z = bmoff >> 3;

	    t = ((map[z] | (map[z + 1] << 8) | (map[z + 2] << 16)) >> (bmoff & 7))
		& ((1 << adfsimg_dr.idlen) - 1);
	    bmoff += adfsimg_dr.idlen;

	    /*
	     * I think that this would be better coded in assembler...
	     */
	    if (bmoff & 7) {
	    	n = map[bmoff >> 3] >> (bmoff & 7);
	    	if (!n)
		    bmoff = (bmoff + 8) & ~7;
		else
		    for (; (n & 1) == 0; n >>= 1, bmoff += 1);
	    }
	    if (!(bmoff & 7)) {
		while ((n = map[bmoff >> 3]) == 0)
		    bmoff += 8;

		for (; (n & 1) == 0; n >>= 1, bmoff += 1);
            }
	    bmoff += 1;
	    if (t == fragment) {
		int length = bmoff - start;
		if (mapoff) {
		    if (length < mapoff) {
			mapoff -= length;
			length = 0;
		    } else {
			start += mapoff;
			length -= mapoff;
		        mapoff = 0;
		    }
		}
		if (length) {
		    if (objlength < length)
			length = objlength;
		    objlength -= length;
		    if (simple)
			return adfsimg_calcblockno (img, zone, start);
		    else
			adfsimg_addmapping(img, zone, start, length);
		    found += 1;
		    if (!objlength)
			return found;
		}
	    }
	}
	zone += 1;
	if (zone == adfsimg_dr.nzones)
	    zone = 0;
	bmoff = 0;
    } while (zone != startzone);
    return found;
}

static int
adfsimg_decenddirs (struct adfsimage *img, int showinfo, unsigned long diraddr, char **ptr)
{
    char *leftptr, *p;
    char dirname[16];
    unsigned char *dir;
    int i, length;

    dir = kmalloc (2048, GFP_KERNEL);
    if (!dir)
	return 0;

    leftptr = *ptr;

    while (1) {
    	unsigned char *dirptr;
	if (!adfsimg_read (dir, img, diraddr, 2048)) {
	    if (showinfo)
		printk (" dir unreadable");
	    kfree (dir);
	    return 0;
	}
	/*
	 * Check that the directory looks feasable
	 */
	if ((strncmp ((char *)dir + 1, "Nick", 4) && strncmp ((char *)dir + 1, "Hugo", 4)) ||
	    strncmp ((char *)dir + 1, (char *)dir + 2043, 4)) {
	    if (showinfo)
		printk (" broken dir");
	    kfree (dir);
	    return 0;
	}

	p = strchr (leftptr, '.');
	if (!p)
	    strcpy (dirname, leftptr);
	else {
	    strncpy (dirname, leftptr, p - leftptr);
	    dirname[p - leftptr] = '\0';
	    leftptr = ++p;
	}

	for (dirptr = dir + 5; *dirptr; dirptr += 26) {
	    for (i = 0; i < 10; i++) {
	    	char c1, c2;
	        if (dirptr[i] == 0x0d || dirname[i] == '\0')
		    break;
		c1 = tolower(dirptr[i]);
		c2 = tolower(dirname[i]);
		if (c1 != c2)
		    break;
	    }
	    if (dirname[i] == '\0' && (i == 10 || dirptr[i] == 0x0d))
		break;
	}

	if (!*dirptr) {
	    if (showinfo)
		printk (" %s '%s' not found\n", p ? "dir" : "file", dirname);
	    kfree (dir);
	    return 0;
	}

	/*
	 * Check that we're not referencing a file when we mean a directory or
	 * vice versa.
	 */
	if ((dirptr[25] & OBJ_DIRECTORY && !p) || (!(dirptr[25] & OBJ_DIRECTORY) && p)) {
	    if (showinfo)
		printk (" %s '%s' is not a %s\n", p ? "file" : "dir", dirname, p ? "dir" : "file");
	    kfree (dir);
	    return 0;
	}

	length = (unsigned long) dirptr[18] | (dirptr[19] << 8) |
		 (dirptr[20] << 16) | (dirptr[21] << 24);
	diraddr = (unsigned long) dirptr[22] | (dirptr[23] << 8) | (dirptr[24] << 16);
	diraddr = adfsimg_scanmap (img, diraddr, length, p ? 1 : 0);
	if (!diraddr) {
	    if (showinfo)
		printk (" internal error");
	    kfree (dir);
	    return 0;
	}
	if (!p) {
	    /*
	     * We must have come to the end of the string.  This entry is the
	     * filename.
	     */
	    kfree (dir);
	    return 1;
	}
    }
}

/*
 * Read in either map, and check the cross-check.  If the first map
 * has an invalid cross-check, try the other map.
 */
static unsigned long
adfsimg_readmap (struct adfsimage *img, int showinfo)
{
    unsigned int map_no, bad, maplen;
    unsigned long mapaddr;

    /* *** 4GB LIMITATION ALERT! *** */
    mapaddr = (((adfsimg_dr.nzones >> 1) * adfsimg_zonesize -
		((adfsimg_dr.nzones > 1) ? RECSIZE * 8 : 0)) << adfsimg_dr.log2bpmb) /
			img->blocksize;
    adfsimg_maplen = adfsimg_dr.nzones << adfsimg_dr.log2secsize;
    maplen = (adfsimg_maplen | (img->blocksize - 1)) / img->blocksize;

    if (!(_adfsimg_map = vmalloc (maplen * img->blocksize))) {
	if (showinfo)
	    printk (" no mem");
	return 0;
    }

    /*
     * We try to read the first map.
     */
    for (map_no = 0; map_no < 2; map_no ++) {
	unsigned long blk_strt;
	unsigned int off, chk;
	unsigned char *map;

	/*
	 * If img->blocksize > 1<<adfsimg_dr.log2secsize, then we could
	 * have problems here when trying the second map.
	 */
	blk_strt = mapaddr + (map_no ? maplen : 0);
	bad = 0;

	if (!adfsimg_read (_adfsimg_map, img, blk_strt, maplen * img->blocksize)) {
	    bad = 1;
	    if (showinfo)
		printk (" map %d unreadable", map_no);
	    break;
	}

	map = (unsigned char *)_adfsimg_map;
	for (off = 3, chk = 0; off < (adfsimg_dr.nzones << adfsimg_dr.log2secsize);
			off += 1 << adfsimg_dr.log2secsize)
	    chk ^= map[off];

	if (chk != 0xff) {
	    bad = 1;
	    if (showinfo)
		printk (" map %d corrupt", map_no + 1);
	} else
	    break;
    }
    if (bad)
    	vfree (_adfsimg_map);

    return bad ? 0 : mapaddr + 2 * maplen;
}

/*
 * Read in the boot sector, and validate it.  Also validate the
 * block size that we are using.
 */
static struct disc_record *
adfsimg_readbootsect (struct adfsimage *img, int showinfo)
{
    unsigned char *boot_sectors;
    int allocation_unit_size;
    struct buffer_head *bh;

    if (!(bh = bread (img->dev, BOOT_SECTOR_ADDRESS / img->blocksize,
		img->blocksize > 512 ? img->blocksize : 512))) {
	if (showinfo)
	    printk (" bread failed");
	return NULL;
    }

    boot_sectors = bh->b_data + BOOT_SECTOR_ADDRESS % img->blocksize;

    if (adfs_checkbblk (boot_sectors)) {
	if (showinfo)
	    printk (" non-adfs");
	brelse (bh);
	return NULL;
    }

    memcpy (&adfsimg_dr, boot_sectors + 0x1c0, sizeof (adfsimg_dr));
    brelse (bh);

    /*
     * The blocksize must not be larger than the sector size specified
     * in the boot sector, or else all this code breaks.
     */
    if (adfsimg_dr.log2secsize > adfsimg_dr.log2bpmb)
	allocation_unit_size = 1 << adfsimg_dr.log2secsize;
    else
	allocation_unit_size = 1 << adfsimg_dr.log2bpmb;
    if (img->blocksize > allocation_unit_size) {
	if (showinfo)
	    printk (" allocation unit size too large - %d > %d", img->blocksize, allocation_unit_size);
	return NULL;
    }

    return &adfsimg_dr;
}

static int adfsimg_init_onedisk (int target, int showinfo)
{
    struct adfsimage *img = adfsimg_images + target;
    unsigned long root_dir_address;
    unsigned int blocksize;
    char *p;

    img->mapdev = MKDEV(UNNAMED_MAJOR, 0);
    if (MAJOR(img->dev) == ADFSIMG_MAJOR)
	return 0;

    if (img->filename == NULL || strlen (img->filename) == 0 || MAJOR(img->dev) == UNNAMED_MAJOR)
	return 0;

    if (showinfo)
	printk (" im%c", target + 'a');

    /*
     * Sort out the sector size that we can support.  Most devices
     * default to 1024 byte blocks, and 512 byte sectors.
     *
     * TODO:
     * However, we do check this against the size wanted in the boot
     * block.  If it says that the sector size is 1024, then we might
     * as well stick with 1024 byte blocks, otherwise switch to 512...
     *
     * Change:
     * We now check the blocksize against the larger of the sector size
     * and the map bit size.
     */
    if (hardsect_size[MAJOR(img->dev)])
	blocksize = hardsect_size[MAJOR(img->dev)][MINOR(img->dev)];
    else
	blocksize = 512;

    set_blocksize (img->dev, blocksize);
    img->blocksize = blocksize;

    /*
     * Now read in the boot sector.  This has to be blocksize
     * independent.
     */
    if (!adfsimg_readbootsect (img, showinfo))
	return 0;

    adfsimg_zonesize   = (8 << adfsimg_dr.log2secsize) - adfsimg_dr.zone_spare;
    adfsimg_idsperzone = adfsimg_zonesize / (adfsimg_dr.idlen + 1);

    if (!(root_dir_address = adfsimg_readmap (img, showinfo)))
	return 0;

    p = img->filename;
    adfsimg_decenddirs (img, showinfo, root_dir_address, &p);

    if (_adfsimg_map != NULL) {
	vfree (_adfsimg_map);
	_adfsimg_map = NULL;
    }

    img->mapdev = img->dev;

#ifdef EXTRA_SAFE_BUT_SLOW
    adfsimg_generatechecksum(img);
#endif

    return img->mapstrt[img->mapsize].strt_sblk + img->mapstrt[img->mapsize].length;
}

/*
 * Get an index into the mapping array for this 'block'.  Once we have found
 * it, then as we map following blocks, we can just increment the index.
 */
static inline int adfsimg_getindex (struct adfsimage *img, unsigned long block)
{
    unsigned int i, low, high;
    const struct map *mp;

    mp = img->mapstrt;

    if (block >= (mp[img->mapsize].strt_sblk + mp[img->mapsize].length)) {
	printk ("%s: access requested to block outside device: blk %lX\n",
		kdevname (img->dev), block);
	return -1;
    }
    for (low = 0, i = (high = img->mapsize) / 2; low != high; i = (high + low) >> 1) {
	if (block < mp[i].strt_sblk)
	    high = i;
	else if (block >= mp[i].strt_sblk + mp[i].length)
	    low = i + 1;
	else break;
    }
#if DEBUG & DEBUG_MAPPING
    printk ("Map entry: %lX [%lX-%lX]->[%lX-%lX]\n", block,
		mp[i].strt_sblk, mp[i].strt_sblk + mp[i].length,
		mp[i].strt_dblk, mp[i].strt_dblk + mp[i].length);
#endif
    return i;
}

#ifdef EXTRA_SAFE_BUT_SLOW
static inline int adfsimg_getindex_check (struct adfsimage *img, unsigned long block, int write)
{
    if (write && adfsimg_checkchecksum(img))
        return -1; /* bad */
    return adfsimg_getindex (img, block);
}
#endif

/*========================================================================================
 * This is the 'device' driver for the image files.
 */

static struct hd_struct adfsimg_parts[256] = { {0, 0}, };

#ifndef EXTRA_SAFE_BUT_SLOW
#define GETINDEX(img,blk,write) adfsimg_getindex ((img), (blk))
#else
#define GETINDEX(img,blk,write) adfsimg_getindex_check ((img), (blk), (write))
#endif

/*
 * Every request that comes in here is a request that needs to be split up into
 * a maximum of 64 separate requests.
 */
static void adfsimg_request (void)
{
    while (1) {
	unsigned int target, minor;
	unsigned long block, length, nsectors, offset;
	struct adfsimage *img;
	int index;

	INIT_REQUEST;

	target = TARGETMAP(CURRENT->rq_dev);
	img = adfsimg_images + target;

	if (MAJOR(img->mapdev) == UNNAMED_MAJOR) {
	    printk (KERN_ERR
		"im%c: request for mapping to unnamed device\n",
			target + 'a');
	    end_request (0);
	    continue;
	}

	if (img->changed) {
	    end_request (0);
	    continue;
	}

	minor = MINOR(CURRENT->rq_dev);
	block = CURRENT->sector;
	nsectors = CURRENT->current_nr_sectors;
	if (block >= adfsimg_parts[minor].nr_sects ||
	    (block + nsectors) > adfsimg_parts[minor].nr_sects) {
	    printk (KERN_ERR
		"im%c: invalid block/size: minor = %u, block = %lu count = %lu\n",
			target + 'a', minor, block, nsectors);
	    end_request (0);
	    continue;
	}

	block += adfsimg_parts[minor].start_sect;
	index = GETINDEX(img, block, (CURRENT->cmd == WRITE));
	if (index < 0) {
	    printk (KERN_ERR
		    "im%c: invalid index: block = %lX\n", target + 'a', block);
	    end_request (0);
	    continue;
	}
#if 0
 	printk (KERN_ERR "Received an adfsimg_request (blk %lu len %lu ", block, nsectors);
#endif
	offset = block - img->mapstrt[index].strt_sblk;
	block = offset + img->mapstrt[index].strt_dblk;
	length = img->mapstrt[index].length - offset;
#if 0
	printk ("-> blk %lu len %lu), blksz=%d ", block, length, img->blocksize);
#endif

	while (nsectors) {
	    struct buffer_head *bh;
	    bh = getblk (img->mapdev, block, img->blocksize);
	    if (!bh) {
		printk (KERN_ERR
			"im%c: getblk(-,%ld,%d) returned NULL",
			target + 'a', block, img->blocksize);
		end_request (0);
		break;
	    }
	    if (!buffer_uptodate(bh) && CURRENT->cmd == READ) {
		ll_rw_block(READ, 1, &bh);
		wait_on_buffer(bh);
		if (!buffer_uptodate(bh)) {
		    brelse(bh);
#if 0
		    printk ("error\n");
#endif
		    end_request (0);
		    break;
		}
	    }
	    if (CURRENT->cmd == READ)
		memcpy (CURRENT->buffer, bh->b_data, img->blocksize);
	    else {
	    	memcpy (bh->b_data, CURRENT->buffer, img->blocksize);
	    	mark_buffer_uptodate(bh, 1);
	    	mark_buffer_dirty(bh, 1);
	    }
	    brelse(bh);
	    CURRENT->buffer += img->blocksize;
	    CURRENT->sector += img->blocksize >> 9;
	    nsectors -= 1;
	    block ++;
	    if (!--length) {
	    	index ++;
		block = img->mapstrt[index].strt_dblk;
		length = img->mapstrt[index].length;
	    }
	    if (nsectors == 0) {
#if 0
		printk ("ok\n");
#endif
		end_request (1);
	    }		
	}
    }
}

static int adfsimg_revalidate_disk (kdev_t dev, int maxusage);
/*
 * TODO:
 *   IOCTL to get mapping filenames
 *   IOCTL to set mapping filenames
 *	- lock device.
 *	- set changed flag on device.
 *	- create new map.
 *	- unlock device.
 */
static int
adfsimg_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    struct hd_geometry *loc = (struct hd_geometry *) arg;
    int err, target;

    if (!inode || !(inode->i_rdev))
	return -EINVAL;

    target = TARGETMAP(inode->i_rdev);

    if (target >= ADFSIMG_NRDEV)
	return -ENODEV;

    switch (cmd) {
    case HDIO_GETGEO:
	if (!loc) return -EINVAL;
	err = verify_area (VERIFY_WRITE, loc, sizeof (*loc));
	if (err)
	    return err;
	put_user (4, (char *)&loc->heads);
	put_user (17, (char *)&loc->sectors);
	put_user (adfsimg_parts[MINOR(inode->i_rdev)].nr_sects / (17 * 4), (short *)&loc->cylinders);
	put_user (adfsimg_parts[MINOR(inode->i_rdev)].start_sect, (long *)&loc->start);
	return 0;
    case BLKRASET:
	if (!suser ()) return -EACCES;
	if (arg > 0xff) return -EINVAL;
	read_ahead[MAJOR(inode->i_rdev)] = arg;
	return 0;
    case BLKRAGET:
	if (!arg)  return -EINVAL;
	err = verify_area (VERIFY_WRITE, (long *)arg, sizeof (long));
	if (err)
	    return err;
	put_user (read_ahead[MAJOR(inode->i_rdev)], (long *)arg);
	return 0;
    case BLKGETSIZE: /* Return device size */
	if (!arg)  return -EINVAL;
	err = verify_area (VERIFY_WRITE, (long *)arg, sizeof (long));
	if (err)
	    return err;
	put_user (adfsimg_parts[MINOR(inode->i_rdev)].nr_sects, (long *)arg);
	return 0;
    case BLKFLSBUF:
	if (!suser ()) return -EACCES;
	fsync_dev (inode->i_rdev);
	invalidate_buffers (inode->i_rdev);
	return 0;

    case BLKRRPART:
	return adfsimg_revalidate_disk (inode->i_rdev, 1);
#if 0
    case IMGFNGET:
	return -EINVAL;

    case IMGFNSET:
	if (!suser ()) return -EACCES;
	return -EINVAL;
#endif

    RO_IOCTLS(inode->i_rdev, arg);
    default:
	return -EINVAL;
    }
}

/*
 * Open an mapped device.  Check for disk changes etc, and open
 * the real device.  This means that if that device wants to, it
 * can do things like locking the drive door.  We force ourself
 * to be busy while opening the real device.
 */
static int adfsimg_open (struct inode *inode, struct file *filp)
{
    struct adfsimage *img;
    unsigned int target;
    int retval = 0;

    target = TARGETMAP(inode->i_rdev);

    if (target >= ADFSIMG_NRDEV)
	return -ENODEV;

    img = adfsimg_images + target;

    if (MAJOR(img->mapdev) == UNNAMED_MAJOR)
	return -ENODEV;
#if 1
    if (filp->f_mode & 2)
	return -EROFS;
#endif
    check_disk_change (inode->i_rdev);

    /*
     * If the device is still changed, then it means that we haven't
     * been able to revalidate it yet.
     */
    if (img->changed)
	return -ENXIO;

    adfsimg_lockdev (img);

    if (img->accesscount == 0) {
	/* was this the first open? */
	struct file *d_filp;
	struct inode *d_inode;
	int retval;

	d_filp = &img->filp;
	d_inode = &img->inode;

	memset (d_filp, 0, sizeof (*d_filp));
	memset (d_inode, 0, sizeof (*d_inode));

	d_inode->i_rdev = img->mapdev;
	d_filp->f_inode = d_inode;
	d_filp->f_mode  = filp->f_mode;

	retval = blkdev_open (d_inode, d_filp);
    }

    if (retval == 0)
	img->accesscount ++;

    adfsimg_unlockdev (img);

    return retval;
}

/*
 * Releasing a block device means we sync() it, so that it can safely
 * be forgotten about...  Close the real device as well if this is the
 * last release.
 */
static void adfsimg_release (struct inode *inode, struct file *filp)
{
    struct adfsimage *img;
    unsigned int target;

    /*
     * Make this image file busy while we unmount the device (we may sleep).
     */

    sync_dev (inode->i_rdev);
    target = TARGETMAP(inode->i_rdev);

    if (target >= ADFSIMG_NRDEV) {
	printk ("im: releasing invalid device %d\n", target);
	return;
    }

    img = adfsimg_images + target;

    adfsimg_lockdev (img);

    img->accesscount --;
    if (img->accesscount == 0 && img->filp.f_op && img->filp.f_op->release)
	img->filp.f_op->release (&img->inode, NULL);

    /*
     * Don't forget to wake up anything on the busy queues.
     */
    adfsimg_unlockdev (img);
}

static void adfsimg_geninit (struct gendisk *dev);

static int adfsimg_sizes[256]      = { 0, };
static int adfsimg_blocksizes[256] = { 0, };
static int adfsimg_hardsizes[256]  = { 0, };

static struct gendisk adfsimg_gendisk = {
    MAJOR_NR,		/* major number */
    "im",		/* major name   */
    ADFSIMG_SHIFT,	/* bits to shift to get real from partition */
    1 << ADFSIMG_SHIFT,	/* number of partitions per real */
    ADFSIMG_NRDEV,	/* maximum number of reals */
    adfsimg_geninit,	/* init function */
    adfsimg_parts,	/* adfsimg_struct */
    adfsimg_sizes,	/* block sizes */
    0,			/* number */
    (void *)NULL,	/* internal */
    NULL		/* next */
};

/*
 * This is the ADFS Image file init entry point
 */
static void adfsimg_geninit (struct gendisk *dev)
{
    int i;

    adfsimg_gendisk.nr_real = ADFSIMG_NRDEV;

    for (i = 0; i < 256; i++) {
	adfsimg_blocksizes[i] = 1024;
	adfsimg_hardsizes[i] = 512;
    }

    blksize_size[MAJOR_NR] = adfsimg_blocksizes;
    hardsect_size[MAJOR_NR] = adfsimg_hardsizes;

    printk (" im :");
    for (i = 0; i < ADFSIMG_NRDEV; i++) {
    	int retval;

    	retval = adfsimg_init_onedisk (i, 1);
    	if (retval > 0) {
	    adfsimg_gendisk.part[i << ADFSIMG_SHIFT].nr_sects = retval;
	    adfsimg_images[i].capacity = retval;
	    adfsimg_images[i].changed = 0;
	} else {
	    adfsimg_gendisk.part[i << ADFSIMG_SHIFT].start_sect = -1;
	    adfsimg_gendisk.part[i << ADFSIMG_SHIFT].nr_sects = 0;
	    adfsimg_images[i].capacity = retval;
	    adfsimg_images[i].changed = 1;
	}
    }
    printk ("\n");
}

/*
 * Check to see if our media that we are mapping has changed, or
 * the mapped file has been altered.  I do wish that there were
 * upcalls...
 *
 * This will always return 1 until:
 * a) ALL opens on the device have been closed.
 * b) The device is re-opened.
 */
static int adfsimg_checkmediachange (kdev_t dev)
{
    struct adfsimage *img;
    int target;

    target = TARGETMAP(dev);

    if (target >= ADFSIMG_NRDEV) {
	printk ("im: adfsimg_checkmediachange: invalid device\n");
	return 0;
    }

    img = adfsimg_images + target;

    if (MAJOR(img->mapdev) == UNNAMED_MAJOR)
 	return 0;

    /*
     * Check our mapped device first.  This way, our mapped device
     * gets invalidated, and we can tell our device that its changed.
     */
    if (check_disk_change(img->mapdev)) {
	/* its changed */
	int i;
	for (i = 0; i < ADFSIMG_NRDEV; i++) {
	    if (img->mapdev == adfsimg_images[i].mapdev) {
		adfsimg_images[i].changed = 1;
#if DEBUG & DEBUG_CHANGED
		printk ("im: checkdiskchange: host device changed [target = %d]\n", i);
#endif
	    }
	}
    }
    if (img->changed)
	return 1;
    return 0;
}

/*
 * Revalidate our partitions.  We do not revalidate the device that
 * we are mapping, since that should have already been done.  However,
 * we do call it the disk change just to make sure.
 *
 * This means that we don't re-read our mapping information, and
 * we keep our size info on the raw image file.
 */
static int adfsimg_revalidate_disk (kdev_t dev, int maxusage)
{
    int target, max_p, start, i;
    struct adfsimage *img;
    unsigned long flags;

    target = TARGETMAP(dev);

    if (target >= ADFSIMG_NRDEV) {
	printk ("im: adfsimg_revalidate: invalid device\n");
	return 0;
    }
#if DEBUG & DEBUG_CHANGED
printk ("im: adfsimg_revalidate: target %d\n", target);
#endif
    img = adfsimg_images + target;

    if (MAJOR(img->mapdev) == UNNAMED_MAJOR)
	return 0;

    save_flags_cli (flags);
    if (img->busy || img->accesscount > maxusage) {
    	restore_flags (flags);
	return -EBUSY;
    }
    img->busy = 1;
    restore_flags (flags);

    start = target << ADFSIMG_SHIFT;
    max_p = adfsimg_gendisk.max_p;

    for (i = adfsimg_gendisk.max_p - 1; i >= 0; i--) {
	int minor = start + i;
	if (adfsimg_gendisk.part[minor].nr_sects > 0) {
	    kdev_t devi = MKDEV(MAJOR_NR, minor);
	    sync_dev (devi);
	    invalidate_inodes (devi);
	    invalidate_buffers (devi);
	}
	adfsimg_gendisk.part[minor].start_sect = 0;
	adfsimg_gendisk.part[minor].nr_sects = 0;
	blksize_size[MAJOR_NR][minor] = 1024;
    }

    if (!check_disk_change (img->mapdev)) {
	img->changed = 0;

	adfsimg_gendisk.part[start].nr_sects = img->capacity;
	resetup_one_dev (&adfsimg_gendisk, target);
    }

    adfsimg_unlockdev (img);
    return 0;
}

static int adfsimg_revalidate (kdev_t dev)
{
    return adfsimg_revalidate_disk (dev, 0);
}

static struct file_operations adfsimg_fops = {
    NULL,		/* lseek - default */
    block_read,		/* read - general block-dev read */
    block_write,	/* write - general block-dev write */
    NULL,		/* readdir - bad */
    NULL,		/* select */
    adfsimg_ioctl,	/* ioctl */
    NULL,		/* mmap */
    adfsimg_open,	/* open */
    adfsimg_release,	/* release */
    block_fsync,	/* fsync */
    NULL,		/* fasync */
    adfsimg_checkmediachange, /* disk change */
    adfsimg_revalidate	/* revalidate */
};

int adfsimg_init (void)
{
    struct gendisk **gdp;

    if (register_blkdev (MAJOR_NR, "im", &adfsimg_fops)) {
	printk ("adfsimg: unable to get major %d for image file system\n", MAJOR_NR);
	return -1;
    }
    blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
    read_ahead[MAJOR_NR] = 0;	/* we rely on the hardware reading ahead */

    for (gdp = &gendisk_head; *gdp; gdp = &((*gdp)->next));
    *gdp = &adfsimg_gendisk;	/* link onto tail of list */
    return 0;
}
