/*
 * bulkmem_itsy.c
 * Implementation of the PCMCIA Card Services "bulk memory"
 * functionality for the Itsy onboard and daughtercard flash, but
 * without the Card Services infrastructure.
 * Tim Mann 3-18-99
 * Copyright (c) Compaq Computer Corporation, 1999
 */

#include <linux/version.h>
#define __NO_VERSION__ 1
#include "k_compat.h"
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/hdreg.h>
#include <linux/interrupt.h>
#include <linux/flinfo.h>
#include <linux/bulkmem_itsy.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>

#if (LINUX_VERSION_CODE >= VERSION(2,1,0))
#include <linux/vmalloc.h>
#endif

#ifdef FTL_ITSY_DEBUG
extern int ftl_itsy_debug;
#define DEBUG(n, args...) if (ftl_itsy_debug>(n)) printk(KERN_DEBUG args)
#define STATIC
int ftl_itsy_check_writes = 0;
MODULE_PARM(ftl_itsy_check_writes, "i");
#else
#define DEBUG(n, args...)
#define STATIC static
#endif

static FlMemInfo *flMemInfo[LAST_BANK+1];
static int valid[LAST_BANK+1];
static int openCount[LAST_BANK+1];
static volatile int needCheck[LAST_BANK+1]; /* BMICheckEraseQueue deferred */
static eraseq_hdr_t eraseQueue[LAST_BANK+1];
static client_reg_t clientReg[LAST_BANK+1];
static void BMIGotCompletionInterrupt(int b);
#ifdef CONFIG_POWERMGR
static void BMIPowerSuspend(int b);
static void BMIPowerResume(int b);
#endif

/* Current operation in progress on each bank */
/* !!Note: we keep only one operation in progress at a time, even
   though there may be multiple sub-banks that can do different things
   in parallel.  */
#define BANK_IDLE 0
#define BANK_READING 1
#define BANK_WRITING 2
#define BANK_ERASING 3
#define BANK_RWE 3 /* mask for reading/writing/erasing */
#define BANK_ERASE_SUSPEND 4  /*can be or'ed with reading or writing */
#define BANK_ERASE_COMPLETE 8 /*can be or'ed with reading, writing, or idle */
#define BANK_POWERDOWN 16 /*can be or'ed with erasing or erase-suspend*/
static volatile int bankState[LAST_BANK+1];

/* bulkmem.h leaves codes 1 to 0x7f for states during the erase */
#define ERASE_STARTED 1

/* !!should be in a .h file */

extern void flash_reg_interrupt(int banknum, void (*interrupt)(int banknum));
extern void flash_unreg_interrupt(int banknum);
#ifdef CONFIG_POWERMGR
extern void flash_reg_powermgr(int banknum,
			       void (*suspend)(int), void (*resume)(int));
extern void flash_unreg_powermgr(int banknum);
#endif
extern FlMemInfo* flash_getinfo_prim(int banknum);

void
BMIFlashError(char* str, int err)
{
  int status;
  void *addr;
  unsigned int data;

  FlMemGetError(&status, &addr, &data);
  printk("Flash hardware error in %s:"
	 " status %d, addr 0x%lx, wrote 0x%x, read 0x%x\n",
	 str, status, (unsigned long)addr, data, *(unsigned int*)addr);
  if (status != err) {
    printk("?error getting above state, err %d != status %d\n", err, status);
  }
}

int
BMIRegisterClient(client_handle_t client, client_reg_t *reg)
{
  int bank = (int) client;
  DEBUG(0, "BMIRegisterClient bank %d\n", bank);
  if (!valid[bank]) {
    flMemInfo[bank] = flash_getinfo_prim(bank);
    if (flMemInfo[bank] == NULL) return -ENXIO;
    valid[bank] = 1;
  }
  clientReg[bank] = *reg; 
  eraseQueue[bank].QueueEntryCnt = 0;
  eraseQueue[bank].QueueEntryArray = NULL;
  return 0;
}

int
BMIDeregisterClient(client_handle_t client)
{
  DEBUG(0, "BMIDeregisterClient bank %d\n", (int)client);
  memset(&clientReg[(int) client], sizeof(client_reg_t), 0);
  valid[(int)client] = 0;
  return 0;
}

STATIC int
BMIAddrToBank(caddr_t vaddr)
{
  int i;
  for (i=FIRST_BANK; i<=LAST_BANK; i++) {
    FlMemInfo *fi = flMemInfo[i];
    if (!valid[i]) continue;
    if (vaddr >= (caddr_t) fi->subBnkAdd[0] &&
	vaddr < (caddr_t) fi->subBnkAdd[0] + fi->size) return i;
  }
  return -1;
}

STATIC int
BMIAddrToSubbank(void* addr, FlMemInfo* fi)
{
  int i;
  for (i=1; i<fi->numSubBnk; i++) {
    if (addr < fi->subBnkAdd[i]) break;
  }
  return i-1;
}

int
BMIGetFirstRegion(client_handle_t handle, region_info_t *region)
{
  FlMemInfo *fi = flMemInfo[(int)handle];
  DEBUG(0, "BMIGetFirstRegion bank %d\n", (int)handle);

  region->Attributes = REGION_TYPE_CM;
  region->AccessSpeed = 0; /* unused */
  region->BlockSize = fi->nlSctSz;
  region->PartMultiple = region->BlockSize;
  region->JedecMfr = 0; /* unused */
  region->JedecInfo = 0; /* unused */
  region->next = (memory_handle_t) 1; /* next index in region table */

  /* First region is always the whole bank.  The final sector is
     left unused in the ftl devices because 256 bytes at the end are
     reserved by the Itsy monitor.  The raw char devices allow access
     to this sector, but beware if you erase it; you must write back
     the final 256 bytes! */
  region->CardOffset = 0;
  region->RegionSize = fi->size - fi->nlSctSz; 

  return 0;
}

int
BMIGetNextRegion(client_handle_t handle, region_info_t *region)
{
  FlMemInfo *fi = flMemInfo[(int)handle];
  region_desc_t *rs =
    (region_desc_t *) (fi->subBnkAdd[0] + fi->size - REGION_SIZES_OFFSET);
  int next = (int)region->next;
  DEBUG(0, "BMIGetNextRegion bank %d, region->next %d\n", (int)handle, next);

  if (next == MAX_REGION || rs[next].start == 0xffff ||
      (rs[next].start == 0 && rs[next].size == 0)) {
    /* No more regions */
    return -1;
  } else {
    region->CardOffset = rs[next].start * fi->nlSctSz;
    region->RegionSize = rs[next].size * fi->nlSctSz;
    region->next = (memory_handle_t) (next + 1);
    return 0;
  }
}

int
BMIOpenMemory(client_handle_t *handle, open_mem_t *req)
{
  int b = (int)*handle;
  DEBUG(0, "BMIOpenMemory bank %d, offset %d\n", b, req->Offset);
  *handle = (client_handle_t) flMemInfo[b]->subBnkAdd[0] + req->Offset;
  if (openCount[b]++ == 0) {
#ifdef CONFIG_POWERMGR
    flash_reg_powermgr(b, BMIPowerSuspend, BMIPowerResume);
#endif
  }
  return 0;
}


int
BMICloseMemory(memory_handle_t handle)
{
  int b = BMIAddrToBank(handle);
  DEBUG(0, "BMICloseMemory address %x (bank %d)\n", (u32)handle, b);
  if (--openCount[b] == 0) {
#ifdef CONFIG_POWERMGR
    flash_unreg_powermgr(b);
#endif
  }
  return 0;
}


/* Suspend erasing on bank b if needed and enter state s. */
/* Succeeds if bank is idle, erasing, or idle with an erase complete.
   Fails with -EWOULDBLOCK otherwise. */
STATIC int
BMIEraseSuspend(int b, int s)
{
  unsigned long flags;
  int err = 0;
  static int msgcount = 0;

  save_flags(flags);
  cli();
  DEBUG(4, "BMIEraseSuspend bank %d state %d req state %d\n",
	b, bankState[b], s);
  switch (bankState[b]) {
  case BANK_ERASING:
    {
      int i = eraseQueue[b].QueueEntryCnt;
      eraseq_entry_t *e = eraseQueue[b].QueueEntryArray;
      FlMemInfo *fi = flMemInfo[b];
      if (++msgcount < 10) {
	DEBUG(0, "BMIEraseSuspend: suspend bank %d (state %d req state %d)\n",
	      b, bankState[b], s);
      }
      while (i--) {
	if (ERASE_IN_PROGRESS(e->State)) {
	  void *ad, *bs;
	  flash_unreg_interrupt(b);
	  ad = e->Handle + e->Offset + e->Done;
	  bs = fi->subBnkAdd[BMIAddrToSubbank(ad, fi)];
	  if (msgcount < 10) {
	    DEBUG(1, "  sctErsSusAlg(%x, %x)\n", (u32)bs, (u32)ad);
	  }
	  err = fi->sctErsSusAlg(bs, ad);
	  if (err == 0) {
	    /* Erase suspended successfully */
	    bankState[b] = s | BANK_ERASE_SUSPEND;
	  } else if (err == -1) {
	    /* Erase was already complete */
	    bankState[b] = s | BANK_ERASE_COMPLETE;
	    needCheck[b] = 1;
	    err = 0;
	  } else {
	    BMIFlashError("erasing (BMIEraseSuspend)", err);
	    e->State = ERASE_FAILED;
	    bankState[b] = s | BANK_ERASE_COMPLETE;
	    err = -EIO;
	  }
	  break;
	}
	e++;
      }
      if (i < 0) {
	bankState[b] = s;
	printk("BMIEraseSuspend BUG: found no erase to suspend\n");
      }
      if (msgcount < 10) {
	DEBUG(1, "  bank %d new state %d\n", b, bankState[b]);
      }
    }
    break;
  case BANK_IDLE | BANK_ERASE_COMPLETE:
    bankState[b] = s | BANK_ERASE_COMPLETE;
    break;
  case BANK_IDLE:
    bankState[b] = s;
    break;
  case BANK_READING:
  case BANK_READING | BANK_ERASE_SUSPEND:
  case BANK_READING | BANK_ERASE_COMPLETE:
  case BANK_WRITING:
  case BANK_WRITING | BANK_ERASE_SUSPEND:
  case BANK_WRITING | BANK_ERASE_COMPLETE:
    /* Possible if the character driver blocked and the bulkmem layer is
       being reentered.  Current caller must wait behind previous caller. */
    err = -EWOULDBLOCK;
    break;
  default:
    /* Bug! */
    printk("BMIEraseSuspend BUG: bad state %d\n", bankState[b]);
    break;
  }
  restore_flags(flags);
  return err;
}

/* Resume erasing on bank b if needed. */
STATIC int
BMIEraseResume(int b)
{
  int i = eraseQueue[b].QueueEntryCnt;
  eraseq_entry_t *e = eraseQueue[b].QueueEntryArray;
  FlMemInfo *fi = flMemInfo[b];
  unsigned long flags;
  int err = 0;
  static int msgcount = 0;

  save_flags(flags);
  cli();
  DEBUG(4, "BMIEraseResume bank %d state %d\n", b, bankState[b]);
  if (bankState[b] & BANK_ERASE_SUSPEND) {
    if (++msgcount < 10) {
      DEBUG(0, "BMIEraseResume: resume bank %d (state %d)\n",
	    b, bankState[b]);
    }
    while (i--) {
      if (ERASE_IN_PROGRESS(e->State)) {
	void *ad, *bs;
	flash_reg_interrupt(b, BMIGotCompletionInterrupt);
	ad = e->Handle + e->Offset + e->Done;
	bs = fi->subBnkAdd[BMIAddrToSubbank(ad, fi)];
	if (msgcount < 10) {
	  DEBUG(1, "  sctErsResAlg(%x, %x)\n", (u32)bs,	(u32)ad);
	}
	err = fi->sctErsResAlg(bs, ad);
	if (err) {
	  /* this can't really happen */
	  BMIFlashError("resuming erase (BMIEraseResume)", err);
	  err = -EIO;
	}
	bankState[b] = BANK_ERASING;
	break;
      }
      e++;
    }
    if (i < 0) {
      bankState[b] = BANK_IDLE;
      printk("BMIEraseResume BUG: found no erase to resume\n");
    }
    if (msgcount < 10) {
      DEBUG(1, "  bank %d new state %d\n", b, bankState[b]);
    }
  } else {
    bankState[b] = BANK_IDLE | (bankState[b] & BANK_ERASE_COMPLETE);
  }
  restore_flags(flags);
  return err;
}

int
BMIReadMemory(memory_handle_t handle, mem_op_t *req, caddr_t buf)
{
  int err = 0;
  int b = BMIAddrToBank(handle);

  DEBUG(3, "BMIReadMemory base %x, offset %x, count %x\n",
	(u32)handle, (u32)req->Offset, (u32)req->Count);
  if (b < 0) return -EIO;

  if (!(req->Attributes & MEM_OP_BUFFER_KERNEL)) {
    err = verify_area(VERIFY_WRITE, (void*)buf, req->Count);
    if (err) return err;
  }
    
  err = BMIEraseSuspend(b, BANK_READING);
  if (err) return err;
  if (req->Attributes & MEM_OP_BUFFER_KERNEL) {
    memcpy(buf, handle + req->Offset, req->Count);
  } else {
    /* Note that this may block! */
    memcpy_tofs(buf, handle + req->Offset, req->Count);
  }
  err = BMIEraseResume(b);

  /* In case handling the erase queue was deferred while we were reading */
  if (needCheck[b]) {
    (void) BMICheckEraseQueue((eraseq_handle_t)
			      BMIAddrToBank((caddr_t) handle));
  }

  return err;
}

int
BMIWriteMemory(memory_handle_t handle, mem_op_t *req, caddr_t buffer)
{
  int err;
  int ret = 0;
  int b = BMIAddrToBank(handle);
  FlMemInfo *fi = flMemInfo[b];
  void* start = handle + req->Offset;
  int sb = BMIAddrToSubbank(start, fi);
  u_int count = req->Count;
  caddr_t buf = buffer;
  int count1;
  char scratch[4];
  int alignment_offset;

  DEBUG(3, "BMIWriteMemory base %x, offset %x, count %x\n",
	(u32)handle, (u32)req->Offset, (u32)req->Count);
  if (b < 0) return -EIO;

  if (count == 0) return 0;

  if (!(req->Attributes & MEM_OP_BUFFER_KERNEL)) {
    err = verify_area(VERIFY_READ, (void*)buf, req->Count);
    if (err) return err;
  }
    
  err = BMIEraseSuspend(b, BANK_WRITING);
  if (err) return err;

  /* Check whether start of area to be programmed is aligned */
  alignment_offset = ((u_long)start) & fi->dataWdthMsk;
  if (alignment_offset != 0) {
    /* Unaligned.  Do read/modify/write for the first word */
    count1 = fi->dataWdthMsk + 1 - alignment_offset;
    if (count1 > count) count1 = count;
    memcpy(scratch, start - alignment_offset, fi->dataWdthMsk + 1);
    if (req->Attributes & MEM_OP_BUFFER_KERNEL) {
      memcpy(scratch + alignment_offset, buf, count1);
    } else {
      memcpy_fromfs(scratch + alignment_offset, buf, count1);
    }
    err =
      fi->prgmAlg(fi->subBnkAdd[sb], start - alignment_offset,
		  scratch, fi->dataWdthMsk + 1);
    if (err) {
      BMIFlashError("programming (BMIWriteMemory)", err);
      ret = -EIO;
      goto finish;
    }
    buf += count1;
    start += count1;
    count -= count1;
  }

  /* Check whether end of area remaining to be programmed is aligned */
  alignment_offset = ((u_long)(start + count)) & fi->dataWdthMsk;

  /* Loop through (possibly) multiple subbanks */
  /* count2 = amount to do this outer iteration */
  /* count1 = amount to do this inner iteration */
  while (count > alignment_offset) {
    int count2 = count;
    if (sb+1 < fi->numSubBnk && start+count > fi->subBnkAdd[sb+1]) {
      /* Avoid crossing subbanks within this iteration */
      count2 = fi->subBnkAdd[sb+1] - start;
    }

    if (((u_long)buf) & fi->dataWdthMsk) {
      /* Source and destination did not have the same alignment, so
	 now that designation is aligned, source is unaligned. */
      count1 = fi->dataWdthMsk + 1;
      if (count2 >= count1) {
	/* This is pretty slow if we really have to do it */
	DEBUG(0, "BMIWriteMemory slow case\n");
      }
      while (count2 >= count1) {
	if (req->Attributes & MEM_OP_BUFFER_KERNEL) {
	  memcpy(scratch, buf, count1);
	} else {
	  memcpy_fromfs(scratch, buf, count1);
	}
	err = fi->prgmAlg(fi->subBnkAdd[sb], start, scratch, count1);
	if (err) {
	  BMIFlashError("programming (BMIWriteMemory)", err);
	  ret = -EIO;
	  goto finish;
	}
	buf += count1;
	start += count1;
	count2 -= count1;
	count -= count1;
      }
    } else {
      /* Fast case: buf and start now are both aligned */
      count1 = count2 & ~fi->dataWdthMsk;
      if (req->Attributes & MEM_OP_BUFFER_KERNEL) {
	err = fi->prgmAlg(fi->subBnkAdd[sb], start, buf, count1);
      } else {
	/* Not so fast after all: must copy from user space */
	caddr_t buf2 = vmalloc(count1);
	memcpy_fromfs(buf2, buf, count1);
	err = fi->prgmAlg(fi->subBnkAdd[sb], start, buf2, count1);
	vfree(buf2);
      }
      if (err) {
	BMIFlashError("programming (BMIWriteMemory)", err);
	ret = -EIO;
	goto finish;
      }
      buf += count1;
      start += count1;
      count2 -= count1;
      count -= count1;
    }

    /* Update subbank number if needed */
    sb = BMIAddrToSubbank(start, fi);
  }

  if (alignment_offset != 0) {
    /* End of area to be programmed is unaligned.  Do a
       read/modify/write for the last 16- or 32-bit word. */
    count1 = fi->dataWdthMsk + 1;
    memcpy(scratch, start, count1);
    if (req->Attributes & MEM_OP_BUFFER_KERNEL) {
      memcpy(scratch, buf, alignment_offset);
    } else {
      memcpy_fromfs(scratch, buf, alignment_offset);
    }
    err = fi->prgmAlg(fi->subBnkAdd[sb], start, scratch, count1);
    if (err) {
      BMIFlashError("programming (BMIWriteMemory)", err);
      ret = -EIO;
      goto finish;
    }
  }

#ifdef FTL_ITSY_DEBUG
  if (ftl_itsy_check_writes &&
      memcmp(buffer, handle + req->Offset, req->Count) != 0) {
    printk("Flash error: data read back does not match data programmed:"
	   " addr 0x%lx, count 0x%x\n",
	   (unsigned long) handle + req->Offset, req->Count);
  }
#endif

 finish:
  err = BMIEraseResume(b);
  if (err && ret == 0) ret = err;

  /* In case handling the erase queue was deferred while we were writing */
  if (needCheck[b]) {
    (void) BMICheckEraseQueue((eraseq_handle_t)
			      BMIAddrToBank((caddr_t) handle));
  }

  return ret;
}


int
BMIRegisterEraseQueue(client_handle_t *handle, eraseq_hdr_t *header)
{
  int bank = (int) *handle;
  DEBUG(0, "BMIRegisterEraseQueue bank %d, queue (%d,%x)\n",
	bank, header->QueueEntryCnt, (u32)header->QueueEntryArray);
  eraseQueue[bank] = *header;
  return 0;
}


int
BMIDeregisterEraseQueue(eraseq_handle_t handle)
{
  int bank = (int) handle;
  DEBUG(0, "BMIDeregisterEraseQueue bank %d\n", bank);
  eraseQueue[bank].QueueEntryCnt = 0;
  eraseQueue[bank].QueueEntryArray = NULL;
  return 0;
}


/* Task queue entries so that the callback to ftl_itsy can run 
   from the scheduler task queue */
static struct tq_struct bmi_tq[LAST_BANK+1];

/* Check what there is to do on this erase queue, and move it along if
   possible.  May be called from interrupt level or not. */
int
BMICheckEraseQueue(eraseq_handle_t handle) 
{
  unsigned long flags;
  int i;
  eraseq_entry_t *e;
  int err = 0;
  int b = (int)handle;
  FlMemInfo *fi;
  void *addr;

  DEBUG(0, "BMICheckEraseQueue handle %x\n", (u32)handle);

  if (!valid[b]) return 0;
  if (eraseQueue[b].QueueEntryArray == NULL) return 0;
  fi = flMemInfo[b];
  save_flags(flags);
  cli();

  DEBUG(1, "  bank %d state %d\n", b, bankState[b]);

  switch (bankState[b]) {
  case BANK_ERASING:
  case BANK_IDLE | BANK_ERASE_COMPLETE:
    /* Find active erase queue entry and check if done */
    i = eraseQueue[b].QueueEntryCnt;
    e = eraseQueue[b].QueueEntryArray;
    while (i--) {
      if (ERASE_IN_PROGRESS(e->State)) {
	addr = e->Handle + e->Offset + e->Done;
	if (bankState[b] == BANK_ERASING) {
	  err = fi->sctErsStat(fi->subBnkAdd[BMIAddrToSubbank(addr, fi)],
			       addr);
	  if (err == -1) {
	    /* Erase still in progress */
	    DEBUG(1, "  erase still in progress\n");
	    break;
	  }
	  flash_unreg_interrupt(b);
	} else {
	  DEBUG(1, "  erase completed earlier\n");
	}
	if (err) {
	  BMIFlashError("erasing (BMICheckEraseQueue)", err);
	  e->State = ERASE_FAILED;
	  bankState[b] = BANK_IDLE;
	  e->Done = e->Size;
	} else {
	  u_int sctsz;
	  /* Check if we just did a short sector (yuck) */
	  if (addr >= (void *)fi->spSctAdd[0] &&
	      addr < (void *)(fi->spSctAdd[fi->numSpSct-1] +
			      fi->spSctSz[fi->numSpSct-1])) {
	    /* Yes, which one (i.e., how big)? */
	    int j;
	    for (j=0; j<fi->numSpSct; j++) {
	      if (addr == (void *)fi->spSctAdd[j]) break;
	    }
	    if (j == fi->numSpSct) {
	      printk("BMICheckEraseQueue BUG: unaligned sector\n");
	      sctsz = e->Size;
	    } else {
	      sctsz = fi->spSctSz[j];
	    }
	  } else {
	    /* No; did a full sector */
	    sctsz = fi->nlSctSz;
	  }
#ifdef FTL_ITSY_DEBUG
	  if (ftl_itsy_check_writes) {
	    unsigned char *p = (unsigned char *) addr;
	    u_int i = sctsz;
	    while (i--) {
	      if (*p != 0xff) {
		printk("Flash error: data was not erased:"
		       " addr 0x%lx, data 0x%x", (unsigned long) p, *p);
		e->State = ERASE_FAILED;
		bankState[b] = BANK_IDLE;
		e->Done = e->Size;
		goto idle;
	      }
	      p++;
	    }
	  }
#endif
	  e->Done += sctsz;
	  DEBUG(1, "  erase %x done, %x left\n", e->Done, e->Size - e->Done);
	  /* Have we done the whole erase now? */
	  if (e->Done == e->Size) {
	    /* Yes */
	    e->State = ERASE_PASSED;
	    bankState[b] = BANK_IDLE;
	  } else {
	    /* No; start the next part */
	    flash_reg_interrupt(b, BMIGotCompletionInterrupt);
	    addr = e->Handle + e->Offset + e->Done;
	    err = fi->sctErsAlg(fi->subBnkAdd[BMIAddrToSubbank(addr, fi)],
				addr);
	    if (err) {
	      /* can't really happen */
	      BMIFlashError("starting erase (BMICheckEraseQueue)", err);
	    }
	    /* Erase again in progress */
	    bankState[b] = BANK_ERASING;
	    break;
	  }
	}
      } else if (e->State == ERASE_FAILED) {
	/* Erase failure discovered when suspending */
	bankState[b] = BANK_IDLE;
	e->Done = e->Size;
      }
      if (bankState[b] == BANK_IDLE) {
	/* Tell ftl_itsy we finished this request.  Use the
	   scheduler task queue to avoid reentering ftl_itsy's
	   request_fn */
	clientReg[b].event_callback_args.info = (void*) e;
	if (clientReg[b].event_handler) {
	  bmi_tq[b].routine = (void (*)(void*)) clientReg[b].event_handler;
	  bmi_tq[b].data = (void*) &clientReg[b].event_callback_args;
	  queue_task(&(bmi_tq[b]), &tq_scheduler);
	}
	break;
      }
      e++;
    }
    needCheck[b] = 0;
    if (bankState[b] == BANK_ERASING) break;
    /* fall thru */

  case BANK_IDLE:
  idle:
    /* Look for an erase to start */
    i = eraseQueue[b].QueueEntryCnt;
    e = eraseQueue[b].QueueEntryArray;
    while (i--) {
      if (e->State == ERASE_QUEUED) {
	e->Done = 0;
	addr = e->Handle + e->Offset + e->Done;
	DEBUG(1, "  starting erase at %x\n", (u32)addr);
	flash_reg_interrupt(b, BMIGotCompletionInterrupt);
	err = fi->sctErsAlg(fi->subBnkAdd[BMIAddrToSubbank(addr, fi)], addr);
	if (err) {
	  /* can't really happen */
	  BMIFlashError("starting erase (BMICheckEraseQueue)", err);
	} else {
	  e->State = ERASE_STARTED;
	  bankState[b] = BANK_ERASING;
	}
	break;
      }
      e++;
    }
    needCheck[b] = 0;
    break;
	
  case BANK_READING:
  case BANK_READING | BANK_ERASE_SUSPEND:
  case BANK_READING | BANK_ERASE_COMPLETE:
  case BANK_WRITING:
  case BANK_WRITING | BANK_ERASE_SUSPEND:
  case BANK_WRITING | BANK_ERASE_COMPLETE:
    /* Can't do anything right now */
    DEBUG(1, "  setting needCheck[%d]\n", b);
    needCheck[b] = 1;
    break;

  default:
    /* Bug! */
    printk("BMICheckEraseQueue BUG: bad state %d\n", bankState[b]);
    break;
  }
  DEBUG(1, "  bank %d new state %d\n", b, bankState[b]);
  restore_flags(flags);
  return 0;
}


STATIC void
BMIGotCompletionInterrupt(int b)
{
  if (valid[b]) {
    DEBUG(1, "BMIGotCompletionInterrupt bank %d\n", b);
    (void) BMICheckEraseQueue((eraseq_handle_t)b);
  }
}

#ifdef CONFIG_POWERMGR
/* Power manager asked us to suspend what we're doing.  If an erase is
   in progress or suspended, assume that the powerdown will kill it,
   and mark it to be redone after resume. */
STATIC void
BMIPowerSuspend(int b)
{
  unsigned long flags;
  DEBUG(0, "BMIPowerSuspend bank %d\n", b);

  if (!valid[b]) return;
  if (eraseQueue[b].QueueEntryArray == NULL) return;
  save_flags(flags);
  cli();

  switch (bankState[b]) {
  case BANK_ERASING:
    DEBUG(1, "  bank %d was erasing\n", b);
    /* Flag bank as needing attention after powerup */
    bankState[b] |= BANK_POWERDOWN;
    flash_unreg_interrupt(b);
    break;

  case BANK_READING | BANK_ERASE_SUSPEND:
  case BANK_WRITING | BANK_ERASE_SUSPEND:
    DEBUG(1, "  bank %d was erase-suspended (state %d)\n",
	  b, bankState[b]);
    /* Flag bank as needing attention after powerup */
    bankState[b] |= BANK_POWERDOWN;
    break;

  case BANK_READING:
  case BANK_READING | BANK_ERASE_COMPLETE:
  case BANK_WRITING:
  case BANK_WRITING | BANK_ERASE_COMPLETE:
  case BANK_IDLE:
  case BANK_IDLE | BANK_ERASE_COMPLETE:
    /* Nothing needs to be done */
    break;
  }

  restore_flags(flags);
}

/* Power manager asked us to resume what we're doing.  Find out if we
   were doing an erase and mark it to be re-done if so.

   !!Assume that a powerdown really happened, causing the chips to be
   reset if there were any erases in progress.  Might be a bit safer
   to forcibly reset them here to make sure, but that would require a
   new primitive in flprgm.c, so we don't do it for now.
*/
STATIC void
BMIPowerResume(int b)
{
  int i;
  eraseq_entry_t *e;

  DEBUG(0, "BMIPowerResume bank %d\n", b);

  if (!valid[b]) return;
  if (eraseQueue[b].QueueEntryArray == NULL) return;
  if (!(bankState[b] & BANK_POWERDOWN)) return;

  switch (bankState[b]) {
  case BANK_POWERDOWN|BANK_ERASING:
    bankState[b] = BANK_IDLE;
    break;
  case BANK_POWERDOWN|BANK_ERASE_SUSPEND|BANK_READING:
    bankState[b] = BANK_READING;
    break;
  case BANK_POWERDOWN|BANK_ERASE_SUSPEND||BANK_WRITING:
    bankState[b] = BANK_WRITING;
    break;
  default:
    printk("BMIPowerResume BUG: bad state %d\n", bankState[b]);
    return;
  }

  /* Find active erase queue entries and force them to restart */
  i = eraseQueue[b].QueueEntryCnt;
  e = eraseQueue[b].QueueEntryArray;
  while (i--) {
    if (ERASE_IN_PROGRESS(e->State)) {
      e->State = ERASE_QUEUED;
    }
    e++;
  }

  (void) BMICheckEraseQueue((eraseq_handle_t)b);
}
#endif
