/*
 * Flash Translation Layer driver for Itsy.  Based on PCMCIA Card
 * Services version, but hacked mercilessly to work on top of 
 * flinfo.h interface instead.
 * Changes Copyright (c) Compaq Computer Corporation, 1999
 *
 * Tim Mann 10-Mar-99
 *
 * Original header comment and copyright notice follow.
 */

/*======================================================================

    A Flash Translation Layer memory card driver

    This driver implements a disk-like block device driver with an
    apparent block size of 512 bytes for flash memory cards.

    ftl_cs.c 1.38 1998/07/30 23:13:19

    The contents of this file are subject to the Mozilla Public
    License Version 1.0 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The initial developer of the original code is David A. Hinds
    <dhinds@hyper.stanford.edu>.  Portions created by David A. Hinds
    are Copyright (C) 1998 David A. Hinds.  All Rights Reserved.
    
======================================================================*/

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#include <linux/version.h>
#include "k_compat.h"

#define PSYCHO_DEBUG 1 /*!!*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/hdreg.h>
#include <linux/bulkmem_itsy.h>
#include <linux/memory.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>

#if (LINUX_VERSION_CODE >= VERSION(2,1,0))
#include <linux/vmalloc.h>
#endif

#include "ftl.h"

/***** Copied from pcmcia/driver_ops.h *****/
#ifndef DEV_NAME_LEN
#define DEV_NAME_LEN    32
#endif

typedef struct dev_node_t {
    char                dev_name[DEV_NAME_LEN];
    u_short             major, minor;
    struct dev_node_t   *next;
} dev_node_t;
/***********************************/

/***** Copied from pcmcia/ds.h *****/
typedef struct dev_link_t {
    dev_node_t          *dev;
    u_int               state, open;
    struct wait_queue   *pending;
    struct timer_list   release;
    client_handle_t     handle;
    /*io_req_t            io;*/
    /*irq_req_t           irq;*/
    /*config_req_t        conf;*/
    /*window_handle_t     win;*/
    void                *priv;
    struct dev_link_t   *next;
} dev_link_t;

/* Flags for device state */
#define DEV_PRESENT             0x01
#define DEV_CONFIG              0x02
#define DEV_STALE_CONFIG        0x04    /* release on close */
#define DEV_STALE_LINK          0x08    /* detach on release */
#define DEV_CONFIG_PENDING      0x10
#define DEV_RELEASE_PENDING     0x20
#define DEV_SUSPEND             0x40
#define DEV_BUSY                0x80

#define DEV_OK(l) \
    ((l) && ((l->state & ~DEV_BUSY) == (DEV_CONFIG|DEV_PRESENT)))
/***********************************/


/*====================================================================*/

/* Parameters that can be set with 'insmod' */
#define DEFAULT_FLASH_MAJOR 43  /* as used by old flash char driver */

/* Major device # for FTL device */
static int major_dev = DEFAULT_FLASH_MAJOR;

static int shuffle_freq = 50;

MODULE_PARM(major_dev, "i");
MODULE_PARM(shuffle_freq, "i");

/*====================================================================*/

/* Funky stuff for setting up a block device */
#define MAJOR_NR		major_dev
#define DEVICE_NAME		"ftl_itsy"
#define DEVICE_REQUEST		do_ftl_request
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#define DEVICE_NR(minor)	((minor)>>5)
#define REGION_NR(minor)	(((minor)>>3)&3)
#define PART_NR(minor)		((minor)&7)
#define MINOR_NR(dev,reg,part)	(((dev)<<5)+((reg)<<3)+(part))

#include BLK_DEV_HDR

#ifdef FTL_ITSY_DEBUG
int ftl_itsy_debug = FTL_ITSY_DEBUG;
MODULE_PARM(ftl_itsy_debug, "i");
#define DEBUG(n, args...) if (ftl_itsy_debug>(n)) printk(KERN_DEBUG args)
static char *version = 
"$Id: ftl_itsy.c,v 2.9 1999/12/23 01:27:47 kerr Exp $";
#define STATIC
#else
#define DEBUG(n, args...)
#define STATIC static
#endif

/*====================================================================*/

/* Maximum number of separate memory devices we'll allow */
#define MAX_DEV		4

/* Maximum number of regions per device */
#define MAX_REGION	4

/* Maximum number of partitions in an FTL region */
#define PART_BITS	3
#define MAX_PART	8

/* Maximum number of outstanding erase requests per socket */
#define MAX_ERASE	8

/* Sector size -- shouldn't need to change */
#define SECTOR_SIZE	512

static void ftl_config(dev_link_t *link);
static void ftl_release(u_long arg);
static int ftl_event(event_callback_args_t *args);
static dev_link_t *ftl_attach(client_handle_t);
static void ftl_detach(dev_link_t *);

/* Block IO request handler stuff */
static void do_ftl_request(void);

/* Each memory region corresponds to a minor device */
typedef struct partition_t {
    dev_node_t		dev;
    u_int		state;
    u_int		*VirtualBlockMap;
    u_int		*VirtualPageMap;
    u_int		FreeTotal;
    struct eun_info_t {
	u_int			Offset;
	u_int			EraseCount;
	u_int			Free;
	u_int			Deleted;
        u_int                   Control;
    } *EUNInfo;
    struct xfer_info_t {
	u_int			Offset;
	u_int			EraseCount;
	u_short			state;
    } *XferInfo;
    u_short		bam_index;
    u_int		*bam_cache;
    u_short		DataUnits;
    u_int		BlocksPerUnit;
    erase_unit_header_t	header;        /* ftl.h */
    region_info_t	region;
    memory_handle_t	handle;
    int			open;
    int                 open_char_read;
    int                 open_char_write;
    int			locked;
    int                 reclaim_unit;  /* next unit to possibly reclaim-ahead */
} partition_t;

/* Partition state flags */
#define FTL_FORMATTED	0x01

/* Transfer unit states */
#define XFER_UNKNOWN	0x00
#define XFER_ERASING	0x01
#define XFER_ERASED	0x02
#define XFER_PREPARED	0x03
#define XFER_FAILED	0x04

typedef struct ftl_dev_t {
    eraseq_handle_t	eraseq_handle;
    eraseq_entry_t	eraseq[MAX_ERASE];
    struct wait_queue	*erase_pending;
    partition_t		minor[MAX_REGION];
} ftl_dev_t;

static dev_link_t *dev_table[MAX_DEV] = { NULL, /* ... */ };

static struct hd_struct ftl_hd[MINOR_NR(MAX_DEV, 0, 0)];
static int ftl_sizes[MINOR_NR(MAX_DEV, 0, 0)];
static int ftl_blocksizes[MINOR_NR(MAX_DEV, 0, 0)];

static struct wait_queue *ftl_wait_open = NULL;

static void ftl_geninit(struct gendisk *);

static struct gendisk ftl_gendisk = {
    0,			/* Major device number */
    "ftl",              /* Prefix used in /dev.  At most 6 chars (!). */
    PART_BITS,		/* Bits in partition # */
    MAX_PART,		/* Partitions per device */
    MAX_DEV*MAX_REGION, /* maximum number of real (unpartitioned) devices */
    ftl_geninit,	/* init function */
    ftl_hd,		/* hd_struct */
    ftl_sizes,		/* block sizes */
    MAX_DEV*MAX_REGION,	/* actual number of real (unpartitioned) devices */
    NULL,
    NULL
};

/*====================================================================*/

static int ftl_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg);
static int ftl_open(struct inode *inode, struct file *file);
static FS_RELEASE_T ftl_close(struct inode *inode, struct file *file);
static int ftl_reread_partitions(int minor);

static struct file_operations ftl_blk_fops = {
    NULL,		/* lseek */
    block_read,		/* read */
    block_write,	/* write */
    NULL,		/* readdir */
    NULL,		/* select */
    ftl_ioctl,		/* ioctl */
    NULL,		/* mmap */
    ftl_open,		/* open */
    ftl_close,		/* release */
    block_fsync		/* fsync */
};

/*====================================================================*/

#define cs_error(func, ret) printk(DEVICE_NAME ": %s error %d\n", #func, ret)

/*======================================================================

    ftl_attach() creates an "instance" of the driver, allocating
    local data structures for one device.

======================================================================*/

STATIC dev_link_t
*ftl_attach(client_handle_t client)
{
    client_reg_t client_reg;
    dev_link_t *link;
    ftl_dev_t *dev;
    eraseq_hdr_t eraseq_hdr;
    int i, ret;
    
    DEBUG(0, DEVICE_NAME ": ftl_attach()\n");

    i = (int) client;
    
    /* Create new memory card device */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    dev_table[i] = link;
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &ftl_release;
    link->release.data = (u_long)link;
    
    dev = kmalloc(sizeof(struct ftl_dev_t), GFP_KERNEL);
    memset(dev, 0, sizeof(ftl_dev_t));
    init_waitqueue(&dev->erase_pending);
    link->priv = dev;

    /* Register with BMI */
    link->handle = client;
    client_reg.event_handler = &ftl_event;
    client_reg.event_callback_args.client_data = link;
    ret = BMIRegisterClient(client, &client_reg);
    if (ret != 0) {
	cs_error(RegisterClient, ret);
	ftl_detach(link);
	return NULL;
    }

    for (i = 0; i < MAX_ERASE; i++)
	dev->eraseq[i].State = ERASE_IDLE;
    eraseq_hdr.QueueEntryCnt = MAX_ERASE;
    eraseq_hdr.QueueEntryArray = dev->eraseq;
    dev->eraseq_handle = (void *)link->handle;
    ret = BMIRegisterEraseQueue(&dev->eraseq_handle, &eraseq_hdr);
    if (ret != 0) {
	cs_error(RegisterEraseQueue, ret);
	dev->eraseq_handle = NULL;
	ftl_detach(link);
	return NULL;
    }

    return link;
} /* ftl_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

STATIC void
ftl_detach(dev_link_t *link)
{
    ftl_dev_t *dev;
    int i;

    DEBUG(0, DEVICE_NAME ": ftl_detach(0x%p)\n", link);
    
    /* Locate device structure */
    for (i = 0; i < MAX_DEV; i++)
	if (dev_table[i] == link) break;
    if (i == MAX_DEV)
	return;

    if (link->state & DEV_CONFIG) {
	ftl_release((u_long)link);
	if (link->state & DEV_STALE_CONFIG) {
	    link->state |= DEV_STALE_LINK;
	    return;
	}
    }

    dev = (ftl_dev_t *)link->priv;
    if (dev->eraseq_handle)
	BMIDeregisterEraseQueue(dev->eraseq_handle);
    if (link->handle)
	BMIDeregisterClient(link->handle);
    
    /* Unlink device structure, free bits */
    dev_table[i] = NULL;
    kfree_s(dev, sizeof(struct ftl_dev_t));
    kfree_s(link, sizeof(struct dev_link_t));
    
} /* ftl_detach */


STATIC void
ftl_config(dev_link_t *link)
{
    ftl_dev_t *dev;
    partition_t *minor;
    region_info_t region;
    dev_node_t **tail;
    int i, ret, nr;

    DEBUG(0, DEVICE_NAME ": ftl_config(0x%p)\n", link);

    /* Configure card */
    link->state |= DEV_CONFIG;

    for (i = 0; i < MAX_DEV; i++)
	if (dev_table[i] == link) break;
    dev = (ftl_dev_t *)link->priv;
    tail = &link->dev;
    minor = dev->minor;
    nr = 0;
    region.Attributes = REGION_TYPE_CM;
    ret = BMIGetFirstRegion(link->handle, &region);
    while (ret == 0) {
	minor->region = region;
	sprintf(minor->dev.dev_name, "ftl%dc%d", i, nr);
	minor->dev.major = major_dev;
	minor->dev.minor = MINOR_NR(i, nr, 0);
	*tail = &minor->dev; tail = &minor->dev.next;
	minor++; nr++;
	ret = BMIGetNextRegion(link->handle, &region);
    }
    *tail = NULL;
    
    link->state &= ~DEV_CONFIG_PENDING;
    
    if (nr == 0)
	printk(KERN_NOTICE DEVICE_NAME ": ftl%d: no regions found!\n", i);
    else {
	printk(KERN_INFO DEVICE_NAME ": ftl%d:", i);
	minor = dev->minor;
	for (i = 0; i < nr; i++) {
	    if (minor[i].region.RegionSize & 0xfffff)
		printk(" %u kb", minor[i].region.RegionSize >> 10);
	    else
		printk(" %u mb", minor[i].region.RegionSize >> 20);
	}
	printk("\n");
    }
    
} /* ftl_config */


STATIC void
ftl_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    int i;
    
    DEBUG(0, DEVICE_NAME ": ftl_release(0x%p)\n", link);

    for (i = 0; i < MAX_DEV; i++)
	if (dev_table[i] == link) break;
    if (link->open) {
	DEBUG(1, DEVICE_NAME ": release postponed, ftl%d still open\n", i);
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    link->dev = NULL;
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
	ftl_detach(link);
    
} /* ftl_release */


STATIC void save_status(eraseq_entry_t *erase);

/* The only event that can happen is erase complete */
STATIC int
ftl_event(event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    ftl_dev_t *dev;

    DEBUG(1, DEVICE_NAME ": ftl_event()\n");
    
    save_status((eraseq_entry_t *)(args->info));
    dev = (ftl_dev_t *)link->priv;
    wake_up(&dev->erase_pending);
#if !CANSLEEP
    do_ftl_request();
#endif

    return 0;
} /* ftl_event */


/*======================================================================

    Scan_header() checks to see if a memory region contains an FTL
    partition.  build_maps() reads all the erase unit headers, builds
    the erase unit map, and then builds the virtual page map.
    
======================================================================*/

STATIC int
scan_header(partition_t *part)
{
    erase_unit_header_t header;
    mem_op_t req;
    int ret;
    

    /* Search first megabyte for a valid FTL header */
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = sizeof(header);
    for (req.Offset = 0;
	 req.Offset < 0x100000;
	 req.Offset += part->region.BlockSize) {
	ret = BMIReadMemory(part->handle, &req, (caddr_t)&header);
	if (ret != 0) {
	    cs_error(ReadMemory, ret);
	    return ret;
	}
	if (strcmp(header.DataOrgTuple+3, "FTL100") == 0) break;
    }
    if (req.Offset == 0x100000) {
	printk(KERN_NOTICE DEVICE_NAME
	       ": FTL header not found. (minor %d)\n", part->dev.minor);
	return -EIO;
    }
    if ((header.NumEraseUnits > 65536) || (header.BlockSize != 9) ||
	(header.EraseUnitSize < 10) || (header.EraseUnitSize > 31) ||
	(header.NumTransferUnits >= header.NumEraseUnits)) {
	printk(KERN_NOTICE DEVICE_NAME
	       ": FTL header corrupt! (minor %d)\n", part->dev.minor);
	return -EIO;
    }
    if (header.NumEraseUnits !=
	part->region.RegionSize/part->region.BlockSize) {
      printk(KERN_NOTICE DEVICE_NAME
	     ": FTL wrong region size.  (minor %d)\n", part->dev.minor);
      return -EIO;
    }
    part->header = header;
    return 0;
}

STATIC int
build_maps(partition_t *part)
{
    erase_unit_header_t header;
    mem_op_t req;
    u_short xvalid, xtrans, i;
    u_int blocks, j;
    int hdr_ok, ret;

    /* Set up erase unit maps */
    part->DataUnits = part->header.NumEraseUnits -
	part->header.NumTransferUnits;
    part->EUNInfo = kmalloc(part->DataUnits * sizeof(struct eun_info_t),
			    GFP_KERNEL);
    for (i = 0; i < part->DataUnits; i++)
	part->EUNInfo[i].Offset = 0xffffffff;
    part->XferInfo =
	kmalloc(part->header.NumTransferUnits * sizeof(struct xfer_info_t),
		GFP_KERNEL);

    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = sizeof(header);
    xvalid = xtrans = 0;
    for (i = 0; i < part->header.NumEraseUnits; i++) {
	req.Offset = ((i + part->header.FirstPhysicalEUN)
		      << part->header.EraseUnitSize);
	ret = BMIReadMemory(part->handle, &req, (caddr_t)&header);
	if (ret != 0) {
	    cs_error(ReadMemory, ret);
	    return ret;
	}
	/* Is this a transfer partition? */
	hdr_ok = (strcmp(header.DataOrgTuple+3, "FTL100") == 0);
	if (hdr_ok && (header.LogicalEUN < part->DataUnits) &&
	    (part->EUNInfo[header.LogicalEUN].Offset == 0xffffffff)) {
	    part->EUNInfo[header.LogicalEUN].Offset = req.Offset;
	    part->EUNInfo[header.LogicalEUN].EraseCount =
		header.EraseCount;
	    xvalid++;
	} else {
	    if (xtrans == part->header.NumTransferUnits) {
		printk(KERN_NOTICE DEVICE_NAME ": format error: too many "
		       "transfer units!\n");
		return -1;
	    }
	    if (hdr_ok && (header.LogicalEUN == 0xffff)) {
		part->XferInfo[xtrans].state = XFER_PREPARED;
		part->XferInfo[xtrans].EraseCount = header.EraseCount;
	    } else {
		part->XferInfo[xtrans].state = XFER_UNKNOWN;
		/* Pick anything reasonable for the erase count */
		part->XferInfo[xtrans].EraseCount =
		    part->header.EraseCount;
	    }
	    part->XferInfo[xtrans].Offset = req.Offset;
	    xtrans++;
	}
    }
    /* Check for format trouble */
    header = part->header;
    if ((xtrans != header.NumTransferUnits) ||
	(xvalid+xtrans != header.NumEraseUnits)) {
	printk(KERN_NOTICE DEVICE_NAME ": format error: erase units "
	       "don't add up!\n");
	return -1;
    }
    
    /* Set up virtual page map */
    blocks = header.FormattedSize >> header.BlockSize;
    part->VirtualBlockMap = vmalloc(blocks * sizeof(u_int));
    memset(part->VirtualBlockMap, 0xff, blocks * sizeof(u_int));
    part->BlocksPerUnit = (1 << header.EraseUnitSize) >> header.BlockSize;
    req.Count = part->BlocksPerUnit * sizeof(u_int);
    part->reclaim_unit = 0;

    part->bam_cache = kmalloc(part->BlocksPerUnit * sizeof(u_int),
			      GFP_KERNEL);
    part->bam_index = 0xffff;
    part->FreeTotal = 0;
    for (i = 0; i < part->DataUnits; i++) {
	part->EUNInfo[i].Free = 0;
	part->EUNInfo[i].Deleted = 0;
	part->EUNInfo[i].Control = 0;
	req.Offset = part->EUNInfo[i].Offset + header.BAMOffset;
	ret = BMIReadMemory(part->handle, &req, (caddr_t)part->bam_cache);
	if (ret != 0) {
	    cs_error(ReadMemory, ret);
	    return ret;
	}
	for (j = 0; j < part->BlocksPerUnit; j++) {
	    if (BLOCK_FREE(part->bam_cache[j])) {
		part->EUNInfo[i].Free++;
		part->FreeTotal++;
	    } else if ((BLOCK_TYPE(part->bam_cache[j]) == BLOCK_DATA) &&
		     (BLOCK_NUMBER(part->bam_cache[j]) < blocks))
		part->VirtualBlockMap[BLOCK_NUMBER(part->bam_cache[j])] =
		    (i << header.EraseUnitSize) + (j << header.BlockSize);
	    else if (BLOCK_DELETED(part->bam_cache[j]))
		part->EUNInfo[i].Deleted++;
	    else if (BLOCK_TYPE(part->bam_cache[j]) == BLOCK_CONTROL)
		part->EUNInfo[i].Control++;
	}
    }
    
    return 0;
    
} /* build_maps */

/*======================================================================

    Erase_xfer() schedules an asynchronous erase operation for a
    transfer unit.
    
======================================================================*/

STATIC int
erase_xfer(ftl_dev_t *dev, partition_t *part,
		      u_short xfernum)
{
    int i, ret;
    struct xfer_info_t *xfer;
    u_short oldstate;

    xfer = &part->XferInfo[xfernum];
    DEBUG(1, DEVICE_NAME ": erasing xfer unit at 0x%x\n", xfer->Offset);
    oldstate = xfer->state;
    xfer->state = XFER_ERASING;
    /* Is there a free erase slot? */
    for (;;) {
	for (i = 0; i < MAX_ERASE; i++)
	    if (!ERASE_IN_PROGRESS(dev->eraseq[i].State)) break;
	if (i < MAX_ERASE) break;
#if CANSLEEP
	DEBUG(0, DEVICE_NAME ": erase queue is full; waiting...\n");
	sleep_on(&dev->erase_pending);
#else
	DEBUG(0, DEVICE_NAME ": erase queue is full; deferring request...\n");
	xfer->state = oldstate;
	return -EWOULDBLOCK;
#endif
    }

    /* Queue the request */
    dev->eraseq[i].State = ERASE_QUEUED;
    dev->eraseq[i].Handle = part->handle;
    dev->eraseq[i].Offset = xfer->Offset;
    dev->eraseq[i].Size = part->region.BlockSize;
    dev->eraseq[i].Optional = part;
    ret = BMICheckEraseQueue(dev->eraseq_handle);
    if (ret != 0) {
	cs_error(CheckEraseQueue, ret);
	return ret;
    }
    xfer->EraseCount++;
    return ret;
} /* erase_xfer */

/*======================================================================

    Prepare_xfer() takes a freshly erased transfer unit and gives
    it an appropriate header.
    
======================================================================*/

STATIC void
save_status(eraseq_entry_t *erase)
{
    partition_t *part;
    struct xfer_info_t *xfer;
    int i;
    
    /* Look up the transfer unit */
    part = (partition_t *)(erase->Optional);
    for (i = 0; i < part->header.NumTransferUnits; i++)
	if (part->XferInfo[i].Offset == erase->Offset) break;
    if (i == part->header.NumTransferUnits) {
	printk(KERN_NOTICE DEVICE_NAME ": internal error: "
	       "erase lookup failed!\n");
	return;
    }
    xfer = &part->XferInfo[i];
    if (erase->State == ERASE_PASSED)
	xfer->state = XFER_ERASED;
    else {
	xfer->state = XFER_FAILED;
	printk(KERN_NOTICE DEVICE_NAME ": erase failed: state = %d\n",
	       erase->State);
    }
}

STATIC void
prepare_xfer(partition_t *part, int i)
{
    erase_unit_header_t header;
    mem_op_t req;
    struct xfer_info_t *xfer;
    int nbam, ret;
    u_int ctl;

    xfer = &part->XferInfo[i];
    xfer->state = XFER_FAILED;
    
    DEBUG(1, DEVICE_NAME ": preparing xfer unit at 0x%x\n", xfer->Offset);

    /* Write the transfer unit header */
    header = part->header;
    header.LogicalEUN = 0xffff;
    header.EraseCount = xfer->EraseCount;
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = sizeof(header);
    req.Offset = xfer->Offset;
    ret = BMIWriteMemory(part->handle, &req, (caddr_t)&header);
    if (ret != 0) {
	cs_error(WriteMemory, ret);
	return;
    }

    /* Write the BAM stub */
    nbam = (part->BlocksPerUnit * sizeof(u_int) +
	    part->header.BAMOffset + SECTOR_SIZE - 1) / SECTOR_SIZE;
    req.Offset = xfer->Offset + part->header.BAMOffset;
    req.Count = sizeof(u_int);
    ctl = BLOCK_CONTROL;
    for (i = 0; i < nbam; i++, req.Offset += sizeof(u_int)) {
	ret = BMIWriteMemory(part->handle, &req, (caddr_t)&ctl);
	if (ret != 0) {
	    cs_error(WriteMemory, ret);
	    return;
	}
    }
    xfer->state = XFER_PREPARED;
    
} /* prepare_xfer */

/*======================================================================

    Copy_erase_unit() takes a full erase block and a transfer unit,
    copies everything to the transfer unit, then swaps the block
    pointers.

    All data blocks are copied to the corresponding blocks in the
    target unit, so the virtual block map does not need to be
    updated.
    
======================================================================*/

STATIC int
copy_erase_unit(partition_t *part, u_short srcunit, u_short xferunit)
{
    u_char buf[SECTOR_SIZE];
    struct eun_info_t *eun;
    struct xfer_info_t *xfer;
    mem_op_t req;
    u_int src, dest, free, i;
    u_short unit;
    int ret;

    eun = &part->EUNInfo[srcunit];
    xfer = &part->XferInfo[xferunit];
    DEBUG(2, DEVICE_NAME ": copying block 0x%x to 0x%x\n",
	  eun->Offset, xfer->Offset);
	
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    
    /* Read current BAM */
    if (part->bam_index != srcunit) {
	req.Offset = eun->Offset + part->header.BAMOffset;
	req.Count = part->BlocksPerUnit * sizeof(u_int);
	ret = BMIReadMemory(part->handle, &req, (caddr_t)part->bam_cache);
	/* mark the cache bad, in case we get an error later */
	part->bam_index = 0xffff;
	if (ret != 0) goto read_error;
    }
    
    /* Write the LogicalEUN for the transfer unit */
    xfer->state = XFER_UNKNOWN;
    req.Count = sizeof(u_short);
    req.Offset = xfer->Offset + 20; /* Bad! */
    unit = 0x7fff;
    ret = BMIWriteMemory(part->handle, &req, (caddr_t)&unit);
    if (ret != 0) goto write_error;
    
    /* Copy all data blocks from source unit to transfer unit */
    src = eun->Offset; dest = xfer->Offset;
    req.Count = SECTOR_SIZE;
    free = 0;
    ret = 0;
    for (i = 0; i < part->BlocksPerUnit; i++) {
	switch (BLOCK_TYPE(part->bam_cache[i])) {
	case BLOCK_CONTROL:
	    /* This gets updated later */
	    break;
	case BLOCK_DATA:
	case BLOCK_REPLACEMENT:
	    req.Offset = src;
	    ret = BMIReadMemory(part->handle, &req, buf);
	    if (ret != 0) goto read_error;
	    req.Offset = dest;
	    ret = BMIWriteMemory(part->handle, &req, buf);
	    if (ret != 0) goto write_error;
	    break;
	default:
	    /* All other blocks must be free */
	    part->bam_cache[i] = 0xffffffff;
	    free++;
	    break;
	}
	src += SECTOR_SIZE;
	dest += SECTOR_SIZE;
    }

    /* Write the BAM to the transfer unit */
    req.Offset = xfer->Offset + part->header.BAMOffset;
#ifndef BUG
    /* This code was missing in ftl_cs.c 3.0.5 */
    req.Count = part->BlocksPerUnit * sizeof(u_int);
#endif
    ret = BMIWriteMemory(part->handle, &req, (caddr_t)part->bam_cache);
    if (ret != 0) goto write_error;
    
    /* All clear? Then update the LogicalEUN again */
    req.Offset = xfer->Offset + 20; /* Bad! */
    req.Count = sizeof(u_short);
    ret = BMIWriteMemory(part->handle, &req, (caddr_t)&srcunit);
    if (ret != 0) goto write_error;

    /* Update the maps and usage stats*/
    i = xfer->EraseCount;
    xfer->EraseCount = eun->EraseCount;
    eun->EraseCount = i;
    i = xfer->Offset;
    xfer->Offset = eun->Offset;
    eun->Offset = i;
    part->FreeTotal -= eun->Free;
    part->FreeTotal += free;
    eun->Free = free;
    eun->Deleted = 0;
    
    /* Now, the cache should be valid for the new block */
    part->bam_index = srcunit;

    return 0;
    
read_error:
    cs_error(ReadMemory, ret);
    return ret;
    
write_error:
    cs_error(WriteMemory, ret);
    return ret;
} /* copy_erase_unit */

/*======================================================================

    reclaim_block() picks a full erase unit and a transfer unit and
    then calls copy_erase_unit() to copy one to the other.  Then, it
    schedules an erase on the expired block.  (If CANSLEEP is true,
    then if no transfer unit is currently available, the function
    returns immediately if ndelay is true or sleeps if it is false.
    If CANSLEEP is false, the function immediately returns
    -EWOULDBLOCK in either case.)

    What's a good way to decide which transfer unit and which erase
    unit to use?  Beats me.  My way is to always pick the transfer
    unit with the fewest erases, and usually pick the data unit with
    the most deleted blocks.  But with a small probability, pick the
    oldest data unit instead.  This means that we generally postpone
    the next reclaimation as long as possible, but shuffle static
    stuff around a bit for wear leveling.
    
======================================================================*/

STATIC int
reclaim_block(ftl_dev_t *dev, partition_t *part, int ndelay)
{
    u_short i, eun, xfer;
    u_int best;
    int queued, ret;

    /* Pick the least erased transfer unit */
    best = 0xffffffff; xfer = 0xffff;
    do {
	queued = 0;
	for (i = 0; i < part->header.NumTransferUnits; i++) {
  	    if (part->XferInfo[i].state == XFER_UNKNOWN) {
		ret = erase_xfer(dev, part, i);
		if (ret != 0) return ret;
	    }
	    if (part->XferInfo[i].state == XFER_ERASING)
		queued = 1;
	    else if (part->XferInfo[i].state == XFER_ERASED)
		prepare_xfer(part, i);
	    if ((part->XferInfo[i].state == XFER_PREPARED) &&
		(part->XferInfo[i].EraseCount <= best)) {
		    best = part->XferInfo[i].EraseCount;
		    xfer = i;
		}
	}
	if (xfer == 0xffff) {
	    if (queued) {
	        if (ndelay) {
		  return -EWOULDBLOCK;
		} else {
#if CANSLEEP
		  DEBUG(1, DEVICE_NAME ": reclaim waiting for transfer "
			"unit to be prepared...\n");
		  sleep_on(&dev->erase_pending);
#else
		  DEBUG(1, DEVICE_NAME ": reclaim deferred because no transfer "
			"unit is prepared...\n");
		  return -EWOULDBLOCK;
#endif
		}
	    } else {
		static int ne = 0;
		if (++ne < 5)
		    printk(KERN_NOTICE DEVICE_NAME ": reclaim failed: no "
			   "suitable transfer units!\n");
		return -ENOSPC;
	    }
	}
    } while (xfer == 0xffff);

    DEBUG(0, DEVICE_NAME ": reclaiming space...\n");
	
    eun = 0;
    if ((jiffies % shuffle_freq) == 0) {
	DEBUG(1, DEVICE_NAME ": recycling freshest block...\n");
	best = 0xffffffff;
	for (i = 0; i < part->DataUnits; i++)
	    if (part->EUNInfo[i].EraseCount <= best) {
		best = part->EUNInfo[i].EraseCount;
		eun = i;
	    }
    } else {
	best = 0;
	for (i = 0; i < part->DataUnits; i++)
	    if (part->EUNInfo[i].Deleted >= best) {
		best = part->EUNInfo[i].Deleted;
		eun = i;
	    }
	if (best == 0) {
	    static int ne = 0;
	    if (++ne < 5)
		printk(KERN_NOTICE DEVICE_NAME ": reclaim failed: "
		       "no free blocks!\n");
	    return -ENOSPC;
	}
    }
    ret = copy_erase_unit(part, eun, xfer);
    if (ret == 0) {
        /* If this returns -EWOULDBLOCK (unlikely), we'll get it later */
        (void) erase_xfer(dev, part, xfer);
    } else {
	printk(KERN_NOTICE DEVICE_NAME ": copy_erase_unit failed! (%d)\n", ret);
    }
    return ret;
} /* reclaim_block */

/*======================================================================

    Find_free() searches for a free block.  If necessary, it updates
    the BAM cache for the erase unit containing the free block.  It
    returns the block index -- the erase unit is just the currently
    cached unit.  If there are no free blocks, it returns 0 -- this
    is never a valid data block because it contains the header.
    
======================================================================*/

#ifdef PSYCHO_DEBUG
STATIC void
dump_lists(partition_t *part)
{
    int i;
    printk(KERN_DEBUG DEVICE_NAME ": Free total = %d\n", part->FreeTotal);
    for (i = 0; i < part->DataUnits; i++)
	printk(KERN_DEBUG DEVICE_NAME ":   unit %d: %d phys, %d free, "
	       "%d deleted\n", i,
	       part->EUNInfo[i].Offset >> part->header.EraseUnitSize,
	       part->EUNInfo[i].Free, part->EUNInfo[i].Deleted);
}
#endif

STATIC u_int
find_free(partition_t *part)
{
    u_short stop, eun;
    u_int blk;
    mem_op_t req;
    int ret;
    
    /* Find an erase unit with some free space */
    stop = (part->bam_index == 0xffff) ? 0 : part->bam_index;
    eun = stop;
    do {
	if (part->EUNInfo[eun].Free != 0) break;
	/* Wrap around at end of table */
	if (++eun == part->DataUnits) eun = 0;
    } while (eun != stop);

    if (part->EUNInfo[eun].Free == 0)
	return 0;
    
    /* Is this unit's BAM cached? */
    if (eun != part->bam_index) {
	/* Invalidate cache */
	part->bam_index = 0xffff;
	req.Attributes = MEM_OP_BUFFER_KERNEL;
	req.Count = part->BlocksPerUnit * sizeof(u_int);
	req.Offset = part->EUNInfo[eun].Offset + part->header.BAMOffset;
	ret = BMIReadMemory(part->handle, &req, (caddr_t)part->bam_cache);
	if (ret != 0) {
	    cs_error(ReadMemory, ret);
	    return 0;
	}
	part->bam_index = eun;
    }

    /* Find a free block */
    for (blk = 0; blk < part->BlocksPerUnit; blk++)
	if (BLOCK_FREE(part->bam_cache[blk])) break;
    if (blk == part->BlocksPerUnit) {
#ifdef PSYCHO_DEBUG
	static int ne = 0;
	if (++ne == 1)
	    dump_lists(part);
#endif
	printk(KERN_NOTICE DEVICE_NAME ": bad free list!\n");
	return 0;
    }
    DEBUG(3, DEVICE_NAME ": found free block at %d in %d\n", blk, eun);
    return blk;
    
} /* find_free */

/*======================================================================

    This gets a memory handle for the region corresponding to the
    minor device number.
    
======================================================================*/

STATIC int
ftl_open(struct inode *inode, struct file *file)
{
    int minor = MINOR(inode->i_rdev);
    dev_link_t *link;
    ftl_dev_t *dev;
    partition_t *partition;
    open_mem_t open;
    int ret;
    
    DEBUG(0, DEVICE_NAME ": ftl_open(%d)\n", minor);

    link = dev_table[DEVICE_NR(minor)];
    if (!DEV_OK(link))
	return -ENODEV;
    
    dev = (ftl_dev_t *)link->priv;
    partition = &dev->minor[REGION_NR(minor)];
    if (partition->region.RegionSize == 0)
	return -ENODEV;
    while (partition->locked) {
        DEBUG(0, DEVICE_NAME ": waiting on locked partition...\n");
	sleep_on(&ftl_wait_open);
    }
    
    /* prevent open as ftl while open for writing as raw flash */
    if (partition->open_char_write) return -EBUSY;

    if (partition->handle == NULL) {
	partition->handle = (memory_handle_t)link->handle;
	open.Attributes = partition->region.Attributes;
	open.Offset = partition->region.CardOffset;
	ret = BMIOpenMemory(&partition->handle, &open);
	if (ret != 0) {
	    cs_error(OpenMemory, ret);
	    return ret;
	}
	if ((scan_header(partition) == 0) &&
	    (build_maps(partition) == 0)) {
	    partition->state = FTL_FORMATTED;
	    ftl_reread_partitions(minor);
#ifdef FTL_ITSY_DEBUG
	    printk(KERN_INFO DEVICE_NAME
		   ": opening %d kb FTL partition (minor %d)\n",
		   partition->header.FormattedSize >> 10,
		   partition->dev.minor);
#endif
	} else {
	    BMICloseMemory(partition->handle);
	    printk(KERN_NOTICE DEVICE_NAME
		   ": FTL partition is invalid. (minor %d)\n",
		   partition->dev.minor);
	    partition->handle = NULL;
	    return -ENODEV;
	}
    }
    
    partition->open++;
    link->open++;
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    return 0;
} /* ftl_open */

/*====================================================================*/

STATIC FS_RELEASE_T
ftl_close(struct inode *inode, struct file *file)
{
    dev_link_t *link;
    int minor = MINOR(inode->i_rdev);
    ftl_dev_t *dev;
    partition_t *part;
    int i;
    
    DEBUG(0, DEVICE_NAME ": ftl_close(%d)\n", minor);

    /* Flush all writes */
    fsync_dev(inode->i_rdev);
    INVALIDATE_INODES(inode->i_rdev);
    invalidate_buffers(inode->i_rdev);
    
    link = dev_table[DEVICE_NR(minor)];
    dev = (ftl_dev_t *)link->priv;
    part = &dev->minor[REGION_NR(minor)];
    
    /* Wait for any pending erase operations to complete */
    for (i = 0; i < part->header.NumTransferUnits; i++) {
        if (part->XferInfo[i].state == XFER_ERASING) {
	    DEBUG(0, DEVICE_NAME ": waiting for erase(s) to complete...\n");
	    sleep_on(&dev->erase_pending);
	}
	if (part->XferInfo[i].state == XFER_ERASED)
	    prepare_xfer(part, i);
    }
    
    link->open--;
    part->open--;
    if (part->open == 0) {
	BMICloseMemory(part->handle);
	part->handle = NULL;
	if (part->VirtualBlockMap) {
	    vfree(part->VirtualBlockMap);
	    part->VirtualBlockMap = NULL;
	}
	if (part->VirtualPageMap) {
	    kfree(part->VirtualPageMap);
	    part->VirtualPageMap = NULL;
	}
	if (part->EUNInfo) {
	    kfree(part->EUNInfo);
	    part->EUNInfo = NULL;
	}
	if (part->XferInfo) {
	    kfree(part->XferInfo);
	    part->XferInfo = NULL;
	}
	if (part->bam_cache) {
	    kfree(part->bam_cache);
	    part->bam_cache = NULL;
	}
    }
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    return (FS_RELEASE_T)0;
} /* ftl_close */

/*======================================================================

    Read a series of sectors from an FTL partition.
    
======================================================================*/

STATIC int
ftl_read(partition_t *part, caddr_t buffer, u_long sector, u_long nblocks)
{
    mem_op_t req;
    u_int log_addr, bsize;
    u_long i;
    int ret;
    
    DEBUG(2, DEVICE_NAME ": ftl_read(0x%p, 0x%lx, %ld)\n",
	  part->handle, sector, nblocks);
    if (!(part->state & FTL_FORMATTED)) {
	printk(KERN_NOTICE DEVICE_NAME ": bad partition\n");
	return -EIO;
    }
    bsize = part->region.BlockSize;
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = SECTOR_SIZE;
    for (i = 0; i < nblocks; i++) {
	if (((sector+i) * SECTOR_SIZE) >= part->header.FormattedSize) {
	    printk(KERN_NOTICE DEVICE_NAME ": bad read offset\n");
	    return -EIO;
	}
	log_addr = part->VirtualBlockMap[sector+i];
	if (log_addr == 0xffffffff)
	    memset(buffer, 0, SECTOR_SIZE);
	else {
	    req.Offset = (part->EUNInfo[log_addr / bsize].Offset
			  + (log_addr % bsize));
	    ret = BMIReadMemory(part->handle, &req, buffer);
	    if (ret != 0) {
		cs_error(ReadMemory, ret);
		return ret;
	    }
	}
	buffer += SECTOR_SIZE;
    }
    return 0;
} /* ftl_read */

/*======================================================================

    Write a series of sectors to an FTL partition
    
======================================================================*/

STATIC int
set_bam_entry(partition_t *part, u_int log_addr, u_int virt_addr)
{
    mem_op_t req;
    u_int bsize, blk;
#ifdef PSYCHO_DEBUG
    u_int old_addr;
#endif
    u_short eun;
    int ret;
    
    DEBUG(3, DEVICE_NAME ": set_bam_entry(0x%p, 0x%x, 0x%x)\n",
	  part->handle, log_addr, virt_addr);
    bsize = part->region.BlockSize;
    eun = log_addr / bsize;
    blk = (log_addr % bsize) / SECTOR_SIZE;
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = sizeof(u_int);
    req.Offset = (part->EUNInfo[eun].Offset + blk * sizeof(u_int) +
		  part->header.BAMOffset);
    
#ifdef PSYCHO_DEBUG
    BMIReadMemory(part->handle, &req, (caddr_t)&old_addr);
    if (((virt_addr == 0xfffffffe) && !BLOCK_FREE(old_addr)) ||
	((virt_addr == 0) && (BLOCK_TYPE(old_addr) != BLOCK_DATA)) ||
	(!BLOCK_DELETED(virt_addr) && (old_addr != 0xfffffffe))) {
	static int ne = 0;
	if (++ne < 5) {
	    printk(KERN_NOTICE DEVICE_NAME ": set_bam_entry() inconsistency!\n");
	    printk(KERN_NOTICE DEVICE_NAME ":   log_addr = 0x%x, old = 0x%x"
		   ", new = 0x%x\n", log_addr, old_addr, virt_addr);
	}
	return -EIO;
    }
#endif
    if (part->bam_index == eun) {
#ifdef PSYCHO_DEBUG
	if (part->bam_cache[blk] != old_addr) {
	    static int ne = 0;
	    if (++ne < 5) {
		printk(KERN_NOTICE DEVICE_NAME ": set_bam_entry() "
		       "inconsistency!\n");
		printk(KERN_NOTICE DEVICE_NAME ":   log_addr = 0x%x, cache"
		       /*" = 0x%x, card = 0x%x\n",  ??what was card??*/
		       " = 0x%x\n",
		       part->bam_cache[blk], old_addr);
	    }
	    return -EIO;
	}
#endif
	part->bam_cache[blk] = virt_addr;
    }

    ret = BMIWriteMemory(part->handle, &req, (caddr_t)&virt_addr);
    if (ret != 0) {
	printk(KERN_NOTICE DEVICE_NAME ": set_bam_entry() failed!\n");
	printk(KERN_NOTICE DEVICE_NAME ":   log_addr = 0x%x, new = 0x%x\n",
	       log_addr, virt_addr);
	cs_error(WriteMemory, ret);
    }
    return ret;
} /* set_bam_entry */

STATIC int
ftl_write(ftl_dev_t *dev, partition_t *part, caddr_t buffer,
	  u_long sector, u_long nblocks)
{
    mem_op_t req;
    u_int bsize, log_addr, virt_addr, old_addr, blk;
    u_long i;
    int ret;

    DEBUG(2, DEVICE_NAME ": ftl_write(0x%p, 0x%lx, %ld)\n",
	  part->handle, sector, nblocks);
    if (!(part->state & FTL_FORMATTED)) {
	printk(KERN_NOTICE DEVICE_NAME ": bad partition\n");
	return -EIO;
    }
    /* !!check for read-only bank? */

    /* See if we need to reclaim space, before we start */
    while (part->FreeTotal < nblocks) {
      ret = reclaim_block(dev, part, 0);
      if (ret != 0) return ret;
    }
    if (part->EUNInfo[part->reclaim_unit].Deleted >=
	part->BlocksPerUnit - part->EUNInfo[part->reclaim_unit].Control) {
      /* Found a unit with all blocks deleted; call reclaim an extra
         time in ndelay mode to (probably) reclaim it. */
      ret = reclaim_block(dev, part, 1);
      if (ret != 0 && ret != -EWOULDBLOCK) return ret;
    }
    part->reclaim_unit = (part->reclaim_unit + 1) % part->DataUnits;
    
    bsize = part->region.BlockSize;
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = SECTOR_SIZE;
    virt_addr = sector * SECTOR_SIZE | BLOCK_DATA;
    for (i = 0; i < nblocks; i++) {
	if (virt_addr >= part->header.FormattedSize) {
	    printk(KERN_NOTICE DEVICE_NAME ": bad write offset\n");
	    return -EIO;
	}

	/* Grab a free block */
	blk = find_free(part);
	if (blk == 0) {
	    static int ne = 0;
	    if (++ne < 5)
		printk(KERN_NOTICE DEVICE_NAME ": internal error: "
		       "no free blocks!\n");
	    return -ENOSPC;
	}

	/* Tag the BAM entry, and write the new block */
	log_addr = part->bam_index * bsize + blk * SECTOR_SIZE;
	part->EUNInfo[part->bam_index].Free--;
	part->FreeTotal--;
	if (set_bam_entry(part, log_addr, 0xfffffffe)) {
#if PSYCHO_DEBUG
	    static int ne = 0;
	    if (++ne < 5) {
		printk(KERN_NOTICE DEVICE_NAME ":   virt_addr = 0x%x\n",
		       virt_addr);
	    }
#endif	    
	    return -EIO;
	}
	part->EUNInfo[part->bam_index].Deleted++;
	req.Offset = (part->EUNInfo[part->bam_index].Offset +
		      blk * SECTOR_SIZE);
	ret = BMIWriteMemory(part->handle, &req, buffer);
	if (ret != 0) {
	    printk(KERN_NOTICE DEVICE_NAME ": block write failed!\n");
	    printk(KERN_NOTICE DEVICE_NAME ":   log_addr = 0x%x, virt_addr"
		   " = 0x%x, Offset = 0x%x\n", log_addr, virt_addr,
		   req.Offset);
	    cs_error(WriteMemory, ret);
	    return ret;
	}
	
	/* Only delete the old entry when the new entry is ready */
	old_addr = part->VirtualBlockMap[sector+i];
	if (old_addr != 0xffffffff) {
	    part->VirtualBlockMap[sector+i] = 0xffffffff;
	    part->EUNInfo[old_addr/bsize].Deleted++;
	    if (set_bam_entry(part, old_addr, 0) != 0) {
#if PSYCHO_DEBUG
	    static int ne = 0;
	    if (++ne < 5) {
		printk(KERN_NOTICE DEVICE_NAME ":   virt_addr = 0x%x\n",
		       virt_addr);
	    }
#endif	    
		return -EIO;
	    }
	}

	/* Finally, set up the new pointers */
	if (set_bam_entry(part, log_addr, virt_addr))
	    return -EIO;
	part->VirtualBlockMap[sector+i] = log_addr;
	part->EUNInfo[part->bam_index].Deleted--;
	
	buffer += SECTOR_SIZE;
	virt_addr += SECTOR_SIZE;
    }
    return 0;
} /* ftl_write */


#ifdef DELETEU
STATIC int
ftl_delete(ftl_dev_t *dev, partition_t *part, u_long sector, u_long nblocks)
{
    mem_op_t req;
    u_int bsize, virt_addr, old_addr;
    u_long i;

    DEBUG(2, DEVICE_NAME ": ftl_delete(0x%p, 0x%lx, %ld)\n",
	  part->handle, sector, nblocks);
    if (!(part->state & FTL_FORMATTED)) {
	printk(KERN_NOTICE DEVICE_NAME ": bad partition\n");
	return -EIO;
    }
    
    bsize = part->region.BlockSize;
    req.Attributes = MEM_OP_BUFFER_KERNEL;
    req.Count = SECTOR_SIZE;
    virt_addr = sector * SECTOR_SIZE | BLOCK_DATA;
    for (i = 0; i < nblocks; i++) {
	if (virt_addr >= part->header.FormattedSize) {
	    printk(KERN_NOTICE DEVICE_NAME ": bad write offset\n");
	    return -EIO;
	}

	/* Delete the old entry */
	old_addr = part->VirtualBlockMap[sector+i];
	if (old_addr != 0xffffffff) {
	    part->VirtualBlockMap[sector+i] = 0xffffffff;
	    if (set_bam_entry(part, old_addr, 0) != 0) {
#if PSYCHO_DEBUG
	      static int ne = 0;
	      if (++ne < 5) {
		printk(KERN_NOTICE DEVICE_NAME ":   virt_addr = 0x%x\n",
		       virt_addr);
	      }
#endif	    
		return -EIO;
	    }
	    if (++part->EUNInfo[old_addr/bsize].Deleted >=
		part->BlocksPerUnit - part->EUNInfo[old_addr/bsize].Control) {
	        /* Optimization: call reclaim ahead of time if we just
		   deleted the last block in an erase unit; this will
		   (usually) erase those blocks and make them ready
		   for reuse on the next write.  (It won't do that if
		   there is no transfer unit currently preparted, or if
		   reclaim happens to pick the least-erased unit instead
		   of the one with the most deleted blocks.) */
	        int ret = reclaim_block(dev, part, 1);
		if (ret == -EWOULDBLOCK) ret = 0;  /* couldn't reclaim now */
		if (ret != 0) return ret;
	    }
	}

	virt_addr += SECTOR_SIZE;
    }
    return 0;
} /* ftl_delete */
#endif

/*======================================================================

    IOCTL calls for getting device parameters.

======================================================================*/

STATIC int
ftl_ioctl(struct inode *inode, struct file *file,
	  u_int cmd, u_long arg)
{
    struct hd_geometry *geo = (struct hd_geometry *)arg;
    struct blk_delete_args *bda = (struct blk_delete_args *)arg;
    int ret = 0, minor = MINOR(inode->i_rdev);
    dev_link_t *link;
    ftl_dev_t *dev;
    partition_t *part;
    u_long sect;

    link = dev_table[DEVICE_NR(minor)];
    if (!DEV_OK(link)) return -ENODEV;
    dev = (ftl_dev_t *)link->priv;
    part = &dev->minor[REGION_NR(minor)];

    DEBUG(0, DEVICE_NAME ": ioctl 0x%x\n", cmd);

    switch (cmd) {
    case HDIO_GETGEO:
	ret = verify_area(VERIFY_WRITE, (long *)arg, sizeof(*geo));
	if (ret) return ret;
	/* Sort of arbitrary: round size down to 4K boundary */
	sect = part->header.FormattedSize/SECTOR_SIZE;
	put_user(1, (char *)&geo->heads);
	put_user(8, (char *)&geo->sectors);
	put_user((sect>>3), (short *)&geo->cylinders);
	put_user(ftl_hd[minor].start_sect, (u_long *)&geo->start);
	break;
    case BLKGETSIZE:
	ret = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long));
	if (ret) return ret;
	put_user(part->header.FormattedSize/SECTOR_SIZE,
		 (long *)arg);
	break;
    case BLKFLSBUF:
	if (!suser()) return -EACCES;
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);
	break;
    case BLKRRPART:
	ret = ftl_reread_partitions(minor);
	break;
    case BLKDELETEU:
    case BLKDELETEZ:
        ret = verify_area(VERIFY_READ, (long *)arg, _IOC_NR(cmd));
	if (ret) return ret;
	if (bda->count == 0) return 0;
	if (bda->start % ftl_blocksizes[minor] ||
	    bda->count % ftl_blocksizes[minor]) {
	  return -EINVAL;
	}
        ll_delete(cmd, inode->i_rdev, bda->start % ftl_blocksizes[minor],
		  bda->start % ftl_blocksizes[minor]);
	break;
    RO_IOCTLS(inode->i_rdev, arg);
    default:
	ret = -EINVAL;
    }

    return ret;
} /* ftl_ioctl */

/*======================================================================

    Handler for block device requests

======================================================================*/

STATIC int
ftl_reread_partitions(int minor)
{
    int d = DEVICE_NR(minor), r = REGION_NR(minor);
    ftl_dev_t *dev = dev_table[d]->priv;
    partition_t *part = &(dev->minor[r]);
    int i, whole;

    DEBUG(0, DEVICE_NAME ": ftl_reread_partition(%d)\n", minor);
    if (part->locked || (part->open > 1))
	return -EBUSY;
    part->locked = 1;
    
    whole = minor & ~(MAX_PART-1);
    for (i = 0; i < MAX_PART; i++) {
	if (ftl_hd[whole+i].nr_sects > 0) {
	    kdev_t rdev = MKDEV(major_dev, whole+i);
	    sync_dev(rdev);
	    INVALIDATE_INODES(rdev);
	    invalidate_buffers(rdev);
	}
	ftl_hd[whole+i].start_sect = 0;
	ftl_hd[whole+i].nr_sects = 0;
    }

    if (scan_header(part) == 0)
	ftl_hd[whole].nr_sects =
	    part->header.FormattedSize/SECTOR_SIZE;
    else 
	ftl_hd[whole].start_sect = -1;

    resetup_one_dev(&ftl_gendisk, whole >> PART_BITS);
    
#ifdef FTL_ITSY_DEBUG
    for (i = 0; i < MAX_PART; i++) {
	if (ftl_hd[whole+i].nr_sects > 0)
	    printk(KERN_INFO "  %d: start %ld size %ld\n", i,
		   ftl_hd[whole+i].start_sect,
		   ftl_hd[whole+i].nr_sects);
    }
#endif
    
    part->locked = 0;
    wake_up(&ftl_wait_open);
    return 0;
}

/*======================================================================

    Handler for block device requests

======================================================================*/

STATIC void
do_ftl_request(void)
#if CHECK_IF_REENTERED
{
  STATIC void do_ftl_request_inner(void);
  static int reentered = 0;
  if (reentered) {
    printk("do_ftl_request reentered!");
  }
  reentered++;
  do_ftl_request_inner();
  reentered--;
}

STATIC void
do_ftl_request_inner(void)
#endif
{
    int ret, minor;
    dev_link_t *link;
    ftl_dev_t *dev;
    partition_t *part;

    do {
        if (CURRENT == &blk_dev[MAJOR_NR].plug) return;
        sti();
	INIT_REQUEST;

	minor = MINOR(DEVICE(CURRENT));
	
	link = dev_table[DEVICE_NR(minor)];
	if (link == NULL) {
	  end_request(0);
	  return;
	}
	dev = (ftl_dev_t *)link->priv;
	part = &dev->minor[REGION_NR(minor)];
	ret = 0;
	
	switch (CURRENT->cmd) {
	    
	case READ:
	    ret = ftl_read(part, CURRENT->buffer,
			   CURRENT->sector+ftl_hd[minor].start_sect,
			   CURRENT->current_nr_sectors);
	    break;
	    
	case WRITE:
	    ret = ftl_write(dev, part, CURRENT->buffer,
			    CURRENT->sector+ftl_hd[minor].start_sect,
			    CURRENT->current_nr_sectors);
	    break;
	    
#ifdef DELETEU
	case DELETEU:
	case DELETEZ:
	    ret = ftl_delete(dev, part,
			     CURRENT->sector+ftl_hd[minor].start_sect,
			     CURRENT->current_nr_sectors);
	    break;
#endif

	default:
	    panic(DEVICE_NAME ": unknown block command!\n");
	    
	}
	if (ret == -EWOULDBLOCK) {
	  return;
	} else {
	  end_request((ret == 0) ? 1 : 0);
	}
    } while (1);
} /* do_ftl_request */

/*====================================================================*/

int flash_chrdev_init(void);
void flash_chrdev_cleanup(void);

int ftl_itsy_init(void)
{
    int i, res;
    dev_link_t *link;

    DEBUG(0, "%s\n", version);
#ifdef FTL_ITSY_DEBUG
    printk(DEVICE_NAME ": debug level %d\n", ftl_itsy_debug);
#endif
    
    res = register_blkdev(major_dev, "ftl_itsy", &ftl_blk_fops);
    if (res < 0) {
	printk(KERN_NOTICE DEVICE_NAME ": unable to grab major "
	       "device number!\n");
	return res;
    }
    if (res) major_dev = res;

    for (i = 0; i < MINOR_NR(MAX_DEV, 0, 0); i++) {
      ftl_blocksizes[i] = 1024;
    }
    for (i = 0; i < MINOR_NR(MAX_DEV, 0, 0); i++) {
      ftl_hd[i].nr_sects = 0;
      ftl_hd[i].start_sect = -1;
    }
    blksize_size[major_dev] = ftl_blocksizes;
    ftl_gendisk.major = major_dev;
    blk_dev[major_dev].request_fn = DEVICE_REQUEST;
#if 0
    /* Not needed, as we are deferring partition check until first open */
    ftl_gendisk.next = gendisk_head;
    gendisk_head = &ftl_gendisk;
#endif
    init_waitqueue(&ftl_wait_open);

    /* Loop through StrongARM bank numbers that can have real flash
       (not a mirror) on Itsy. */
    for (i = FIRST_BANK; i <= LAST_BANK; i++) {
      link = ftl_attach((client_handle_t)i);
      if (link) {
	ftl_config(link);
	link->state |= DEV_PRESENT;
      }
    }

    return flash_chrdev_init();
}

/* Not actually called, as we aren't linking into the gendisk list */
STATIC void
ftl_geninit(struct gendisk *gd)
{
  int i;
  for (i = 0; i < MINOR_NR(MAX_DEV, 0, 0); i++) {
    /* Prevent partition checks during bootup; wait for first open */
    ftl_hd[i].nr_sects = 0;
    ftl_hd[i].start_sect = -1;
  }
}

#ifdef MODULE
int init_module(void)
{
  return ftl_itsy_init();
}

void cleanup_module(void)
{
    int i;
    dev_link_t *link;
    struct gendisk *gd, **gdp;

    DEBUG(0, DEVICE_NAME ": unloading\n");

    if (major_dev != 0)
	unregister_blkdev(major_dev, "ftl_itsy");
    for (i = 0; i < MAX_DEV; i++) {
	link = dev_table[i];
	if (link) {
	    if (link->state & DEV_CONFIG)
		ftl_release((u_long)link);
	    ftl_detach(link);
	}
    }
    blk_dev[major_dev].request_fn = NULL;
    blksize_size[major_dev] = NULL;
    for (gdp = &gendisk_head; *gdp; gdp = &((*gdp)->next)) {
	if (*gdp == &ftl_gendisk) {
	    gd = *gdp; *gdp = gd->next;
	    break;
	}
    }

    flash_chrdev_cleanup();
}
#endif

/*********************************************************************
 *
 *  Character device support (/dev/flash*).  Bypasses ftl and directly
 *  addresses either an individual region or an entire bank.  Minor
 *  device bitfields DEVICE_NR and REGION_NR are used as with the ftl
 *  devices, while PART_NR must be 0.
 *
 *********************************************************************/

#define FLASH_TIMEOUT_JIFFIES (HZ)
#define CHRDEV_NAME "flash"

static int chr_dev = DEFAULT_FLASH_MAJOR;
MODULE_PARM(chr_dev, "i");

/* Decode old-style minors 0, 1, 2, 3 for backward compatibility */
int
compat_minor(int minor)
{
  if (minor < MAX_DEV) minor = MINOR_NR(minor, 0, 0);
  return minor;
}

STATIC int
flash_lseek(struct inode *inode, struct file *file, off_t offset, int whence)
{
  long pos;
  dev_link_t *link;
  ftl_dev_t *dev;
  partition_t *part;
  int devsize;
  int minor = compat_minor(MINOR(inode->i_rdev));

  link = dev_table[DEVICE_NR(minor)];
  if (!DEV_OK(link)) return -ENXIO;
  dev = (ftl_dev_t *)link->priv;
  part = &dev->minor[REGION_NR(minor)];
  if (PART_NR(minor) != 0) return -ENXIO;

  /* find size */
  devsize = part->region.RegionSize;
  if (REGION_NR(minor) == 0) {
    /* !!kludge to allow char device to access last erase sector -- ugh */
    devsize += part->region.RegionSize;
  }

  switch (whence) {
  case 0: 
    /* SEEK_SET */
    pos = offset;
    break;
  case 1:
    /* SEEK_CUR */
    pos = file->f_pos + offset;
    break;
  case 2:
    /* SEEK_END */
    pos = devsize + offset;
  default:
    return -EINVAL;
  }

  /* sanity check */
  if (pos < 0)
    return(-EINVAL);

  /* update file position */
  file->f_pos = pos;
  return file->f_pos;
}


STATIC int
flash_read(struct inode *inode, struct file *file, char *buffer, int count)
{
  dev_link_t *link;
  ftl_dev_t *dev;
  partition_t *part;
  int err, nleft, nread, devsize;
  mem_op_t req;
  int minor = compat_minor(MINOR(inode->i_rdev));

  DEBUG(0, DEVICE_NAME ": flash_read: minor=%d file->f_pos=%u\n",
	minor, (u32)file->f_pos);

  link = dev_table[DEVICE_NR(minor)];
  if (!DEV_OK(link)) return -ENXIO;
  dev = (ftl_dev_t *)link->priv;
  part = &dev->minor[REGION_NR(minor)];
  if (PART_NR(minor) != 0) return -ENXIO;
  
  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_WRITE, (void *) buffer, count)) != 0) {
    DEBUG(1, DEVICE_NAME ": flash_read: verify_area() failed\n");
    /*stats->read_err++;*/
    return err;
  }

  /* compute size of read */
  devsize = part->region.RegionSize;
  if (REGION_NR(minor) == 0) {
    /* !!kludge to allow char device to access last erase sector -- ugh */
    devsize += part->region.RegionSize;
  }
  nleft = devsize - file->f_pos;
  nread = (nleft < count) ? nleft : count;

  /* fail if invalid count */
  if (nread <= 0) {
    /*stats->read_err++;*/
    return -EINVAL;
  }
  
  req.Attributes = 0;
  req.Count = nread;
  req.Offset = file->f_pos;
  for (;;) {
    err = BMIReadMemory(part->handle, &req, buffer);
    if (err == 0) {
      break;
    } else if (err == EWOULDBLOCK) {
      /* Another process is blocked inside a flash access, presumably
	 because it took a page fault while reading or writing its
	 buffer.  Waiting on erase_pending is thus not really the
	 right thing to do, so we use a timeout to be sure we wake up.
	 This is sloppy, but it should not be a performance problem, as
	 blocking here should be extremely rare. */
      DEBUG(0, DEVICE_NAME ": flash_read blocking due to contention\n");
      current->timeout = jiffies + FLASH_TIMEOUT_JIFFIES;
      interruptible_sleep_on(&dev->erase_pending);
    } else {
      cs_error(ReadMemory, err);
      return err;
    }
  }
  file->f_pos += nread;
  return nread;
}

STATIC int
flash_write(struct inode *inode, struct file *file,
	    const char *buffer, int count)
{
  dev_link_t *link;
  ftl_dev_t *dev;
  partition_t *part;
  int err, nleft, nwrite, devsize;
  mem_op_t req;
  int minor = compat_minor(MINOR(inode->i_rdev));

  DEBUG(0, DEVICE_NAME ": flash_write: minor=%d file->f_pos=%u\n",
	minor, (u32)file->f_pos);

  link = dev_table[DEVICE_NR(minor)];
  if (!DEV_OK(link)) return -ENXIO;
  dev = (ftl_dev_t *)link->priv;
  part = &dev->minor[REGION_NR(minor)];
  if (PART_NR(minor) != 0) return -ENXIO;
  
  /* !!check for read-only bank? */

  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_READ, (void *) buffer, count)) != 0) {
    DEBUG(1, DEVICE_NAME ": flash_write: verify_area() failed\n");
    /*stats->write_err++;*/
    return err;
  }

  /* compute size of write */
  devsize = part->region.RegionSize;
  if (REGION_NR(minor) == 0) {
    /* !!kludge to allow char device to access last erase sector -- ugh */
    devsize += part->region.RegionSize;
  }
  nleft = devsize - file->f_pos;
  nwrite = (nleft < count) ? nleft : count;

  /* fail if invalid count */
  if (nwrite <= 0) {
    /*stats->write_err++;*/
    return -EINVAL;
  }
  
  req.Attributes = 0;
  req.Count = nwrite;
  req.Offset = file->f_pos;
  for (;;) {
    err = BMIWriteMemory(part->handle, &req, (char*)buffer);
    if (err == 0) {
      break;
    } else if (err == EWOULDBLOCK) {
      /* Another process is blocked inside a flash access, presumably
	 because it took a page fault while reading or writing its
	 buffer.  Waiting on erase_pending is thus not really the
	 right thing to do, so we use a timeout to be sure we wake up.
	 This is sloppy, but it should not be a performance problem, as
	 blocking here should be extremely rare. */
      DEBUG(0, DEVICE_NAME ": flash_write blocking due to contention\n");
      current->timeout = jiffies + FLASH_TIMEOUT_JIFFIES;
      interruptible_sleep_on(&dev->erase_pending);
    } else {
      cs_error(WriteMemory, err);
      return err;
    }
  }
  file->f_pos += nwrite;
  return nwrite;
}

/* Card Services memory_cs ioctls, to support ftl_format program */
STATIC int
flash_ioctl(struct inode *inode, struct file *file,
	    unsigned int cmd, unsigned long arg)
{
  dev_link_t *link;
  ftl_dev_t *dev;
  partition_t *part;
  int err;
  int minor = compat_minor(MINOR(inode->i_rdev));

  link = dev_table[DEVICE_NR(minor)];
  if (!DEV_OK(link)) return -ENXIO;
  dev = (ftl_dev_t *)link->priv;
  part = &dev->minor[REGION_NR(minor)];
  if (PART_NR(minor) != 0) return -ENXIO;
  
  switch (cmd) {
  case MEMGETINFO: {
    DEBUG(0, DEVICE_NAME ": flash_ioctl: MEMGETINFO minor=%d\n", minor);
    err = verify_area(VERIFY_WRITE, (void*)arg, sizeof(region_info_t));
    if (err != 0) {
      DEBUG(1, DEVICE_NAME ": flash_ioctl: verify_area() failed\n");
      /*stats->ioctl_err++;*/
      return err;
    }
    memcpy_tofs((void*)arg, &part->region, sizeof(region_info_t));
    break; }

  case MEMERASE: {
    struct erase_info_t erase;
    int devsize, i;

    /* !!check for read-only bank? */

    err = verify_area(VERIFY_READ, (void*)arg, sizeof(erase_info_t));
    if (err != 0) {
      DEBUG(1, DEVICE_NAME ": flash_ioctl: verify_area() failed\n");
      /*stats->ioctl_err++;*/
      return err;
    }
    memcpy_fromfs(&erase, (void*)arg, sizeof(erase_info_t));
    DEBUG(0, DEVICE_NAME
	  ": flash_ioctl: MEMERASE minor=%d Offset=0x%08x Size=0x%08x\n",
	  minor, (u32)erase.Offset, (u32)erase.Size);
    
    /* check alignment: only a round number of full-size erase sectors
       may be requested. */
    if ((erase.Offset % part->region.BlockSize) ||
	(erase.Size % part->region.BlockSize)) {
      return -EINVAL;
    }

    /* check size */
    devsize = part->region.RegionSize;
    if (REGION_NR(minor) == 0) {
      /* !!kludge to allow char device to access last erase sector -- ugh */
      devsize += part->region.RegionSize;
    }
    if (erase.Offset + erase.Size > devsize) return -EINVAL;

    /* Wait for a free erase slot and no other erases on this region */
    for (;;) {
      for (i = 0; i < MAX_ERASE; i++)
	if (!ERASE_IN_PROGRESS(dev->eraseq[i].State)) break;
      if (i == MAX_ERASE) {
	DEBUG(0, DEVICE_NAME ": erase queue is full; waiting...\n");
	sleep_on(&dev->erase_pending);
	continue;
      }
      if (part->XferInfo[0].state == XFER_ERASING) {
	DEBUG(0, DEVICE_NAME ": erase ioctl contention; waiting...\n");
	current->timeout = jiffies + FLASH_TIMEOUT_JIFFIES;
	interruptible_sleep_on(&dev->erase_pending);
	continue;
      }
      break;
    }

    /* Pretend this is a ftl transfer unit to make save_status() happy */
    part->header.NumTransferUnits = 1;
    part->XferInfo[0].Offset = erase.Offset;
    part->XferInfo[0].state = XFER_ERASING;

    /* Queue the request */
    dev->eraseq[i].State = ERASE_QUEUED;
    dev->eraseq[i].Handle = part->handle;
    dev->eraseq[i].Offset = erase.Offset;
    dev->eraseq[i].Size = erase.Size;
    dev->eraseq[i].Optional = part;
    err = BMICheckEraseQueue(dev->eraseq_handle);
    if (err != 0) {
	cs_error(CheckEraseQueue, err);
	return err;
    }

    /* Await completion */
    while (part->XferInfo[0].state == XFER_ERASING) {
      sleep_on(&dev->erase_pending);
    }

    /* Success or failure? */
    if (part->XferInfo[0].state != XFER_ERASED) return -EIO;
    break; }

  default:
    return -EINVAL;
  }
  return 0;
}

STATIC int
flash_open(struct inode *inode, struct file *file)
{
  int minor = compat_minor(MINOR(inode->i_rdev));
  dev_link_t *link;
  ftl_dev_t *dev;
  partition_t *partition;
  open_mem_t open;
  int ret;
    
  DEBUG(0, DEVICE_NAME ": flash_open(%d)\n", minor);

  link = dev_table[DEVICE_NR(minor)];
  if (!DEV_OK(link)) return -ENXIO;
  dev = (ftl_dev_t *)link->priv;
  partition = &dev->minor[REGION_NR(minor)];
  if (partition->region.RegionSize == 0) return -ENXIO;
  if (PART_NR(minor) != 0) return -ENXIO;
    
  /* prevent open for writing as raw flash while open as ftl */
  if ((file->f_mode & FMODE_WRITE) &&
      (partition->open >
       partition->open_char_read + partition->open_char_write)) return -EBUSY;

  if (partition->handle == NULL) {
    partition->handle = (memory_handle_t)link->handle;
    open.Attributes = partition->region.Attributes;
    open.Offset = partition->region.CardOffset;
    ret = BMIOpenMemory(&partition->handle, &open);
    if (ret != 0) {
      cs_error(OpenMemory, ret);
      return ret;
    }
  }
  partition->open++;
  link->open++;
  if (file->f_mode & FMODE_WRITE) {
    partition->open_char_write++;
    partition->header.NumTransferUnits = 0;
  } else {
    partition->open_char_read++;
  }

#ifdef	MODULE
  MOD_INC_USE_COUNT;
#endif

  return 0;
}

STATIC void
flash_release(struct inode *inode, struct file *file)
{
  dev_link_t *link;
  int minor = compat_minor(MINOR(inode->i_rdev));
  ftl_dev_t *dev;
  partition_t *part;
  int i;
    
  DEBUG(0, DEVICE_NAME ": flash_release(%d)\n", minor);

  link = dev_table[DEVICE_NR(minor)];
  dev = (ftl_dev_t *)link->priv;
  part = &dev->minor[REGION_NR(minor)];
    
  /* Wait for any pending erase operations to complete */
  for (i = 0; i < part->header.NumTransferUnits; i++) {
    while (part->XferInfo[i].state == XFER_ERASING) {
      DEBUG(0, DEVICE_NAME ": waiting for erase(s) to complete...\n");
      sleep_on(&dev->erase_pending);
    }
  }
  link->open--;
  part->open--;
  if (part->open == 0) {
    BMICloseMemory(part->handle);
    part->handle = NULL;
    /* Note that the region may have been open as both block and character */
    if (part->VirtualBlockMap) {
      vfree(part->VirtualBlockMap);
      part->VirtualBlockMap = NULL;
    }
    if (part->VirtualPageMap) {
      kfree(part->VirtualPageMap);
      part->VirtualPageMap = NULL;
    }
    if (part->EUNInfo) {
      kfree(part->EUNInfo);
      part->EUNInfo = NULL;
    }
    if (part->XferInfo) {
      kfree(part->XferInfo);
      part->XferInfo = NULL;
    }
    if (part->bam_cache) {
      kfree(part->bam_cache);
      part->bam_cache = NULL;
    }
  }
    
  if (file->f_mode & FMODE_WRITE) {
    part->open_char_write--;
  } else {
    part->open_char_read--;
  }

#ifdef	MODULE
  MOD_DEC_USE_COUNT;
#endif
}

static struct file_operations flash_fops =
{
  flash_lseek,		
  flash_read,
  flash_write,
  NULL,			/* no special readdir */
  NULL,			/* no special select */
  flash_ioctl,
  NULL,			/* no special mmap */
  flash_open,
  flash_release,
  NULL,			/* no special fsync */
  NULL,			/* no special fasync */
  NULL,			/* no special check_media_change */
  NULL			/* no special revalidate */
};

STATIC int
flash_chrdev_init()
{
  int res;
  /* register device */
  res = register_chrdev(chr_dev, CHRDEV_NAME, &flash_fops);
  if (res < 0) {
    /* log error and fail */
    DEBUG(0, DEVICE_NAME ": unable to register major device %d\n", chr_dev);
    return res;
  }
  if (res) chr_dev = res;
  return 0;  
}

#ifdef MODULE
STATIC void
flash_chrdev_cleanup()
{
  /* unregister driver */
  unregister_chrdev(chr_dev, CHRDEV_NAME);
}
#endif
