/*
 * linux/arch/arm/drivers/char/con_brutus.c
 *
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Larry Brakmo.
 *
 * Low level console (framebuffer) stuff
 *
 * Changelog:
 * 22-01-1998	lsb	Created
 *
 * $Log: con_brutus.c,v $
 * Revision 2.6  1999/12/23  01:27:50  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.5  1999/10/09  03:00:33  brakmo
 * Added backlight support (ioctl to set on/off and colormap fix).
 *
 * Revision 2.4  1999/06/03 16:58:01  brakmo
 * Format change in /proc/lcd so it fits better on Itsy's LCD screen.
 *
 * Revision 2.3  1999/05/12 18:00:59  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.2.2.2  1999/05/12  03:06:41  brakmo
 * Fixed dynamic framebuffer mmap (needed to reserve the pages).
 *
 * Revision 2.2.2.1  1999/04/21 01:41:49  brakmo
 * Added session support to console, fb and touchscreen. Renamed ts.c to itsy_ts.c
 *
 * Revision 2.2  1999/01/20  20:57:05  brakmo
 * Added /proc/lcd, /proc/registers and support for interrupt processing in
 * lcdOn and lcdStatic which enables them to return before LCD change is
 * finalized.
 *
 * Revision 2.1  1998/07/02  01:49:41  brakmo
 * Added kernel support for new ggi graphics library.
 *
 */

#define VirtAdd(PhAdd) (io_p2v(PhAdd))

#include <linux/config.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/io_reg.h>
#include <asm/proc/page.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/proc/pgtable.h>
#include <linux/itsy_console.h>
#include <asm/arch/sa1100.h>
#include <asm/arch/Itsy.h>
#include <asm/arch/mmap.h>
#include <linux/itsy_fb.h>
#include <asm/arch/irqs.h>
#include <asm/irq.h>
#include <linux/interrupt.h>

/*
 * LCD_DBG is used for debugging purposes, it should be undefined
 * in most cases.
 */
/* #define LCD_DBG */
#ifdef LCD_DBG
#include "lcdtest.h" 
extern int lcdDebugFlag;
#define IF_LCD_DBG(flag) if (lcdDebugFlag & LCD_DBG_##flag)
#define IF_LCD_DBG2(flag1,flag2) if ((lcdDebugFlag & LCD_DBG_##flag1) || \
				     (lcdDebugFlag & LCD_DBG_##flag2))
#else
#define IF_LCD_DBG(flag) if (0)
#define IF_LCD_DBG2(flag1,flag2) if (0)
#endif

#ifdef CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>

static int lcd_get_info(char *, char **, off_t, int, int);
static int reg_get_info(char *, char **, off_t, int, int);

static struct proc_dir_entry lcd_proc_entry = {
  0, 3, "lcd", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, lcd_get_info, };
static struct proc_dir_entry reg_proc_entry = {
  0, 9, "registers", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, reg_get_info, };
#endif

volatile struct FrameBuffer *FrameBuffer;
volatile struct FrameBuffer *FrameBufferNB;
volatile struct FrameBuffer *FrameBufferB;

#include "fb.h"

int FrameBufferSize;
volatile struct FrameBuffer *CurFrameBuffer;
volatile void *CurFrameBufferPhys;
volatile int CurFbn;
volatile int backlightFlag = 0;

int LcdState=LCD_STATE_OFF;

struct FbLcdParamsStruct FbLcdParams;
static char *lcd_name = "LCD driver";

extern int setup_arm_irq(int, struct irqaction *);
static int lcdirq_flag=0;
static int lcdirq_state=0;

void lcdDbg (char *s)
{
  IF_LCD_DBG2(PRINT,GOSTATIC)
    printk(s);
  IF_LCD_DBG(GOSTATIC) {
    lcdOn();
    lcdStatic(1);
  }
}

static void
setlcdirqs (void)
{
  unsigned long v;

  IF_LCD_DBG(SETIRQ)
    printk("setlcdirqs - LCSR:0x%x\n ICMR:0x%lx\n ICLR:0x%lx\n", LCSR, 
	   *IRQ_ICMR_V, *IRQ_ICLR_V); 
  LCCR0 |= LCCR0_LDM + LCCR0_BAM + LCCR0_ERM;  
  v = LCSR;
  LCSR = v;
  enable_irq(IRQ_LCD);
}

static void
clearlcdirqs (void)
{
  disable_irq(IRQ_LCD);
}

void
lcdCBoff (int flag)
{
  if (flag) {
    clearlcdirqs();
    lcdirq_flag = 0;
  }
  else {
    setlcdirqs();
    lcdirq_flag = 1;
  }
}

static void
set_lcd_bau_int (int state)
{
  lcdirq_state = state;
  LCCR0 = LCCR0 & ~LCCR0_BAM;
  DBAR1 = (Address) virt_to_phys((unsigned long)CurFrameBuffer); 
}

static void
clear_lcd_bau_int (void)
{
  
  LCCR0 = LCCR0 | LCCR0_BAM;
  DBAR1 = (Address) virt_to_phys((unsigned long)CurFrameBuffer);  
}

static void
lcd_inter_handler (int irq, void *dev_id, struct pt_regs *regs)
{
  unsigned long v;

  IF_LCD_DBG(IRQH)
    printk("lcd_int ICPR:0x%lx\n", *IRQ_ICPR_V); 

  v = LCSR;
  LCSR = v;

  if ((v & LCSR_LDD) && lcdirq_state == 5) {
    clear_lcd_bau_int();
    lcdirq_state = 0;
  }
  
  if ((v & LCSR_BAU) == 0) 
    return;
  
  LCSR = LCSR_BAU;
  DBAR1 = (Address) virt_to_phys((unsigned long)CurFrameBuffer); 

  switch (lcdirq_state) {
  case 0:
    break;
  case 1:
    clear_lcd_bau_int();
    GPCR = GPIO_AUXLCDEN ;
    GPSR = GPIO_LCDEN ; 
    lcdirq_state = 0;
    break;
  case 3:
    lcdirq_state = 4;
    break;
  case 4:
    lcdirq_state = 5;
    LCCR0 = LCCR0_LDM + LCCR0_BAM + LCCR0_ERM ;
    break;

  default:
    IF_LCD_DBG(WARNING)
      printk("lcdint - default!\n");
  }
}


/*
 * Allocate space for the FrameBuffer.
 */
#define FBSIZE ((sizeof(struct FrameBuffer) + 4095) & ~0xfff)

static char FrameBufferArea1[FBSIZE + 4095];
static char FrameBufferAreas[FBSIZE*FB_MIN + 4095];
static char FbwmAreaAlloc[((sizeof(struct fbwmStruct)+4095) & ~0xfff) + 4095];
volatile struct fbwmStruct *FbwmArea, *FbwmAreaNB;

void initDisplay (void);
void lsb_flash(int);
static void lcd_powermgr_register(void);

#ifdef CONFIG_ITSY
#define DftBlckLvl      8
#define PixPerNum       (1200/400)
#define PixPerDen       (1000000000/400)
#define PixClkDiv(FreqMul) \
  ((PixPerNum*(FreqMul)*CrystlFreq + PixPerDen - 1) / \
   PixPerDen)
  
static const unsigned int PixClkDivTbl [] =
  {
    LCCR3_CeilPixClkDiv (PixClkDiv (16)),
    LCCR3_CeilPixClkDiv (PixClkDiv (20)),
    LCCR3_CeilPixClkDiv (PixClkDiv (24)),
    LCCR3_CeilPixClkDiv (PixClkDiv (28)),
    LCCR3_CeilPixClkDiv (PixClkDiv (32)),
    LCCR3_CeilPixClkDiv (PixClkDiv (36)),
    LCCR3_CeilPixClkDiv (PixClkDiv (40)),
    LCCR3_CeilPixClkDiv (PixClkDiv (44)),
    LCCR3_CeilPixClkDiv (PixClkDiv (48)),
    LCCR3_CeilPixClkDiv (PixClkDiv (52)),
    LCCR3_CeilPixClkDiv (PixClkDiv (56))
  } ;
#endif /* CONFIG_ITSY */

int fbLcdEnd=0;

void lcdOn (void)
{
  unsigned long t=OSCR;
  
#ifdef CONFIG_ITSY
  LCCR0 = LCCR0_LDM + LCCR0_BAM + LCCR0_ERM ;
  GPDR |= (GPIO_LCDEN + GPIO_AUXLCDEN);
  GPCR = GPIO_LCDEN;
  if (LcdState != LCD_STATE_STATIC)
    GPCR = GPIO_AUXLCDEN ;
  PGSR &= ~(GPIO_LCDEN + GPIO_AUXLCDEN) ;

  LCCR1 = LCCR1_DisWdth (320) + LCCR1_HorSnchWdth (2) +
    	  LCCR1_EndLnDel (2) + LCCR1_BegLnDel (2) ;
  LCCR2 = LCCR2_DisHght (200) + LCCR2_VrtSnchWdth (1) +
    	  LCCR2_EndFrmDel (0) + LCCR2_BegFrmDel (0) ;
  LCCR3 = PixClkDivTbl [(PPCR & FMsk (PPCR_CCF)) >> FShft (PPCR_CCF)] +
          LCCR3_ACBsDiv (512) + LCCR3_ACBsCntOff +
	  LCCR3_VrtSnchH + LCCR3_HorSnchH + LCCR3_PixFlEdg ;
  DBAR1 = (Address) virt_to_phys((unsigned long)CurFrameBuffer); 

  if (fbLcdEnd == 0)
    LCCR0 = LCCR0_LEN + LCCR0_Mono + LCCR0_Sngl + LCCR0_LDM +
      LCCR0_BAM + LCCR0_ERM + LCCR0_Pas + LCCR0_LtlEnd +
	LCCR0_8PixMono + LCCR0_DMADel (0) ;
  else
    LCCR0 = LCCR0_LEN + LCCR0_Mono + LCCR0_Sngl + LCCR0_LDM +
      LCCR0_BAM + LCCR0_ERM + LCCR0_Pas + LCCR0_BigEnd +
	LCCR0_8PixMono + LCCR0_DMADel (0) ;

  LCSR = LCSR_BAU ; 
  if (lcdirq_flag) {
    set_lcd_bau_int(1);
    IF_LCD_DBG(ON)
      printk("lcdOn with CB\n");
  }
  else {
    DBAR1 = (Address) virt_to_phys((unsigned long)CurFrameBuffer);
    while (!(LCSR & LCSR_BAU)) /* Wait */ ;
    GPCR = GPIO_AUXLCDEN ;
    GPSR = GPIO_LCDEN ;
    IF_LCD_DBG(ON)
      printk("lcdOn slow\n");
  }
/*   PGSR |= GPIO_LCDEN ; */
#else
  {
    volatile struct lcd_regs *lcdreg; 
    
    lcdreg->Control0 = 0;
    lcdreg->DMA1Base = virt_to_phys(CurFrameBuffer);
    lcdreg->Control1 = (40<< LCD1_V_BLW) | (100 << LCD1_V_ELW) | (2 << LCD1_V_HSW)
      | ((320-1) << LCD1_V_PPL);
    lcdreg->Control2 = (0 << LCD2_V_BFW) | (0 << LCD2_V_EFW) | (0 << LCD2_V_VSW) |
      ((240-1) << LCD2_V_LPP);
    lcdreg->Control3 = (0 << LCD3_V_OEP) | (0 << LCD3_V_PCP) | (0 << LCD3_V_HSP) |
      (0 << LCD3_V_VSP) | (0 << LCD3_V_API)
	| (0 << LCD3_V_ACB) | (20 << LCD3_V_PCD);
    
    /* Fire up the LCD */
    lcdreg->Control0 = (0 << LCD0_V_FDD) | (0 << LCD0_V_DPD) | (0 << LCD0_V_BLE) |
      (0 << LCD0_V_PAS) 
	| (0 << LCD0_V_SDS) | (0 << LCD0_V_CMS) | (1 << LCD0_V_LEN);
  }
#endif

  LcdState = LCD_STATE_ON;
  IF_LCD_DBG(TIME) {
    t = OSCR - t;
    printk("lcdOn timing: %us\n", (int) (t*1000/3686));
  }
}

static int staticOffFlag=0;

void lcdStaticOff (int flag)
{
  staticOffFlag = flag;
}

void lcdStatic (int type)
{
  if (staticOffFlag)
    return;
  
  if (!(LCCR0 & LCCR0_LEN))
    return;
  GPSR = GPIO_AUXLCDEN ;
  PGSR |= (GPIO_LCDEN + GPIO_AUXLCDEN) ;
  LCSR = LCSR_BAU ; 
  if (type == 0 && lcdirq_flag) {
    IF_LCD_DBG(STATIC)
      printk("  calling set_lcd_bau(3)\n"); 
    set_lcd_bau_int(3);
  }
  else {
    IF_LCD_DBG(STATIC)
      printk("  doing slow static change \n");
    DBAR1 = (Address) virt_to_phys((unsigned long) CurFrameBuffer);
    while (!(LCSR & LCSR_BAU)) /* Wait */ ;
    LCSR = LCSR_BAU ;
    DBAR1 = (Address) virt_to_phys((unsigned long) CurFrameBuffer);
    while (!(LCSR & LCSR_BAU)) /* Wait */ ;
    LCCR0 = LCCR0_LDM + LCCR0_BAM + LCCR0_ERM ;
    while (!(LCSR & LCSR_LDD)) /* Wait */ ;
    IF_LCD_DBG(STATIC)
      printk("  done with slow static change \n");
  }
  LcdState = LCD_STATE_STATIC;
}

void lcdOff (void)
{
  lcdirq_state = 0;
  LCCR0 = LCCR0_LDM + LCCR0_BAM + LCCR0_ERM ;
/*   if (!lcdDebugFlag & LCD_DBG_TIME_TEST) { */
/*     while (!(LCSR & LCSR_LDD)) ; */
/*   } */
  GPCR = GPIO_LCDEN + GPIO_AUXLCDEN ;
  PGSR &= ~(GPIO_LCDEN + GPIO_AUXLCDEN) ;

  LcdState = LCD_STATE_OFF;
}

void backlightFixColormap (volatile struct FrameBuffer *fb)
{
  int k;

#if BITS_PER_PIXEL==4
  fb->palette[0] = (15 - (fb->palette[0] & 0xf)) | LCD_PBS_PALETTE;
  for (k=1; k<16; k++) 
    fb->palette[k] = 15 - (fb->palette[k] & 0x000f);
#endif
}

void lcdSwitchFB (struct FrameBuffer *fb, int fbn)
{
  if (backlightFlag)
    backlightFixColormap(CurFrameBuffer);

  CurFrameBuffer = fb;

  if (backlightFlag)
    backlightFixColormap(CurFrameBuffer);

  CurFbn = fbn;
  DBAR1 = (Address) virt_to_phys((unsigned long) CurFrameBuffer);
  CurFrameBufferPhys = (void *) virt_to_phys((unsigned long) CurFrameBuffer);
}

static struct irqaction lcdirq = {lcd_inter_handler, SA_INTERRUPT, 0, "lcd",
				    NULL, NULL};

void lcdSetIrq (void)
{
  unsigned long v;
  static int firstTime=0;
  
  if (firstTime)
    return;

  firstTime = 1;
  v = LCSR;
  LCSR = v;
  v = setup_arm_irq(IRQ_LCD, &lcdirq);
  
  if (v) { 
    printk("ERROR: Could not allocate lcd IRQ!\n");
  }
  else {
    lcdirq_flag = 1;
    setlcdirqs(); 
  }
}
  
void initDisplay (void)
{
  char *cp;
  int i, k, h, v;
  volatile struct lcd_regs *lcdreg;
  volatile struct FrameBuffer *fb1;
  static int initDisplayFlag=1;

/* serial_echo_print_wait_C_V('a'); */
  if (!initDisplayFlag)
    return;

  initDisplayFlag = 0;

  FrameBufferSize = i = FBSIZE;
  i += 4095;				/* add enough to ensure we can */
  					/* move it to a page boundary */
  cp = (char *) (((int) (FrameBufferArea1 + 4095)) & ~0xfff);
  CurFrameBuffer = FrameBuffer = (struct FrameBuffer *) cp;
  CurFbn = 0;
  printk("console: fb physical address: 0x%x\n", 
	 (int) virt_to_phys((unsigned long)CurFrameBuffer));
  FrameBufferNB = (struct FrameBuffer *) allocate_vspace((unsigned long) virt_to_phys((unsigned long)FrameBuffer), PGDIR_SIZE, 1, 0); 
  /* LSB TODO need to change to use bufferable!*/
  /*   FrameBufferB = FrameBufferNB;  */
  FrameBufferB =  (struct FrameBuffer *) allocate_vspace((unsigned long) virt_to_phys((unsigned long)FrameBuffer), PGDIR_SIZE, 1, 1);  
  if (FrameBufferNB == NULL  ||  FrameBufferB == NULL)
    printk("allocate_vspace for FrameBuffer{B,NB} fails!\n");
  
  fb1 = (struct FrameBuffer *) FRAME_BUFFER_ADDR; /* FrameBuffer */

  cp = (char *) (((int) (FbwmAreaAlloc + 4095)) & ~0xfff);
  FbwmArea = (struct fbwmStruct *) cp;

  FbwmAreaNB = (struct fbwmStruct *) allocate_vspace((unsigned long) virt_to_phys((unsigned long)cp), PGDIR_SIZE, 1, 0);  


  FbwmArea->visibleFB = 0;
  FbwmArea->maxCnt = FB_MAX;
  FbwmArea->size = (sizeof(struct fbwmStruct) + 4095) & ~0xfff;
  FbwmArea->fbArray[0].pid = 1;
  FbwmArea->fbArray[0].addr = (char *) FrameBuffer;
  FbwmArea->fbArray[0].flags = INUSE_FLAG | ALLOC_FLAG;
  strcpy((char *)FbwmArea->fbArray[0].execname, "Console");
  FbwmArea->fbArray[0].alloc_sid = 0;
  FbwmArea->fbArray[0].mmap_count = 1;
  FbwmArea->fbArray[0].open_count = 0;

  cp = (char *)(((int) (FrameBufferAreas + 4095)) & ~0xfff);
  for (k=0; k<FB_MIN; k++) {
    FbwmArea->fbArray[k+1].pid = 0;
    FbwmArea->fbArray[k+1].addr = cp;
    FbwmArea->fbArray[k+1].flags = ALLOC_FLAG;
    FbwmArea->fbArray[k+1].alloc_sid = 0;
    FbwmArea->fbArray[k+1].mmap_count = 0;
    FbwmArea->fbArray[k+1].open_count = 0;
    strcpy((char *)FbwmArea->fbArray[k+1].execname, "not used");
    cp += FBSIZE;
  }

#if 0
  /* previously hardcoded for old graphics library */
  FbwmArea->fbArray[1].flags = (ALLOC_FLAG | INUSE_FLAG);
  FbwmArea->fbArray[1].alloc_sid = 0;
  FbwmArea->fbArray[1].mmap_count = 1;
  strcpy(FbwmArea->fbArray[1].execname, "Hard");
  
  FbwmArea->fbArray[2].flags = (ALLOC_FLAG | INUSE_FLAG);
  FbwmArea->fbArray[2].alloc_sid = 0;
  FbwmArea->fbArray[2].mmap_count = 1;
  strcpy(FbwmArea->fbArray[2].execname, "Hard");
#endif

  FbwmArea->fbArray[FB_NONE].pid = 1;
  FbwmArea->fbArray[FB_NONE].flags = INUSE_FLAG | ALLOC_FLAG;
  FbwmArea->fbArray[FB_NONE].alloc_sid = 0;
  FbwmArea->fbArray[FB_NONE].mmap_count = 1;
  strcpy((char *)FbwmArea->fbArray[FB_NONE].execname, "Blank");

  lcdreg = (struct lcd_regs *) io_p2v(VIDEO_BASE_P);

#ifdef CONFIG_ITSY
  for (i=0; i<PALETTE_NUM_ENTRIES; i++) {
#if BITS_PER_PIXEL==4
    v = i;
#else
#if BITS_PER_PIXEL==8
    v = (i<<8) | (i<<4) | i;
#else 
    v = 0;
#endif
#endif
    fb1->palette[i] = v;
    for (k=1; FbwmArea->fbArray[k].flags & ALLOC_FLAG; k++)
      ((struct FrameBuffer *)FbwmArea->fbArray[k].addr)->palette[i] = v;
  }
  fb1->palette[0] |= LCD_PBS_PALETTE;
  for (k=1; FbwmArea->fbArray[k].flags & ALLOC_FLAG; k++)
    ((struct FrameBuffer *)FbwmArea->fbArray[k].addr)->palette[0] |= 
      LCD_PBS_PALETTE;
#else
  fb1->palette[0] = 0x1F00;
  fb1->palette[1] = 0x1FFF;
  for (i=2; i<PALETTE_NUM_ENTRIES; i++) {
        fb1->palette[i] = 0x0F0;
  }
  fb1->palette[LINUX_WHITE_INDEX] = 0xFFF;
#endif
  for (v=0; v<SCREEN_SIZE_V; v++) {
    for (h=0; h<(SCREEN_SIZE_H); h++) {
      fb1->pixel[v][h] = 0;
    }
  }
  
  lcdOn();
  
  lcd_powermgr_register();

  FbLcdParams.screenSizeH = SCREEN_SIZE_H;
  FbLcdParams.screenSizeV = SCREEN_SIZE_V;
  FbLcdParams.frameBufferSizeH = FRAME_BUFFER_SIZE_H;
  FbLcdParams.frameBufferSize = FrameBufferSize;
  FbLcdParams.pixelDataOffset = PALETTE_OFFSET;
  FbLcdParams.bitsPerPixel = 4;
  FbLcdParams.reserveTopRows = RESERVE_TOP_PIXEL_ROWS;
  FbLcdParams.reserveBottomRows = RESERVE_BOTTOM_PIXEL_ROWS;
  FbLcdParams.paletteSize = 16;
  FbLcdParams.colorType = FB_LCD_GRAY;
  FbLcdParams.bitsPerColor = 4;
  FbLcdParams.flags = FB_LCD_FLAG_PALETTE_FIXED;

#ifdef CONFIG_PROC_FS
  proc_register_dynamic(&proc_root, &lcd_proc_entry);
  proc_register_dynamic(&proc_root, &reg_proc_entry);
#endif
}

#include "con_info.h"

extern void ll_write_char (unsigned long, unsigned long);


void con_putc_fb (int row, int col, int v)
{
  int r, c;

  if (row < 0)
    r = NUM_ROWS + row;
  else
    r = row;
  if (r < 0)
    r = 0;
  else if (r >= NUM_ROWS)
    r = NUM_ROWS - 1;
  
  if (col < 0)
    c = NUM_COLS + col;
  else
    c =  col;
  if (c < 0)
    c = 0;
  else if (c >= NUM_COLS)
    c = NUM_COLS - 1;
  
  ll_write_char((unsigned long) (&FrameBufferB->pixel[r*FONT_CHAR_HEIGHT_PIXELS]
                [c*(FONT_CHAR_WIDTH_PIXELS*BITS_PER_PIXEL)/8]),
		(unsigned long) v);

}

void con_info (int type, int state)
{
  char *cp, str[100];
  int  i;
  static char timer_str[]={'|', '/', '-', '\\'};
  static int  timer_cnt=0;

  return;
  
/*   msetleds(6); */
  switch (type) {
  case CON_INFO_TIMER:		/* Uses pos 0 */
    con_putc_fb(-1, 4, timer_str[timer_cnt++] | (15 << 8) );
    timer_cnt &= 0x3;
    break;
  case CON_INFO_KBSTATE:	/* Uses pos 2-6 */	
    strcpy(str, "[a  ]");
    if (state & CON_INFO_KBSTATE_CAPS)
      str[1] = 'A';
    if (state & CON_INFO_KBSTATE_NUML)
      str[2] = '1';
    if (state & CON_INFO_KBSTATE_SCROLLOCK)
      str[3] = 'S';
    for (i=6,cp=str; *cp; cp++)
      con_putc_fb(-1, i++, *cp | (15 << 8));
    break;
  default:			/* Uses pos -1 */
    printk("LSB default in con_info(%d,%d)\n", type, state);
    con_putc_fb(-1, -3, '*' | (15 << 8));
/*     lsb_flash(0); */
    return;
    break;
  }
  con_putc_fb(-1, 0, 0x11 | (15 << 8));
  con_putc_fb(-1, -1, 0x10 | (15 << 8));
  con_putc_fb(-1, 2, 0x1c | (15 << 8));
/*   msetleds(7); */
}



#ifndef CONFIG_ITSY
void lsb_pause(void)
{
  int i;
  for (i=0; i<5000000; i++);
}

void lsb_flash(int num)
{
  int i;
  static int last_num=0;

  return; /* LSB */
  if (num == 0) {
    last_num = (last_num + 1) & 0x7;
    if (last_num == 0)
      last_num++;
    num = last_num;
  }
  for (i=0; i<12; i++) {
    msetleds(0);
    lsb_pause();
    msetleds(num);
    lsb_pause();
  }
}

#else /* CONFIG_ITSY */


void lsb_flash(int num)
{
}

#ifdef CONFIG_POWERMGR
/*
 * Power management functions
 *
 */

#include <linux/powermgr.h>

static int lcd_suspend_check (void *unused, int idle)
{
  return 0;
}

static int lcd_suspend (void *unused)
{
  IF_LCD_DBG(SUSPEND)
    printk("lcdSuspend ICMR=0x%lx\n", *IRQ_ICMR_V);
  IF_LCD_DBG(TIME_TEST)
    lcdOff();
  else if (LcdState == LCD_STATE_ON) {
    IF_LCD_DBG(SUSPEND_NOCHANGE)
      ;
    else
      lcdStatic(1);
  }
/*   if (LcdState != LCD_STATE_OFF) { */
/*     lcdOn(); */
/*     lcdStatic(1); */
/*   } */
/*   if (staticOffFlag) { */
/*     staticOffFlag = 0; */
/*     lcdStatic(1); */
/*   } */
  return 1;
}

static int lcd_resume (void *unused, int resume_flags)
{
/*   if (LcdState != LCD_STATE_OFF) */
/*   lcdOn(); */
  IF_LCD_DBG(SUSPEND)
    printk("lcd_resume\n");
/*   printk("lcd_resume GPSR=0x%x, PGSR=0x%x\n", GPSR, PGSR); */
  if (lcdirq_flag)
    setlcdirqs(); 

  if (LcdState != LCD_STATE_ON) {
    IF_LCD_DBG(RESUME_NOCHANGE)
      ;
    else
      lcdOn();
  }  
  return 1;
}

static void lcd_powermgr_register (void)
{
  powermgr_client client;
  powermgr_id lcd_powermgr_id;

  client.suspend_check = lcd_suspend_check;
  client.suspend = lcd_suspend;
  client.resume = lcd_resume;
  client.cookie = 0;
  client.name = lcd_name;
  client.uamps_idle = 10000;
  client.uamps_sleep = 0;
  if ((lcd_powermgr_id = powermgr_register(&client)) == POWERMGR_ID_INVALID) {
    printk("Failed to register %s modeule for power management!!\n",
	   client.name);
  }
  return;
}

#endif /* CONFIG_POWERMGR */

#ifdef CONFIG_PROC_FS
static char *LcdStateStr[] = {"Off", "On", "Static", "Error"};
static char *LcdColorStr[] = {"?", "Gray", "Color8", "Color12", "Color16",
				"Color24", "ERROR"};
static char *LcdFlagStr[] = {"Not installed", "Installed", "?"};

extern int fb_get_proc_stats(char *);

static int
lcd_get_info (char *buf, char **start, off_t offset, int length, int unused)
{
  int len, i, n, m;

  for (i=n=m=0; i<FB_MAX; i++) {
    if ((FbwmAreaNB->fbArray[i].flags & INUSE_FLAG))
      n++;
    if ((FbwmAreaNB->fbArray[i].flags & ALLOC_FLAG))
      m++;
  }
  
  /* format stats */
  len = sprintf(buf,
		"screen size: %dx%d\n"
		"bits/color : %d\n"
		"color type : %s\n"
		"LCD state  : %s\n"
		"current FB : %d (p:0x%p)\n"
		"allocated  : %d buffers\n"
		"in use     : %d buffers\n"
		"LCD IRQ    : %s\n" 
		"",
		FbLcdParams.screenSizeH, FbLcdParams.screenSizeV,
		FbLcdParams.bitsPerColor,
		LcdColorStr[FbLcdParams.colorType],
		LcdStateStr[LcdState],
		/* 		FbwmAreaNB->visibleFB,  */
		CurFbn, CurFrameBufferPhys,
		m, n,
		LcdFlagStr[lcdirq_flag]
		);

  for (i=0; i<FB_MAX; i++) {
    if ((FbwmAreaNB->fbArray[i].flags & INUSE_FLAG)) {
      len += sprintf(buf+len, "FB %2d op:%2d mm:%2d SID:%4d "
		     /* 		     "flags:0x%x (0x%x) "  */
		     "%s\n", i, 
		     FbwmAreaNB->fbArray[i].open_count,
		     FbwmAreaNB->fbArray[i].mmap_count,
		     FbwmAreaNB->fbArray[i].alloc_sid,
		     /* 		     FbwmAreaNB->fbArray[i].flags, */
		     /* 		     FbwmArea->fbArray[i].flags, */
		     FbwmAreaNB->fbArray[i].execname);
    }
  }

#ifdef CONFIG_ITSY_SESSIONS
  len += fb_get_proc_stats(buf+len);
#endif

  return(len);
}

static unsigned long *regArray[]={
#include "reg-vals.h"
};

static int
reg_get_info (char *buf, char **start, off_t offset, int length, int unused)
{
  int len=0, n;

  length -= 20;
  for (n=0; regArray[n]!=NULL; n++) {
    len += sprintf(buf+len, "0x%p 0x%lx\n", regArray[n],
		   *((volatile unsigned long *)(io_p2v((int) regArray[n]))));
    if (len > length)
      break;
  }
  
  return(len);
}

#endif  /* CONFIG_PROC_FS */

#endif /* CONFIG_ITSY */


