/*
 * Itsy Flash Memory Device Driver
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Carl Waldspurger and Tim Mann
 *
 * This file now includes only the device initialization, first-level
 * interrupt handling, and /proc/flash implementation.  Both the FTL
 * block driver and the raw char driver are now in ftl_itsy.c.
 *
 * $Log: flash.c,v $
 * Revision 2.15  1999/12/23  01:27:51  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.14  1999/12/11  04:45:35  mann
 * Power manager support for flash.
 * Powerbutton bugfix: don't try to go to sleep from interrupt level, as
 * drivers may be in an unclean state.
 *
 * Revision 2.13  1999/11/23 04:53:58  mann
 * Rewrote char flash driver to be fully compatible with block ftl driver.
 * Both now live in arch/arm/drivers/block/ftl_itsy.c.
 * The flash.c file in this directory still contains low-level support
 * and the /proc/flash implementation.
 *
 * Removed some of the linux code from flprgm.c that is no longer used after
 * the rewrite.
 *
 * Revision 2.12  1999/10/09 00:17:19  mann
 * Added support for 32Mbit flash devices.
 * Added /dev/flash[0-3]o (minor devices 4-7), which write without erasing
 * first.
 *
 * Revision 2.11  1999/09/25 03:36:21  mann
 * Flash write errors now print the faulty address and data pattern, not
 * just the error code number, to /var/log/messages.
 *
 * Revision 2.10  1999/09/04 04:02:39  mann
 * Added an entry point for use by bulkmem_itsy
 *
 * Revision 2.9  1999/04/29 02:49:23  mann
 * Itsy FTL (flash translation layer) support, initial working version.
 *
 * Revision 2.8  1999/02/03 20:32:16  kramer
 * Exported the flash_read/write_prim symbols so that frd.c can call them
 *
 * Revision 2.7  1999/01/30 02:36:18  kramer
 * Split out read/write_flash into read/write_flash_prim to allow the block
 * driver to call the same functions.
 *
 * Revision 2.6  1999/01/18 18:07:17  caw
 * Minor changes to eliminate compiler warnings.
 *
 * Revision 2.5  1999/01/15  01:33:40  caw
 * Report version number during initialization.  Other cosmetic improvements.
 *
 * Revision 2.4  1998/12/15  23:12:28  caw
 * Removed hardcoded flash size, use staticmem_info_pointers[] instead.
 *
 * Revision 2.3  1998/11/25  01:28:01  kerr
 * Added clock speed change capability (with loadable module ioctl interface).
 * Does not depend on sleep mode for implementation.
 *
 * Revision 2.2  1998/08/04  01:23:34  caw
 * Added preliminary support for detection and handling of mirrored banks.
 *
 * Revision 2.1  1998/07/09  23:11:53  caw
 * Substantial modifications for interrupt-driven flash programming.
 * Currently blocks for sector erase and polls for word programming.
 * Dynamically allocate all transient buffers.  Improved /proc/flash support.
 *
 * Revision 2.0  1998/06/24  21:42:48  kramer
 * Major Bump: Move all version numbers up to 2.0 for external release.
 * -drew
 *
 * Revision 1.4  1998/06/24  02:28:07  caw
 * Initial support for writing (programming) flash memory.
 * Added statistics, other minor improvements.
 *
 * Revision 1.3  1998/06/16 19:19:36  kramer
 * Added copyright information
 *
 * Revision 1.2  1998/04/23  20:58:12  caw
 * Added lseek, fixed bad offset bug in read.
 *
 * Revision 1.1  1998/04/23  02:13:07  caw
 * Initial revision.
 * */

/*
 * debugging 
 *
 */

#define FLASH_DEBUG			(1)
#define FLASH_DEBUG_VERBOSE		(0)
/*#define FLASH_DEBUG_VERBOSE		(1)*/

/*
 * includes
 *
 */

#include <linux/config.h>

#ifdef	CONFIG_MODULES
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/major.h>
#include <linux/ctype.h>
#include <linux/errno.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/arch/mmu.h>
#include <asm/arch/mmap.h>
#include <asm/procinfo.h>

#include <asm/arch/irq.h>
#include <asm/arch/irqs.h>
#include <asm/irq.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef	CONFIG_POWERMGR
#include <linux/powermgr.h>
#endif

#include <linux/flinfo.h>
#include <linux/bulkmem_itsy.h>
#include <linux/memory.h>

/*
 * constants
 *
 */

/* names */
#define	FLASH_NAME			"flash"
#define	FLASH_NAME_VERBOSE		"itsy flash driver"
#define	FLASH_VERSION_STRING		"$Revision: 2.15 $"

/* flash id area size (bytes) */
#define	FLASH_ID_SIZE			(32)

/* flash id offsets (16-bit values, from end of flash, big-endian) */
#define	FLASH_ID_CORE			(FLASH_ID_SIZE - 4)
#define	FLASH_ID_GPIO0			(FLASH_ID_SIZE - 8)
#define	FLASH_ID_GPIO1			(FLASH_ID_SIZE - 6)
#define	FLASH_ID_GPIO_INVALID		(0xff)

/* maximum flash memory size (bytes) */
#define	FLASH_BANK_MAX_SIZE		(32 << 20)

/* number of banks supported by StrongARM */
#define	FLASH_NBANK			(4)

/* option flag */
#define	FLASH_CHECK_MIRROR		(1)

/*
 * types
 *
 */

/*!!fixme: stats are mostly unimplemented right now; not updated by (or even
  accessible to) real drivers in ftl_itsy.c */
typedef struct {
  /* read stats */
  ulong read_sys;
  ulong read_err;
  ulong read_bytes;

  /* write stats */
  ulong write_sys;
  ulong write_err;
  ulong write_bytes;
} flash_stats;

typedef struct {
  /* bank parameters */
  FlMemInfo flMemInfo;
  u8  read_only;
  u8  dcard;
  u16 msc_config;

  /* mirroring */
  u8  mirror;
  u8  mirror_bank;

  /* mapping */
  u32 paddr;
  u32 vaddr;

  /* id structure */
  u32 id_paddr;
  u32 id_vaddr;
  u16 id_core;
  u8  id_gpio_ready[2];
   
  /* validity */
  u8  valid;

  /* interrupts */
  int interrupt_armed;  /* >0 if gpio edge detector is on */
  u32 gpio_mask;
  u32 irq_mask;
  void (*external_interrupt)(int banknum);

#ifdef CONFIG_POWERMGR
  /* power management */
  void (*external_suspend)(int banknum);
  void (*external_resume)(int banknum);
#endif

  /* interrupt stats */
  ulong interrupts;
  ulong timeouts;

  /* statistics */
  flash_stats stats;

} flash_bank;

typedef struct {
  /* per-bank data */
  flash_bank bank[FLASH_NBANK];

  /* combined gpio, irq masks */
  u32 gpio_mask;
  u32 irq_mask;

  /* aggregate stats */
  ulong interrupts;
  ulong timeouts;

  /* memory-mapped registers */
  volatile u32 *mem_config_0;
  volatile u32 *mem_config_1;
  volatile u32 *gpio_levels;
  volatile u32 *gpio_rising;
  volatile u32 *gpio_falling;
  volatile u32 *gpio_detect;
  volatile u32 *gpio_pin_dir;

#ifdef	CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
#endif
} flash_state;

/*
 * globals
 *
 */

static flash_state flash_global;

#ifdef	CONFIG_PROC_FS
static int flash_get_info(char *, char **, off_t, int, int);
static struct proc_dir_entry flash_proc_entry = {
  0, 5, "flash", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, flash_get_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int flash_get_info(char *buf, 
			  char **start,
			  off_t offset,
			  int length,
			  int unused)
{
  flash_state *f = &flash_global;
  int i, len, nvalid;

  /* initialize */
  len = 0;

  /* count valid banks */
  nvalid = 0;
  for (i = 0; i < FLASH_NBANK; i++)
    {
      flash_bank *b = &f->bank[i];
      if (b->valid)
	nvalid++;
    }

  /* done if no valid banks */
  if (nvalid < 1)
    {
      len += sprintf(buf + len, "no valid flash banks\n");
      return(len);
    }

  /* report per-bank mappings */
  len += sprintf(buf + len, 
		 "bank  physical  virtual    "
		 "size   sctsz  bits  mode  loc  mirror\n");
  for (i = 0; i < FLASH_NBANK; i++)
    {
      flash_bank *b = &f->bank[i];
      if (b->valid)
	{
	  len += sprintf(buf + len,
			 "%4d  %08x  %08x  "
			 "%5dk  %4dk  %1dx%2d  %4s   %2s  ", 
			 b->flMemInfo.stMemBnkNb,
			 b->paddr, b->vaddr,
			 (b->flMemInfo.size >> 10), 
			 (b->flMemInfo.nlSctSz >> 10),
			 b->flMemInfo.devsPerWord, 
			 b->flMemInfo.bitsPerDev,
			 b->read_only ? "r "  : "rw",
			 b->dcard     ? "dc"  : "mb");
	  if (b->mirror)
	    len += sprintf(buf + len, "bank %d\n", b->mirror_bank);
	  else
	    len += sprintf(buf + len, "no\n");
	}
    }

  /* report per-bank I/O stats */
  len += sprintf(buf + len, 
		 "\n"
		 "bank  "
		 " reads       bytes  errs  "
		 "writes       bytes  errs\n");
  for (i = 0; i < FLASH_NBANK; i++)
    {
      flash_bank *b = &f->bank[i];
      flash_stats *s = &b->stats;
      if (b->valid)
	{
	  len += sprintf(buf + len,
			 "%4d  "
			 "%6ld  %10ld  %4ld  "
			 "%6ld  %10ld  %4ld\n",
			 b->flMemInfo.stMemBnkNb,
			 s->read_sys,  s->read_bytes,  s->read_err,
			 s->write_sys, s->write_bytes, s->write_err);
	}
    }

  /* report per-bank interrupt info */
  len += sprintf(buf + len, 
		 "\n"
		 "bank  "
		 "msc   gpiomask   irqmask  "
		 "ry0  ry1  "
                 "   intrs  timeouts\n");
  for (i = 0; i < FLASH_NBANK; i++)
    {
      flash_bank *b = &f->bank[i];
      if (b->valid)
	len += sprintf(buf + len,
		       "%4d  "
		       "%4x  %8x  %8x  %3d  %3d  "
		       "%8ld  %8ld\n",
		       b->flMemInfo.stMemBnkNb,
		       b->msc_config, b->gpio_mask, b->irq_mask,
		       b->id_gpio_ready[0], b->id_gpio_ready[1],
                       b->interrupts, b->timeouts);
    }

  /* report aggregate information */
  len += sprintf(buf + len, 
		 "\n"
		 "bank  gpiomask   irqmask     intrs  timeouts\n");
  len += sprintf(buf + len, 
		 "%4s  "
		 "%8x  %8x  "
		 "%8ld  %8ld\n",
		 "all",
		 f->gpio_mask, f->irq_mask,
		 f->interrupts, f->timeouts);

  return(len);
}
#endif	/* CONFIG_PROC_FS */


FlMemInfo* flash_getinfo_prim(int banknum)
{
  /* Return info about the specified bank, or NULL if
     bank is invalid or mirrors another bank */
  if (!flash_global.bank[banknum].valid || flash_global.bank[banknum].mirror) {
    return NULL;
  }
  return &flash_global.bank[banknum].flMemInfo;
}

/*
 * interrupt routines 
 *
 */

static inline int flash_gpio_to_irq(int n)
{
  /*
   * modifies: nothing
   * effects:  Returns the IRQ interrupt line for detecting
   *           edge transitions on GPIO n.
   *
   */

  /* SA1100 GPIOs [00..10] have their own IRQs */
  if (n < 11)
    return(n);

  /* SA1100 GPIOs [11..27] share a single IRQ */
  return(11);
}

static void flash_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  /*
   * modifies: flash_global (global)
   * effects:  Wakes up all processes waiting on the specified irq
   *           for a signal that a flash operation has completed.
   *           Calls external handler if registered.
   */

  flash_state *f = &flash_global;
  int i;
  u32 detect;

  /* debugging */
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": flash_interrupt: irq=%d, enabled=%lx\n",
	   irq, get_enabled_irqs());

  /* sanity check */
  if ((f->irq_mask & (1 << irq)) == 0) {
    /* log message and fail */
    printk(FLASH_NAME ": flash_interrupt: unexpected irq=%d\n", irq);
    return;
  }

  /* update stats */
  f->interrupts++;

  /* which gpio(s) interrupted? */
  detect = *(f->gpio_detect);

  /* reset GPIO edge detect flag(s) that we are handling */
  /* Note: the sa1100 resets the bits that we write 1s to and leaves
     the others alone.  This may seem odd at first glance, but it
     avoids the race condition that would exist if we could just store
     to the register. */
  *(f->gpio_detect) = (detect & f->gpio_mask);

  /* wake up waiters for relevant banks */
  for (i = 0; i < FLASH_NBANK; i++) {
    /* convenient abbrev */
    flash_bank *b = &f->bank[i];

    if (b->valid && (b->irq_mask & (1 << irq)) && (detect & b->gpio_mask)) {
      /* debugging */
      if (FLASH_DEBUG_VERBOSE)
	printk(FLASH_NAME
	       ": flash_interrupt: irq=%d, bank=%d, detect=%x"
	       ", bmask=%x, fmask=%x\n", irq, b->flMemInfo.stMemBnkNb,
	       detect, b->gpio_mask, f->gpio_mask);
	  
      /* update stats */
      b->interrupts++;

      /* call external handler, if any */
      if (b->external_interrupt) {
	b->external_interrupt(b->flMemInfo.stMemBnkNb);
      }
    }
  }
}

static inline int flash_bank_contains(const flash_bank *b, u32 vaddr)
{
  /*
   * modifies: nothing
   * effects:  Returns TRUE iff vaddr is mapped into the specified
   *           flash memory bank.
   *
   */

  /* invalid bank contains nothing */
  if (!b->valid)
    return(0);

  return(((vaddr >= b->vaddr) &&
	  (vaddr <= (b->vaddr + b->flMemInfo.size))));
}

static void flash_update_gpio(void)
{
  /* effects:  Sets the gpio edge detectors to deliver interrupts on
   *           busy to ready transitions for just the banks that are
   *           currently marked as armed.
   */
  flash_state *f = &flash_global;
  int i;
  u32 gpio_mask = 0, old, new;
  unsigned long flags;

  for (i = 0; i < FLASH_NBANK; i++) {
    flash_bank *b = &f->bank[i];
    if (b->interrupt_armed) {
      gpio_mask |= b->gpio_mask;
    }
  }

  save_flags(flags);
  cli();
  old = *(f->gpio_rising);
  *(f->gpio_rising) = new = (old & ~f->gpio_mask) | gpio_mask;
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": flash_update_gpio: old=%x new=%x\n", old, new);
  restore_flags(flags);
}

static void flash_arm_interrupt(flash_bank *b)
{
  /* modifies: b
   * effects:  Sets gpio edge detector to generate an interrupt
   *           whenever bank b transitions from busy to ready state.
   *           Arm/disarm requests can be nested, and we also deal with
   *           the case of multiple banks sharing the same gpio, though
   *           neither of these features should currently be needed.
   */
  b->interrupt_armed++;
  if (b->interrupt_armed < 0) {
    printk(FLASH_NAME ": flash_arm_interrupt counter overflow!\n");
  }
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": flash_arm_interrupt: bank=%d, new count=%d\n",
	   b->flMemInfo.stMemBnkNb, b->interrupt_armed);
  flash_update_gpio();
}


static void flash_disarm_interrupt(flash_bank *b)
{
  /* modifies: b
   * effects:  Undo flash_arm_interrupt, disarming the interrupt if the
   *           count reaches zero.
   */
  b->interrupt_armed--;
  if (b->interrupt_armed < 0) {
    printk(FLASH_NAME ": flash_disarm_interrupt counter underflow!\n");
  }
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": flash_disarm_interrupt: bank=%d, new count=%d\n",
	   b->flMemInfo.stMemBnkNb, b->interrupt_armed);
  flash_update_gpio();
}

void flash_reg_interrupt(int banknum, void (*interrupt)(int))
{
  /* Register an external interrupt handler for the given bank */
  flash_bank *b = &flash_global.bank[banknum];
  b->external_interrupt = interrupt;
  flash_arm_interrupt(b);
}

void flash_unreg_interrupt(int banknum)
{
  /* Unregister an external interrupt handler for the given bank */
  flash_bank *b = &flash_global.bank[banknum];
  flash_disarm_interrupt(b);
  b->external_interrupt = NULL;
}

#ifdef CONFIG_POWERMGR
void flash_reg_powermgr(int banknum, void (*suspend)(int), void (*resume)(int))
{
  /* Register suspend and resume handlers for the given bank */
  flash_bank *b = &flash_global.bank[banknum];
  b->external_suspend = suspend;
  b->external_resume = resume;
}

void flash_unreg_powermgr(int banknum)
{
  /* Unegister suspend and resume handlers for the given bank */
  flash_bank *b = &flash_global.bank[banknum];
  b->external_suspend = NULL;
  b->external_resume = NULL;
}
#endif

static int flash_configure_interrupts(flash_state *f)
{
  /*
   * modifies: f
   * effects:  Set up interrupt handler for all irqs associated with
   *           valid flash memory banks in f.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  u32 detect, irq_mask;
  unsigned long flags;
  int i;

  /* request all appropriate interrupt lines */  
  irq_mask = 0;
  for (i = 0; i < 32; i++)
    if (f->irq_mask & (1 << i))
      {
	int err;

	/* request single interrupt line */
	err = request_irq(i,
			  flash_interrupt,
			  SA_INTERRUPT|SA_SHIRQ, 
			  FLASH_NAME "-ready",
			  &flash_global);	

	/* fail if unable to acquire line */
	if (err != 0)
	  {
	    int j;

	    /* log warning message */
	    if (FLASH_DEBUG)
	      printk(FLASH_NAME 
		     ": flash_configure_interrupts failed (irq %d): %d\n",
		     i,
		     err);

	    /* free acquired lines */
	    for (j = 0; j < i; j++)
	      if (irq_mask & (1 << j))
		free_irq(j, &flash_global);
	    
	    return(err);
	  }

	/* keep track of acquired lines */
	irq_mask |= (1 << i);
      }

  /* Initially turn off edge detectors for all our gpio lines,
     effectively disabling interrupts. */
  save_flags(flags);
  cli();
  detect = *(f->gpio_rising);
  *(f->gpio_rising) = detect & (~f->gpio_mask);
  detect = *(f->gpio_falling);
  *(f->gpio_falling) = detect & (~f->gpio_mask);
  restore_flags(flags);

  /* debugging */
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": flash_configure_interrupts: "
	   "l=%x, r=%x, f=%x, d=%x, p=%x, i=%lx\n",
	   *(f->gpio_levels),
	   *(f->gpio_rising),
	   *(f->gpio_falling),
	   *(f->gpio_detect),
	   *(f->gpio_pin_dir),
	   get_enabled_irqs());

  /* everything OK */
  return(0);
}

#ifdef MODULE
static void flash_deconfigure_interrupts(flash_state *f)
{
  /*
   * modifies: f
   * effects:  Removes interrupt handlers for all irqs associated with
   *           valid flash memory banks in f.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  int i;

  /* reset GPIO registers */
  *(f->gpio_rising)  &= ~f->gpio_mask;
  *(f->gpio_falling) &= ~f->gpio_mask;

  /* free interrupt lines */
  for (i = 0; i < 32; i++)
    if (f->irq_mask & (1 << i))
      free_irq(i, &flash_global);
}
#endif MODULE


static int flash_bank_init(flash_state *f, flash_bank *b, int n)
{
  /*
   * modifies: b
   * effects:  Initializes bank b by querying the appropriate static
   *           memory configuration registers and non-volatile memory
   *           identification structures for static memory bank n.
   *           Returns zero iff successful, otherwise returns error code.
   *
   */

  extern u32 staticmem_info_pointers[4];

  u16 *addr, data;
  u8 ready;
  int err;

  /* initialize: default to invalid */
  memset((void *) b, 0, sizeof(flash_bank));

  /* lookup bank-specific configuration */
  switch (n) {
  case 0:
    b->paddr = FLASH0;
    b->msc_config = *(f->mem_config_0) & 0xffff;
    break;
  case 1:
    b->paddr = FLASH1;
    b->msc_config = *(f->mem_config_0) >> 16;
    break;
  case 2:
    b->paddr = FLASH2;
    b->msc_config = *(f->mem_config_1) & 0xffff;
    break;
  case 3:
    b->paddr = FLASH3;
    b->msc_config = *(f->mem_config_1) >> 16;
    break;
  default:
    /* invalid bank number */
    return(-EINVAL);
  }

  /* monitor sets config for missing banks to MSC_INVALID */
  if (b->msc_config == MSC_INVALID)
    return(-EINVAL);
    
  /* valid bank */
  b->valid = 1;
  
  /* map flash memory bank */
  b->vaddr = allocate_vspace(b->paddr, FLASH_BANK_MAX_SIZE, 1, 0);
  if (b->vaddr == 0)
    {
      /* log error and fail */
      printk(FLASH_NAME ": unable to map flash bank %d\n", n);
      return(-ENOMEM);
    }

  /* lookup static memory id address */
  b->id_paddr = staticmem_info_pointers[n];

  /* sanity check (expect id area at end of bank) */
  if (b->id_paddr < b->paddr)
    {
      /* log error and fail */
      printk(FLASH_NAME ": invalid id area for bank %d (%x < %x)\n",
	     n, b->id_paddr, b->paddr);
      return(-EINVAL);
    }
  b->id_vaddr = b->vaddr + (b->id_paddr - b->paddr);

  /* read core id structure */
  addr = (u16 *) (b->id_vaddr + FLASH_ID_CORE);
  b->id_core = *addr;

  /* parse core id structure */
  b->read_only = (b->id_core >>  8) & 1;
  b->dcard     = (b->id_core >>  9) & 1;
  /*b->width_16  = (b->id_core >> 10) & 1;  **Now comes from flMemInfo */
  /*b->size      = (1 << (b->id_core >> 11));  **Now comes from flMemInfo */

  /* identify chip type in use */
  err = FlMemIdentify(n, b->vaddr, &b->flMemInfo);
  if (err) {
    printk(FLASH_NAME ": error in FlMemIdentify: %d\n", err);
  }

  /* debugging */
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": flash_bank_init: "
	   "n=%d, id_paddr=%x, id_vaddr=%x, id_core=%x, size=%x\n",
	   n, b->id_paddr, b->id_vaddr, b->id_core, b->flMemInfo.size);
  
  /* initialize gpio, irq masks */
  b->gpio_mask = 0;
  b->irq_mask  = 0;

  /* read, parse gpio0 info */
  addr = (u16 *) (b->id_vaddr + FLASH_ID_GPIO0);
  data = *addr;
  ready = data >> 8;
  b->id_gpio_ready[0] = ready;
  if (ready != FLASH_ID_GPIO_INVALID)
    {
      b->gpio_mask |= (1 << ready);
      b->irq_mask  |= (1 << flash_gpio_to_irq(ready));
    }

  /* read, parse gpio1 info (skip if 16-bit memory) */
  if (b->flMemInfo.bitsPerWord == 32)
    {
      addr = (u16 *) (b->id_vaddr + FLASH_ID_GPIO1);
      data = *addr;
      ready = data >> 8;
      b->id_gpio_ready[1] = ready;
      if (ready != FLASH_ID_GPIO_INVALID)
	{
	  b->gpio_mask |= (1 << ready);
	  b->irq_mask  |= (1 << flash_gpio_to_irq(ready));
	}
    }

  /* update gpio, irq masks */
  f->gpio_mask |= b->gpio_mask;
  f->irq_mask  |= b->irq_mask;

  /* everything OK */
  return(0);
}

static int flash_bank_identical(const flash_bank *a, const flash_bank *b)
{
  /*
   * modifies: nothing
   * effects:  Attempts to detect low-level flash "mirroring" of banks.
   *           Returns TRUE iff flash banks a and b have identical interrupt 
   *           masks and identical contents.  
   *
   */

  /* quick test: different if use different interrupts */
  if ((a->irq_mask != b->irq_mask) || (a->gpio_mask != b->gpio_mask))
    return(0);

  /* quick test: different if have different sizes */
  if (a->flMemInfo.size != b->flMemInfo.size)
    return(0);

  /* slow test: compare contents */
  if (memcmp((const void *) a->vaddr, (const void *) b->vaddr,
	     a->flMemInfo.size) != 0)
    return(0);

  /* banks are identical */
  return(1);
}

static int flash_check_mirror(flash_state *f, int bank)
{
  /*
   * modifies: f
   * effects:  Checks if specified flash bank mirror another flash bank in f.
   *           Updates f to reflect mirroring, if any.  Returns 0 iff
   *           successful, otherwise error code.
   *
   */

  int i, nmirror;

  nmirror = 0;
  for (i = 0; i < FLASH_NBANK; i++)
    if (i != bank)
      if (flash_bank_identical(&f->bank[bank], &f->bank[i]))
	{
	  /* debugging */
	  if (FLASH_DEBUG)
	    printk(FLASH_NAME ": bank %d mirrors bank %d\n", bank, i);

	  /* update bank to reflect mirroring */
	  f->bank[bank].mirror = 1;
	  f->bank[bank].mirror_bank = i;

	  /* count mirrors */
	  nmirror++;
	}

  /* fail if bank has multiple mirrors */
  if (nmirror > 1)
    return(-EINVAL);

  /* everything OK */
  return(0);
}

static char *flash_version(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current cvs revision number.
   * storage:  Return value is statically-allocated.
   *
   */

  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = FLASH_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return(version);
}

#ifdef  CONFIG_MODULES
static struct symbol_table flash_syms = {
#include <linux/symtab_begin.h>
        X(flash_reg_interrupt),
        X(flash_unreg_interrupt),
#ifdef CONFIG_POWERMGR
        X(flash_reg_powermgr),
        X(flash_unreg_powermgr),
#endif
	X(flash_getinfo_prim),
	X(FlMemGetError), /* !!from flinfo.h */
#include <linux/symtab_end.h>
};
#endif  /* CONFIG_MODULES */


/* Power manager operations */

#ifdef CONFIG_POWERMGR
static int
flash_suspend_check(void *ignore, int idle_jiffies)
{
  /*
   * effects:  Always returns zero; i.e. willing to sleep immediately.
   *
   */

  return 0;
}

static int
flash_suspend(void *ignore)
{
  int i;
  /*
   * effects:  Saves state in preparation for sleep mode.
   *           Always returns zero (ignored).
   *
   */

  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": itsy_flash_suspend\n");

  for (i = 0; i < FLASH_NBANK; i++) {
    flash_bank *b = &flash_global.bank[i];
    if (b->valid && b->external_suspend) {
      b->external_suspend(i);
    }
  }

  /* return value ignored */
  return 0;
}

static int
flash_resume(void *ignore, int resume_flags)
{
  int i;
  flash_state *f = &flash_global;
  u32 old;
  /*
   * effects:  Restores state after wakeup from sleep mode.
   *           Always returns zero (ignored).
   *
   */

  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": itsy_flash_resume\n");

  /* Re-enable interrupts */
  for (i = 0; i < 32; i++) {
    if (f->irq_mask & (1 << i)) {
      if (FLASH_DEBUG_VERBOSE)
	printk(FLASH_NAME ": enable_irq(%d)\n", i);
      enable_irq(i);
    }
  }

  for (i = 0; i < FLASH_NBANK; i++) {
    flash_bank *b = &f->bank[i];
    if (b->valid && b->external_resume) {
      b->external_resume(i);
    }
  }

  /* return value ignored */
  return 0;
}

static const powermgr_client flash_powermgr = {
  /* callback functions */
  flash_suspend_check,
  flash_suspend,
  flash_resume,
  
  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  FLASH_NAME_VERBOSE,

  /* power-consumption info */
  0,
  0
};
#endif	/* CONFIG_POWERMGR */

int itsy_flash_init(void)
{
  flash_state *f = &flash_global;
  int err, i;
  static int init_done = 0;

  if (init_done) return 0;
  init_done = 1;

#ifdef  CONFIG_MODULES
/* export symbols */
(void) register_symtab(&flash_syms);
#endif  /* CONFIG_MODULES */

  /* verbose logging */
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": itsy_flash_init\n");

  /* initialize */
  memset((void *) f, 0, sizeof(flash_state));

  /* memory-mapped registers */
  f->mem_config_0 = (volatile u32 *) MSC0_V;
  f->mem_config_1 = (volatile u32 *) MSC1_V;
  f->gpio_levels  = (volatile u32 *) GPLR_V;
  f->gpio_rising  = (volatile u32 *) GRER_V;
  f->gpio_falling = (volatile u32 *) GFER_V;
  f->gpio_detect  = (volatile u32 *) GEDR_V;
  f->gpio_pin_dir = (volatile u32 *) GPDR_V;

  /* initialize per-bank data */
  for (i = 0; i < FLASH_NBANK; i++)
    flash_bank_init(f, &f->bank[i], i);

  /* detect mirrored bank 0 */
  if (FLASH_CHECK_MIRROR)
    if ((err = flash_check_mirror(f, 0)) != 0)
      {
	/* log error and fail */
	printk(FLASH_NAME ": invalid flash bank mirror\n");
	return(err);
      }

#ifdef	CONFIG_POWERMGR
  /* initialize powermgr state */
  f->pm_id = POWERMGR_ID_INVALID;
#endif	/* CONFIG_POWERMGR */

  /* enable interrupts */
  if ((err = flash_configure_interrupts(f)) != 0)
    {
      /* log error and fail */
      printk(FLASH_NAME ": unable to configure interrupts\n");
      return(err);
    }    

#ifdef	CONFIG_POWERMGR
  /* register with power manager */
  flash_global.pm_id = powermgr_register(&flash_powermgr);
  if (flash_global.pm_id == POWERMGR_ID_INVALID)
    if (FLASH_DEBUG)
      printk(FLASH_NAME ": unable to register with power manager\n");
#endif	/* CONFIG_POWERMGR */

#ifdef	CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &flash_proc_entry);
#endif	/* CONFIG_PROC_FS */

  /* log device registration */
  printk(FLASH_NAME_VERBOSE " version %s initialized\n", flash_version());

  /* everything OK */
  return(0);
}

#ifdef	MODULE
int init_module(void)
{
  if (FLASH_DEBUG_VERBOSE)
    printk(FLASH_NAME ": init_module\n");

  return(itsy_flash_init());
}

void cleanup_module(void)
{
  flash_state *f = &flash_global;
  int err;

  /* deconfigure flash interrupts */
  flash_deconfigure_interrupts(f);

#ifdef	CONFIG_POWERMGR
  /* unregister with power manager */
  if (powermgr_unregister(flash_global.pm_id) != POWERMGR_SUCCESS)
    printk(FLASH_NAME ": unable to unregister with power manager\n");
#endif	/* CONFIG_POWERMGR */

#ifdef	CONFIG_PROC_FS
  /* unregister procfs entry */
  if ((err = proc_unregister(&proc_root, flash_proc_entry.low_ino)) != 0)
    if (FLASH_DEBUG)
      printk(FLASH_NAME ": cleanup_module: proc_unregister(flash) %d\n", err);
#endif	/* CONFIG_PROC_FS */

  /* log device unload */
  printk(FLASH_NAME_VERBOSE " unloaded\n");
}
#endif	/* MODULE */
