/*
 * SA1100 GPIO Driver
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * Author: Jason Flinn
 * 
 * This interfaces with the "unused" pins of the SA1100 GPIO.  These can
 * be used for various purposes, such as supplementing profile information.
 *
 * Note: This shares the GPIO 11-27 interrupt with rocknscroll, which
 * should not be a problem.  
 *
 * Implementation borrowed from clkspeed.c.
 *
 */

#include <linux/config.h>

#ifdef	MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/powermgr.h>

#include <asm/io_reg.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/arch/mmu.h>
#include <asm/arch/mmap.h>
#include <linux/miscdevice.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#include <asm/arch/SA-1100.h>

#include <linux/gpio-sa1100.h>

/*
 * debugging 
 *
 */

#define	GPIO_DEBUG			(1)
#define	GPIO_DEBUG_VERBOSE		(0)

/* names */
#define	GPIO_NAME			"GPIO-itsy"
#define	GPIO_NAME_VERBOSE		"itsy gpio pin driver"
#define	GPIO_VERSION_STRING		"$Revision: 2.4 $"

static int uniqueid;
static struct wait_queue* waitq = NULL;

/* Sets the direction of GPIO pins 11-13 */
static int
gpio_set_pin_direction (struct gpio_t* pdirs)
{
    uint gpdr = inl(GPDR_V);

    if (pdirs->gpio_11 == GPIO_DIR_INPUT) {
	gpdr &= 0xfffff7ff;
    } else if (pdirs->gpio_11 == GPIO_DIR_OUTPUT) {
	gpdr |= 0x00000800;
    }

    if (pdirs->gpio_12 == GPIO_DIR_INPUT) {
	gpdr &= 0xffffefff;
    } else if (pdirs->gpio_12 == GPIO_DIR_OUTPUT) {
	gpdr |= 0x00001000;
    }

    if (pdirs->gpio_13 == GPIO_DIR_INPUT) {
	gpdr &= 0xffffdfff;
    } else if (pdirs->gpio_13 == GPIO_DIR_OUTPUT) {
	gpdr |= 0x00002000;
    }

    outl(gpdr, GPDR_V);

    return (0);
}

/* Sets the values of GPIO pins 11-13.  An input value other than 0
or 1 results for a pin results in the value of that pin being unchanged */
static int
gpio_set_pin_values (struct gpio_t* pvals)
{
    if (pvals->gpio_11 == 1) {
	outl (0x0000800, GPSR_V);
    } else if (pvals->gpio_11 == 0) {
	outl (0x0000800, GPCR_V);
    }

    if (pvals->gpio_12 == 1) {
	outl (0x0001000, GPSR_V);
    } else if (pvals->gpio_12 == 0) {
	outl (0x0001000, GPCR_V);
    }

    if (pvals->gpio_13 == 1) {
	outl (0x0002000, GPSR_V);
    } else if (pvals->gpio_13 == 0) {
	outl (0x0002000, GPCR_V);
    }

    return (0);
}

/* Gets the values of GPIO pins 11-13. */
static int
gpio_get_pin_values (struct gpio_t* pvals)
{
    uint gplr = inl(GPLR_V);

    pvals->gpio_11 = (gplr & 0x0000800) >> 11;
    pvals->gpio_12 = (gplr & 0x0001000) >> 12;
    pvals->gpio_13 = (gplr & 0x0002000) >> 13;

    return (0);
}

/* Blocks until GPIO pins 11-13 match the specified values.  An input
value other than 0 or 1 specifies "don't care" for that pin. */
static int
gpio_wait_for_pin_values (struct gpio_t* pvals)
{
    struct gpio_t curvals;
    uint gfer,  grer;

    do {

	gpio_get_pin_values (&curvals);

	if (!((pvals->gpio_11 == 0 && curvals.gpio_11 == 1) ||
	      (pvals->gpio_11 == 1 && curvals.gpio_11 == 0) ||
	      (pvals->gpio_12 == 0 && curvals.gpio_12 == 1) ||
	      (pvals->gpio_12 == 1 && curvals.gpio_12 == 0) ||
	      (pvals->gpio_13 == 0 && curvals.gpio_13 == 1) ||
	      (pvals->gpio_13 == 1 && curvals.gpio_13 == 0))) {
	    
	    /* All pins that user cares about match specified values */
	    break;
	}
	      
	      /* Enable edge-detection for pins that we care about */
	      gfer = inl(GFER_V);
	      grer = inl(GRER_V);
	      
	      if (pvals->gpio_11 == 0 || pvals->gpio_11 == 1) {
		  gfer |= 0x00000800;
		  grer |= 0x00000800;
	      } 
	      if (pvals->gpio_12 == 0 || pvals->gpio_12 == 1) {
		  gfer |= 0x00001000;
		  grer |= 0x00001000;
	      } 
	      if (pvals->gpio_13 == 0 || pvals->gpio_13 == 1) {
		  gfer |= 0x00002000;
		  grer |= 0x00002000;
	      } 
	      
	      outl(gfer, GFER_V);
	      outl(grer, GRER_V);
	      
	      interruptible_sleep_on (&waitq);

    } while (1);

    return (0);
}

static int 
gpio_ioctl (struct inode *inode,
	    struct file *file,
	    uint command,
	    ulong arg)
{
    if (GPIO_DEBUG_VERBOSE)
	printk(GPIO_NAME ": gpio_ioctl: command=%d, arg=%lu\n", command, arg);
    
    /* dispatch based on command */
    switch (command) {
        case GPIO_SET_PIN_DIRECTION: {
	    return (gpio_set_pin_direction ((struct gpio_t *) arg));
	}
        case GPIO_SET_PIN_VALUES: {
	    return (gpio_set_pin_values ((struct gpio_t *) arg));
	}
        case GPIO_GET_PIN_VALUES: {
	    return (gpio_get_pin_values ((struct gpio_t *) arg));
	}
        case GPIO_WAIT_FOR_PIN_VALUES: {
	    return (gpio_wait_for_pin_values ((struct gpio_t *) arg));
	}
    }

    /* invalid command */
    return(-EINVAL);
}

/* When a process is waiting for certain values on GPIO 11-13, it
   waits on a wait queue.  This interrupt handler wakes up any 
   waiters every time GPIO pin values are changed.  The waiting
   process is responsible for verifying whether or not the new
   pin values match the desired values */

static void 
gpio_interrupt (int irq, void *dev_id, struct pt_regs *regs ) {

    /* Clear edge-detect register for pins 11-13 */
    outl (0x00003800, GEDR_V);

   /* Wake up any waiters */
   wake_up_interruptible( &waitq );
}

static int 
gpio_open (struct inode *inode,
	   struct file *file)
{
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

static void 
gpio_close(struct inode *inode, 
	   struct file *file)
{
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
}

static struct file_operations gpio_fops =
{
    NULL,			/* no special lseek */
    NULL,			/* no special read */
    NULL,			/* no special write */
    NULL,			/* no special readdir */
    NULL,			/* no special select */
    gpio_ioctl,
    NULL,			/* no special mmap */
    gpio_open,
    gpio_close,
    NULL,			/* no special fsync */
    NULL,			/* no special fasync */
    NULL,			/* no special check_media_change */
    NULL			/* no special revalidate */
};

static struct miscdevice gpio_misc = {
    SA1100_GPIO_MINOR, "gpio", &gpio_fops, NULL, NULL
};

int gpio_sa1100_init(void)
{
    if (GPIO_DEBUG_VERBOSE)
	printk(GPIO_NAME ": gpio_sa1100_init\n");
    
    /* register device */
    misc_register(&gpio_misc);
    
    /* log device registration */
    printk(GPIO_NAME_VERBOSE " initialized\n");
    if (GPIO_DEBUG_VERBOSE)
	printk (GPIO_NAME ": %s\n", GPIO_VERSION_STRING);

    /* everything OK */
    return(0);
}

#ifdef	MODULE
int init_module(void)
{
    if (GPIO_DEBUG_VERBOSE)
	printk(GPIO_NAME ": init_module\n");
    
    /* Get the IRQ associated with edge detection for GPIOs 11-27 */
    if (request_irq (IRQ_GPIO11_27, gpio_interrupt, 
		     SA_INTERRUPT | SA_SHIRQ, GPIO_NAME, 
		     &uniqueid) != 0) {
	printk( "GPIO: unable to allocate IRQ_GPIO11_27\n" );
	return -1;
    }

    return (gpio_sa1100_init());
}

void cleanup_module(void)
{
    /* unregister driver */
    misc_deregister(&gpio_misc);

    /* Release shared interrupt */
    free_irq (IRQ_GPIO11_27, &uniqueid);

    /* log device unload */
    printk (GPIO_NAME_VERBOSE " unloaded\n");
}
#endif	/* MODULE */


