/*
 * DS2437 Smart Battery Monitor Commands
 * Copyright (c) Compaq Computer Corporation, November 1999
 * Platform: Itsy V2.0 with SA-1100 processor
 * Author: Jennifer Anderson
 *
 * Issue commands to DS2437 Smart Battery Monitor.
 *		
 * $Log: itsy_battmon.c,v $
 * Revision 2.5  2000/01/19  00:10:28  brakmo
 * Added ioctls to control charging current.
 *
 * Revision 2.4  1999/12/23 01:27:54  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.3  1999/11/24  00:04:05  jennifer
 * Added support for powermgr
 *
 * Revision 2.2  1999/11/23  19:38:03  jennifer
 * Only open device if itsy_major_version >= 2.
 *
 * Revision 2.1  1999/11/23  19:15:38  jennifer
 * Smart battery monitor driver (battmon).
 *
 */

#include <linux/config.h>

#ifdef	MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef CONFIG_POWERMGR
#include <linux/powermgr.h>
#endif

#include <linux/kernel.h>
#include <linux/timer.h>
#include <asm/arch/io_reg.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <linux/itsy_version.h>

#include <asm/arch/Itsy.h>	/* brakmo */

#define BATTMON_DRIVER
#include <linux/itsy_battmon.h>

#define BATTMON_MAJOR                    46

/*
 * debugging 
 *
 */
#define	BATTMON_DEBUG			(0) 
#define	BATTMON_DEBUG_VERBOSE		(0) 

/* names */
#define	BATTMON_NAME			"battery-monitor"
#define	BATTMON_NAME_VERBOSE		"Itsy battery monitor driver"
#define	BATTMON_VERSION_STRING		"$Revision: 2.5 $"


#define MIN(a,b)          ((a) < (b) ? (a) : (b))
#define READ_CLOCK_SPEED  (PPCR_MASK & *(int *)PPCR_V)

#define FALSE             0
#define TRUE              1

/* -----------------------------------------------------------------
   External Procedure Declarations
   ----------------------------------------------------------------- */

/* From battmon_io.S */
extern int battmon_send_reset(int);
extern void battmon_write_one(int);
extern void battmon_write_zero(int);
extern int battmon_read_bit(int);

/* -----------------------------------------------------------------
   Private Type Declarations
   ----------------------------------------------------------------- */

typedef int Boolean;

/* -----------------------------------------------------------------
   Private Variable Declarations
   ----------------------------------------------------------------- */

/* This table is used to verify the DS2437's CRC.  This table was
 * copied from Dallas Semiconductor's Application Note #27: "Understanding
 * and Using Cyclic Redundancy Checks with Dallas Seminconductor iButton
 * Products."
 */

Byte crc_table[256] = {
    0, 94, 188, 226, 97, 63, 221, 131, 194, 156, 126, 32, 163, 253, 31,
    65, 157, 195, 33, 127, 252, 162, 64, 30, 95, 1, 227, 189, 62, 96, 130,
    220, 35, 125, 159, 193, 66, 28, 254, 160, 225, 191, 93, 3, 128, 222,
    60, 98, 190, 224, 2, 92, 223, 129, 99, 61, 124, 34, 192, 158, 29, 67,
    161, 255, 70, 24, 250, 164, 39, 121, 155, 197, 132, 218, 56, 102, 229,
    187, 89, 7, 219, 133, 103, 57, 186, 228, 6, 88, 25, 71, 165, 251, 120,
    38, 196, 154, 101, 59, 217, 135, 4, 90, 184, 230, 167, 249, 27, 69,
    198, 152, 122, 36, 248, 166, 68, 26, 153, 199, 37, 123, 58, 100, 134,
    216, 91, 5, 231, 185, 140, 210, 48, 110, 237, 179, 81, 15, 78, 16,
    242, 172, 47, 113, 147, 205, 17, 79, 173, 243, 112, 46, 204, 146, 211,
    141, 111, 49, 178, 236, 14, 80, 175, 241, 19, 77, 206, 144, 114, 44,
    109, 51, 209, 143, 12, 82, 176, 238, 50, 108, 142, 208, 83, 13, 239,
    177, 240, 174, 76, 18, 145, 207, 45, 115, 202, 148, 118, 40, 171, 245,
    23, 73, 8, 86, 180, 234, 105, 55, 213, 139, 87, 9, 235, 181, 54, 104,
    138, 212, 149, 203, 41, 119, 244, 170, 72, 22, 233, 183, 85, 11, 136,
    214, 52, 106, 43, 117, 151, 201, 74, 20, 246, 168, 116, 42, 200, 150,
    21, 75, 169, 247, 182, 232, 10, 84, 215, 137, 107, 53
};

/* Reading voltage and temperature are implemented as two-step operations,
 * first we issue a command to initiate the voltage/temperature conversion
 * and then issue another command to read the value.  The following
 * variables are used to put the process to sleep with a timer while 
 * the voltage/temperature conversion takes place.  
 */
struct wait_queue *temp_waitq = NULL;
struct wait_queue *voltage_waitq = NULL;
struct timer_list temp_timer;
struct timer_list voltage_timer;
static int get_temp_count = 0;
static int get_voltage_count = 0;

/* Used for locking to ensure that only one
 * command is being issued to the battery monitor at a time 
 */
static unsigned long command_lock = 0;
struct wait_queue *command_waitq = NULL;

#ifdef CONFIG_PROC_FS
static Word64 device_id = 0;
#endif /* CONFIG_PROC_FS */

#ifdef	CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
#endif


/* -----------------------------------------------------------------
   Private Procedure Declarations
   ----------------------------------------------------------------- */

/* 
 * Routines exported via ioctl 
 */

static int battmon_detect(Word64 *rom_code);

static int battmon_get_temp(int *temp);
static int battmon_get_voltage(int *voltage);
static int battmon_get_vad_voltage(int *voltage);

static int battmon_read_current(int *curr);
static int battmon_read_net_current(Byte *curr);
static int battmon_write_net_current(Byte *curr);
static int battmon_read_charging_current(Word16 *curr);
static int battmon_write_charging_current(Word16 *curr);
static int battmon_read_discharging_current(Word16 *curr);
static int battmon_write_discharging_current(Word16 *curr);

static int battmon_read_RTC(Word32 *timestamp);
static int battmon_write_RTC(Word32 *timestamp);
static int battmon_read_disconnect_time(Word32 *timestamp);
static int battmon_read_EOC_time(Word32 *timestamp);


/*
 * Helper routines
 */

static Boolean battmon_configure(Boolean enable_ICA, 
				 Boolean enable_CCA_DCA, 
				 Boolean shadow_CCA_DCA, 
				 Boolean use_input_voltage);
/* Read/write  */
static Byte battmon_read_byte(void);
static void battmon_write_byte(Byte data);
static Boolean battmon_read_page(int page, int bytes, Byte *data);
static Boolean battmon_write_page(int page, int bytes, Byte *data, 
				  Boolean busy_wait);
#if 0
static Boolean battmon_write_page_busy(void);
#endif

/* Initiating and reading temperature and voltage */
static Boolean battmon_fill_temp(void);
static Boolean battmon_fill_temp_busy(void);
static int battmon_read_temp(int *temp);
static Boolean battmon_fill_voltage(void);
static Boolean battmon_fill_voltage_busy(void);
static int battmon_read_voltage(int *voltage);
static void timer_timedout(unsigned long ptr);

/* Translate to/from DS2437 data formats */
static int battmon_translate_temp(Byte lsb, Byte msb);
static int battmon_translate_voltage(Byte lsb, Byte msb);
static int battmon_translate_current(Byte lsb, Byte msb);
static int battmon_translate_CCA_DCA(Byte lsb, Byte msb);
static unsigned int battmon_translate_timestamp(Byte *data);


/*
 * Routines that issue commands to DS2437
 */

/* Reset */
static Boolean battmon_reset(void);

/* ROM commands */
static Word64 battmon_read_ROM(void);
static void battmon_skip_ROM(void);
#if 0
static void battmon_search_ROM(void);
static void battmon_match_ROM(Word64 rom_code);
#endif

/* Memory commands */
static void battmon_write_scratchpad(int page, int bytes, Byte *data);
static void battmon_read_scratchpad(int page,int bytes,Byte *data,Byte *crc);
static void battmon_copy_scratchpad(int page, Boolean busy_wait);
static void battmon_recall_memory(int page);

/* Register commands */
static void battmon_convert_temp(Boolean busy_wait);
static void battmon_convert_voltage(Boolean busy_wait);


/*
 * Device driver routines
 */

int battmon_init(void);
static int battmon_open(struct inode *inode, struct file *file);
static void battmon_close(struct inode *inode, struct file *file);
static int battmon_ioctl(struct inode *inode, struct file *file, uint command,
			 ulong arg);
static void free_command_lock(void);
static void get_command_lock(void);

#ifdef MODULE
void cleanup_module(void);
int init_module(void);
#endif 

#ifdef CONFIG_POWERMGR
static int battmon_suspend_check(void *ignore, int idle_jiffies);
static int battmon_suspend(void *ignore);
static int battmon_resume(void *ignore, int resume_flags);
#endif 

#ifdef CONFIG_PROC_FS
static int battmon_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry  battmon_proc_entry = {
   0, 7, "battmon", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, battmon_procfs_info,
};
#endif


/* -----------------------------------------------------------------
   Private Procedure Implementations
   ----------------------------------------------------------------- */

/*
 * PRIVATE PROCEDURE: battmon_detect
 *
 * Purpose
 *     Detect the DS2437 and get its 64-bit ROM code
 *
 * Input
 *     rom_code                  Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful; error code otherwise
 *
 * Modified
 *     rom_code                  Set to 64-bit ROM code
 *     
 */
static int
battmon_detect(Word64 *rom_code)
{
    Word64 val;


    if (!battmon_reset()) {
	return (-ENODEV);
    }

    val = battmon_read_ROM();
    if (val == 0) {
	return (-ENODEV);
    }

    memcpy_tofs((void *)rom_code, (const void *)&val, 8);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_configure
 *
 * Purpose
 *     Configure the DS2437 by writing to its Status/Configuration 
 *     register
 *
 * Input
 *     enable_ICA           Set the IAD bit.  Enables the ICA (Integrated 
 *                             Current Accumulator) and the current A/D 
 *                             converter, and takes current measurements 
 *                             at rate of 32 Hz.
 *     enable_CCA_DCA       Set the CA bit.  Enables CCA (Charging Current 
 *                             Accumulator) and DCA (Discharging Current 
 *                             Accumulator).
 *     shadow_CCA_DCA       Set the E2? bit.  Shadows the CCA and DCA counters
 *                            to EEPROM.  The registers are written each 
 *                            time the counter is incremented by .32C. 
 *     use_input_voltage    Set the AD bit.  Uses the battery input voltage
 *                            VDD as the input for the DS2437's voltage
 *                            A/D converter.  If not set, then VAD is used.
 * Output
 *     Returns TRUE if configuration was successful, FALSE otherwise
 *
 * Modified
 *     None
 *     
 */
static Boolean 
battmon_configure(Boolean enable_ICA, 
		  Boolean enable_CCA_DCA, 
		  Boolean shadow_CCA_DCA, 
		  Boolean use_input_voltage)
{
    Byte status = 0;
    Byte check;

    status = (enable_ICA) |
	     (enable_CCA_DCA << 1) |
	     (shadow_CCA_DCA << 2) |
	     (use_input_voltage << 3);

    if (!battmon_reset())  return FALSE;
    battmon_skip_ROM();
    battmon_write_scratchpad(0, 1, &status);

    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_read_scratchpad(0, 1, &check, NULL);

    /* MSb of status/configuration register is don't care */
    if ((status & 0x7F) != (check & 0x7F)) {
        return FALSE;
    }

    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_copy_scratchpad(0, TRUE);

    return TRUE;
}



/*
 * PRIVATE PROCEDURE: battmon_read_page
 *
 * Purpose
 *     This command reads the contents of a memory page on the DS2437.  
 *     The data read always begins at address 0 of the selected page.  
 *     Any reserved bits will all be read as logic 1's.  Valid page numbers 
 *     for reading are 0 - 7.  If 8 bytes are read, then the CRC can also 
 *     be read from the DS2437; otherwise the CRC is invalid.
 *
 * Input
 *     page                 Page number to read (0 - 7)
 *     bytes                Number of bytes to read (1 - 8 bytes)
 *     data                 Data read
 * 
 * Output
 *     Returns TRUE if page could be read; FALSE otherwise
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_read_page(int page, int bytes, Byte *data)
{
    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_recall_memory(page);

    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_read_scratchpad(page, bytes, data, NULL);
    
    return TRUE;
}



/*
 * PRIVATE PROCEDURE: battmon_write_page
 *
 * Purpose
 *     This command writes to a memory page on the DS2437.  The entire 
 *     8-byte page may be written, but all writing begins with the byte 
 *     present at address 0 of the selected scratchpad.  Valid page numbers 
 *     for writing are 0 - 7.  This routine first writes to the scratchpad
 *     read the data back to verify, and then copies the page to memory.
 *     If busy_wait is TRUE, then this routine will block until the copy
 *     to memory is complete.  If FALSE, then the routine returns after
 *     issuing the copy.  In the latter case, battmon_write_page_busy can 
 *     be called to check if the copy is complete. 
 *
 * Input
 *     page                 Page number to write (0 - 7)
 *     bytes                Number of bytes written (1 - 8 bytes)
 *     data                 Data to write
 *     busy_wait            Busy wait until write is complete
 * 
 * Output
 *     Returns TRUE if write succeeded, FALSE otherwise
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_write_page(int page, int bytes, Byte *data, Boolean busy_wait)
{
    int i;
    Byte check[8];

    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_write_scratchpad(page, bytes, data);

    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_read_scratchpad(page, bytes, check, NULL);

    for (i = 0; i < bytes; i++) {
	if (data[i] != check[i]) return FALSE;
    }

    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_copy_scratchpad(page, busy_wait);

    return TRUE;
}


#if 0
/*
 * PRIVATE PROCEDURE: battmon_write_page_busy
 *
 * Purpose
 *     Checks if a page write operation is in progress.  The
 *     NVB bit of the Status/Configuration is set while the write is
 *     in progress, and is cleared when the write is complete.  
 *
 * Input
 *     None
 * 
 * Output
 *     Returns TRUE is a write operation is in progress (by
 *     checking the NVB bit of the Status/Configuration register); returns
 *     FALSE otherwise
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_write_page_busy(void)
{
    Byte status;


    if (!battmon_reset()) return FALSE;

    battmon_skip_ROM();
    battmon_read_scratchpad(0, 1, &status, NULL);

    if (status & 0x20) return TRUE;

    return FALSE;
}
#endif


/*
 * PRIVATE PROCEDURE: battmon_fill_temp
 *
 * Purpose
 *     Initiate a temperature conversion.  The current temperature value
 *     is then placed in the DS2437's temperature registers.  The temperature 
 *     conversion may take up to 1 second.  
 *
 * Input
 *     None
 * 
 * Output
 *     TRUE if temperature conversion can be issued, FALSE otherwise.
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_fill_temp(void)
{
    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_convert_temp(FALSE);

    return TRUE;
}


/*
 * PRIVATE PROCEDURE: battmon_fill_temp_busy
 *
 * Purpose
 *     Checks if a temperature conversion operation is in progress.  The
 *     TB bit of the Status/Configuration is set while a temperature
 *     conversion is in progress, and is cleared when the conversion
 *     is complete.
 *
 * Input
 *     None
 * 
 * Output
 *     Returns TRUE if a temp conversion operation is in progress (by
 *     checking the TB bit of the Status/Configuration register); returns
 *     FALSE otherwise
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_fill_temp_busy(void)
{
    Byte status;


    if (!battmon_reset()) return FALSE;

    battmon_skip_ROM();
    battmon_read_scratchpad(0, 1, &status, NULL);

    if (status & 0x10) return TRUE;

    return FALSE;
}



/*
 * PRIVATE PROCEDURE: battmon_read_temp
 *
 * Purpose
 *     Read the value of the temperature registers and set "*temp" to
 *     a numerical temperature value.  The temperature returned is in
 *     units of 1/32 degrees C, i.e. 802 = 25.0625 C.  The valid range
 *     of temperatures is -55C to 125C in .03125 increments.  Returns
 *     0 if the read was successful; -EIO otherwise.
 *
 * Input
 *     temp                 Pointer to returned value
 * 
 * Output
 *      Returns 0 if successful; error code otherwise
 *
 * Modified
 *     temp                 Set to temperature value (units of 1/32 degrees C)
 *      
 */
static int
battmon_read_temp(int *temp)
{
    Byte data[3];


    if (!battmon_read_page(0, 3, data)) {
	return (-EIO);
    }
 
    *temp = battmon_translate_temp(data[1], data[2]);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_get_temp
 *
 * Purpose
 *     Initiate a temperature conversion, wait until it's complete,
 *     and then read the value of the temperature registers.
 *     The temperature returned is in units of 1/32 degrees C, 
 *     i.e. 802 = 25.0625 C.  The valid range of temperatures
 *     is -55C to 125C in .03125 increments.  Returns 0 if the read was
 *     successful; -EIO otherwise.
 *
 * Input
 *     temp                 Pointer to returned value
 * 
 * Output
 *      Returns 0 if successful; error code otherwise
 *
 * Modified
 *     temp                 Set to temperature value (units of 1/32 degrees C)
 *     
 */

static int 
battmon_get_temp(int *temp)
{
    int val;
    Boolean busy;


    if (get_temp_count > 0) {
	return (-EBUSY);
    }

    get_temp_count++;      

    /* Issue command to initiate temperature conversion */
    if (!battmon_fill_temp()) {
      get_temp_count--;
      return (-EIO);
    }

    /* Free command lock to allow other commands to issue
     * while current process waits for temperature conversion 
     */
    free_command_lock();

    /* Sleep while waiting for temperature conversion to complete.
     * Temp conversion can take up to 1 sec.  We wait in 0.5 second
     * increments before checking if the result is ready. 
     */
    while (TRUE) {
        init_timer(&temp_timer);
	temp_timer.function = timer_timedout;
	temp_timer.data = (unsigned long) &temp_waitq;
	temp_timer.expires = jiffies + (0.5 * HZ);
	add_timer(&temp_timer);
	interruptible_sleep_on(&temp_waitq);

	get_command_lock();
	busy = battmon_fill_temp_busy();
	free_command_lock();
	if (!busy) break;
    }
    
    /* Now read the temperature */
    get_command_lock();
    get_temp_count--;

    if (battmon_read_temp(&val) < 0) {
	return (-EIO);
    }
	
    put_user(val, temp);
    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_fill_voltage
 *
 * Purpose
 *     Initiate a voltage analog-to-digital conversion cycle.
 *     The current voltage value is then placed in the DS2437's voltage 
 *     registers.  The voltage conversion may take up to 2 ms. 
 *
 * Input
 *     None
 * 
 * Output
 *     TRUE if voltage conversion can be issued, FALSE otherwise.
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_fill_voltage(void)
{
    if (!battmon_reset()) return FALSE;
    battmon_skip_ROM();
    battmon_convert_voltage(FALSE);

    return TRUE;
}


/*
 * PRIVATE PROCEDURE: battmon_fill_voltage_busy
 *
 * Purpose
 *     Checks if a voltage conversion operation is in progress.  The
 *     ADB bit of the Status/Configuration is set while a voltage A/D
 *     conversion is in progress, and is cleared when the conversion
 *     is complete.
 *
 * Input
 *     None
 * 
 * Output
 *     Returns TRUE if a voltage conversion operation is in progress (by
 *     checking the ADB bit of the Status/Configuration register); returns
 *     FALSE otherwise
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_fill_voltage_busy(void)
{
    Byte status;


    if (!battmon_reset()) return FALSE;

    battmon_skip_ROM();
    battmon_read_scratchpad(0, 1, &status, NULL);

    if (status & 0x40) return TRUE;

    return FALSE;
}


/*
 * PRIVATE PROCEDURE: battmon_read_voltage
 *
 * Purpose
 *     Read the value of the voltage registers and set "*voltage" to a
 *     numerical voltage value.  The voltage returned is in units of
 *     10mV, i.e. 330 = 3.3V.  The valid range of voltages 0V to
 *     10.23V, in 10mV increments. Returns 0 if the read was
 *     successful; -EIO otherwise.
 *
 *
 * Input
 *     voltage              Pointer to return value
 * 
 * Output
 *     Returns 0 if successful; error code otherwise
 *
 * Modified
 *     voltage              Set to numeric voltage value (units of 10mV)
 *     
 */
static int
battmon_read_voltage(int *voltage)
{
    Byte data[5];

    
    if (!battmon_read_page(0, 5, data))
        return (-EIO);

    *voltage = (battmon_translate_voltage(data[3], data[4]));
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_get_voltage
 *
 * Purpose
 *     Initiate a voltage conversion, wait until it's complete and then
 *     read the value of the voltage registers.  The voltage returned is 
 *     in units of 10mV, i.e. 330 = 3.3V.  The valid range of voltages 
 *     0V to 10.23V, in 10mV increments. Returns 0 if the read was 
 *     successful; -EIO otherwise.
 *
 * Input
 *     voltage              Pointer to return value
 * 
 * Output
 *     Returns 0 if successful; error code otherwise
 *
 * Modified
 *     voltage              Set to numeric voltage value (units of 10mV)
 *     
 */
static int 
battmon_get_voltage(int *voltage)
{
    int val;
    Boolean busy;


    if (get_voltage_count > 0) {
	return (-EBUSY);
    }

    get_voltage_count++;

    /* Issue command to initiate voltage conversion */
    if (!battmon_fill_voltage()) {
	get_voltage_count--;
	return (-EIO);
    }


    /* Free command lock to allow other commands to issue
     * while current process waits for temperature conversion 
     */
    free_command_lock();

    /* Sleep while waiting for voltage conversion to complete.
     * Voltage conversion can take up to 2 ms, so we set the timer to
     * go off after 10 ms (minimum resolution of timer)
     */
    while (TRUE) {
        init_timer(&voltage_timer);
	voltage_timer.function = timer_timedout;
	voltage_timer.data = (unsigned long) &voltage_waitq;
	voltage_timer.expires = jiffies + 1;
	add_timer(&voltage_timer);
	interruptible_sleep_on(&voltage_waitq);

	get_command_lock();
	busy = battmon_fill_voltage_busy();
	free_command_lock();
	if (!busy) break;
    }
    
    /* Now read the voltage */
    get_command_lock();
    get_voltage_count--;

    if (battmon_read_voltage(&val) < 0) {
	return (-EIO);
    }

    put_user(val, voltage);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_get_vad_voltage
 *
 * Purpose
 *     Set the battery monitor configuration to read the Vad voltage.
 *     Then, initiate a voltage conversion, wait until it's complete and then
 *     read the value of the voltage registers.  After the read is
 *     complete the configuration is reset to read the Vdd voltage (default).
 *     The voltage returned is in units of 10mV, i.e. 330 = 3.3V.  
 *     The valid range of voltages 0V to 10.23V, in 10mV increments.  
 *     Returns 0 if the read was successful; -EIO otherwise.
 *
 * Input
 *     voltage              Pointer to return value
 * 
 * Output
 *     Returns 0 if successful; error code otherwise
 *
 * Modified
 *     voltage              Set to numeric voltage value (units of 10mV)
 *     
 */
static int 
battmon_get_vad_voltage(int *voltage)
{
    int val;
    int err;
    Boolean busy;


    if (get_voltage_count > 0) {
	return (-EBUSY);
    }

    get_voltage_count++;

    /* Reconfigure battery monitor to read Vad voltage (default is Vdd) */
    if (!battmon_configure(TRUE, TRUE, TRUE, FALSE)) {
	get_voltage_count--;
	return (-EIO);
    }

    /* Issue command to initiate voltage conversion */
    if (!battmon_fill_voltage()) {
	get_voltage_count--;
	return (-EIO);
    }


    /* Free command lock to allow other commands to issue
     * while current process waits for temperature conversion 
     */
    free_command_lock();

    /* Sleep while waiting for voltage conversion to complete.
     * Voltage conversion can take up to 2 ms, so we set the timer to
     * go off after 10 ms (minimum resolution of timer)
     */
    while (TRUE) {
        init_timer(&voltage_timer);
	voltage_timer.function = timer_timedout;
	voltage_timer.data = (unsigned long) &voltage_waitq;
	voltage_timer.expires = jiffies + 1;
	add_timer(&voltage_timer);
	interruptible_sleep_on(&voltage_waitq);

	get_command_lock();
	busy = battmon_fill_voltage_busy();
	free_command_lock();
	if (!busy) break;
    }

    /* Now read the voltage */
    get_command_lock();
    get_voltage_count--;

    err = battmon_read_voltage(&val);

    /* Reconfigure battery monitor to read Vdd voltage.
     * We expect Vdd to be the command case, so we reset the 
     * configure here.
     */
    if (!battmon_configure(TRUE, TRUE, TRUE, TRUE)) {
	get_voltage_count--;
	return (-EIO);
    }

    if (err < 0) {
        return (-EIO);
    }

    put_user(val, voltage);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_read_current
 *
 * Purpose
 *     Read the value of the current register and set "*curr" to the numerical 
 *     current value.  The current is in units of 1/205 C (.004883C)
 *     i.e. 256 = 1.25C.   Returns 0 if the read is successful; -EIO otherwise.
 *     The valid range of current values is -2.0C to +2.0C in .004883C 
 *     increments).  Note: for Itsy V2, 1C = 500 mA.
 *
 * Input
 *     curr                 Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     curr                 Set to current (in units of 1/205C = 2.44 mA)
 *     
 */
static int
battmon_read_current(int *curr)
{
    Byte data[7];

    if (!battmon_read_page(0, 7, data)) {
      return (-EIO);
    }

    put_user(battmon_translate_current(data[5], data[6]), curr);
    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_read_net_current
 *
 * Purpose
 *     Read the value of the ICA register and set "*curr" to the numerical 
 *     current value.  The ICA maintains a net accumulated total of 
 *     current flowing into and out of the battery; therefore the reading 
 *     in this register is an indication of the remaining capacity in the
 *     battery.  The current is in units of .01C, where full 
 *     battery capacity is 1C.  Thus ICA=100 represents a fully charged
 *     battery and ICA=0 represents fully discharged.  Returns 0 if the
 *     read is successful; -EIO otherwise. The valid range of current 
 *     values is 0C to +2.55C in .01 increments.  Note: for Itsy V2, 1C=500 mA.
 *
 * Input
 *     curr                 Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     curr                 Set to net current (in units of .01C = 5 mA)
 * 
 */
static int
battmon_read_net_current(Byte *curr)
{
    Byte data[5];


    if (!battmon_read_page(1, 5, data)) {
        return (-EIO);
    }

    put_user(data[4], curr);
    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_write_net_current
 *
 * Purpose
 *     Write the value of the ICA register. The ICA maintains a net 
 *     accumulated total of current flowing into and out of the battery; 
 *     therefore the reading in this register is an indication of the 
 *     remaining capacity in the battery.  The current is in units of 
 *     .01C, where full battery capacity is 1C.  Thus ICA=100 represents 
 *     a fully charged battery and ICA=0 represents fully discharged.  
 *     battery capacity is 1C.  Thus ICA=100 represents a fully charged
 *     battery and ICA=0 represents fully discharged.  Since charging
 *     typically provides the battery with more than its rated capacity,
 *     the ICA should be reset to 100 when charging is complete.  This
 *     helps insure the accuracy of later gas gauge measurements.
 *     Returns 0 if the write succeeded; -EIO otherwise.
 *
 * Input
 *     curr                Pointer to net current value to write into ICA
 * 
 * Output
 *     Returns 0 if write succeeded, error code otherwise
 *
 * Modified
 *     None
 * 
 */
static int
battmon_write_net_current(Byte *curr)
{
    Byte data[5];


    if (!battmon_read_page(1, 5, data)) {
        return (-EIO);
    }

    data[4] = get_user(curr);

    if (!battmon_write_page(1, 5, data, FALSE)) {
        return (-EIO);
    }

    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_read_charging_current
 *
 * Purpose
 *     Read the value of the CCA register and set "*curr" to the
 *     current value.  The CCA represents the total charging current
 *     the battery has encountered in its lifetime.  It is only updated
 *     when the current is positive, i.e., the battery is being charged.
 *     The CCA registers are non-volatile, so the information will
 *     accumulate over the lifetime of the battery pack.  
 *     The current is in units of .32C, where full battery capacity is 1C.
 *     The valid range is 0C to 20971C in .32 increments.
 *     Returns 0 if the read is successful; -EIO otherwise.
 *
 * Input
 *     curr                 Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     curr                 Set to charging current (in units of .32C)
 * 
 */
static int
battmon_read_charging_current(Word16 *curr)
{
    Byte data[6];


    if (!battmon_read_page(7, 6, data)) {
        return (-EIO);
    }

    put_user(battmon_translate_CCA_DCA(data[4], data[5]), curr);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_write_charging_current
 *
 * Purpose
 *     Write the value of the CCA register.  The CCA represents the 
 *     total charging current the battery has encountered in its lifetime.
 *     It is only updated when the current is positive, i.e., the 
 *     battery is being charged.  The CCA registers are non-volatile, 
 *     so the information will accumulate over the lifetime of the battery 
 *     pack.  The current is in units of .32C, where full battery capacity 
 *     is 1C.  Returns 0 if the write succeeded; -EIO otherwise.
 *
 * Input
 *     curr                 Charging current value to write into CCA
 * 
 * Output
 *     Returns 0 if write succeeded, error code otherwise
 *
 * Modified
 *     None
 * 
 */
static int
battmon_write_charging_current(Word16 *curr)
{
    Byte data[6];
    Word16 charging_current = get_user(curr);

    if (!battmon_read_page(7, 6, data)) {
        return (-EIO);
    }

    data[4] = charging_current & 0xFF;    /* LSB */
    data[5] = charging_current & 0xFF00;  /* MSB */

    if (!battmon_write_page(7, 6, data, FALSE)) {
        return (-EIO);
    }

    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_read_discharging_current
 *
 * Purpose
 *     Read the value of the DCA register and set "*curr" to the numerical 
 *     current value.  The DCA represents the total discharging current
 *     the battery has encountered in its lifetime.  It is only updated
 *     when the current is negative, i.e., the battery is being discharged.
 *     The DCA registers are non-volatile, so the information will
 *     accumulate over the lifetime of the battery pack.  
 *     The current is in units of .32C, where full battery capacity is 1C.
 *     The valid range is 0C to 20971C in .32 increments.
 *     Returns 0 If the DCA is read successfully; -EIO otherwise.
 *
 * Input
 *     curr                 Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     curr                 Set to charging current (in units of .32C)
 * 
 */
static int
battmon_read_discharging_current(Word16 *curr)
{
    Byte data[8];


    if (!battmon_read_page(7, 8, data)) {
        return (-EIO);
    }

    put_user(battmon_translate_CCA_DCA(data[6], data[7]), curr);
    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_write_discharging_current
 *
 * Purpose
 *     Write the value of the DCA register.  The DCA represents the 
 *     total discharging current the battery has encountered in its lifetime.
 *     It is only updated when the current is negative, i.e., the 
 *     battery is being discharged.  The DCA registers are non-volatile, 
 *     so the information will accumulate over the lifetime of the battery 
 *     pack.  The current is in units of .32C, where full battery capacity 
 *     is 1C.  Returns 0 if the write was successful; -EIO otherwise.
 *
 * Input
 *     curr                 Pointer to discharging current to write into DCA
 * 
 * Output
 *     Returns 0 if write succeeded, error code otherwise
 *
 * Modified
 *     None
 * 
 */
static int
battmon_write_discharging_current(Word16 *curr)
{
    Byte data[8];
    Word16 discharging_current = get_user(curr);

    if (!battmon_read_page(7, 8, data)) {
        return (-EIO);
    }

    data[6] = discharging_current & 0xFF;    /* LSB */
    data[7] = discharging_current & 0xFF00;  /* MSB */

    if (!battmon_write_page(7, 8, data, FALSE)) {
        return (-EIO);
    }

    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_read_RTC
 *
 * Purpose
 *     Read the value of the RTC (real-time-clock) registers and set
 *     "*timestamp" to a numeric timestamp value.  The timestamp is in units 
 *     of 1 sec.  Returns 0 if successful; -EIO otherwise.
 *
 * Input
 *     timestamp            Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     timestamp            Returns the timestamp (in units of 1 sec)
 * 
 */
static int
battmon_read_RTC(Word32 *timestamp)
{
    Byte data[4];

    if (!battmon_read_page(1, 4, data)) {
        return (-EIO);
    }

    put_user(battmon_translate_timestamp(data), timestamp);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_write_RTC
 *
 * Purpose
 *     Write the value of the RTC (real-time-clock) registers.
 *     The timestamp is in units of 1 sec.  Returns 0 if the write was
 *     sucessful; -EIO otherwise.
 *
 * Input
 *     timestamp            Pointer to timestamp to write into RTC registers
 * 
 * Output
 *     Returns 0 if write succeeded, error code otherwise
 *
 * Modified
 *     None
 * 
 */
static int
battmon_write_RTC(Word32 *timestamp)
{
    Byte data[4];
    int i;
    Word32 ts = get_user(timestamp);


    for (i = 0; i < 4; i++) {
      data[i] = ts & 0xFF;
      ts = ts >> 8;
    }

    if (!battmon_write_page(1, 4, data, FALSE)) {
        return (-EIO);
    }

    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_read_disconnect_time
 *
 * Purpose
 *     Read the value of the Disconnect-Timestamp registers and set
 *     "*timestamp" to the numeric timestamp value.  The Disconnect-Timestamp
 *     is written to by the DS2437 whenever it senses that the DQ line has 
 *     been low for more than 1 second.  This condition signals that the 
 *     battery pack has been removed from the system.  The timestamp is in 
 *     units of 1 sec.  Returns 0 if the read was successful, -EIO otherwise.
 *
 * Input
 *     timestamp            Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     timestamp            Set to the timestamp (in units of 1 sec)
 * 
 */
static int
battmon_read_disconnect_time(Word32 *timestamp)
{
    Byte data[8];

    if (!battmon_read_page(2, 4, data)) {
        return -EIO;
    }

    put_user(battmon_translate_timestamp(data), timestamp);
    return (0);
}


/*
 * PRIVATE PROCEDURE: battmon_read_EOC_time
 *
 * Purpose
 *     Read the value of the EOC (end-of-charge) registers and set "*timestamp"
 *     to the numeric timestamp value.  The end-of-charge timestamp is written 
 *     by the DS2437 whenever it senses that charging is finished (when
 *     the current changes direction).  The timestamp is in units of 1 sec.
 *     Returns 0 if the read was successful, -EIO otherwise.
 *
 * Input
 *     timestamp            Pointer to returned value
 * 
 * Output
 *     Returns 0 if successful, error code otherwise
 *
 * Modified
 *     timestamp            Set to the timestamp (in units of 1 sec)
 * 
 */
static int
battmon_read_EOC_time(Word32 *timestamp)
{
    Byte data[4];

    if (!battmon_read_page(2, 8, data)) {
        return (-EIO);
    }

    put_user(battmon_translate_timestamp(&(data[4])), timestamp);
    return (0);
}



/*
 * PRIVATE PROCEDURE: battmon_reset
 *
 * Purpose
 *     Reset the smart battery monitor 
 *
 * Input
 *     None
 * 
 * Output
 *     Returns TRUE if battery monitor reset was successful, FALSE
 *     if reset failed.
 *
 * Modified
 *     None
 *     
 */
static Boolean
battmon_reset(void)
{
    int clock_freq_num = READ_CLOCK_SPEED;

    return ((Boolean) battmon_send_reset(clock_freq_num));
}



/*
 * PRIVATE PROCEDURE: battmon_read_ROM
 *
 * Purpose
 *     Issue the "Read ROM" command [0x33].  This command allows the bus
 *     master to read the DS2437's 8-bit family code, unique 48-bit serial
 *     number and an 8-bit CRC.  These values form the DS2437's 64-bit
 *     ROM code. This command can only be used if there
 *     is a single DS2437 on the bus.  If more than one slave is present
 *     on the bus, a data collision will occur when all slaves try to 
 *     transmit at the same time.
 *
 * Input
 *     None
 *
 * Output
 *     Returns the DS2437's 64-bit ROM code if successful, 0 otherwise
 *
 * Modified
 *     None
 *     
 */
static Word64
battmon_read_ROM(void)
{
    int i;
    Word64 rom_code = 0;
    Byte crc = 0;
    Byte data;
    
    battmon_write_byte(0x33);
    
    for (i = 0; i < 64; i += 8) {
        data = battmon_read_byte();
        rom_code |= (Word64) data << i;

	if (i < 56) {
	    /* Calculate CRC on first 7 bytes */
	    crc = crc_table[crc ^ data];
	} else {
	    /* Confirm CRC on last byte */
	    if (data != crc) return (0);
	}
    }

    return (rom_code);
}


/*
 * PRIVATE PROCEDURE: battmon_skip_ROM
 *
 * Purpose
 *     Issue the "Skip ROM" command [0xCC].  This command can save time in a
 *     single drop bus system by allowing the bus master to access the memory 
 *     functions without providing the 64-bit ROM code.  If more than one 
 *     slave is present on the bus and a read command is issued following the
 *     "Skip ROM" command, data collision will occur on the bus as multiple
 *     slaves transmit simultaneously.
 *
 * Input
 *     None
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_skip_ROM(void)
{
    battmon_write_byte(0xCC);
}


#if 0
/*
 * PRIVATE PROCEDURE: battmon_match_ROM
 *
 * Purpose
 *     Issue the "Match ROM" command [0x55].  The "Match ROM" command, followed
 *     by a 64-bit ROM sequence allows the bus master to address a specific
 *     DS2437 on a multidrop bus.  Only the DS2437 that exactly matches the 
 *     64-bit ROM sequence will respond to the following memory function
 *     command.  All slaves that do not match the 64-bit ROM sequence will 
 *     wait for the a reset pulse.  This command can be used with a single
 *     or multiple devices on the bus.
 *
 * Input
 *     rom_code             64-bit code to be matched
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_match_ROM(u64 rom_code)
{
    int i;
    Byte data;

    battmon_write_byte(0x55);
    
    for (i = 0; i < 64; i+= 8) {
        data = (Byte) ((rom_code >> i) & 0xFF);
	battmon_write_byte(data);
    }
}


/*
 * PRIVATE PROCEDURE: battmon_search_ROM
 *
 * Purpose
 *     Issue the "Search ROM" command [0xF0].  When a system is initially
 *     brought up, the bus master might not know the number of devices on the 
 *     1-Wire bus or their 64-bit ROM codes.  The "Search ROM" command 
 *     allows the bus master to use a process of elimination to identify 
 *     the 64-bit ROM codes of all slave devices on the bus.
 *
 * Input
 *     None
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_Search_ROM(void)
{
    battmon_write_byte(0xF0);

    /* XXX  Search process not implemented XXX */
}
#endif


/*
 * PRIVATE PROCEDURE: battmon_write_scratchpad
 *
 * Purpose
 *     Issue the "Write Scratchpad" command [0x4E,0xYY]. This command writes
 *     to the scratchpad page 0xYY of the DS2437.  The entire 8-byte scratchpad
 *     space may be written, but all writing begins with the byte present at
 *     address 0 of the selected scratchpad.  Valid page numbers for writing 
 *     are 0 - 7.
 *
 * Input
 *     page                 Page number to write (0 - 7)
 *     bytes                Number of bytes written (1 - 8 bytes)
 *     data                 Data to write
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_write_scratchpad(int page, int bytes, Byte *data)
{
    int i;

    if ((page < 0) || (page > 7) || (bytes < 1)) return;

    battmon_write_byte(0x4E);
    battmon_write_byte(page);

    for (i = 0; i < MIN(8, bytes); i++) {
	battmon_write_byte(data[i]);
    }
}



/*
 * PRIVATE PROCEDURE: battmon_read_scratchpad
 *
 * Purpose
 *     Issue the "Read Scratchpad" command [0xBE,0xYY]. This command reads
 *     the contents of the scratchpad page 0xYY on the DS2437.  The data read
 *     always begins at address 0 of the selected scratchpad.  Any reserved
 *     bits will all be read as logic 1's.  Valid page numbers for reading
 *     are 0 - 7.  If 8 bytes are read, then the CRC is also read from
 *     the DS2437; otherwise the CRC is invalid.
 *
 * Input
 *     page                 Page number to read (0 - 7)
 *     bytes                Number of bytes to read (1 - 8 bytes)
 *     data                 Data read
 *     crc                  Pointer to returned value (or NULL)
 * 
 * Output
 *     None
 *
 * Modified
 *     crc                  Set to 8-bit CRC if non-NULL and 8 bytes read, 
 *                            otherwise invalid
 *     
 */
static void
battmon_read_scratchpad(int page, int bytes, Byte *data, Byte *crc)
{
    int i;

    if ((page < 0) || (page > 7) || (bytes < 1)) return;

    battmon_write_byte(0xBE);
    battmon_write_byte(page);

    for (i = 0; i < MIN(8, bytes); i++) {
       data[i] = battmon_read_byte();
    }

    if ((bytes > 7) && (crc != NULL)) {
	*crc = battmon_read_byte();
    }

}



/*
 * PRIVATE PROCEDURE: battmon_copy_scratchpad
 *
 * Purpose
 *     Issue the "Copy Scratchpad" command [0x48,0xYY]. This command copies
 *     the scratchpad page 0xYY into the EEPROM / SRAM memory page 0xYY of
 *     the DS2437.  Valid page numbers are 0 - 7.  This command sets the 
 *     NVB flag in the Status/Configuration register, and the flag is 
 *     cleared when the copy is done.  The copy may take up to 50 ms 
 *     (2-10 ms typical).
 *
 * Input
 *     page                 Page number to copy (0 - 7)
 *     busy_wait            Busy wait until copy is complete
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_copy_scratchpad(int page, Boolean busy_wait)
{
    int clock_freq_num;

    if ((page < 0) || (page > 7)) return;

    battmon_write_byte(0x48);
    battmon_write_byte(page);

    if (busy_wait) {
	clock_freq_num = READ_CLOCK_SPEED;
	while (battmon_read_bit(clock_freq_num) == 0) ;
    }
}



/*
 * PRIVATE PROCEDURE: battmon_recall_memory
 *
 * Purpose
 *     Issue the "Recall Memory" command [0xB8,0xYY]. This command recalls
 *     the stored values in EEPROM/SRAM page 0xYY to the scratchpad page
 *     0xYY.  This command must preceed a "Read Scratchpad" command in 
 *     order to read any page of memory on the DS2437.  No memory data is 
 *     directly available with a "Read Scratchpad" command.  Valid 
 *     page numbers are 0 - 7.  
 *
 * Input
 *     page                 Page number to recall (0 - 7)
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_recall_memory(int page)
{
    if ((page < 0) || (page > 7)) return;

    battmon_write_byte(0xB8);
    battmon_write_byte(page);
}



/*
 * PRIVATE PROCEDURE: battmon_convert_temp
 *
 * Purpose
 *     Issue the "Convert T" command [0x44]. This command begins a 
 *     temperature conversion.  This command sets the TB flag in the 
 *     Status/Configuration register, and the flag is cleared when the 
 *     temperature conversion is done.  The temperature conversion may take
 *     up to 1 second.  If busy_wait is TRUE, then this routine will block 
 *     until the conversion is complete.  If FALSE, then the routine returns 
 *     after issuing the "Convert T" command.  In the latter case, 
 *     battmon_fill_temp_busy can be called to check if the conversion
 *     is complete. 
 *
 * Input
 *     busy_wait            Busy wait until temp conversion is complete
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_convert_temp(Boolean busy_wait)
{
    int clock_freq_num;


    battmon_write_byte(0x44);

    if (busy_wait) {
      clock_freq_num = READ_CLOCK_SPEED;
      while (battmon_read_bit(clock_freq_num) == 0) ;
    }
}



/*
 * PRIVATE PROCEDURE: battmon_convert_voltage
 *
 * Purpose
 *     Issue the "Convert V" command [0xB4]. This command begins a 
 *     voltage analog-to-digital conversion cycle.  The voltage supply
 *     that is measure is defined by the AD bit of the Status/Configuration
 *     register.  If AD = 1, then V_DD is measured, otherwise V_AD is 
 *     measured.  This command sets the ADB flag, and the flag is 
 *     cleared when the A/D conversion is done.  The current voltage value
 *     is then placed in the Voltage Register of page 0.  The voltage
 *     conversion may take up to 2 ms. If busy_wait is TRUE, then this 
 *     routine will block until the conversion is complete.  If FALSE, then 
 *     the routine returns after issuing the "Convert V" command.  In the 
 *     latter case, battmon_fill_voltage_busy can be called to check if the 
 *     conversion is complete. 
 *
 * Input
 *     busy_wait            Busy wait until voltage conversion is complete
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void
battmon_convert_voltage(Boolean busy_wait)
{
    int clock_freq_num;

    battmon_write_byte(0xB4);

    if (busy_wait) {
      clock_freq_num = READ_CLOCK_SPEED;
      while (battmon_read_bit(clock_freq_num) == 0) ;
    }
}


/*
 * PRIVATE PROCEDURE: battmon_translate_temp
 *
 * Purpose
 *     Translate the two bytes read from the DS2437's temperature
 *     registers into a numerical temperature value 
 *     The temperature returned is in units of 1/32 degrees C, i.e.
 *     802 = 25.0625 C.  The data in the temperature registers is
 *     stored as two's complement. The data layout is shown below:
 *
 *          -------------------------------------------------
 *     LSB: | 2^-1| 2^-2| 2^-3| 2^-4| 2^-5|  0  |  0  |  0  |
 *          -------------------------------------------------
 *
 *          -------------------------------------------------
 *     MSB: |sign | 2^6 | 2^5 | 2^4 | 2^3 | 2^2 | 2^1 | 2^0 |
 *          -------------------------------------------------
 *
 * Input
 *     lsb                  Least significant byte of temperature
 *     msb                  Most significant byte of temperature
 * 
 * Output
 *     Returns numeric temperature value (units of 1/32 degrees C)
 *     
 * Modified
 *     None
 *     
 */
static int
battmon_translate_temp(Byte lsb, Byte msb)
{
    int temp;

    if (0x80 & msb) {  /* temp is negative */

	/* XXX There's probably a more clever way to do this XXX */ 
	temp = (int) ((Byte) ~msb + 1);
	temp = (temp << 5);
	temp = -temp;
    }
    else {  /* temp is positive */
	temp = (msb << 5);
    }

    temp += (int) ((Byte) (lsb >> 3) & 0x1F);
    return temp;
}



/*
 * PRIVATE PROCEDURE: battmon_translate_voltage
 *
 * Purpose
 *     Translate the two bytes read from the DS2437's voltage
 *     registers into a numerical voltage value.  The voltage
 *     returned is in units of 10mV, i.e. 330 = 3.3V.  The
 *     data layout of the voltage registers is shown below:
 *
 *          -------------------------------------------------
 *     LSB: | 2^7 | 2^6 | 2^5 | 2^4 | 2^3 | 2^2 | 2^1 | 2^0 |
 *          -------------------------------------------------
 *
 *          -------------------------------------------------
 *     MSB: |  0  |  0  |  0  |  0  |  0  |  0  | 2^9 | 2^8 |
 *          -------------------------------------------------
 *
 * Input
 *     lsb                  Least significant byte of voltage
 *     msb                  Most significant byte of voltage
 * 
 * Output
 *     Returns numeric voltage value (in 10mV units)
 *     
 * Modified
 *     None
 *     
 */
static int
battmon_translate_voltage(Byte lsb, Byte msb)
{
    return (int) ((msb << 8) | lsb);
}



/*
 * PRIVATE PROCEDURE: battmon_translate_current
 *
 * Purpose
 *     Translate the two bytes read from the DS2437's current
 *     registers into a numerical current value.  The current
 *     returned is in units of .004883C, i.e. 205 = 1.00C.  The
 *     data layout of the current registers is shown below:
 *
 *          -------------------------------------------------
 *     LSB: | 2^7 | 2^6 | 2^5 | 2^4 | 2^3 | 2^2 | 2^1 | 2^0 |
 *          -------------------------------------------------
 *
 *          -------------------------------------------------
 *     MSB: |sign |sign |sign |sign |sign |sign |sign | 2^8 |
 *          -------------------------------------------------
 *
 * Input
 *     lsb                  Least significant byte of current
 *     msb                  Most significant byte of current
 * 
 * Output
 *     Returns numeric current value (in units of .004883C)
 *     
 * Modified
 *     None
 *     
 */
static int
battmon_translate_current(Byte lsb, Byte msb)
{
    int curr;

    curr = (int) (msb << 8) | lsb;

    if (0x80 & msb) {  /* current is negative */

	/* XXX There's probably a more clever way to do this XXX */ 
	curr = (~curr & 0x1FF) + 1;
	curr = -curr;
    }

    return curr;
}




/*
 * PRIVATE PROCEDURE: battmon_translate_CCA_DCA
 *
 * Purpose
 *     Translate the two bytes read from the DS2437's CCA
 *     (Charging Current Accumulator) or DCA (Discharging
 *     Current Accumulator) into a numeric value.  The value
 *     returned is in units of .32C.  The data layout of the CCA/DCA
 *     registers is shown below:
 *
 *          -------------------------------------------------
 *     LSB: | 2^7 | 2^6 | 2^5 | 2^4 | 2^3 | 2^2 | 2^1 | 2^0 |
 *          -------------------------------------------------
 *
 *          -------------------------------------------------
 *     MSB: | 2^15| 2^14| 2^13| 2^12| 2^11| 2^10| 2^9 | 2^8 |
 *          -------------------------------------------------
 *
 * Input
 *     lsb                  Least significant byte of current
 *     msb                  Most significant byte of current
 * 
 * Output
 *     Returns numeric current value (in units of .32C)
 *     
 * Modified
 *     None
 *     
 */
static int
battmon_translate_CCA_DCA(Byte lsb, Byte msb)
{
    return (int) ((msb << 8) | lsb);
}


/*
 * PRIVATE PROCEDURE: battmon_translate_timestamp
 *
 * Purpose
 *     Translate the four bytes read from the DS2437's timestamp
 *     registers (Real-time-clock, End-of-charge or Disconnect)
 *     into a numeric value.  The value returned is in units of 1 sec.  
 *     The data layout of the timestamp registers is shown below:
 *
 *          -------------------------------------------------
 *     LSB: | 2^7 | 2^6 | 2^5 | 2^4 | 2^3 | 2^2 | 2^1 | 2^0 |
 *          -------------------------------------------------
 *          -------------------------------------------------
 *          | 2^15| 2^14| 2^13| 2^12| 2^11| 2^10| 2^9 | 2^8 |
 *          -------------------------------------------------
 *          -------------------------------------------------
 *          | 2^23| 2^22| 2^21| 2^20| 2^19| 2^18| 2^17| 2^16|
 *          -------------------------------------------------
 *          -------------------------------------------------
 *     MSB: | 2^31| 2^30| 2^29| 2^28| 2^27| 2^26| 2^25| 2^24|
 *          -------------------------------------------------
 *
 *
 * Input
 *     data                 Array of 4 data bytes; byte 0 is LSB, byte 3 is MSB
 * 
 * Output
 *     Returns numeric timestamp value (in units of 1 sec)
 *     
 * Modified
 *     None
 *     
 */

static Word32
battmon_translate_timestamp(Byte *data)
{
    Word32 timestamp = 0;

    timestamp = ((data[3] << 24) |
                 (data[2] << 16) |
                 (data[1] << 8) |
                 (data[0]));

    return (timestamp);
}


/*
 * PRIVATE PROCEDURE: battmon_read_byte
 *
 * Purpose
 *     Read a byte from the smart battery monitor 
 *
 * Input
 *     None
 * 
 * Output
 *     Returns the byte read
 *
 * Modified
 *     None
 *     
 */
static Byte 
battmon_read_byte(void)
{
    int i;
    Byte data = 0;
    int val;
    int clock_freq_num = READ_CLOCK_SPEED;

    for (i = 0; i < 8; i++) {
	val = battmon_read_bit(clock_freq_num);
	val = val << i;
	data = data | (Byte) val;
    }

    return data;
}



/*
 * PRIVATE PROCEDURE: battmon_write_byte
 *
 * Purpose
 *     Write a byte to the smart battery monitor 
 *
 * Input
 *     data                  The byte to write
 * 
 * Output
 *     None
 *
 * Modified
 *     None
 *     
 */
static void 
battmon_write_byte(Byte data)
{
    int i;
    Byte val = data;
    int clock_freq_num = READ_CLOCK_SPEED;

    for (i = 0; i < 8; i++) {
	if ((0x1 & val) == 0) {
	    battmon_write_zero(clock_freq_num);
	} else {
	    battmon_write_one(clock_freq_num);
	}
	val = (val >> 1);
    }
}



/*
 * powermgr operations
 */

#ifdef CONFIG_POWERMGR
static int
battmon_suspend_check(void *ignore, int idle_jiffies)
{
    /* Willing to sleep immediately */
    return (0);
}

static int
battmon_suspend(void *ignore)
{
    /* Nothing to do before suspending */
    return (0);
}

static int 
battmon_resume(void *ignore, int resume_flags)
{
    /* Nothing to do before resuming */
    return (0);
}

static const powermgr_client battmon_powermgr = {
  /* callback functions */
  battmon_suspend_check,
  battmon_suspend,
  battmon_resume,
  
  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  BATTMON_NAME_VERBOSE,

  /* power-consumption info */
  0,
  0
};
#endif /* CONFIG_POWER */



/*
 * Wake up processes waiting on the queue given by "ptr".
 */
static void
timer_timedout(unsigned long ptr)
{
    wake_up_interruptible((struct wait_queue **) ptr);
}

/*
 * Used to ensure that only one process at a time can
 * issue commands to the battery monitor.  Processes
 * that fail to get the lock are put to sleep. 
 */
static void get_command_lock(void)
{
    while (set_bit(0, &command_lock) != 0) {
        interruptible_sleep_on(&command_waitq);
    }
}

/*
 * Free the lock and wake up processes that are sleeping
 * while waiting for the lock.
 */
static void free_command_lock(void)
{
    clear_bit(0, &command_lock);
    wake_up_interruptible(&command_waitq);
}


static int 
battmon_ioctl(struct inode *inode,
	      struct file *file,
	      uint command,
	      ulong arg)
{
    int err = 0; 
    int size = _IOC_SIZE(command);


    if (BATTMON_DEBUG_VERBOSE) {
	printk(BATTMON_NAME ": battmon_ioctl: cmd=%u, arg=%lu, size=%d\n", 
	       command, arg, size);
    }

    /* Verify that the command is for this device. */
    if (_IOC_TYPE(command) != BATTMON_IOCTL_TYPE) return -EINVAL;
    if (_IOC_NR(command) > BATTMON_IOCTL_MAXNR) return -EINVAL;

    /* Check direction and size of data transfers */
    if (_IOC_DIR(command) & _IOC_READ)
	err = verify_area( VERIFY_WRITE, (void *)arg, size );
    else if (_IOC_DIR(command) & _IOC_WRITE)
	err = verify_area( VERIFY_READ, (void *)arg, size );
    if (err) return err;

    /* grab lock */
    get_command_lock();
    
    /* 
     * dispatch based on command 
     */
    switch (command) {
        case BATTMON_DETECT: {
	    err = (battmon_detect((Word64 *)arg));
	    break;
	}
        case BATTMON_GET_TEMP: {
	    err = (battmon_get_temp((int *) arg));
	    break;
        }
        case BATTMON_GET_VOLTAGE: {
	    err = (battmon_get_voltage((int *) arg));
	    break;
	}
        case BATTMON_GET_VAD_VOLTAGE: {
	    err = (battmon_get_vad_voltage((int *) arg));
	    break;
	}
	case BATTMON_READ_CURRENT: {
	    err = (battmon_read_current((int *) arg));
	    break;
	}
	case BATTMON_READ_NET_CURRENT: {
	    err = (battmon_read_net_current((Byte *) arg));
	    break;
	}
	case BATTMON_WRITE_NET_CURRENT: {
	    err = (battmon_write_net_current((Byte *) arg));
	    break;
	}
	case BATTMON_READ_CHARGING_CURRENT: {
	    err = (battmon_read_charging_current((Word16 *) arg));
	    break;
	}
	case BATTMON_WRITE_CHARGING_CURRENT: {
	    err = (battmon_write_charging_current((Word16 *) arg));
	    break;
	}
	case BATTMON_READ_DISCHARGING_CURRENT: {
	    err = (battmon_read_discharging_current((Word16 *) arg));
	    break;
	}
	case BATTMON_WRITE_DISCHARGING_CURRENT: {
	    err = (battmon_write_discharging_current((Word16 *) arg));
	    break;
	}
	case BATTMON_READ_RTC: {
	    err = (battmon_read_RTC((Word32 *) arg));
	    break;
	}
	case BATTMON_WRITE_RTC: {
	    err = (battmon_write_RTC((Word32 *) arg));
	    break;
	}
	case BATTMON_READ_DISCONNECT_TIME: {
	    err = (battmon_read_disconnect_time((Word32 *) arg));
	    break;
	}
	case BATTMON_READ_EOC_TIME: {
	    err = (battmon_read_EOC_time((Word32 *) arg));
	    break;
	}
        case BATTMON_SET_CHARGE_RATE: { 	/* LSB */
            if (arg == 0) {	/* slow rate */
	      if (GPIO_2SlowChg == 0)
		GPCR = GPIO_2FSTCHG;
	      else
		GPSR = GPIO_2FSTCHG;
            }
            else {		/* fast rate */
	      if (GPIO_2FastChg == 0)
		GPCR = GPIO_2FSTCHG;
	      else
		GPSR = GPIO_2FSTCHG;
            } 
	    break;
        }
        case BATTMON_GET_CHARGE_RATE: { /* LSB */
	    int t;
	    t = (GPLR & GPIO_2FSTCHG) == GPIO_2FastChg;
	    memcpy_tofs((void *)arg, &t, sizeof(int));
	    break;
	}

        default:
	  err = -EINVAL;
    }

    /* free lock for next command */
    free_command_lock();

    return err;
}


/*
 * Open the device.  Since this procedure issues commands to the battery
 * monitor, we must grab (and release) the lock.
 */
static int 
battmon_open(struct inode *inode,
	     struct file *file)
{
    Word64 rom_code = 0;

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    if (itsy_major_rev() < 2) {
      return (-ENODEV);
    }
    
    get_command_lock();

    /* Reset and detect the battery monitor */
    if (!battmon_reset() || ((rom_code = battmon_read_ROM()) == 0)) {
	free_command_lock();
        printk(BATTMON_NAME ": Failed to detect device\n");
	return (-ENODEV);
    }

    /* Initialize configuration register */
    if (!battmon_configure(TRUE, TRUE, TRUE, TRUE)) {
	free_command_lock();
	printk(BATTMON_NAME ": Failed to configure device\n");
	return (-ENODEV);
    }

    free_command_lock();

    if (BATTMON_DEBUG_VERBOSE) {
	/* XXX Is there a format for printing long longs? XXX */
	printk(BATTMON_NAME ": Opened device with id=0x%08x%08x\n",
	       (Word32) (rom_code >> 32) & 0xFFFFFFFF, 
	       (Word32) rom_code & 0xFFFFFFFF);
    }

#ifdef CONFIG_PROC_FS
    if (device_id == 0) {
        device_id = rom_code;
    }
#endif

    return 0;
}


static void 
battmon_close(struct inode *inode, 
	      struct file *file)
{
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
}


static struct file_operations battmon_fops =
{
    NULL,			/* no special lseek */
    NULL,			/* no special read */
    NULL,			/* no special write */
    NULL,			/* no special readdir */
    NULL,			/* no special select */
    battmon_ioctl,
    NULL,			/* no special mmap */
    battmon_open,
    battmon_close,
    NULL,			/* no special fsync */
    NULL,			/* no special fasync */
    NULL,			/* no special check_media_change */
    NULL			/* no special revalidate */
};

int 
battmon_init(void)
{
    int  err;


    if (BATTMON_DEBUG_VERBOSE) {
	printk(BATTMON_NAME ": initializing device\n");
    }
    
   /* Register the device driver */
   if ((err = register_chrdev(BATTMON_MAJOR, BATTMON_NAME, 
			      &battmon_fops )) != 0) {
      printk("battmon: unable to get major device = %i, err = %i\n",
	      BATTMON_MAJOR, err );
      return err;
   }

#ifdef	CONFIG_POWERMGR
   /* register with power manager */
   pm_id = powermgr_register(&battmon_powermgr);
   if (pm_id == POWERMGR_ID_INVALID) {
       if (BATTMON_DEBUG) {
           printk(BATTMON_NAME ": unable to register with power manager\n");
       }
   }
#endif

#ifdef  CONFIG_PROC_FS
   /* Register procfs device */
   proc_register_dynamic(&proc_root, &battmon_proc_entry);
#endif

    /* Log device registration */
    printk(BATTMON_NAME_VERBOSE " initialized\n");
    if (BATTMON_DEBUG_VERBOSE) {
	printk (BATTMON_NAME ": version %s\n", BATTMON_VERSION_STRING);
    }

    /* Everything OK */
    return(0);
}


#ifdef	MODULE
int 
init_module(void)
{
    if (BATTMON_DEBUG_VERBOSE) {
	printk(BATTMON_NAME ": initializing module\n");
    }
    
    return (battmon_init());
}


void 
cleanup_module(void)
{
    int err;


    /* Unregister the device driver */
    if ((err = unregister_chrdev(BATTMON_MAJOR, BATTMON_NAME )) != 0) {
	printk(BATTMON_NAME ": unregister_chrdev returns: %i\n", err );
    }

    /* log device unload */
    printk(BATTMON_NAME_VERBOSE " unloaded\n");

#ifdef	CONFIG_POWERMGR
  /* unregister with power manager */
  if (powermgr_unregister(pm_id) != POWERMGR_SUCCESS) {
      printk(BATTMON_NAME ": unable to unregister with power manager\n");
  }
#endif

#ifdef  CONFIG_PROC_FS
   /* Register procfs device */
    if ((err = proc_unregister(&proc_root, battmon_proc_entry.low_ino)) != 0) {
	if (BATTMON_DEBUG_VERBOSE) {
	    printk(BATTMON_NAME ": cleanup_module: proc_unregister %d\n", err);
	}
    }
#endif 

}
#endif	/* MODULE */



/*
 * procfs operations
 */

#ifdef CONFIG_PROC_FS
static int 
battmon_procfs_info(char *buf, char **start, off_t offset, 
		    int length, int unused) 
{
   int len = 0;

   /* Format device information */
   len += sprintf(buf+len, "itsy_major_rev: %i\n", itsy_major_rev()); 
   len += sprintf(buf+len, "battery monitor device id: 0x%08x%08x\n",
		  (Word32) (device_id >> 32) & 0xFFFFFFFF, 
		  (Word32) device_id & 0xFFFFFFFF);
   return(len);
}
#endif  /* CONFIG_PROC_FS */



