/*
 * Itsy Buttons Device
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * Author: Carl Waldspurger
 *
 * $Log: itsy_buttons.c,v $
 * Revision 2.17  1999/12/23  01:27:55  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.16  1999/11/08  22:41:40  caw
 * Added support for Itsy V2 encoder wheel.  The encoder wheel is modeled
 * as two separate buttons (A = clockwise, B = counter-clockwise).  Moving
 * the encoder wheel a step in one direction generates two button events:
 * button down followed by button up.
 *
 * Revision 2.15  1999/10/21 22:55:36  caw
 * Minor cleanup of V2 changes.  Removed some debugging code.
 *
 * Revision 2.14  1999/10/12 01:48:44  caw
 * Bug fixes for V2: (1) invert button sense, (2) clear interrupts after read.
 *
 * Revision 2.13  1999/10/05 01:34:41  caw
 * Updated to support both V1 and V2 hardware (only tested on V1.5).
 *
 * Revision 2.12  1999/09/10 21:41:14  caw
 * Updated to reflect changes to itsy_event_header field names.
 *
 * Revision 2.11  1999/06/10  17:39:46  caw
 * Removed debugging output in buttons_write().
 *
 * Revision 2.10  1999/06/10  17:05:00  caw
 * Convert ITSY_SESSION_ID_FOREGROUND to foreground sid for synthetic events.
 *
 * Revision 2.9  1999/06/10  00:53:32  caw
 * Defined write operation for fabricating synthetic buttons events.
 *
 * Revision 2.8  1999/06/09  23:56:32  caw
 * Added support for fasync.  Also properly handle blocking/nonblocking reads.
 *
 * Revision 2.7  1999/05/12  18:01:01  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.6.2.8  1999/05/04  23:09:35  caw
 * Implemented improved approach to prevent "lost" interrupts.  Seems stable.
 *
 * Revision 2.6.2.7  1999/05/04  21:24:02  caw
 * Changes to interrupt-related routines to prevent "lost" interrupts.
 *
 * Revision 2.6.2.6  1999/04/27  19:53:07  caw
 * Reset current button state after last client release.
 *
 * Revision 2.6.2.5  1999/04/27  15:01:00  brakmo
 * Fixed bug in release code that would freeze touchscreen.
 *
 * Revision 2.6.2.4  1999/04/21 01:01:45  caw
 * Buf fix to suspend/resume to avoid codec operations when disabled.
 *
 * Revision 2.6.2.3  1999/04/19  23:14:48  caw
 * Register session manager callback to handle death of foreground session.
 * Completed initial support for per-client buttons grabs/ungrabs.
 *
 * Revision 2.6.2.2  1999/04/19  21:21:22  caw
 * Fixed redundant cleanup code that could result in hangs.
 * Added preliminary support for grab operations.
 *
 * Revision 2.6.2.1  1999/04/15  22:03:31  caw
 * Assume itsy sessions support, removed conditional compilation, parameters.
 * Enable codec only when necessary.  Added missing module use counts.
 *
 * Revision 2.6  1999/04/01  02:22:22  caw
 * Bug fix to correct updating of client->buf_next in buttons_read().
 * Removed unnecessary codec reset code; now done in "mcp_common".
 *
 * Revision 2.5  1999/04/01  00:29:14  caw
 * Defined raw and sid-filtered polling devices (minor 2 and 3).
 * Minor changes for sessions, including use of itsy_session_from_task().
 *
 * Revision 2.4  1999/03/31  03:05:15  caw
 * Use mcp_common_{enable,disable}_irq() interface for well-behaved sharing.
 * Removed unnecessary operations, other minor improvements.
 *
 * Revision 2.3  1999/03/29  23:44:43  caw
 * Moved primitive mcp/codec operations to "itsy_codec.h".
 *
 * Revision 2.2  1999/03/29  22:25:01  caw
 * Reduced debugging output.  Added module parameter documentation.
 *
 * Revision 2.1  1999/03/29  19:55:02  caw
 * Initial revision (still experimental).
 *
 */

/*
 * Overview
 *
 * This driver supports interrupt-driven button input on the Itsy platform,
 * which uses a StrongARM SA1100 coupled with a Philips UCB1200 codec.  The
 * UCB1200 is configured to detect rising/falling edge transitions on the
 * hardware button input lines, causing an MCP interrupt on Itsy GPIO 6.
 *
 */

/*
 * debugging 
 *
 */

#define	BUTTONS_DEBUG			(1)
#define	BUTTONS_DEBUG_VERBOSE		(0)

/*
 * includes
 *
 */

#include <linux/config.h>

#ifdef	MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/types.h>
#include <linux/ctype.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef	CONFIG_POWERMGR
#include <linux/powermgr.h>
#include <asm/arch/irq.h>
#endif

#include <asm/segment.h>
#include <asm/io_reg.h>

#include <linux/itsy_session.h>
#include <linux/itsy_codec.h>
#include <linux/mcp_common.h>
#include <linux/itsy_buttons.h>
#include <linux/buttons.h>
#include <linux/itsy_version.h>

#include "itsy_util.h"

/*
 * constants
 *
 */

/* names */
#define	BUTTONS_NAME			"itsy-buttons"
#define	BUTTONS_NAME_VERBOSE		"itsy buttons driver"
#define	BUTTONS_VERSION_STRING		"$Revision: 2.17 $"

/* device numbers */
#define	BUTTONS_MAJOR			(62)
#define	BUTTONS_RAW_MINOR		(0)
#define	BUTTONS_SESSION_MINOR		(1)
#define BUTTONS_POLL_RAW_MINOR		(2)
#define	BUTTONS_POLL_SESSION_MINOR	(3)

/* interrupt numbers */
#define	BUTTONS_IRQ			(CODEC_IRQ)

/* buffer size */
#define	BUTTONS_BUFFER_LG_NEVENTS	(10)
#define	BUTTONS_BUFFER_NEVENTS		(1 << BUTTONS_BUFFER_LG_NEVENTS)
#define	BUTTONS_BUFFER_INDEX_MASK	(BUTTONS_BUFFER_NEVENTS - 1)

/* empty state (nothing pressed) */
#define	BUTTONS_STATE_DEFAULT		(0)

/*
 * types
 *
 */

typedef struct buttons_client {
  /* linked list */
  struct buttons_client *prev, *next;

  /* identity */
  int id;
  int sid;

  /* asynchronous notifications */
  struct fasync_struct *asyncq;

  /* buffer management */
  int buf_next;

  /* grab operations */
  ushort grab_mask;
  ushort grab_state;

  /* statistics */
  int drop_count;
} buttons_client;

typedef struct {
  /* memory-mapped regs */
  gpio_reg *gpio;
  mcp_reg *mcp;

  /* mcp common id */
  mcp_common_id mcp_id;

  /* current state */
  ushort state_raw;
  ushort state_ungrab;

  /* asynchronous notifications (raw) */
  struct fasync_struct *asyncq;

  /* buffer management */
  struct wait_queue *waitq;  
  itsy_buttons_event *buf;
  int buf_events;
  int buf_bytes;

  /* buffer indexing */
  int buf_mask;
  int buf_next;

  /* clients */
  buttons_client *clients;
  int client_id_next;
  int nclients;

  /* grab operations */
  ushort grab_mask;

  /* statistics */
  ulong interrupt_count;
  ulong interrupt_loop_count;
  ulong event_count;

  /* session management */
  int sm_id;
  int sm_active;

#ifdef	CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
  ulong last_active;
#endif

  /* v2-specific data */
  ushort v2_cursor;
  ushort v2_other;
  ushort v2_encoder;
  ushort v2_enc_a;
  ushort v2_enc_b;
} buttons_dev;

/*
 * globals
 *
 */

static buttons_dev buttons_global;

#ifdef	CONFIG_PROC_FS
static int buttons_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry buttons_proc_entry = {
  0, 7, "buttons", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, buttons_procfs_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * macros
 *
 */

static inline ushort codec_itsy_io_buttons(void)
{
  /* return appropriate button mask for hardware version */
  return(is_itsy_v1() ? CODEC_ITSY_V1_IO_BUTTONS : CODEC_ITSY_V2_IO_BUTTONS);
}

/*
 * instantiate list operations
 *
 */

INSTANTIATE_LIST_INSERT(buttons_client)
INSTANTIATE_LIST_REMOVE(buttons_client)

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int buttons_procfs_info(char *buf, 
			       char **start,
			       off_t offset,
			       int length,
			       int unused)
{
  buttons_dev *dev = &buttons_global;
  int len = 0;

#ifdef	BUTTONS_DEBUG
  if (is_itsy_v2())
    len += sprintf(buf + len,
		   "v2-specific:\n"
		   "    %04x cursor\n"
		   "    %04x other\n"
		   "    %04x encoder\n"
		   "    %04x enc a\n"
		   "    %04x enc b\n",
		   dev->v2_cursor,
		   dev->v2_other,
		   dev->v2_encoder,
		   dev->v2_enc_a,
		   dev->v2_enc_b);
#endif	/* BUTTONS_DEBUG */

  /* format current state */
  len += sprintf(buf + len, 
		 "status:\n"
		 "  %06x raw state\n"
		 "  %06x ungrab state\n"
		 "  %06x grab mask\n",
		 dev->state_raw,
		 dev->state_ungrab,
		 dev->grab_mask);

  /* format statistics */
  len += sprintf(buf + len, 
		 "statistics:\n"
		 "  %6d interrupts\n"
		 "  %6d intr loops\n"
		 "  %6d events\n"
		 "  %6d clients\n",
		 dev->interrupt_count,
		 dev->interrupt_loop_count,
		 dev->event_count,
		 dev->nclients);

  /* format session state */
  len += sprintf(buf + len,
		 "sessions:\n"
		 "  %6d foreground\n",
		 dev->sm_active);

  /* format client info */
  if (dev->clients != NULL)
    {
      buttons_client *c;

      len += sprintf(buf + len, "clients:\n");
      for (c = dev->clients; c != NULL; c = c->next)
	len += sprintf(buf + len, 
		       "  %6d %06x %6d\n",
		       c->id,
		       c->grab_mask,
		       c->drop_count);
    }

  return(len);  
}
#endif	/* CONFIG_PROC_FS */

/*
 * codec operations
 *
 */

static inline void codec_buttons_clear(mcp_reg *mcp)
{
  /*
   * modifies: mcp
   * effects:  Clears pending codec interrupts due to buttons.
   *
   */

  /* clear interrupts with 0->1 transition */
  codec_write(mcp, CODEC_REG_INT_STATUS, 0);
  codec_write(mcp, CODEC_REG_INT_STATUS, codec_itsy_io_buttons());
}

static inline void gpio_buttons_clear(gpio_reg *gpio)
{
  /* setup GPIO registers */
  gpio->edge_detect |= (1 << BUTTONS_IRQ);
}

static void buttons_interrupt_clear(buttons_dev *dev)
{
  ushort status;

  /* clear interrupts at codec */
  codec_buttons_clear(dev->mcp);

  /* clear gpio interrupt only if no non-button interrupts pending */
  status = codec_read(dev->mcp, CODEC_REG_INT_STATUS);
  if ((status & ~codec_itsy_io_buttons()) == 0)
    gpio_buttons_clear(dev->gpio);
}

static ushort codec_read_buttons_v2(buttons_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Returns current button state from GPIO input ports.
   *           Bit set = "button down", bit clear = "button up".
   * caveats:  Specialized for Itsy V2 hardware, which multiplexes
   *           eight physical buttons on four GPIOs.
   *
   */

  mcp_reg *mcp = dev->mcp;
  ushort raw, data, cursor, other, cvt;

  /* snapshot data (with aliased buttons) */
  raw  = codec_read(mcp, CODEC_REG_IO_DATA);
  data = raw & ~CODEC_ITSY_V2_IO_BUTTONS_MUX;

  /* snapshot cursor buttons (disable both, then disable other) */
  codec_write(mcp, CODEC_REG_IO_DATA, data | CODEC_ITSY_V2_IO_BUTTONS_MUX);
  codec_write(mcp, CODEC_REG_IO_DATA, data | CODEC_ITSY_V2_IO_BUTTONS_MUX1);
  cursor = ~codec_read(mcp, CODEC_REG_IO_DATA) & CODEC_ITSY_V2_IO_BUTTONS;

  /* snapshot other buttons (disable both, then disable cursor) */
  codec_write(mcp, CODEC_REG_IO_DATA, data | CODEC_ITSY_V2_IO_BUTTONS_MUX);
  codec_write(mcp, CODEC_REG_IO_DATA, data | CODEC_ITSY_V2_IO_BUTTONS_MUX0);
  other  = ~codec_read(mcp, CODEC_REG_IO_DATA) & CODEC_ITSY_V2_IO_BUTTONS;

  /* re-enable aliased buttons */
  codec_write(mcp, CODEC_REG_IO_DATA, data);

  /* clear button interrupts */
  buttons_interrupt_clear(dev);

  /* capture v2-specific data */
  dev->v2_enc_a = ~raw & (1 << 4);
  dev->v2_enc_b = ~raw & (1 << 5);
  dev->v2_cursor = cursor;
  dev->v2_other  = other;

  /* XXX for now, convert button data to look like V1 (ugh!) */
  cvt = 0;
  if (cursor & (1 << 6)) cvt |= BUTTON_CURSOR_RIGHT;
  if (cursor & (1 << 7)) cvt |= BUTTON_CURSOR_LEFT;  
  if (cursor & (1 << 8)) cvt |= BUTTON_CURSOR_DOWN;  
  if (cursor & (1 << 9)) cvt |= BUTTON_CURSOR_UP;  
  if (other  & (1 << 6)) cvt |= BUTTON_FRONT_TOP;	/* encoder */
  if (other  & (1 << 7)) cvt |= BUTTON_SIDE_TOP;	/* side right */
  if (other  & (1 << 8)) cvt |= BUTTON_FRONT_BOTTOM;	/* side middle */
  if (other  & (1 << 9)) cvt |= BUTTON_SIDE_BOTTOM;	/* side left */

  /*
   * treat v2 encoder wheel as two buttons
   *
   *    A,  B => pressed   (confirmed)
   *    A, !B => A pressed (tentative)
   *   !A,  B => B pressed (tentative)
   *   !A, !B => released
   *
   */

  if (dev->v2_enc_a && dev->v2_enc_b)
    {
      /* confirmed press => encode button down */
      if (dev->v2_encoder == dev->v2_enc_a) cvt |= BUTTON_ENCODER_A;
      if (dev->v2_encoder == dev->v2_enc_b) cvt |= BUTTON_ENCODER_B;
    }
  else if (!dev->v2_enc_a && !dev->v2_enc_b)
    {
      /* release => encode button release, reset */
      dev->v2_encoder = 0;
    }
  else if (dev->v2_encoder == 0)
    {
      /* tentative press => save direction */
      if (dev->v2_enc_a) dev->v2_encoder |= dev->v2_enc_a;
      if (dev->v2_enc_b) dev->v2_encoder |= dev->v2_enc_b;
    }
  
  return(cvt);
}

static inline ushort codec_read_buttons(buttons_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Returns current button state from GPIO input ports.
   *           Bit set = "button down", bit clear = "button up".
   *
   */

  /* on V1, simply snapshot button state */
  if (is_itsy_v1())
    return((~codec_read(dev->mcp, CODEC_REG_IO_DATA)) &
	   CODEC_ITSY_V1_IO_BUTTONS);

  /* on V2, need to handle mux'd signals */
  return(codec_read_buttons_v2(dev));
}

static void codec_buttons_enable(mcp_reg *mcp)
{
  ushort enable;

  /* clear any pending interrupts */
  codec_buttons_clear(mcp);

  /* request interrupts on button up transitions */
  enable  = codec_read(mcp, CODEC_REG_RISE_INT_ENABLE);
  enable |= codec_itsy_io_buttons();
  codec_write(mcp, CODEC_REG_RISE_INT_ENABLE, enable);

  /* request interrupts on button down transitions */
  enable  = codec_read(mcp, CODEC_REG_FALL_INT_ENABLE);
  enable |= codec_itsy_io_buttons();
  codec_write(mcp, CODEC_REG_FALL_INT_ENABLE, enable);
}

static void codec_buttons_disable(mcp_reg *mcp)
{
  ushort enable;

  /* ignore interrupts on button up transitions */
  enable  = codec_read(mcp, CODEC_REG_RISE_INT_ENABLE);
  enable &= ~codec_itsy_io_buttons();
  codec_write(mcp, CODEC_REG_RISE_INT_ENABLE, enable);

  /* ignore interrupts on button down transitions */
  enable  = codec_read(mcp, CODEC_REG_FALL_INT_ENABLE);
  enable &= ~codec_itsy_io_buttons();
  codec_write(mcp, CODEC_REG_FALL_INT_ENABLE, enable);
}

/*
 * buttons_client operations
 *
 */

static buttons_client *buttons_client_create(int id, int sid)
{
  /*
   * modifies: nothing
   * effects:  Returns a new, initialized buttons_client, or
   *	       NULL if unable to allocate storage.
   */

  buttons_client *client;

  /* allocate storage, fail if unable */
  if ((client = KOBJ_ALLOC(buttons_client)) == NULL)
    return(NULL);
  KOBJ_INIT(client, buttons_client);

  /* set identity */
  client->id  = id;
  client->sid = sid;

  /* everything OK */
  return(client);
}

static void buttons_client_destroy(buttons_client *client)
{
  /*
   * modifies: client
   * effects:  Reclaims all storage associated with client.
   *
   */

  kfree(client);
}

/*
 * buttons_dev operations
 *
 */

static int buttons_dev_init(buttons_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Initializes buttons device dev.
   *	       Returns 0 iff successful, otherwise error code.
   *
   */

  /* clear everything */
  KOBJ_INIT(dev, buttons_dev);
  
  /* allocate buffer, fail if unable */
  dev->buf_events = BUTTONS_BUFFER_NEVENTS;
  dev->buf_bytes  = dev->buf_events * sizeof(itsy_buttons_event);
  dev->buf = (itsy_buttons_event *) kmalloc(dev->buf_bytes, GFP_KERNEL);
  if (dev->buf == NULL)
    return(-ENOMEM);

  /* buffer indexing */
  dev->buf_mask = BUTTONS_BUFFER_INDEX_MASK;
  dev->buf_next = 0;
  
  /* initialize gpio */
  dev->gpio = GPIO_BASE_V;

  /* initialize mcp */
  dev->mcp = MCP_BASE_V;
  dev->mcp_id = MCP_COMMON_ID_INVALID;

  /* initialize session manager state */
  dev->sm_id = ITSY_SESSION_NOTIFY_ID_INVALID;
  dev->sm_active = ITSY_SESSION_ID_INVALID;

#ifdef	CONFIG_POWERMGR
  /* initialize powermgr state */
  dev->pm_id = POWERMGR_ID_INVALID;
  dev->last_active = jiffies;
#endif	/* CONFIG_POWERMGR */

  /* v2-specific data */
  dev->v2_cursor  = 0;
  dev->v2_other   = 0;
  dev->v2_encoder = 0;
  dev->v2_enc_a   = 0;
  dev->v2_enc_b   = 0;

  /* everything OK */
  return(0);
}

static void buttons_dev_event(buttons_dev *dev,
			      int sid,
			      ushort state, 
			      unsigned char flags)
{
  /*
   * modifies: buf
   * effects:  Creates event with specified sid, button state, and flag.
   *           Event is added to the buffer associated with dev.  Wakes up
   *           any clients that may be blocked waiting for the event.
   *
   */

  itsy_buttons_event *next;
  buttons_client *c;

  /* use next slot in circular buffer */
  next = &dev->buf[dev->buf_next & dev->buf_mask];
  
  /* generic session event */
  do_gettimeofday(&next->common.timestamp);

  next->common.event_class = ITSY_EVENT_CLASS_BUTTONS;
  next->common.event_type  = ITSY_EVENT_BUTTONS_STATE;
  next->common.size  = sizeof(itsy_buttons_event);
  next->common.flags = flags;

  /* specific buttons event */
  next->sid = sid;
  next->state = state;
  
  /* advance buffer index */
  dev->buf_next++;

  /* wakeup waiters, if any */
  wake_up_interruptible(&dev->waitq);

  /* raw asynchronous notifications */
  if (dev->asyncq != NULL)
    kill_fasync(dev->asyncq, SIGIO);

  /* sid-filtered asynchronous notifications */
  for (c = dev->clients; c != NULL; c = c->next)
    if ((c->sid == sid) && (c->asyncq != NULL))
      kill_fasync(c->asyncq, SIGIO);
    
  /* update statistics */
  dev->event_count++;
}

/*
 * interrupt processing
 *
 */

static inline int buttons_interrupt_pending(buttons_dev *dev)
{
  ushort status = codec_read(dev->mcp, CODEC_REG_INT_STATUS);
  return(status & codec_itsy_io_buttons());
}

static void buttons_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  buttons_dev *dev = &buttons_global;
  ushort state_raw, state_ungrab;
  unsigned char flags;
  buttons_client *c;

  if (BUTTONS_DEBUG_VERBOSE)
    if (dev->interrupt_count < 20)
      printk(BUTTONS_NAME ": buttons_interrupt %ld\n", dev->interrupt_count);

  do {
    /* clear button interrupts */
    buttons_interrupt_clear(dev);

    /* read current buttons state */
    state_raw = codec_read_buttons(dev);
    state_ungrab = state_raw & ~dev->grab_mask;

    /* update raw state */
    dev->state_raw = state_raw;

    /* buffer ungrabbed event if state changed */
    flags = ITSY_EVENT_FLAG_NONE;
    if (state_ungrab != dev->state_ungrab)
      {
	buttons_dev_event(dev, dev->sm_active, state_ungrab, flags);
	dev->state_ungrab = state_ungrab;
      }

    /* buffer grabbed events if state changed */
    flags = ITSY_EVENT_FLAG_GRABBED;
    if (dev->grab_mask != 0)
      for (c = dev->clients; c != NULL; c = c->next)
	if (c->grab_mask != 0)
	  {
	    ushort grab_mask  = dev->grab_mask & ~c->grab_mask;
	    ushort grab_state = state_raw & ~grab_mask;
	    
	    if (grab_state != c->grab_state)
	      {
		buttons_dev_event(dev, c->sid, grab_state, flags);
		c->grab_state = grab_state;
	      }
	  }

    /* update statistics */
    dev->interrupt_loop_count++;
  } while (buttons_interrupt_pending(dev));

  /* update statistics */
  dev->interrupt_count++;
}


/*
 * device operations
 *
 */

static inline int buttons_is_raw_device(const struct inode *inode)
{
  /*
   * modifies: nothing
   * effects:  Returns TRUE iff inode is a "raw" buttons device.
   *
   */

  int minor = MINOR(inode->i_rdev);
  return((minor == BUTTONS_RAW_MINOR) || (minor == BUTTONS_POLL_RAW_MINOR));
}

static inline int buttons_is_poll_device(const struct inode *inode)
{
  /*
   * modifies: nothing
   * effects:  Returns TRUE iff inode is a "polling" buttons device.
   *
   */

  int minor = MINOR(inode->i_rdev);
  return((minor == BUTTONS_POLL_SESSION_MINOR) ||
	 (minor == BUTTONS_POLL_RAW_MINOR));
}

static int buttons_dev_get_grab_all(buttons_dev *dev, ushort *mask)
{
  /* copyout reply value */
  put_user(dev->grab_mask, mask);
  return(0);
}

static int buttons_dev_get_grab(buttons_dev *dev,
				buttons_client *client,
				ushort *mask)
{
  /* copyout reply value */
  put_user(client->grab_mask, mask);
  return(0);
}

static int buttons_dev_set_grab(buttons_dev *dev, 
				buttons_client *client,
				ushort mask)
{
  ushort conflict_mask;

  /* fail if conflict with existing grabs */
  conflict_mask = dev->grab_mask & ~client->grab_mask;
  if (mask & conflict_mask)
    return(-EBUSY);

  /* update grab masks */
  dev->grab_mask &= ~client->grab_mask;
  client->grab_mask = mask;
  dev->grab_mask |= mask;
  return(0);
}

static int buttons_ioctl(struct inode *inode,
			 struct file *file,
			 uint command,
			 ulong arg)
{
  buttons_client *client = (buttons_client *) file->private_data;
  buttons_dev *dev = &buttons_global;
  int size, err;

  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_ioctl: cmd=%d, arg=%lu\n", command, arg);

  /* sanity checks */
  if (_IOC_TYPE(command) != BUTTONS_IOCTL_TYPE)
    return(-EINVAL);
  if (_IOC_NR(command) > BUTTONS_IOCTL_MAXNR)
    return(-EINVAL);

  /* verify transfer memory */
  size = _IOC_SIZE(command);
  if (_IOC_DIR(command) & _IOC_READ)
    if ((err = verify_area(VERIFY_WRITE, (void *) arg, size)) != 0)
      return(err);
  if (_IOC_DIR(command) & _IOC_WRITE)
    if ((err = verify_area(VERIFY_READ,  (void *) arg, size)) != 0)
      return(err);

  /* dispatch based on command */
  switch (command) {
  case BUTTONS_GET_GRAB_ALL:
    return(buttons_dev_get_grab_all(dev, (ushort *) arg));
  case BUTTONS_GET_GRAB:
    return(buttons_dev_get_grab(dev, client, (ushort *) arg));
  case BUTTONS_SET_GRAB:
    return(buttons_dev_set_grab(dev, client, (ushort) (arg & 0xffff)));
  }

  /* invalid command */
  return(-EINVAL);
}

static int buttons_read_poll(struct inode *inode,
			     struct file *file,
			     char *buffer,
			     int count)
{
  /*
   * modifies: buffer
   * effects:  Copies the current buttons state to buffer.
   *	       If the specified device is not a raw device, and
   *           the caller is not in the foreground session,
   *           then the current state is defined as zero.
   *
   */

  buttons_client *client = (buttons_client *) file->private_data;
  buttons_dev *dev = &buttons_global;
  ulong state = dev->state_raw;

  /* fail if not polling device */
  if (!buttons_is_poll_device(inode))
    return(0);

  /* fail if insufficient buffer space */
  if (count < sizeof(state))
    return(0);

  /* update client session */
  client->sid = itsy_session_from_task(current);

  /* XXX not quite right; need to deal w/ fg session grabs */

  /* use empty state if neither raw nor foreground session */
  if (!buttons_is_raw_device(inode) && (client->sid != dev->sm_active))
    state = BUTTONS_STATE_DEFAULT;

  /* copyout button state */
  put_user(state, (ulong *) buffer);
  return(sizeof(state));
}

static int buttons_read(struct inode *inode,
			struct file *file,
			char *buffer,
			int count)
{
  /*
   * modifies: buffer
   * effects:  Copies available itsy button events to buffer.
   *           Returns the number of bytes copied.
   *
   */

  const int event_bytes = sizeof(itsy_buttons_event);

  buttons_client *client = (buttons_client *) file->private_data;
  buttons_dev *dev = &buttons_global;
  int total, err, buf_first, raw;
  
  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_read\n");

  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_WRITE, (void *) buffer, count)) != 0)
    return(err);

  /* special case: polling-based interface */
  if (buttons_is_poll_device(inode))
    return(buttons_read_poll(inode, file, buffer, count));

  /* initialize */
  raw = buttons_is_raw_device(inode);
  total = 0;

  while (total == 0)
    {
      /* update client session */
      client->sid = itsy_session_from_task(current);
      
      /* determine first available buffered event */
      buf_first = MAX(0, dev->buf_next - dev->buf_events);

      /* skip over dropped events */
      if (client->buf_next < buf_first)
	{
	  client->buf_next = buf_first;
	  client->drop_count++;
	}

      /* transfer integral number of matching events */
      for (;
	   (client->buf_next < dev->buf_next) && (count >= event_bytes);
	   client->buf_next++)
	{
	  itsy_buttons_event *event;
	  
	  /* examine next buffered event */
	  event = &dev->buf[client->buf_next & dev->buf_mask];
	  if (raw || (event->sid == client->sid))
	    {
	      /* copyout matching event */
	      memcpy_tofs((void *) buffer, (const void *) event, event_bytes);
	      
	      /* update counts */
	      total  += event_bytes;
	      buffer += event_bytes;
	      count  -= event_bytes;
	    }
	}

      /* handle "no data ready" */
      if (total == 0)
	{
	  /* nonblocking => try again later */
	  if (file->f_flags & O_NONBLOCK)
	    return(-EAGAIN);

	  /* wait for data */
	  interruptible_sleep_on(&dev->waitq);

	  /* handle interrupted system call */
	  if (current->signal & ~current->blocked)
	    return(-EINTR);
	} 
    }

  /* everything OK */
  return(total);
}

static int buttons_write(struct inode *inode,
			 struct file *file,
			 const char *buffer,
			 int count)
{
  const int event_bytes = sizeof(itsy_buttons_event);

  buttons_dev *dev = &buttons_global;
  unsigned char syn_flags;
  const char *data;
  int total, err;

  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_write, count=%d\n", count);

  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_READ, (void *) buffer, count)) != 0)
    return(err);

  /* initialize */
  data = buffer;
  total = 0;  

  /* compute synthetic event flags */
  syn_flags = ITSY_EVENT_FLAG_SYNTHETIC;
  if (suser())
    syn_flags |= ITSY_EVENT_FLAG_SUSER;

  /* copy integral number of events */
  while (count >= event_bytes)
    {
      const itsy_buttons_event *next;
      int sid;

      /* queue synthetic event */
      next = (const itsy_buttons_event *) data;
      sid = next->sid;
      if (sid == ITSY_SESSION_ID_FOREGROUND)
	sid = dev->sm_active;
      buttons_dev_event(dev, sid, next->state, next->common.flags | syn_flags);

      /* update counts */
      total += event_bytes;
      data  += event_bytes;
      count -= event_bytes;
    }

  return(total);
}

static int buttons_fasync(struct inode *inode, struct file *file, int mode)
{
  buttons_client *client = (buttons_client *) file->private_data;
  buttons_dev *dev = &buttons_global;
  struct fasync_struct **aq;

  /* add client to appropriate notification queue */
  aq = buttons_is_raw_device(inode) ? &dev->asyncq : &client->asyncq;
  return(fasync_helper(inode, file, mode, aq));
}

static int buttons_select(struct inode *inode,
			  struct file *file,
			  int mode,
			  select_table *table)
{
  buttons_client *client = (buttons_client *) file->private_data;
  buttons_dev *dev = &buttons_global;

  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_select\n");

  /* handle select on input */
  if (mode == SEL_IN)
    {
      /* polling device always ready */
      if (buttons_is_poll_device(inode))
	return(1);

      if (buttons_is_raw_device(inode))
	{
	  /* any raw events buffered? */
	  if (client->buf_next < dev->buf_next)
	    return(1);

	  /* wait for raw event */
	  if (BUTTONS_DEBUG_VERBOSE)
	    printk(BUTTONS_NAME ": buttons_select: wait (raw)...\n");
	  select_wait(&dev->waitq, table);
	  return(0);
	}
      else
	{
	  /* update client session */
	  client->sid = itsy_session_from_task(current);

	  /* any sid-specific events buffered? */
	  if (client->buf_next < dev->buf_next)
	    {
	      /* scan backwards in buffer looking for match */
	      int buf_first, next;
	      buf_first = MAX(0, dev->buf_next - dev->buf_events);	      
	      buf_first = MAX(buf_first, client->buf_next);
	      for (next = buf_first; next < dev->buf_next; next++)
		if (dev->buf[next & dev->buf_mask].sid == client->sid)
		  return(1);
	    }

	  /* wait for any event (next invocation will determine match) */
	  if (BUTTONS_DEBUG_VERBOSE)
	    printk(BUTTONS_NAME ": buttons_select: wait (sid %d)...\n", 
		   client->sid);
	  select_wait(&dev->waitq, table);
	  return(0);
	}
    }

  /* SEL_OUT, SEL_EX not supported */
  return(0);
}

static void buttons_hardware_init(buttons_dev *dev)
{
  /* turn on the MCP */
  (void) mcp_common_enable(dev->mcp_id);

  /* enable button interrupts */
  (void) mcp_common_enable_irq(dev->mcp_id);
  codec_buttons_enable(dev->mcp);
}

static void buttons_hardware_shutdown(buttons_dev *dev)
{
  /* disable button interrupts */
  codec_buttons_disable(dev->mcp);
  (void) mcp_common_disable_irq(dev->mcp_id);

  /* turn off the MCP */
  (void) mcp_common_disable(dev->mcp_id);
}

static int buttons_open(struct inode *inode, struct file *file)
{
  buttons_dev *dev = &buttons_global;
  int nclients = dev->nclients;
  buttons_client *client;
  int id, sid, err;

  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_open\n");

  /* add new client, fail if unable */
  id = dev->client_id_next++;
  sid = itsy_session_from_task(current);
  if ((client = buttons_client_create(id, sid)) == NULL)
    return(-ENOMEM);

  /* skip over old events */
  client->buf_next = dev->buf_next;

  /* add to client list */
  buttons_client_insert(&dev->clients, client);
  dev->nclients++;

  /* attach client state to file */
  file->private_data = client;

#ifdef	MODULE
  MOD_INC_USE_COUNT;
#endif

  /* initialize IRQ, hardware if necessary */
  if (nclients == 0)
    {
      /* initialize shared IRQ */
      if ((err = request_irq(BUTTONS_IRQ,
			     buttons_interrupt,
			     SA_INTERRUPT | SA_SHIRQ,
			     BUTTONS_NAME,
			     dev)) != 0)
	{
	  /* log error and fail */
	  printk(BUTTONS_NAME ": request_irq failed: %d\n", err);
	  return(err);
	}
      
      /* initialize hardware */
      buttons_hardware_init(dev);
    }

  /* everything OK */
  return(0);
}

static void buttons_release(struct inode *inode, struct file *file)
{
  buttons_client *client = (buttons_client *) file->private_data;
  buttons_dev *dev = &buttons_global;

  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_release\n");

  /* remove from asynchronous notification queues */
  buttons_fasync(inode, file, 0);

  /* release client grabs, if any */
  dev->grab_mask &= ~client->grab_mask;

  /* remove from client list */
  buttons_client_remove(&dev->clients, client);
  dev->nclients--;

  /* reclaim storage */
  buttons_client_destroy(client);
  file->private_data = NULL;

  /* reclaim IRQ, shutdown hardware if no clients */
  if (dev->nclients == 0)
    {
      /* shutdown hardware */
      buttons_hardware_shutdown(dev);

      /* free appropriate interrupt line */
      free_irq(BUTTONS_IRQ, dev);

      /* reset initial button state */
      dev->state_raw    = BUTTONS_STATE_DEFAULT;
      dev->state_ungrab = BUTTONS_STATE_DEFAULT;
    }

#ifdef	MODULE
  MOD_DEC_USE_COUNT;
#endif
}

#ifdef	CONFIG_POWERMGR
/*
 * power-management operations
 *
 */

static int buttons_suspend_check(void *ignore, int idle_jiffies)
{
  /*
   * modifies: nothing
   * effects:  Always returns zero; i.e. willing to sleep immediately.
   *
   */

  return(0);
}

static int buttons_suspend(void *ignore)
{
  /*
   * modifies: buttons_global (global)
   * effects:  Saves button state in preparation for sleep mode.
   *           Always returns zero (ignored).
   *
   */

  buttons_dev *dev = &buttons_global;  

  /* debugging */
  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_suspend\n");

  /* disable button interrupts, if necessary */
  if (dev->nclients > 0)
    {
      codec_buttons_disable(dev->mcp);
      (void) mcp_common_disable_irq(dev->mcp_id);
    }

  /* return value ignored */
  return(0);
}

static int buttons_resume(void *ignore, int resume_flags)
{
  /*
   * modifies: buttons_global (global)
   * effects:  Resumes buttons state after wakeup from sleep mode.
   *           Always returns zero (ignored).
   *
   */

  buttons_dev *dev = &buttons_global;

  /* debugging */
  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_resume\n");

  /* re-enable button interrupts, if necessary */
  if (dev->nclients > 0)
    {
      (void) mcp_common_enable_irq(dev->mcp_id);
      codec_buttons_enable(dev->mcp);
    }

  /* return value ignored */
  return(0);
}

static const powermgr_client buttons_powermgr = {
  /* callback functions */
  buttons_suspend_check,
  buttons_suspend,
  buttons_resume,
  
  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  BUTTONS_NAME_VERBOSE,

  /* power-consumption info */
  0,
  0
};
#endif	/* CONFIG_POWERMGR */

/*
 * session management operations
 *
 */

static void buttons_session_switchto(void *private, int sid)
{
  /* update active session id */
  buttons_dev *dev = (buttons_dev *) private;
  dev->sm_active = sid;

  /* debugging */
  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_session_switchto: sid=%d\n", sid);
}

static void buttons_session_destroy(void *private, int sid)
{
  /* reset active session id, if necessary */
  buttons_dev *dev = (buttons_dev *) private;
  if (dev->sm_active == sid)
    dev->sm_active = itsy_session_get_foreground();

  /* debugging */
  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": buttons_session_destroy: sid=%d, active=%d\n",
	   sid, dev->sm_active);
}

/*
 * initialization
 *
 */

static struct file_operations buttons_fops =
{
  NULL,			/* no special lseek */
  buttons_read,
  buttons_write,	
  NULL,			/* no special readdir */
  buttons_select,
  buttons_ioctl,
  NULL,			/* no special mmap */
  buttons_open,
  buttons_release,
  NULL,			/* no special fsync */
  buttons_fasync,
  NULL,			/* no special check_media_change */
  NULL			/* no special revalidate */
};

static char *buttons_version(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current cvs revision number.
   * storage:  Return value is statically-allocated.
   *
   */

  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = BUTTONS_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return(version);
}

int itsy_buttons_init(void)
{
  static int initialized = 0;

  buttons_dev *dev = &buttons_global;
  int err;

  /* initialize only once */
  if (initialized++)
    return(-EPERM);

  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": itsy_buttons_init\n");

  /* initialize global state */
  buttons_dev_init(dev);

  /* register device */
  if ((err = register_chrdev(BUTTONS_MAJOR, BUTTONS_NAME, &buttons_fops)) != 0)
    {
      /* log error and fail */
      printk(BUTTONS_NAME ": unable to register major device %d\n",
	     BUTTONS_MAJOR);
      return(err);
    }
      
  /* register with mcp common module */
  if ((dev->mcp_id = mcp_common_register()) == MCP_COMMON_ID_INVALID)
    {
      /* log error and fail */
      printk(BUTTONS_NAME ": unable to register with mcp module\n");
      return(-EBUSY);
    }

#ifdef	CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &buttons_proc_entry);
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_POWERMGR
  /* register with power manager */
  dev->pm_id = powermgr_register(&buttons_powermgr);
  if (dev->pm_id == POWERMGR_ID_INVALID)
    if (BUTTONS_DEBUG)
      printk(BUTTONS_NAME ": unable to register with power manager\n");
#endif	/* CONFIG_POWERMGR */

  /* register with session manager */
  dev->sm_id = itsy_session_notify_register((void *) dev,
					    BUTTONS_NAME,
					    NULL,
					    buttons_session_destroy,
					    buttons_session_switchto,
					    NULL);
  if (dev->sm_id == ITSY_SESSION_NOTIFY_ID_INVALID)
    if (BUTTONS_DEBUG)
      printk(BUTTONS_NAME ": unable to register with session manager\n");
  dev->sm_active = itsy_session_get_foreground();

  /* log device registration */
  printk(BUTTONS_NAME_VERBOSE " version %s initialized\n", buttons_version());

  /* everything OK */
  return(0);
}

#ifdef	MODULE
int init_module(void)
{
  if (BUTTONS_DEBUG_VERBOSE)
    printk(BUTTONS_NAME ": init_module\n");

  return(itsy_buttons_init());
}

void cleanup_module(void)
{
  buttons_dev *dev = &buttons_global;
  int err;

  /* unregister with mcp common module */
  if (mcp_common_unregister(dev->mcp_id) != 0)
    printk(BUTTONS_NAME ": unable to unregister with mcp module\n");

  /* unregister with session manager */
  if (itsy_session_notify_unregister(dev->sm_id) != 0)
    printk(BUTTONS_NAME ": unable to unregister with session manager\n");

#ifdef	CONFIG_POWERMGR
  /* unregister with power manager */
  if (powermgr_unregister(dev->pm_id) != POWERMGR_SUCCESS)
    printk(BUTTONS_NAME ": unable to unregister with power manager\n");
#endif	/* CONFIG_POWERMGR */

#ifdef	CONFIG_PROC_FS
  /* unregister procfs entry */
  if ((err = proc_unregister(&proc_root, buttons_proc_entry.low_ino)) != 0)
    if (BUTTONS_DEBUG)
      printk(BUTTONS_NAME ": cleanup_module: proc_unregister(buttons) %d\n",
	     err);
#endif	/* CONFIG_PROC_FS */

  /* unregister driver */
  unregister_chrdev(BUTTONS_MAJOR, BUTTONS_NAME);

  /* log device unload */
  printk(BUTTONS_NAME_VERBOSE " unloaded\n");
}
#endif	/* MODULE */
