/*
 * Itsy LED Device
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * Author: Carl Waldspurger
 *
 * $Log: itsy_led.c,v $
 * Revision 2.2  1999/12/23  01:27:56  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.1  1999/11/09  01:13:32  caw
 * Initial revision.
 *
 */

/*
 * Overview
 *
 * This driver supports simple operations on the Itsy V2 LED 
 * (white LED, located near encoder wheel).  
 *
 */

/*
 * debugging 
 *
 */

#define	LED_DEBUG			(1)
#define	LED_DEBUG_VERBOSE		(0)

/*
 * includes
 *
 */

#include <linux/config.h>

#ifdef	MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/types.h>
#include <linux/ctype.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef	CONFIG_POWERMGR
#include <linux/powermgr.h>
#include <asm/arch/irq.h>
#endif

#include <asm/segment.h>
#include <asm/io_reg.h>

#include <linux/itsy_session.h>
#include <linux/itsy_codec.h>
#include <linux/mcp_common.h>
#include <linux/itsy_led.h>
#include <linux/itsy_version.h>

#include "itsy_util.h"

/*
 * constants
 *
 */

/* names */
#define	LED_NAME			"itsy-led"
#define	LED_NAME_VERBOSE		"itsy led driver"
#define	LED_VERSION_STRING		"$Revision: 2.2 $"

/* device numbers */
#define	LED_MAJOR			(121)
#define	LED_MINOR			(0)

/* codec */
#define	CODEC_LED_MASK			(0x0001)

/*
 * types
 *
 */

typedef struct {
  /* memory-mapped regs */
  mcp_reg *mcp;

  /* mcp common id */
  mcp_common_id mcp_id;

  /* statistics */
  ulong nclients;
  ulong set_on_count;
  ulong set_off_count;
  ulong get_status_count;

#ifdef	CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
  int pm_state;
#endif
} led_dev;

/*
 * globals
 *
 */

static led_dev led_global;

#ifdef	CONFIG_PROC_FS
static int led_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry led_proc_entry = {
  0, 3, "led", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, led_procfs_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * codec operations
 *
 */

static int codec_get_led(led_dev *dev)
{
  mcp_reg *mcp = dev->mcp;
  ushort data;

  /* enable codec I/O */
  (void) mcp_common_enable(dev->mcp_id);

  /* snapshot data */
  data = codec_read(mcp, CODEC_REG_IO_DATA);

  /* disable codec I/O */
  (void) mcp_common_disable(dev->mcp_id);  

  /* return 1 iff led is on */
  return(data & CODEC_LED_MASK ? 0 : 1);
}

static void codec_set_led(led_dev *dev, int enable)
{
  mcp_reg *mcp = dev->mcp;
  ushort data;

  /* enable codec I/O */
  (void) mcp_common_enable(dev->mcp_id);

  /* snapshot data */
  data = codec_read(mcp, CODEC_REG_IO_DATA);

  /* codec gpio 0 is !LEDEN */
  if (enable)
    codec_write(mcp, CODEC_REG_IO_DATA, data & ~CODEC_LED_MASK);
  else
    codec_write(mcp, CODEC_REG_IO_DATA, data | CODEC_LED_MASK);

  /* disable codec I/O */
  (void) mcp_common_disable(dev->mcp_id);  
}

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int led_procfs_info(char *buf,
			   char **start,
			   off_t offset,
			   int length,
			   int unused)
{
  led_dev *dev = &led_global;
  int len = 0;
  int status;

  /* query led status */
  status = codec_get_led(dev);

  /* format statistics */
  len += sprintf(buf + len, 
		 "status:\n"
		 "  %6d (%s)\n"
		 "statistics:\n"
		 "  %6d clients\n"
		 "  %6d set on\n"
		 "  %6d set off\n"
		 "  %6d get status\n",
		 status,
		 status ? "on" : "off",
		 dev->nclients,
		 dev->set_on_count,
		 dev->set_off_count,
		 dev->get_status_count);

  return(len);  
}
#endif	/* CONFIG_PROC_FS */

/*
 * led_dev operations
 *
 */

static int led_dev_init(led_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Initializes led device dev.
   *	       Returns 0 iff successful, otherwise error code.
   *
   */

  /* clear everything */
  KOBJ_INIT(dev, led_dev);
  
  /* initialize mcp */
  dev->mcp = MCP_BASE_V;
  dev->mcp_id = MCP_COMMON_ID_INVALID;

#ifdef	CONFIG_POWERMGR
  /* initialize powermgr state */
  dev->pm_id = POWERMGR_ID_INVALID;
  dev->pm_state = 0;
#endif	/* CONFIG_POWERMGR */

  /* everything OK */
  return(0);
}

/*
 * device operations
 *
 */

static int led_ioctl(struct inode *inode,
		     struct file *file,
		     uint command,
		     ulong arg)
{
  led_dev *dev = &led_global;
  int size, err;

  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": led_ioctl: cmd=%d, arg=%lu\n", command, arg);

  /* sanity checks */
  if (_IOC_TYPE(command) != LED_IOCTL_TYPE)
    return(-EINVAL);
  if (_IOC_NR(command) > LED_IOCTL_MAXNR)
    return(-EINVAL);

  /* verify transfer memory */
  size = _IOC_SIZE(command);
  if (_IOC_DIR(command) & _IOC_READ)
    if ((err = verify_area(VERIFY_WRITE, (void *) arg, size)) != 0)
      return(err);
  if (_IOC_DIR(command) & _IOC_WRITE)
    if ((err = verify_area(VERIFY_READ,  (void *) arg, size)) != 0)
      return(err);

  /* dispatch based on command */
  switch (command) {
  case LED_SET_ON:
    dev->set_on_count++;
    codec_set_led(dev, 1);
    return(0);
  case LED_SET_OFF:
    dev->set_off_count++;
    codec_set_led(dev, 0);
    return(0);
  case LED_GET_STATUS:
    dev->get_status_count++;
    put_user(codec_get_led(dev), (int *) arg);
    return(0);
  }

  /* invalid command */
  return(-EINVAL);
}

static int led_open(struct inode *inode, struct file *file)
{
  led_dev *dev = &led_global;

  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": led_open\n");

  /* update statistics */
  dev->nclients++;

#ifdef	MODULE
  MOD_INC_USE_COUNT;
#endif

  /* everything OK */
  return(0);
}

static void led_release(struct inode *inode, struct file *file)
{
  led_dev *dev = &led_global;

  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": led_release\n");

  /* update statistics */
  dev->nclients--;

#ifdef	MODULE
  MOD_DEC_USE_COUNT;
#endif
}

#ifdef	CONFIG_POWERMGR
/*
 * power-management operations
 *
 */

static int led_suspend_check(void *ignore, int idle_jiffies)
{
  /*
   * modifies: nothing
   * effects:  Always returns zero; i.e. willing to sleep immediately.
   *
   */

  return(0);
}

static int led_suspend(void *ignore)
{
  /*
   * modifies: led_global (global)
   * effects:  Saves led state in preparation for sleep mode.
   *           Always returns zero (ignored).
   *
   */

  led_dev *dev = &led_global;  

  /* debugging */
  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": led_suspend\n");

  /* save current state */
  dev->pm_state = codec_get_led(dev);

  /* return value ignored */
  return(0);
}

static int led_resume(void *ignore, int resume_flags)
{
  /*
   * modifies: led_global (global)
   * effects:  Resumes led state after wakeup from sleep mode.
   *           Always returns zero (ignored).
   *
   */

  led_dev *dev = &led_global;

  /* debugging */
  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": led_resume\n");

  /* restore previous state */
  codec_set_led(dev, dev->pm_state);

  /* return value ignored */
  return(0);
}

static const powermgr_client led_powermgr = {
  /* callback functions */
  led_suspend_check,
  led_suspend,
  led_resume,
  
  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  LED_NAME_VERBOSE,

  /* power-consumption info */
  0,
  0
};
#endif	/* CONFIG_POWERMGR */

/*
 * initialization
 *
 */

static struct file_operations led_fops =
{
  NULL,			/* no special lseek */
  NULL,			/* no special read */
  NULL,			/* no special write */
  NULL,			/* no special readdir */
  NULL,			/* no special select */
  led_ioctl,
  NULL,			/* no special mmap */
  led_open,
  led_release,
  NULL,			/* no special fsync */
  NULL,			/* no special fasync */
  NULL,			/* no special check_media_change */
  NULL			/* no special revalidate */
};

static char *led_version(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current cvs revision number.
   * storage:  Return value is statically-allocated.
   *
   */

  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = LED_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return(version);
}

int itsy_led_init(void)
{
  static int initialized = 0;

  led_dev *dev = &led_global;
  int err;

  /* device exists only on Itsy V2s */
  if (!is_itsy_v2())
    return(-ENXIO);

  /* initialize only once */
  if (initialized++)
    return(-EPERM);

  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": itsy_led_init\n");

  /* initialize global state */
  led_dev_init(dev);

  /* register device */
  if ((err = register_chrdev(LED_MAJOR, LED_NAME, &led_fops)) != 0)
    {
      /* log error and fail */
      printk(LED_NAME ": unable to register major device %d\n", LED_MAJOR);
      return(err);
    }
      
  /* register with mcp common module */
  if ((dev->mcp_id = mcp_common_register()) == MCP_COMMON_ID_INVALID)
    {
      /* log error and fail */
      printk(LED_NAME ": unable to register with mcp module\n");
      return(-EBUSY);
    }

#ifdef	CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &led_proc_entry);
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_POWERMGR
  /* register with power manager */
  dev->pm_id = powermgr_register(&led_powermgr);
  if (dev->pm_id == POWERMGR_ID_INVALID)
    if (LED_DEBUG)
      printk(LED_NAME ": unable to register with power manager\n");
#endif	/* CONFIG_POWERMGR */

  /* log device registration */
  printk(LED_NAME_VERBOSE " version %s initialized\n", led_version());

  /* everything OK */
  return(0);
}

#ifdef	MODULE
int init_module(void)
{
  if (LED_DEBUG_VERBOSE)
    printk(LED_NAME ": init_module\n");

  return(itsy_led_init());
}

void cleanup_module(void)
{
  led_dev *dev = &led_global;
  int err;

  /* unregister with mcp common module */
  if (mcp_common_unregister(dev->mcp_id) != 0)
    printk(LED_NAME ": unable to unregister with mcp module\n");

#ifdef	CONFIG_POWERMGR
  /* unregister with power manager */
  if (powermgr_unregister(dev->pm_id) != POWERMGR_SUCCESS)
    printk(LED_NAME ": unable to unregister with power manager\n");
#endif	/* CONFIG_POWERMGR */

#ifdef	CONFIG_PROC_FS
  /* unregister procfs entry */
  if ((err = proc_unregister(&proc_root, led_proc_entry.low_ino)) != 0)
    if (LED_DEBUG)
      printk(LED_NAME ": cleanup_module: proc_unregister(led) %d\n", err);
#endif	/* CONFIG_PROC_FS */

  /* unregister driver */
  unregister_chrdev(LED_MAJOR, LED_NAME);

  /* log device unload */
  printk(LED_NAME_VERBOSE " unloaded\n");
}
#endif	/* MODULE */
