/*
 * Itsy Session Manager Device
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * Author: Carl Waldspurger
 *
 * $Log: itsy_session.c,v $
 * Revision 2.14  1999/12/23  01:27:57  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.13  1999/09/10  21:41:15  caw
 * Updated to reflect changes to itsy_event_header field names.
 *
 * Revision 2.12  1999/06/10  20:15:50  caw
 * Added support for session manager requests.
 *
 * Revision 2.11  1999/05/21 21:30:04  caw
 * Removed temporary debugging code (MANAGER_DEATH_HACK).
 *
 * Revision 2.9  1999/05/20  22:39:58  caw
 * Updated to reflect new "override" option to itsy_session_set_name().
 *
 * Revision 2.8  1999/05/12  18:01:02  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.7.2.7  1999/04/30  00:33:57  caw
 * Implemented SESSION_REAP operation.
 *
 * Revision 2.7.2.6  1999/04/22  21:51:28  caw
 * Uncommented call to console_session_register().
 *
 * Revision 2.7.2.5  1999/04/22  16:53:23  caw
 * Export more general itsy_session_signal() operation.
 * Added ioctls to terminate/kill session tasks.
 *
 * Revision 2.7.2.4  1999/04/22  00:18:37  caw
 * Implemented kernel and ioctl interfaces to stop/continue session tasks.
 *
 * Revision 2.7.2.3  1999/04/21  01:00:39  caw
 * Minor bug fix to prevent redundant initializations.
 *
 * Revision 2.7.2.2  1999/04/19  23:13:00  caw
 * Updated itsy_session_destroy() to remove client from session list
 * before invoking notification callbacks.
 *
 * Revision 2.7.2.1  1999/04/15  22:04:56  caw
 * When CONFIG_ITSY_SESSION not defined, create single "default" session.
 * Updated to use itsy_session_from_task().
 *
 * Revision 2.7  1999/03/30  01:37:10  caw
 * Updated to include macros from "itsy_util.h".
 *
 * Revision 2.6  1999/02/05 01:31:34  caw
 * Completed sid-filtered session_select() code.  Disabled verbose debugging.
 *
 * Revision 2.5  1999/02/04  23:08:01  caw
 * Bug fixes: initialize ntotal in session_get_ids(), fixed session_select().
 * Added support for NAME and DATA events.  Limited testing (as opposed to
 * almost no testing with early checkins) indicates this version is stable.
 *
 * Revision 2.4  1999/02/04  19:23:01  caw
 * Bug fix: properly check result of session_instance_lookup().
 *
 * Revision 2.3  1999/02/03  18:36:30  caw
 * Several bug fixes.  Updated ioctl interface.  Exported kernel symbols.
 * Removed module-specific code, since must compile into kernel.
 *
 * Revision 2.2  1999/02/02  03:04:31  caw
 * Updated to reflect modified ioctl interface.  Improved procfs info.
 *
 * Revision 2.1  1999/02/02  01:09:37  caw
 * Initial revision.
 *
 */

/*
 * "to do" list
 *
 *   - change GFP_KERNEL to GFP_ATOMIC where necessary
 *   - more rigorous testing
 *   - better access control
 *
 */

/*
 * debugging 
 *
 */

#define	SESSION_DEBUG			(1)
#define	SESSION_DEBUG_VERBOSE		(0)

/*
 * includes
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/ctype.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#include <asm/segment.h>

#include <linux/itsy_session.h>
#include "itsy_util.h"

/*
 * constants
 *
 */

/* names */
#define	SESSION_NAME			"itsy-session"
#define	SESSION_NAME_VERBOSE		"itsy session manager"
#define	SESSION_VERSION_STRING		"$Revision: 2.14 $"

/* device numbers */
#define	SESSION_MAJOR			(61)
#define	SESSION_RAW_MINOR		(0)

/* buffer size */
#define	SESSION_BUFFER_LG_NEVENTS	(8)
#define	SESSION_BUFFER_NEVENTS		(1 << SESSION_BUFFER_LG_NEVENTS)
#define	SESSION_BUFFER_INDEX_MASK	(SESSION_BUFFER_NEVENTS - 1)

/* reap size */
#define	SESSION_REAP_MAX		(16)

/*
 * types
 *
 */

typedef struct notify_client {
  /* linked list */
  struct notify_client *prev, *next;

  /* identity */
  int id;
  char *name;

  /* notification callbacks */
  session_notify_callback create;
  session_notify_callback destroy;
  session_notify_callback switchto;
  session_query_callback  query;

  /* uninterpreted data */
  void *private_data;
} notify_client;

typedef struct session_client {
  /* linked list */
  struct session_client *prev, *next;

  /* identity */
  int id;

  /* buffer management */
  int buf_next;

  /* access control */
  int manager;

  /* statistics */
  int drop_count;
} session_client;

typedef struct session_instance {
  /* linked list */
  struct session_instance *prev, *next;

  /* identity */
  int id;
  char *name;
  int name_override;

  /* uninterpreted data */
  void *data;
  int data_size;

  /* buffer management */
  struct wait_queue *waitq;

  /* status */
  int foreground;

  /* statistics */
  int switch_count;
} session_instance;

typedef struct {
  /* active sessions */
  session_instance *sessions;

  /* buffer management */
  struct wait_queue *waitq;  
  itsy_session_event *buf;
  int buf_events;
  int buf_bytes;

  /* buffer indexing */
  int buf_mask;
  int buf_next;

  /* notifications */
  notify_client *notify;
  int notify_id_next;

  /* clients */
  session_client *clients;
  int client_id_next;

  /* statistics */
  int switch_count;
} session_dev;

/*
 * globals
 *
 */

static session_dev session_global;

#ifdef	CONFIG_PROC_FS
static int session_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry session_proc_entry = {
  0, 7, "session", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, session_procfs_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * instantiate list operations
 *
 */

INSTANTIATE_LIST_LOOKUP(session_instance)
INSTANTIATE_LIST_INSERT(session_instance)
INSTANTIATE_LIST_REMOVE(session_instance)

INSTANTIATE_LIST_LOOKUP(notify_client)
INSTANTIATE_LIST_INSERT(notify_client)
INSTANTIATE_LIST_REMOVE(notify_client)

INSTANTIATE_LIST_INSERT(session_client)
INSTANTIATE_LIST_REMOVE(session_client)

/*
 * utility routines
 *
 */

static char *kstrdup(const char *s)
{
  char *dup;

  /* special case: empty string */
  if (s == NULL)
    return(NULL);

  /* allocate storage */
  if ((dup = kmalloc(strlen(s) + 1, GFP_KERNEL)) == NULL)
    return(NULL);

  /* clone string */
  strcpy(dup, s);
  return(dup);
}

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int session_procfs_info(char *buf, 
			       char **start,
			       off_t offset,
			       int length,
			       int unused)
{
  session_dev *dev = &session_global;
  int len = 0;

  /* debugging */
  len += sprintf(buf + len, 
		 "statistics:\n"
		 "%5d events\n"
		 "%5d switches\n",
		 dev->buf_next,
		 dev->switch_count);

  /* format session info */
  if (dev->sessions != NULL)
    {
      session_instance *s;
      notify_client *n;

      len += sprintf(buf + len, "sessions:\n");
      for (s = dev->sessions; s != NULL; s = s->next)
	{
	  len += sprintf(buf + len,
			 "%5d %c %5d %-32s\n",
			 s->id,
			 (s->foreground ? 'F' : ' '),
			 s->switch_count,
			 SAFE_STRING(s->name));

	  if (SESSION_DEBUG)
	    for (n = dev->notify; n != NULL; n = n->next)
	      if (n->query != NULL)
		{
		  char *result = (*(n->query))(n->private_data, s->id);
		  len += sprintf(buf + len,
				 "    %5d %-32s %s\n",
				 n->id,
				 SAFE_STRING(n->name),
				 SAFE_STRING(result));
		}
	}
    }

  /* format notifier info */
  if (dev->notify != NULL)
    {
      notify_client *n;

      len += sprintf(buf + len, "notify clients:\n");
      for (n = dev->notify; n != NULL; n = n->next)
	len += sprintf(buf + len,
		       "%5d %-32s\n",
		       n->id, 
		       (n->name ? n->name : ""));
    }

  /* format client info */
  if (dev->clients != NULL)
    {
      session_client *c;

      len += sprintf(buf + len, "session clients:\n");
      for (c = dev->clients; c != NULL; c = c->next)
	len += sprintf(buf + len, 
		       "%5d %c %5d\n",
		       c->id,
		       (c->manager ? 'M' : ' '),
		       c->drop_count);
    }
  
  return(len);  
}
#endif	/* CONFIG_PROC_FS */

/*
 * session_client operations
 *
 */

static session_client *session_client_create(int id)
{
  /*
   * modifies: nothing
   * effects:  Returns a new, initialized session_client, or
   *	       NULL if unable to allocate storage.
   */

  session_client *client;

  /* allocate storage, fail if unable */
  if ((client = KOBJ_ALLOC(session_client)) == NULL)
    return(NULL);
  KOBJ_INIT(client, session_client);

  /* set identity */
  client->id  = id;

  /* everything OK */
  return(client);
}

static void session_client_destroy(session_client *client)
{
  /*
   * modifies: client
   * effects:  Reclaims all storage associated with client.
   *
   */

  kfree(client);
}

/*
 * session_dev operations
 *
 */

static int session_dev_init(session_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Initializes session device dev.
   *	       Returns 0 iff successful, otherwise error code.
   *
   */

  /* clear everything */
  KOBJ_INIT(dev, session_dev);
  
  /* allocate buffer, fail if unable */
  dev->buf_events = SESSION_BUFFER_NEVENTS;
  dev->buf_bytes  = dev->buf_events * sizeof(itsy_session_event);
  dev->buf = (itsy_session_event *) kmalloc(dev->buf_bytes, GFP_KERNEL);
  if (dev->buf == NULL)
    return(-ENOMEM);

  /* buffer indexing */
  dev->buf_mask = SESSION_BUFFER_INDEX_MASK;
  dev->buf_next = 0;
  
  /* everything OK */
  return(0);
}

static void session_dev_event_prim(session_dev *dev,
				   int sid,
				   unsigned char event_type,
				   unsigned char flags,
				   itsy_session_request *r)
{
  /*
   * modifies: dev
   * effects:  Creates event with specified sid, event_type, and request.
   *           The event is added to the buffer associated with dev. 
   *           Wakes up any clients blocked waiting for the event.
   *           Note that the request r may be NULL.
   *
   */

  itsy_session_event *next;
  session_instance *s;

  /* use next slot in circular buffer */
  next = &dev->buf[dev->buf_next & dev->buf_mask];
  
  /* generic session event */
  do_gettimeofday(&next->common.timestamp);
  next->common.event_class = ITSY_EVENT_CLASS_SESSION;
  next->common.size  = sizeof(itsy_session_event);
  next->common.flags = 0;

  /* specific session event */
  next->common.event_type = event_type;
  next->sid = sid;

  /* session request data */
  if (r == NULL)
    memset(&next->request, 0, sizeof(itsy_session_request));    
  else
    memcpy(&next->request, (const void *) r, sizeof(itsy_session_request));
  
  /* advance buffer index */
  dev->buf_next++;

  /* wakeup waiters for raw, if any */
  wake_up_interruptible(&dev->waitq);
  
  /* wakeup waiters for sid, if any */
  if ((s = session_instance_lookup(dev->sessions, sid)) != NULL)
    wake_up_interruptible(&s->waitq);
}

static inline void session_dev_event(session_dev *dev,
				     int sid,
				     unsigned char event_type)
{
  /*
   * modifies: dev
   * effects:  Simplified wrapper for session_dev_event_prim().
   *
   */

  session_dev_event_prim(dev, sid, event_type, ITSY_EVENT_FLAG_NONE, NULL);
}

/*
 * exported session interface 
 *
 */

int itsy_session_signal(int sid, ulong sig)
{
  /*
   * modifies: nothing
   * effects:  Sends signal sig to all tasks associated with sid.
   *           Returns count of tasks successfully signalled.
   *
   */

  struct task_struct *t;
  int nsignal = 0;

  for_each_task(t)
    {
      /* signal each task in session */
      if (itsy_session_from_task(t) == sid)
	if (send_sig(sig, t, 1) == 0)
	  nsignal++;
    }

  return(nsignal);
}

const char *itsy_session_get_name(int sid)
{
  /*
   * modifies: nothing
   * effects:  Returns the session name associated with sid, or
   *           NULL if no such association exists.
   *
   */ 

  session_dev *dev = &session_global;
  session_instance *s;  

  /* lookup session, fail if not found */
  if ((s = session_instance_lookup(dev->sessions, sid)) == NULL)
    return(NULL);

  /* everything OK */
  return(s->name);
}

int itsy_session_set_name(int sid, const char *name, int override)
{
  /*
   * modifies: session_global
   * effects:  Sets the name associated with specified session.
   *           Fails if a name has already been set with override,
   *           unless the override flag is again specified.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  session_dev *dev = &session_global;
  session_instance *s;  

  /* lookup session, fail if not found */
  if ((s = session_instance_lookup(dev->sessions, sid)) == NULL)
    return(-EINVAL);

  /* must specify override to set overridden name */
  if (s->name_override && !override)
    return(-EPERM);

  /* reclaim storage */
  if (s->name != NULL)
    kfree(s->name);

  /* associate new name */
  s->name = kstrdup(name);
  s->name_override = override;
  
  /* handle NAME event */
  session_dev_event(dev, sid, ITSY_EVENT_SESSION_NAME);

  /* everything OK */
  return(0);
}

static int itsy_session_set_data_prim(int sid,
				      const void *data,
				      int data_size,
				      int user_space)
{
  /*
   * requires: data is a valid region of at least data_size bytes
   *           user_space is set iff data is a user-space pointer
   * modifies: session_global
   * effects:  Sets uninterpreted data associated with specified session.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  session_dev *dev = &session_global;
  session_instance *s;  

  /* lookup session, fail if not found */
  if ((s = session_instance_lookup(dev->sessions, sid)) == NULL)
    return(-EINVAL);

  /* reclaim storage */
  if (s->data != NULL)
    kfree(s->data);
  s->data = NULL;
  s->data_size = 0;

  /* associate new data */
  if ((data_size > 0) && (data != NULL))
    {
      /* allocate storage */
      if ((s->data = kmalloc(data_size, GFP_KERNEL)) == NULL)
	return(-ENOMEM);
      s->data_size = data_size;

      /* copy data */
      if (user_space)
	memcpy_fromfs(s->data, data, data_size);
      else
	memcpy(s->data, data, data_size);
    }
  
  /* handle DATA event */
  session_dev_event(dev, sid, ITSY_EVENT_SESSION_DATA);

  /* everything OK */
  return(0);   
}

int itsy_session_set_data(int sid, const void *data, int data_size)
{
  /*
   * requires: data is a valid region of at least data_size bytes
   * modifies: session_global
   * effects:  Sets uninterpreted data associated with specified session.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  /* invoke primitive; data is kernel-space pointer */
  return(itsy_session_set_data_prim(sid, data, data_size, 0));
}

int itsy_session_create(int sid, const char *name)
{
  session_dev *dev = &session_global;
  session_instance *s;  
  notify_client *n;
  
  /* allocate storage, fail if unable */
  if ((s = KOBJ_ALLOC(session_instance)) == NULL)
    return(-ENOMEM);

  /* initialize session */
  KOBJ_INIT(s, session_instance);
  s->name_override = 0;
  s->name = kstrdup(name);
  s->id = sid;

  /* add to session list */
  session_instance_insert(&dev->sessions, s);
  
  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": itsy_session_create(%d, %s)\n",
	   sid,
	   SAFE_STRING(name));

  /* invoke notification callbacks */
  for (n = dev->notify; n != NULL; n = n->next)
    if (n->create != NULL)
      (*(n->create))(n->private_data, sid);

  /* handle CREATED event */
  session_dev_event(dev, sid, ITSY_EVENT_SESSION_CREATED);

  /* everything OK */
  return(0);
}

int itsy_session_destroy(int sid)
{
  session_dev *dev = &session_global;
  session_instance *s;  
  notify_client *n;

  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": itsy_session_destroy(%d)\n", sid);

  /* lookup session, fail if not found */
  if ((s = session_instance_lookup(dev->sessions, sid)) == NULL)
    return(-EINVAL);

  /* remove client from session list */
  session_instance_remove(&dev->sessions, s);

  /* invoke notification callbacks */
  for (n = dev->notify; n != NULL; n = n->next)
    if (n->destroy != NULL)
      (*(n->destroy))(n->private_data, sid);

  /* reclaim storage */
  if (s->name != NULL)
    kfree(s->name);
  kfree(s);

  /* handle DESTROYED event */
  session_dev_event(dev, sid, ITSY_EVENT_SESSION_DESTROYED);

  /* everything OK */
  return(0);
}

static session_instance *session_get_foreground_prim(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current foreground session instance, or
   *           NULL if none exists.
   *
   */

  session_dev *dev = &session_global;
  session_instance *s;
  
  /* search for foreground session */
  for (s = dev->sessions; s != NULL; s = s->next)
    if (s->foreground)
      return(s);

  /* not found */
  return(NULL);
}

int itsy_session_get_foreground(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the session id of the current foreground session,
   *           or ITSY_SESSION_ID_NONE if none exists.
   *
   */

  session_instance *s;
  
  /* lookup foreground session */
  if ((s = session_get_foreground_prim()) != NULL)
    return(s->id);

  /* none found */
  return(ITSY_SESSION_ID_NONE);
}

int itsy_session_set_foreground(int sid)
{
  /*
   * modifies: nothing
   * effects:  Sets the foreground session to the specified sid.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  session_dev *dev = &session_global;
  session_instance *old, *s;
  notify_client *n;

  /* lookup session, fail if not found */
  if ((s = session_instance_lookup(dev->sessions, sid)) == NULL)
    return(-EINVAL);

  /* lookup old foreground session */
  old = session_get_foreground_prim();

  /* avoid unnecessary work */
  if (s == old)
    return(0);

  /* invoke notification callbacks */
  for (n = dev->notify; n != NULL; n = n->next)
    if (n->switchto != NULL)
      (*(n->switchto))(n->private_data, sid);

  /* clear old foreground session */
  if (old != NULL)
    {
      old->foreground = 0;
      session_dev_event(dev, old->id, ITSY_EVENT_SESSION_BACKGROUND);
    }      
  
  /* set new foreground session */
  s->foreground = 1;
  session_dev_event(dev, sid, ITSY_EVENT_SESSION_FOREGROUND);

  /* update stats */
  dev->switch_count++;
  s->switch_count++;

  /* everything OK */
  return(0);
}


/*
 * exported session notification interface 
 *
 */

int itsy_session_notify_register(void *private, 
				 const char *name,
				 session_notify_callback create,
				 session_notify_callback destroy,
				 session_notify_callback switchto,
				 session_query_callback  query)
{
  /*
   * modifies: nothing
   * effects:  Register as a session notification client.  Session
   *           notifications occur via the client-specified callbacks.
   *           Returns a unique identifier for use with future calls,
   *           or ITSY_SESSION_NOTIFY_ID_INVALID if unable to register.
   *
   */

  session_dev *dev = &session_global;
  notify_client *client;

  /* allocate client, fail if unable */
  if ((client = KOBJ_ALLOC(notify_client)) == NULL)
    return(ITSY_SESSION_NOTIFY_ID_INVALID);

  /* initialize client */
  KOBJ_INIT(client, notify_client);
  client->private_data = private;
  client->create   = create;
  client->destroy  = destroy;
  client->switchto = switchto;
  client->query    = query;
  client->name = kstrdup(name);
  client->id   = dev->notify_id_next++;

  /* add to notification list */
  notify_client_insert(&dev->notify, client);

  /* everything OK */
  return(client->id);
}

int itsy_session_notify_unregister(int id)
{
  /*
   * modifies: nothing
   * effects:  Deregisters id as a session notification client.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  session_dev *dev = &session_global;
  notify_client *client;  

  /* lookup client */
  if ((client = notify_client_lookup(dev->notify, id)) == NULL)
    return(-EINVAL);

  /* remove client from notification list */
  notify_client_remove(&dev->notify, client);

  /* reclaim storage */
  if (client->name != NULL)
    kfree(client->name);
  kfree(client);

  /* everything OK */
  return(0);
}

/*
 * device operations
 *
 */

static inline int session_raw_device(const struct inode *inode)
{
  /*
   * modifies: nothing
   * effects:  Returns TRUE iff inode is the "raw" session device.
   *
   */

  return(MINOR(inode->i_rdev) == SESSION_RAW_MINOR);
}

static int session_signal_safe(const session_client *client, 
			       int sid, 
			       ulong sig)
{
  /*
   * modifies: nothing
   * effects:  Sends signal sig to all tasks associated with sid.
   *	       Fails if client is not a session manager.
   *           Returns 0 iff successful, else error code.
   *
   */

  /* only manager can signal sessions */
  if (!client->manager)
    return(-EPERM);

  /* send signal */
  itsy_session_signal(sid, sig);

  return(0);
}

static session_client *session_get_manager(session_dev *dev)
{
  /*
   * modifies: nothing
   * effects:  Returns the session manager client for dev, or
   *           NULL if no such manager exists.
   *
   */

  session_client *c;

  /* search for client acting as manager, if any */
  for (c = dev->clients; c != NULL; c = c->next)
    if (c->manager)
      return(c);

  /* no manager found */
  return(NULL);
}

static int session_set_manager(session_dev *dev,
			       session_client *client,
			       int enable)
{
  /*
   * modifies: client
   * effects:  Attempts to set client manager flag to enable.
   *           Only one client in dev can be marked as a manager.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  session_client *m;

  /* disable always OK */
  if (!enable)
    {
      client->manager = 0;
      return(0);
    }

  /* done if already manager */
  if (client->manager)
    return(0);

  /* fail if another client already manager */
  m = session_get_manager(dev);
  if ((m != NULL) && (m != client))
    return(-EPERM);

  /* enable OK */
  client->manager = 1;
  return(0);
}

static int session_manager_request(session_dev *dev,
				   const session_client *client,
				   const itsy_session_request *r)
{
  /*
   * modifies: dev
   * effects:  Sends event for request r to current session manager, if any.
   *           Returns 0 iff successful, or ESRCH if no manager exists.
   *
   */

  itsy_session_request copy;
  unsigned char flags;
  session_client *m;

  /* lookup manager, fail if none */
  if ((m = session_get_manager(dev)) == NULL)
    return(-ESRCH);

  /* copyin request */
  memcpy_fromfs((void *) &copy, (void *) r, sizeof(itsy_session_request));

  /* ensure correct sender */
  copy.sender = itsy_session_from_task(current);

  /* queue request event */
  flags = suser() ? ITSY_EVENT_FLAG_SUSER : ITSY_EVENT_FLAG_NONE;
  session_dev_event_prim(dev,
			 m->id,
			 ITSY_EVENT_SESSION_MANAGER_REQUEST,
			 flags,
			 &copy);
  return(0);
}

static int session_get_id(const session_dev *dev,
			  const session_client *client,
			  int *sid)
{
  /*
   * requires: sid is a user-space pointer
   * modifies: sid
   * effects:  Sets sid to the caller's session id.
   * caveats:  This operation is currently equivalent to the getsid syscall,
   *           but this is not guaranteed to remain true in future versions.
   *
   */

  put_user(itsy_session_from_task(current), sid);
  return(0);
}

static int session_get_foreground(const session_dev *dev, 
				  const session_client *client,
				  int *sid)
{
  /*
   * requires: sid is a user-space pointer
   * modifies: sid
   * effects:  Sets sid to the foreground session id.
   *
   */

  /* invoke primitive, copyout reply value */
  put_user(itsy_session_get_foreground(), sid);
  return(0);
}

static int session_set_foreground(session_dev *dev,
				  const session_client *client,
				  int sid)
{
  /*
   * modifies: dev (indirectly)
   * effects:  Attempts to set the foreground session to sid.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  /* only manager can switch foreground session */
  if (!client->manager)
    return(-EPERM);

  /* invoke primitive */
  return(itsy_session_set_foreground(sid));
}

static int session_get_info(session_dev *dev,
			    session_client *client,
			    itsy_session_info *info)
{
  /*
   * requires: info is a user-space pointer
   * modifies: info
   * effects:  Sets info->name to session name associated with info->sid.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  itsy_session_info copy;
  session_instance *s;

  /* copyin request */
  memcpy_fromfs((void *) &copy, (void *) info, sizeof(itsy_session_info));

  /* lookup session associated with specified sid */
  if ((s = session_instance_lookup(dev->sessions, copy.sid)) == NULL)
    return(-EINVAL);

  /* initialize name */
  KOBJ_INIT(copy.name, itsy_session_name);

  /* make temporary copy of name */
  if (s->name != NULL)
    {
      int len;

      len = strlen(s->name);
      if (len > sizeof(itsy_session_name))
	len = sizeof(itsy_session_name);
      strncpy(copy.name, (const char *) s->name, len);
    }

  /* copyout reply */
  memcpy_tofs((void *) info, (void *) &copy, sizeof(itsy_session_info));

  /* everything OK */
  return(0);
}

static int session_set_info(session_dev *dev,
			    session_client *client,
			    const itsy_session_info *info)
{
  /*
   * requires: info is a user-space pointer
   * modifies: dev (indirectly)
   * effects:  Attempts to set session name for info->sid to info->name.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  itsy_session_info copy;

  /* copyin request */
  memcpy_fromfs((void *) &copy, (void *) info, sizeof(itsy_session_info));

  /* forcibly null-terminate name */
  copy.name[ITSY_SESSION_NAME_LEN - 1] = '\0';

  /* primitive access control */
  if ((itsy_session_from_task(current) != copy.sid) && !client->manager)
    return(-EPERM);

  /* attempt to associate name with sid */
  return(itsy_session_set_name(copy.sid, copy.name, 1));
}

static int session_get_data(session_dev *dev,
			    session_client *client,
			    itsy_session_data *data)
{
  /*
   * requires: data is a user-space pointer
   * modifies: data
   * effects:  Attempts to copy session data for data->sid to data.
   *           Sets data->ndata to actual bytes copied to data->data.
   *           Sets data->ntotal to total bytes available to be copied.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  itsy_session_data copy;
  session_instance *s;
  int err;
  
  /* copyin request */
  memcpy_fromfs((void *) &copy, (void *) data, sizeof(itsy_session_data));

  /* lookup session associated with specified sid */
  if ((s = session_instance_lookup(dev->sessions, copy.sid)) == NULL)
    return(-EINVAL);

  /* determine transfer size */
  copy.ntotal = s->data_size;
  if (copy.ntotal < copy.ndata)
    copy.ndata = copy.ntotal;

  /* copyout uninterpreted data */
  if (copy.ndata > 0)
    {
      if ((err = verify_area(VERIFY_WRITE, copy.data, copy.ndata)) != 0)
	return(err);
      memcpy_tofs((void *) copy.data, s->data, copy.ndata);
    }

  /* copyout reply */
  memcpy_tofs((void *) data, (void *) &copy, sizeof(itsy_session_data));
  
  /* everything OK */
  return(0);
}

static int session_set_data(session_dev *dev,
			    session_client *client,
			    const itsy_session_data *data)
{
  /*
   * requires: data is a user-space pointer
   * modifies: dev (indirectly)
   * effects:  Attempts to set session data for data->sid to data->data.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  itsy_session_data copy;
  int err;
  
  /* copyin request */
  memcpy_fromfs((void *) &copy, (void *) data, sizeof(itsy_session_data));

  /* primitive access control */
  if ((itsy_session_from_task(current) != copy.sid) && !client->manager)
    return(-EPERM);

  /* verify supplied data region */
  if (copy.ndata > 0)
    if ((err = verify_area(VERIFY_READ, copy.data, copy.ndata)) != 0)
      return(err);    

  /* attempt to associate data with sid */
  return(itsy_session_set_data_prim(copy.sid, copy.data, copy.ndata, 1));
}

static int session_get_ids(session_dev *dev,
			   session_client *client,
			   itsy_session_ids *ids)
{
  /*
   * requires: ids is a user-space pointer
   * modifies: ids
   * effects:  Attempts to copy list of active sessions to ids.
   *           Sets ids->nsids to actual number of ids copied to ids->sids.
   *           Sets ids->ntotal to total number of active session ids.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

  itsy_session_ids copy;
  session_instance *s;
  int err, nalloc, nbytes;
  
  /* copyin request */
  memcpy_fromfs((void *) &copy, (void *) ids, sizeof(itsy_session_ids));
  nalloc = copy.nsids;

  /* verify space for ids */
  nbytes = nalloc * sizeof(int);
  if ((err = verify_area(VERIFY_WRITE, (void *) copy.sids, nbytes)) != 0)
    return(err);

  /* copy sessions to ids */
  copy.nsids  = 0;
  copy.ntotal = 0;
  for (s = dev->sessions; s != NULL; s = s->next)
    {
      copy.ntotal++;
      if (copy.nsids < nalloc)
	put_user(s->id, &copy.sids[copy.nsids++]);
    }

  /* copyout ids data */
  memcpy_tofs((void *) ids, (void *) &copy, sizeof(itsy_session_ids));
  
  /* everything OK */
  return(0);
}

static int session_exists(int sid)
{
  struct task_struct *t;

  /* search all tasks for match */
  for_each_task(t)
    {
      /* done if find match */
      if (itsy_session_from_task(t) == sid)
	return(1);
    }
  
  /* not found */
  return(0);
}

static int session_reap(session_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Reaps all sessions associated with dev that no longer exist.
   * caveats:  This operation would be unnecessary if the kernel exit code
   *           didn't seem to sometimes miss an exiting sessions.  :-(
   *
   */

  session_instance *reap[SESSION_REAP_MAX];
  int nreap;
  
  do {
    session_instance *s;
    int i;

    /* initialize */
    nreap = 0;

    /* identify sessions to reap */
    for (s = dev->sessions;
	 (s != NULL) && (nreap < SESSION_REAP_MAX); 
	 s = s->next)
      if (!session_exists(s->id))
	reap[nreap++] = s;
  
    /* destroy non-existent sessions */
    for (i = 0; i < nreap; i++)
      itsy_session_destroy(reap[i]->id);

    /* debugging */
    if (SESSION_DEBUG && (nreap > 0))
      printk(SESSION_NAME ": session_reap: reaped %d sessions\n", nreap);
  } while (nreap == SESSION_REAP_MAX);
      
  /* everything OK */
  return(0);
}

static int session_ioctl(struct inode *inode,
			 struct file *file,
			 uint command,
			 ulong arg)
{
  session_client *client = (session_client *) file->private_data;
  session_dev *dev = &session_global;
  int size, err;

  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": session_ioctl: cmd=%d, arg=%lu\n", command, arg);

  /* sanity checks */
  if (_IOC_TYPE(command) != SESSION_IOCTL_TYPE)
    return(-EINVAL);
  if (_IOC_NR(command) > SESSION_IOCTL_MAXNR)
    return(-EINVAL);

  /* verify transfer memory */
  size = _IOC_SIZE(command);
  if (_IOC_DIR(command) & _IOC_READ)
    if ((err = verify_area(VERIFY_WRITE, (void *) arg, size)) != 0)
      return(err);
  if (_IOC_DIR(command) & _IOC_WRITE)
    if ((err = verify_area(VERIFY_READ,  (void *) arg, size)) != 0)
      return(err);

  /* dispatch based on command */
  switch (command) {
  case SESSION_GET_ID:
    return(session_get_id(dev, client, (int *) arg));
  case SESSION_GET_FOREGROUND:
    return(session_get_foreground(dev, client, (int *) arg));
  case SESSION_SET_FOREGROUND:
    return(session_set_foreground(dev, client, (int) arg));
  case SESSION_GET_INFO:
    return(session_get_info(dev, client, (itsy_session_info *) arg));
  case SESSION_SET_INFO:
    return(session_set_info(dev, client, (const itsy_session_info *) arg));
  case SESSION_GET_DATA:
    return(session_get_data(dev, client, (itsy_session_data *) arg));
  case SESSION_SET_DATA:
    return(session_set_data(dev, client, (const itsy_session_data *) arg));
  case SESSION_SET_MANAGER:
    return(session_set_manager(dev, client, (int) arg));
  case SESSION_GET_IDS:
    return(session_get_ids(dev, client, (itsy_session_ids *) arg));
  case SESSION_SIG_SUSPEND:
    return(session_signal_safe(client, (int) arg, SIGSTOP));
  case SESSION_SIG_CONTINUE:
    return(session_signal_safe(client, (int) arg, SIGCONT));
  case SESSION_SIG_TERMINATE:
    return(session_signal_safe(client, (int) arg, SIGTERM));
  case SESSION_SIG_KILL:
    return(session_signal_safe(client, (int) arg, SIGKILL));
  case SESSION_REAP:
    return(session_reap(dev));
  case SESSION_MANAGER_REQUEST:
    return(session_manager_request(dev,
				   client,
				   (const itsy_session_request *) arg));
  }

  /* invalid command */
  return(-EINVAL);
}

static int session_read(struct inode *inode,
			struct file *file,
			char *buffer,
			int count)
{
  /*
   * modifies: buffer
   * effects:  XXX
   *
   */

  const int event_bytes = sizeof(itsy_session_event);

  session_client *client = (session_client *) file->private_data;
  session_dev *dev = &session_global;
  int total, err, buf_first, raw, sid, next;
  
  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": session_read\n");

  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_WRITE, (void *) buffer, count)) != 0)
    return(err);

  /* initialize */
  raw = session_raw_device(inode);
  sid = itsy_session_from_task(current);
  total = 0;

  /* determine first available buffered event */
  buf_first = MAX(0, dev->buf_next - dev->buf_events);

  /* skip over dropped events */
  if (client->buf_next < buf_first)
    {
      client->buf_next = buf_first;
      client->drop_count++;
    }

  /* transfer integral number of matching events */
  for (next = client->buf_next;
       (next < dev->buf_next) && (count >= event_bytes);
       next++)
    {
      itsy_session_event *event;
      
      /* examine next buffered event */
      event = &dev->buf[next & dev->buf_mask];
      if (raw || (event->sid == sid))
	{
	  /* copyout matching event */
	  memcpy_tofs((void *) buffer, (const void *) event, event_bytes);
	  client->buf_next++;

	  /* update counts */
	  total  += event_bytes;
	  buffer += event_bytes;
	  count  -= event_bytes;
	}
    }

  /* everything OK */
  return(total);
}

static int session_select(struct inode *inode,
			  struct file *file,
			  int mode,
			  select_table *table)
{
  session_client *client = (session_client *) file->private_data;
  session_dev *dev = &session_global;

  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": session_select\n");

  /* handle select on input */
  if (mode == SEL_IN)
    {
      if (session_raw_device(inode))
	{
	  /* any raw events buffered? */
	  if (client->buf_next < dev->buf_next)
	    return(1);

	  /* wait for raw event */
	  if (SESSION_DEBUG_VERBOSE)
	    printk(SESSION_NAME ": session_select: wait (raw)...\n");
	  select_wait(&dev->waitq, table);
	  return(0);
	}
      else
	{
	  int sid = itsy_session_from_task(current);
	  session_instance *s;

	  /* any sid-specific events buffered? */
	  if (client->buf_next < dev->buf_next)
	    {
	      /* scan backwards in buffer looking for match */
	      int buf_first, next;
	      buf_first = MAX(0, dev->buf_next - dev->buf_events);	      
	      buf_first = MAX(buf_first, client->buf_next);
	      for (next = buf_first; next < dev->buf_next; next++)
		if (dev->buf[next & dev->buf_mask].sid == sid)
		  return(1);
	    }

	  /* lookup sid, fail if not found */
	  if ((s = session_instance_lookup(dev->sessions, sid)) == NULL)
	    {
	      if (SESSION_DEBUG)
		printk(SESSION_NAME ": session_select: no sid %d\n", sid);
	      return(0);
	    }

	  /* wait for matching event */
	  if (SESSION_DEBUG_VERBOSE)
	    printk(SESSION_NAME ": session_select: wait (sid %d)...\n", sid);
	  select_wait(&s->waitq, table);
	  return(0);
	}
    }

  /* SEL_OUT, SEL_EX not supported */
  return(0);
}

static int session_open(struct inode *inode, struct file *file)
{
  session_dev *dev = &session_global;
  session_client *client;
  int id;

  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": session_open\n");

  /* add new client, fail if unable */
  id = dev->client_id_next++;
  if ((client = session_client_create(id)) == NULL)
    return(-ENOMEM);

  /* skip over old events */
  client->buf_next = dev->buf_next;

  /* add to client list */
  session_client_insert(&dev->clients, client);

  /* attach client state to file */
  file->private_data = client;

  /* everything OK */
  return(0);
}

static void session_release(struct inode *inode, struct file *file)
{
  session_client *client = (session_client *) file->private_data;
  session_dev *dev = &session_global;

  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": session_release\n");

  /* remove from client list */
  session_client_remove(&dev->clients, client);

  /* reclaim storage */
  session_client_destroy(client);
  file->private_data = NULL;
}

/*
 * initialization
 *
 */

static struct file_operations session_fops =
{
  NULL,			/* no special lseek */
  session_read,
  NULL,			/* no special write */
  NULL,			/* no special readdir */
  session_select,
  session_ioctl,
  NULL,			/* no special mmap */
  session_open,
  session_release,
  NULL,			/* no special fsync */
  NULL,			/* no special fasync */
  NULL,			/* no special check_media_change */
  NULL			/* no special revalidate */
};

static char *session_version(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current cvs revision number.
   * storage:  Return value is statically-allocated.
   *
   */

  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = SESSION_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return(version);
}

/* exported symbols */
static struct symbol_table itsy_session_syms = {
#include <linux/symtab_begin.h>
        X(itsy_session_create),
        X(itsy_session_destroy),
        X(itsy_session_get_foreground),
        X(itsy_session_set_foreground),
        X(itsy_session_get_name),
        X(itsy_session_set_name),
        X(itsy_session_notify_register),
        X(itsy_session_notify_unregister),
#include <linux/symtab_end.h>
};

static int session_init_special(void)
{
  extern void console_session_register(void);

  /* console driver */
  console_session_register();

  /* everything OK */
  return(0);
}

int itsy_session_init(void)
{
  static int initialized = 0;
  int err;

  /* initialize only once */
  if (initialized++)
    return(-EPERM);

  if (SESSION_DEBUG_VERBOSE)
    printk(SESSION_NAME ": itsy_session_init\n");

  /* initialize global state */
  session_dev_init(&session_global);

  /* register device */
  if ((err = register_chrdev(SESSION_MAJOR, SESSION_NAME, &session_fops)) != 0)
    {
      /* log error and fail */
      printk(SESSION_NAME ": unable to register major device %d\n",
	     SESSION_MAJOR);
      return(err);
    }
      
#ifdef	CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &session_proc_entry);
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_MODULES
  /* export symbols */
  (void) register_symtab(&itsy_session_syms);
#endif	/* CONFIG_MODULES */

  /* device-specific initialization */
  if (session_init_special() != 0)
    printk(SESSION_NAME ": unable to intitialize special devices\n");

#ifndef	CONFIG_ITSY_SESSIONS
  /* create single foreground session */
  if (itsy_session_create(ITSY_SESSION_ID_NONE, "default") != 0)
    printk(SESSION_NAME ": unable to create single default session\n");
#endif	/* CONFIG_ITSY_SESSIONS */

  /* log device registration */
  printk(SESSION_NAME_VERBOSE " version %s initialized\n", session_version());

  /* everything OK */
  return(0);
}
