/*
 * Touchscreen driver
 *
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Larry Brakmo.
 *
 * $Log: itsy_ts.c,v $
 * Revision 2.8  1999/12/23  01:27:58  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.7  1999/10/22  23:11:16  brakmo
 * Fixed Itsy freezing problem when LCD was turned off by disabling ADC_SYNC_ENA.
 *
 * Revision 2.6  1999/10/12 00:39:55  brakmo
 * Interim fix for V2 (should also work for V1.*)
 *
 * Revision 2.5  1999/10/07 01:22:06  brakmo
 * Made it V2 compatible (no /dev/battery), cleaned rate per session code.
 *
 * Revision 2.4  1999/09/10 21:41:16  caw
 * Updated to reflect changes to itsy_event_header field names.
 *
 * Revision 2.3  1999/06/10 20:14:59  brakmo
 * Added rate control per session and async support.
 *
 * Revision 2.2  1999/05/25 17:49:58  brakmo
 * Got rid of a printk in adc ioctl.
 *
 * Revision 2.1  1999/05/12 18:01:02  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 1.1.2.8  1999/05/12  03:07:16  brakmo
 * Minor changes.
 *
 * Revision 1.1.2.7  1999/05/05 20:33:48  brakmo
 * Changing default calibration will now change it on existing clients using default calibration.
 *
 * Revision 1.1.2.6  1999/05/04 19:41:38  brakmo
 * Small change in the interrupt handler routine.
 *
 * Revision 1.1.2.5  1999/04/27 15:01:20  brakmo
 * Minor change.
 *
 * Revision 1.1.2.4  1999/04/23 18:54:58  brakmo
 * Fixed suspend/resume problem (Itsy would lock on resume).
 *
 * Revision 1.1.2.3  1999/04/23 00:13:34  brakmo
 * Fixed revision number.
 *
 * Revision 2.4  1999/04/23 00:03:34  brakmo
 * Got rid of a spurious printk.
 *
 * Revision 1.1.2.1  1999/04/21 01:41:54  brakmo
 * Added session support to console, fb and touchscreen. Renamed ts.c to itsy_ts.c
 *
 * Revision 2.3  1999/01/20  20:57:06  brakmo
 * Added /proc/lcd, /proc/registers and support for interrupt processing in
 * lcdOn and lcdStatic which enables them to return before LCD change is
 * finalized.
 *
 * Revision 2.2  1998/11/25  01:08:02  brakmo
 * Fixed problem with resume after suspend (powermgr)
 *
 * Revision 2.1  1998/07/02  01:49:48  brakmo
 * Added kernel support for new ggi graphics library.
 *
 * Revision 1.10  1998/06/16  19:19:44  kramer
 * Added copyright information
 *
 * Revision 1.9  1998/03/28  01:24:42  brakmo
 * Added powermgr support for touchscreen.
 *
 * Revision 1.8  1998/03/18  03:49:08  brakmo
 * Fixed a power related bug introduced with the last modification.
 *
 * Revision 1.7  1998/03/17  02:09:29  brakmo
 * We now have 3 framebuffers by default. One is for the console, the other
 * two for applications.
 *
 * Revision 1.6  1998/03/11  16:49:43  brakmo
 * Added mmap of frame buffer, two built in buffers, plus more
 *
 * Revision 1.5  1998/03/06  19:14:40  brakmo
 * New Touchscreen driver and modifications to the console.
 *
 * Revision 2.0  1998/06/24  21:43:44  kramer
 * Major Bump: Move all version numbers up to 2.0 for external release.
 * -drew
 *
 */

/*
 * Overview
 *
 * The general behavior of the driver is as follows
 *
 * 1) If the touchscreen is not being pressed set an interrupt to notify us
 *    when it is pressed.
 * 2) If the touchscreen is being pressed, get touchscreen information
 *    (pressure, x coord, y coord) at the current rate (default is 50/sec).
 *
 * The driver has 2 interrupt handlers. One is for an OS timer (OS timer 1) and
 * the other one to handle some codec (UCB1200) interrupts which arrive to
 * the sa1100 through GPIO6 on Itsy. The codec interrupts that are used are:
 * 1) detect touchscreen press
 *
 * The process for reading the touchscreen is:
 * 1) power specific pins (start)
 * 2) wait until things become stable, around 200us (delay)
 * 3) do adc conversion (read)
 *
 * I use the interrupts of an OS timer (timer1) to implement the delay
 * and a finite state machine to keep track of the state. 
 *
 * The states are:
 *
 * TS_FSM_INIT:		- do adc conversions if necessary, state++
 * TS_FSM_START_CHECK_PEN 	- start check if pen down and delay, state++
 * TS_FSM_CHECK_PEN 	- check if pen down
 *			    if up, start set pen inter, state=TS_FSM_SET_PEN_INT
 *			    if down, start read x and delay, state++
 * TS_FSM_READ_X	- read x, start read y and delay, state++
 * TS_FSM_READ_Y	- read y, 
 * TS_FSM_READ_P	- read pressure, add randomness to system, wake
 *			    procs because we have new data, power pins down,
 *			    state++
 * TS_FSM_DELAY		- disable mcp to conserve power, wait rate_delay,
 *			    state = TS_FSM_INIT
 * TS_FSM_SET_PEN_INT	- set pen inter, disable mcp, stop timer, 
 *			    state = TS_FSM_INIT
 *
 * There is a circular event queue to store touchscreen events. The x and y
 * coordinates are stored in raw format (as read from the adc) and are
 * mapped to the LCD (calibrated) when read by a client. Each client
 * has its own calibration information, and they are initialized to a
 * global default value.
 * Areas of the touchscreen can be "grabbed" by opening the touchscreen
 * grab device (usually /dev/tsgrab) and specifying areas of interests
 * with the TS_ADD_GRAB_AREA ioctl. Touchscreen events on grabbed areas
 * are sent to the grab client regardless of whether it is the currently
 * active session. These events are not seen by the active session unless
 * it opened the "raw" device.
 * A process can request to get ALL touchscreen events by openning the
 * "raw" device (usually /dev/tsraw). In this case it also gets grabbed
 * events, in which case the "grabbed" bit in the event common header
 * flag is set.
 * A process can create synthetic events by writing events to the
 * touchscreen device. It should set the session id of the receiving
 * session or set the sid to ITSY_SESSION_ID_FOREGROUND in which case the
 * currently active session will get them.
 *
 */

/*
 * includes
 */

/* #include "ts_i.h" */

#include <linux/config.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/types.h>
#include <linux/ctype.h>

#ifdef CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef CONFIG_POWERMGR
#include <linux/powermgr.h>
#endif

#include <linux/itsy_session.h>

#include <linux/random.h>
#include <linux/itsy_codec.h>
#include <linux/fs.h>
#include <linux/mm.h>    /* for verify_area */
#include <linux/errno.h> /* for -EBUSY */
#include <asm/segment.h> /* for put_user_byte */

#include <asm/arch/sa1100.h>
#include <asm/arch/SA-1100.h>

#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/io_reg.h>
#include <asm/proc/page.h>
#include <asm/proc/pgtable.h>

#include <asm/arch/mmap.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/arch/ostimers.h>

#define VirtAdd(PhAdd) (io_p2v(PhAdd))
#include <asm/arch/Itsy.h>

#include <linux/itsy_ts.h>
#include <linux/itsy_adc.h>
#include <linux/mcp_common.h>
#include "itsy_util.h"
#include <linux/itsy_version.h>

/*
 * debug options
 */

/* #define TS_DEBUG 1  */

#ifdef TS_DEBUG
#define IF_DEBUG(x) (x)
#else
#define IF_DEBUG(x)
#endif


/* 
 * constants
 */

/* names */
#define TS_NAME 		"ts"
#define TS_NAME_VERBOSE		"touchscreen driver"
#define TS_VERSION_STRING	"$Revision: 2.8 $"

/* device numbers */
#define TS_MAJOR		(44)
#define TS_MINOR		(0)
#define TS_RAW_MINOR		(1)
#define TS_GRAB_MINOR		(2)
#define TS_USTROKE_MINOR	(3)
#define TS_BATTERY_MINOR	(4)
#define TS_ADC_MINOR		(5)

/* #define TS_SESSION_MINOR	(6) */
/* #define TS_SESSION_RAW_MINOR	(7) */

#define TS_RATE			(50)
#define TS_STABILIZE_DELAY	(736) /* NEW  (2000) */
/* in OS timer ticks */
/* interrupt numbers */
#define TS_IRQ			(CODEC_IRQ)
#define TS_TIMER		(IRQ_OST1)

/* buffer size */
#define TS_BUFFER_LG_NEVENTS	(10)
#define TS_BUFFER_NEVENTS	(1 << TS_BUFFER_LG_NEVENTS)
#define TS_BUFFER_INDEX_MASK	(TS_BUFFER_NEVENTS - 1)

/* OS timer */
#define OS_TIMER_RATE		3686400

/* Finite state machine */
#define TS_FSM_INIT		0
#define TS_FSM_START_CHECK_PEN	1
#define TS_FSM_CHECK_PEN	2
#define TS_FSM_READ_X		3
#define TS_FSM_READ_Y		4
#define TS_FSM_READ_P		5
#define TS_FSM_DELAY		6
#define TS_FSM_SET_PEN_INT	7

static int mcpOffFlag=1;

/*
 * macros
 */
#ifndef ABS
#define ABS(x)  ( (x) < 0 ? (-(x)) : (x) )
#endif

#define SWAP_IF_GT(a, b, t) \
  ((a) > (b) ? (t) = (a), (a) = (b), (b) = (t) : (b))
	      
#define MCP_ENABLE(id) \
  (mcpOffFlag ? \
   (mcp_common_enable(id), mcpOffFlag = 0) : 1 \
  )

#define MCP_DISABLE(id) \
  (mcpOffFlag == 0 ? \
   (mcp_common_disable(id), mcpOffFlag = 1): 0 \
  )

/*
 * types
 */

typedef struct ts_client {
  /* linked list */
  struct ts_client *prev, *next;

  /* identity */
  int id;

  /* session id */
  int sid;

  /* asynchronous notifications */
  struct fasync_struct *asyncq;

  /* buffer management */
  struct wait_queue *waitq;
  int buf_next;

  /* statistics */
  int drop_count;

  /* calibration parameters */
  ts_calibration cal;

  /* session rate (points per second */
  /*   int rate; */
  /*   int rate_delay; */

} ts_client;

typedef struct ts_area {
  /* linked list */
  struct ts_area *prev, *next;

  int id;
  
  /* boundaries */
  ts_grab_area garea;

} ts_area;

typedef struct {
  int ts_rate;			/* ts data rate (per second) */
  int ts_rate_delay;		/* OS timer rate (3686400) / ts_rate */
  int ts_stabilize_delay;	/* in OS timer units */

  int ts_fg_rate_delay;		/* rate delay to use (from current foreground*/

  /* touchscreen stuff */
  int minp;
  int p;
  int x;
  int y;
  
  /* battery and ADC stuff */
  int adc_read_flags;		/* desired adc conversions */
  ushort adc_values[4];
  ulong  adc_times[4];		/* time (in jiffies) of last conversion */
  struct wait_queue *adc_waitq[4];	
  
  /* There is a finite state machine, this is the current state */
  int fsm_state;

  /* pen down flag */
  int pen_down;

  /* timer1 alloc flag */
  int timer1_alloc;

  /* mcp_common_id */
  mcp_common_id mcp_id;

  /* buffer management */
  struct wait_queue *raw_waitq;
  ts_event *buf;
  int buf_events;
  int buf_bytes;

  /* buffer indexing */
  int buf_mask;
  int buf_next;
  
  /* grab areas */
  ts_area *areas;
  int area_id_next;
  
  /* clients */
  ts_client *grabber;
  ts_client *rawclients;
  ts_client *clients;
  int client_id_next;

  /* asynchronous notifications */
  struct fasync_struct *raw_asyncq;

  /* statistics */
  ulong interrupt_count;
  ulong event_count;

#ifdef CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
  ulong last_active;
#endif

  /* session stuff */
  int sm_id;
  int sm_active;
} ts_dev;

typedef struct ts_sid_map {
  struct ts_sid_map *prev, *next;
  int id;
  int rate;
  int rate_delay;
} ts_sid_map;

ts_sid_map *tsSidList = NULL;
ts_sid_map *tsSidFreeList = NULL;

INSTANTIATE_LIST_INSERT(ts_sid_map)
INSTANTIATE_LIST_REMOVE(ts_sid_map)
INSTANTIATE_LIST_LOOKUP(ts_sid_map)


/* 
 * globals
 */

static ts_dev ts_global;
static int Reg10=0;		/* copy of codec Reg10 */
static mcp_reg *mcp = (mcp_reg *) MCP_BASE_V;
static gpio_reg *gpio = (gpio_reg *) GPIO_BASE_V;

#define TS_MODE_TIMER		1
#define TS_MODE_INT		2
#define TS_TIMER_CNT_DBG	50
static int ts_mode = 0;
static int ts_timer_cnt = 0;
static int ts_ih_cnt = 0;
static int grab_sel_cnt=0;
static int grab_rd_cnt=0;
static int recovery_cnt=0;
static int addRandomFlag=1;

#ifdef CONFIG_ITSY
static ts_calibration default_cal={97, -31, -65, 227, 1, 50};
#else
static ts_calibration default_cal={101, -55, 80, -29, 1, 50};
#endif

#ifdef CONFIG_PROC_FS
static int ts_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry ts_proc_entry = {
  0, 2, "ts", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, ts_procfs_info,
};
#endif

/*
 * function declarations
 */

static void ts_raw_to_cal (ts_calibration *, ts_point *, ts_point *);
static inline void ts_start_measurements (void);
static void ts_dev_event (int, ushort, ushort, ushort, unsigned char flags);
static inline void ts_clear_codec_int (void);

/*
 * instatiate list operations
 */

INSTANTIATE_LIST_INSERT(ts_client)
INSTANTIATE_LIST_REMOVE(ts_client)
INSTANTIATE_LIST_INSERT(ts_area)
INSTANTIATE_LIST_REMOVE(ts_area)
INSTANTIATE_LIST_LOOKUP(ts_area)
  
/*
 * procfs operations
 */
#ifdef CONFIG_PROC_FS
  
static int ts_procfs_info(char *buf,
			  char **start,
			  off_t offset,
			  int length,
			  int unused)
{
  ts_dev *dev = &ts_global;
  ts_client *client;
  ts_area *ap;
  ts_point p0, p1; 
  int len=0;
  int ngrabclients=0, nrawclients=0, nclients=0, nareas=0;
  int mcpOff;
  uint reg4, icip;
  
  if (dev->grabber) {
    ngrabclients = 1;
    for (ap = dev->areas; ap != NULL; ap = ap->next)
      nareas++;
  }

  for (client = dev->rawclients; client != NULL; client = client->next)
    nrawclients++;

  for (client = dev->clients; client != NULL; client = client->next)
    nclients++;

  mcpOff = mcpOffFlag;
  MCP_ENABLE(ts_global.mcp_id);
  
  /* debugging info */
  len += sprintf(buf+len,
		 "debugging:\n"
		 "  %08x ICIM mask\n"
		 "  %08x ICPR pending\n"
		 "  %08x ICIP IRQ pending\n"
		 "  %08x codec reg2\n"
		 "  %08x codec reg3\n"
		 "  %08x codec reg4\n"
		 "  %08x codec reg9\n"
		 "  %08x codec reg10\n\n",
		 *(uint *) IRQ_ICMR_V,
		 *(uint *) IRQ_ICIP_V,
		 icip = *(uint *) IRQ_ICPR_V,
		 codec_read(mcp, CODEC_REG_RISE_INT_ENABLE),
		 codec_read(mcp, CODEC_REG_FALL_INT_ENABLE),
		 reg4 = codec_read(mcp, CODEC_REG_INT_STATUS),
		 codec_read(mcp, CODEC_REG_TS_CTL),
		 codec_read(mcp, CODEC_REG_ADC_CTL));

  /* shared info */
  len += sprintf(buf+len,
		 "ts shared:\n"
		 "device major: %5d\n"
		 "default rate: %5d/sec\n"
		 "current rate: %5d/sec\n"
		 "mode:         %5s\n"
		 "ts_timer_cnt: %5d\n"
		 "ts_ih_cnt:    %5d\n"
		 "mcpOff:       %5d\n"
		 "recoveries:   %5d\n"
		 "fsm_state:    %5d\n"
		 "raw pressure: %5d\n"
		 "raw x:        %5d\n"
		 "raw y:        %5d\n"
		 "%5d events\n"
		 "%5d grab_sel_cnt\n"
		 "%5d grab_rd_cnt\n"
		 "nclients:     %5d\n"
		 "  grabber:    %5d\n"
		 "  raw:        %5d\n"
		 "  normal:     %5d\n"
		 "grab areas:   %5d\n",
		 TS_MAJOR,
		 dev->ts_rate,
		 OS_TIMER_RATE / dev->ts_fg_rate_delay,
		 (ts_mode == 2 ? "INT" : (ts_mode == 1 ? "TIMER" : "INIT")),
		 ts_timer_cnt,
		 ts_ih_cnt,
		 mcpOff,
		 recovery_cnt,
		 dev->fsm_state,
		 dev->p, dev->x, dev->y,
		 dev->buf_next,
		 grab_sel_cnt,
		 grab_rd_cnt,
		 ngrabclients + nrawclients + nclients,
		 ngrabclients,
		 nrawclients,
		 nclients,
		 nareas);

  if (nareas) {
    for (ap = dev->areas; ap != NULL; ap = ap->next) {
      len += sprintf(buf+len,
		     "  (%d,%d), (%d,%d)  ",
		     ap->garea.p0.x, ap->garea.p0.y,
		     ap->garea.p1.x, ap->garea.p1.y);
      ts_raw_to_cal(&dev->grabber->cal, &ap->garea.p0, &p0); 
      ts_raw_to_cal(&dev->grabber->cal, &ap->garea.p1, &p1);
      len += sprintf(buf+len,
		     "[ (%d,%d), (%d,%d) ]\n",
 		     p0.x, p0.y, p1.x, p1.y); 
    }
  }

  if (mcpOff)
    MCP_DISABLE(ts_global.mcp_id);
  
  /* for test purposes, check if we lost an interrupt! */
  if ((reg4 & (TSPX_INT | TSMX_INT))  &&   !(icip & (1 << TS_IRQ))) {
    recovery_cnt++;
    ts_clear_codec_int();
    ts_start_measurements();
  }
    
  return len;
}
#endif /* CONFIG_PROC_FS */

static int
calibrations_are_equal (ts_calibration *cal1, ts_calibration *cal2)
{
  if (cal1->xscale != cal2->xscale  ||
      cal1->xtrans != cal2->xtrans  ||
      cal1->yscale != cal2->yscale  ||
      cal1->ytrans != cal2->ytrans  ||
      cal1->xyswap != cal2->xyswap)
    return 0;
  else
    return 1;
}


/*
 * Codec related stuff 
 *
 */

static inline void
do_delay (void) 
{
  ts_global.fsm_state++;
  reset_timer1(ts_global.ts_stabilize_delay);
}

static inline void
set_pen_down_inter0 (void)
{
  MCP_ENABLE(ts_global.mcp_id);
  codec_write(mcp, CODEC_REG_ADC_CTL, 0);
  codec_write(mcp, CODEC_REG_TS_CTL, (TSPX_POW | TSMX_POW |
			      TSPY_GND | TSMY_GND | TSC_MODE_INT));
}

static void
set_pen_down_inter1 (void)
{
  int v;
  
  MCP_ENABLE(ts_global.mcp_id);
  /*   printk("setting pen down inter\n"); */
  ts_clear_codec_int();
  set_pen_down_inter0();

  /* do I need to clear interrupts (reg 4)? */

  v = codec_read(mcp, CODEC_REG_FALL_INT_ENABLE);
  codec_write(mcp, CODEC_REG_FALL_INT_ENABLE, v | TSPX_INT | TSMX_INT);

  ts_global.fsm_state = TS_FSM_SET_PEN_INT;
  reset_timer1(ts_global.ts_stabilize_delay);
}

#define ADC_ENA_TYPE_NEW ((LCCR0 & LCCR0_LEN) ? (ADC_ENA_TYPE) : (ADC_ENA))

static inline void
set_read_pressure (void)
{
  MCP_ENABLE(ts_global.mcp_id);
  codec_write(mcp, CODEC_REG_ADC_CTL, (Reg10=ADC_ENA_TYPE_NEW));
  codec_write(mcp, CODEC_REG_TS_CTL,  (TSPX_POW | TSMX_POW | TSPY_GND |
			       TSMY_GND | TSC_MODE_PRESSURE | TSC_BIAS_ENA));
}
	
static inline void
set_read_resistance_x (void)
{			
  MCP_ENABLE(ts_global.mcp_id);
  codec_write(mcp, CODEC_REG_ADC_CTL, (Reg10=ADC_ENA_TYPE_NEW));
  codec_write(mcp, CODEC_REG_TS_CTL, (TSPX_POW | TSMX_GND |
			      TSC_MODE_PRESSURE | TSC_BIAS_ENA));
}
	
static inline void
set_read_resistance_y (void)
{			
  MCP_ENABLE(ts_global.mcp_id);
  codec_write(mcp, CODEC_REG_ADC_CTL, (Reg10=ADC_ENA_TYPE_NEW));
  codec_write(mcp, CODEC_REG_TS_CTL, (TSPY_POW | TSMY_GND |
			      TSC_MODE_PRESSURE | TSC_BIAS_ENA));
}

static inline void
set_read_x_pos (int tsy)
{
  MCP_ENABLE(ts_global.mcp_id);
  codec_write(mcp, CODEC_REG_ADC_CTL, Reg10=(ADC_ENA_TYPE_NEW | tsy));
  /* NEW */
  codec_write(mcp, CODEC_REG_TS_CTL, (TSPY_POW | TSMX_GND |
			      TSC_MODE_PRESSURE | TSC_BIAS_ENA));
  codec_read(mcp, CODEC_REG_RISE_INT_ENABLE);

  codec_write(mcp, CODEC_REG_TS_CTL, (TSPX_POW | TSMX_GND |
			      TSC_MODE_POSITION | TSC_BIAS_ENA));
}

static inline void
set_read_y_pos (int tsx)
{
  MCP_ENABLE(ts_global.mcp_id);
  codec_write(mcp, CODEC_REG_ADC_CTL, Reg10=(ADC_ENA_TYPE_NEW | tsx));
  /* NEW */
  codec_write(mcp, CODEC_REG_TS_CTL, (TSMX_POW | TSMY_GND | 
				      TSC_MODE_PRESSURE | TSC_BIAS_ENA)); 
  codec_read(mcp, CODEC_REG_RISE_INT_ENABLE);

  codec_write(mcp, CODEC_REG_TS_CTL, (TSPY_POW | TSMY_GND |
			      TSC_MODE_POSITION | TSC_BIAS_ENA));
}

static inline void
set_read_adc (int ad)
{
  MCP_ENABLE(ts_global.mcp_id);
  ad = (ad + 4) << 2;
  codec_write(mcp, CODEC_REG_ADC_CTL, Reg10=(ADC_ENA_TYPE_NEW | ad));
  codec_write(mcp, CODEC_REG_TS_CTL, 0);
}

static int
read_adc (void)
{
  /*
   * Starts previously set ADC coversion, waits until done
   * and returns the value.
   * This is used to read ts and adc.
   */
  int v, p;

/*   save_flags_cli(flags); */
  codec_write(mcp, CODEC_REG_ADC_CTL, Reg10 | ADC_START);
  /*   if ((Reg10 | ADC_START) & ADC_SYNC_ENA)  */
  /*     printk("ts_driver: ADC_SYNC_ENA is set!\n");  LSB */
  while (((v = codec_read(mcp, CODEC_REG_ADC_DATA)) & ADC_DAT_VAL) == 0)
    ;

  p = GET_ADC_DATA(v);
/*   restore_flags(flags); */
  return p;
}
  
static void
ts_pin_power_off (void)
{
  MCP_ENABLE(ts_global.mcp_id);
  IF_DEBUG( if (ts_timer_cnt < TS_TIMER_CNT_DBG)
	   printk(TS_NAME ": ts_pin_power_off\n") );
  
  /* unpower all touchscreen related pins */
  codec_write(mcp, 9, 0); 

  /* unpower all ADC related pins */
  codec_write(mcp, 10, 0); 

}

/*
 * interrupt processing
 */

static inline void ts_enable_timer_int (void)
{
  *IRQ_ICMR_V |= 0x08000000;
}

static inline void ts_disable_timer_int (void)
{
  int s;
  ulong flags;

  save_flags_cli(flags);   
  s = *IRQ_ICMR_V;
  *IRQ_ICMR_V = s & ~0x08000000;
  restore_flags(flags); 
}

static inline void ts_clear_codec_int (void)
{
  int s0, s;
  
  MCP_ENABLE(ts_global.mcp_id);
  s0 = codec_read(mcp, CODEC_REG_INT_STATUS);
  s  = s0 & (TSPX_INT | TSMX_INT | ADC_INT);

  /* clear interrupts at codec */
  if (s) {
    codec_write(mcp, CODEC_REG_INT_STATUS, 0);
    codec_write(mcp, CODEC_REG_INT_STATUS, s);
  }

  /* clear interrupt TS_IRQ */
  if (s == s0)
  gpio->edge_detect |= (1 << TS_IRQ);
}

static inline void ts_disable_codec_int (void)
{
  int s;
  ulong flags;

  MCP_ENABLE(ts_global.mcp_id);
  
  /* disable touschreen related interrupts at codec */
  save_flags_cli(flags); 
  s = codec_read(mcp, CODEC_REG_FALL_INT_ENABLE);
  s &= ~(TSPX_INT | TSMX_INT);
  codec_write(mcp, CODEC_REG_FALL_INT_ENABLE, s);
  restore_flags(flags); 
}

static void
ts_disable_all_int (void)
{
  IF_DEBUG( if (ts_timer_cnt < TS_TIMER_CNT_DBG)
	   printk(TS_NAME ": ts_disable_all_int\n")
	   );

  ts_disable_codec_int();
  ts_disable_timer_int();
  mcp_common_disable_irq(ts_global.mcp_id);

}

static inline void ts_start_measurements (void)
{
  /* set state of finite state machine */
  ts_global.fsm_state = TS_FSM_INIT;
  
  setup_timer1(ts_global.ts_stabilize_delay);
  ts_mode = TS_MODE_TIMER;
}

static void 
ts_timer_int_handler (int irq, void *dev_id, struct pt_regs *regs)
{
  int k, v;
  
  ts_timer_cnt++;
  IF_DEBUG( if (ts_timer_cnt < TS_TIMER_CNT_DBG)
	   printk(TS_NAME ": timer_ih fsm_state=%d\n", ts_global.fsm_state)
	   );
  
  MCP_ENABLE(ts_global.mcp_id);
  
  switch (ts_global.fsm_state) {
  case TS_FSM_INIT:
    if (ts_global.adc_read_flags != 0) {
      IF_DEBUG( printk(TS_NAME ": adc_read_flags=0x%x in fsm_init\n",
		       ts_global.adc_read_flags) );
      for (k=0; k<4; k++) {
	if (ts_global.adc_read_flags & (1 << k)) {
	  set_read_adc(k);
	  ts_global.adc_values[k] = read_adc();
	  ts_global.adc_times[k] = jiffies;
	  wake_up_interruptible(&ts_global.adc_waitq[k]);
	}
      }
      ts_global.adc_read_flags = 0;
    }
    do_delay();
    break;
  case TS_FSM_START_CHECK_PEN:
    set_pen_down_inter0();
    do_delay(); 
    break; 
  case TS_FSM_CHECK_PEN:
    v = codec_read(mcp, CODEC_REG_TS_CTL);
    if ((v & (TSPX_LOW | TSMX_LOW)) != 0) {	/* pen is up */
      ts_global.p = 0;
      if (ts_global.pen_down)
	ts_dev_event (ts_global.sm_active, ts_global.p, ts_global.x,
		      ts_global.y, (unsigned char) 0);
      ts_global.pen_down = 0;
      set_pen_down_inter1();
    }
    else {					/* pen is down */
      ts_global.pen_down = 1;
      /*       pen_down_time = jiffies; */
      set_read_x_pos(ADC_INPUT_TSMY); 
      /*       set_read_x_pos(ADC_INPUT_TSPY); */
      do_delay();
    }
    break;  
  case TS_FSM_READ_X:
    ts_global.x = read_adc();
    set_read_y_pos(ADC_INPUT_TSMX); 
    /*     set_read_y_pos(ADC_INPUT_TSPX); */
    do_delay();
    break;
  case TS_FSM_READ_Y:
    ts_global.y = read_adc();
    set_read_pressure();
    do_delay();
    break;  
  case TS_FSM_READ_P:
    ts_global.p = read_adc();
    /* LSB */
    /*
    if (addRandomFlag) {
      add_mouse_randomness((__u32) (ts_global.x | (ts_global.y<<10) | 
				    (ts_global.p<<20)));
      addRandomFlag = 0;
    }
    */
    if (ts_global.p > ts_global.minp) { 
      ts_dev_event (ts_global.sm_active, ts_global.p, ts_global.x,
		    ts_global.y, (unsigned char) 0);
    }
    ts_pin_power_off();
    do_delay();
    break;
  case TS_FSM_DELAY:
    MCP_DISABLE(ts_global.mcp_id);
    ts_global.fsm_state = TS_FSM_INIT;
    reset_timer1(ts_global.ts_fg_rate_delay);
    break;
  case TS_FSM_SET_PEN_INT:
    ts_global.fsm_state = TS_FSM_INIT;
    MCP_DISABLE(ts_global.mcp_id);
    cleanup_timer1();
    ts_mode = TS_MODE_INT;
    break;
  default:
    printk(TS_NAME ": ERROR in ts_timer_int_handler, fsm_state=%d\n",
	   ts_global.fsm_state);
  }
}

static void
ts_codec_int_handler (int irq, void *dev_id, struct pt_regs *regs)
{
  ts_ih_cnt++;
  IF_DEBUG( if (ts_ih_cnt < 15)
	   printk(TS_NAME ": ts_codec_int_handler, cnt=%d\n", ts_ih_cnt)
	   );
  
  addRandomFlag = 1;

  ts_clear_codec_int();  
  ts_disable_codec_int();
  ts_start_measurements();
}

/*
 * coordinate transformations
 */

static void
ts_raw_to_cal (ts_calibration *cal, ts_point *from, ts_point *to)
{
  /*
   * modifies: to
   * effect:   converts point "from" from raw screen coordinates to
   *	       calibrated coordinates and stores them in point "to"
   */
  int fx = from->x, fy = from->y;
  int tx, ty;
  
  if (cal->xyswap) {
    tx = ((fy * cal->xscale) >> 8) + cal->xtrans;
    ty = ((fx * cal->yscale) >> 8) + cal->ytrans;
  }
  else {
    tx = ((fx * cal->xscale) >> 8) + cal->xtrans;
    ty = ((fy * cal->yscale) >> 8) + cal->ytrans;
  }
  if (tx < 0)
    tx = 0;
  if (ty < 0)
    ty = 0;
  to->x = tx;
  to->y = ty;
}

static void
ts_cal_to_raw (ts_calibration *cal, ts_point *from, ts_point *to)
{
  /*
   * modifies: to
   * effect:   converts point "from" from calibrated screen coordinates to
   *	       raw coordinates and stores them in point "to"
   */
  int fx = from->x, fy = from->y;
  int tx, ty;
  
  if (cal->xyswap) {
    tx = ((fy - cal->ytrans) << 8) / cal->yscale;
    ty = ((fx - cal->xtrans) << 8) / cal->xscale;
  }
  else {
    tx = ((fx - cal->xtrans) << 8) / cal->xscale;
    ty = ((fy - cal->ytrans) << 8) / cal->yscale;
  }
  if (tx < 0)
    tx = 0;
  if (ty < 0)
    ty = 0;
  to->x = tx;
  to->y = ty;
}

/*
 * grab area operations
 */
static ts_area *
ts_area_create (int id)
{
  /*
   * modifies: nothing
   * effects:  returns a new initialized ts_area, or
   *	       NULL if unable to allocate storage.
   */

  ts_area *area;

  /* allocate storage, fail if unable */
  if ((area = KOBJ_ALLOC(ts_area)) == NULL)
    return (NULL);
  KOBJ_INIT(area, ts_area);

  /* set identity */
  area->id = id;
  area->garea.id = id;
  
  /* everything OK */
  return (area);
}

static void
ts_area_destroy (ts_area *area)
{
  /*
   * modifies: area
   * effects:  recalims all storage associated with area.
   */
  kfree(area);
}

static void
ts_area_insert_in_order (ts_area **areasp, ts_area *area)
{
  /*
   * modifies: *areasp or one of the areas contained in the list *areasp
   * effect:   inserts the new area into the list so the areas are
   *	       sorted by increasing area->garea.p0.y
   */
  ts_area *ap=*areasp;
  int	  y = area->garea.p0.y;

  if (ap == NULL  ||  y < ap->garea.p0.y) 
    ts_area_insert(areasp, area);
  else {
    for ( ; ap->next != NULL  &&  y > ap->next->garea.p0.y;
	 ap = ap->next)
      ;
    area->prev = ap;
    area->next = ap->next;
    if (ap->next != NULL)
      ap->prev = area;
    ap->next = area;
  }
}

/*
 * ts_client operations
 */

static ts_client *
ts_client_create (int id, int sid)
{
  /*
   * modifies: nothing
   * effects:  returns a new initialized ts_client, or
   *	       NULL if unable to allocate storage.
   */

  ts_client *client;

  /* allocate storage, fail if unable */
  if ((client = KOBJ_ALLOC(ts_client)) == NULL)
    return (NULL);
  KOBJ_INIT(client, ts_client);

  /* set identity */
  client->id = id;
  client->sid = sid;

  client->cal = default_cal;
  
  /*   client->rate = ts_global.ts_rate; */
  /*   client->rate_delay = ts_global.ts_rate_delay; */
  
  /* everything OK */
  return (client);
}

static void
ts_client_destroy (ts_client *client)
{
  /*
   * modifies: client
   * effects:  recalims all storage associated with client.
   */
  kfree(client);
}

/*
 * ts_dev operations
 */

static int
ts_dev_init (ts_dev *dev)
{
  /* clear everything */
  KOBJ_INIT(dev, ts_dev);
  
  dev->ts_stabilize_delay = TS_STABILIZE_DELAY;
  dev->ts_rate = TS_RATE;
  dev->ts_rate_delay = OS_TIMER_RATE / dev->ts_rate;   
  dev->ts_fg_rate_delay = dev->ts_rate_delay;

  dev->fsm_state = TS_FSM_INIT;
  dev->p = 0;
  dev->minp = 70;

  dev->timer1_alloc = 0;
  dev->mcp_id = MCP_COMMON_ID_INVALID;

  dev->raw_waitq = NULL;

  /* allocate buffer */
  dev->buf_events = TS_BUFFER_NEVENTS;
  dev->buf_bytes = dev->buf_events * sizeof(ts_event);
  dev->buf = (ts_event *) kmalloc(dev->buf_bytes, GFP_KERNEL);
  if (dev->buf == NULL)
    return (-ENOMEM);

  /* buffer indexing */
  dev->buf_mask = TS_BUFFER_INDEX_MASK;
  dev->buf_next =0;

  dev->areas = NULL;
  dev->grabber = NULL;
  dev->clients = NULL;
  dev->client_id_next = 1;

#ifdef CONFIG_POWRMGR
  dev->pm_id = POWERMGR_ID_INVALID;
  dev->last_active = 0;
#endif

  dev->sm_id = ITSY_SESSION_NOTIFY_ID_INVALID;
  dev->sm_active = itsy_session_get_foreground();

  return 0;
}

static void
ts_dev_event (int sid, ushort p, ushort x, ushort y, unsigned char flags)
{
  /*
   * modifies: buf
   * effects:  creates event with specified sid and ts state, and
   * 	       adds it to the buffer associated with dev. The x and y
   *	       coordinates are assumed to be raw coordinates.
   *	       Wakes up any clients that may be blocked waiting
   *	       for the event.
   */

  ts_dev *dev = &ts_global;
  ts_event *ev;
  ts_client *client;
  ts_area *area;
  
  IF_DEBUG( printk(TS_NAME ": ts_dev_event p=%d,x=%d,y=%d,f=0x%x\n",
		   (int)p, (int)x, (int)y, (int)flags) );
	   
  /* use next slot in circular buffer */
  ev = &dev->buf[dev->buf_next & dev->buf_mask];

  /* generic session event */
  do_gettimeofday(&ev->common.timestamp);

  ev->common.event_class = ITSY_EVENT_CLASS_TOUCHSCREEN;
  ev->common.event_type  = ITSY_EVENT_TOUCHSCREEN;
  ev->common.size  = sizeof(ts_event);
  ev->common.flags = flags;

  /* specific ts event */
  ev->sid = sid;
  ev->pressure  = p;
  ev->point.x   = x;
  ev->point.y   = y;

  /* advance buffer index */
  dev->buf_next++;

  /* update statistics */
  dev->event_count++;

  /* check if it belongs to the grabber, and wake him if so*/
  if (dev->grabber) {
/*     printk("dev_ev x=%d y=%d\n", x, y); */
    
    /* depending on the sign of xscale and yscale, p0.x may be > p1.x */
    if (dev->grabber->cal.xscale * dev->grabber->cal.yscale > 0) {
      for (area=dev->areas; area != NULL  &&  y >= area->garea.p0.y;
	   area = area->next) {
	if (y <= area->garea.p1.y  &&  x >= area->garea.p0.x  &&
	    x <= area->garea.p1.x) {
	  /* 	  printk("g: %d, %d\n", x, y); */
	  ev->common.flags |= ITSY_EVENT_FLAG_GRABBED;
	  wake_up_interruptible(&dev->grabber->waitq);
	  /* asynchronous notifications */
	  if (dev->grabber->asyncq != NULL)
	    kill_fasync(dev->grabber->asyncq, SIGIO);

	  /* wake up raw waiters, if any */
	  wake_up_interruptible(&dev->raw_waitq);
	  /* asynchronous notifications */
	  if (dev->raw_asyncq != NULL)
	    kill_fasync(dev->raw_asyncq, SIGIO);
	  
	  return;
	}
      }
    }
    else { /* xscale*yscale < 0 */
      for (area=dev->areas; area != NULL  &&  y >= area->garea.p0.y;
	   area = area->next) {
	if (y <= area->garea.p1.y  &&  x >= area->garea.p1.x  &&
	    x <= area->garea.p0.x) {
	  /* 	  printk("g: %d, %d\n", x, y); */
	  ev->common.flags |= ITSY_EVENT_FLAG_GRABBED;
	  wake_up_interruptible(&dev->grabber->waitq);
	  /* asynchronous notifications */
	  if (dev->grabber->asyncq != NULL)
	    kill_fasync(dev->grabber->asyncq, SIGIO);
	  
	  /* wake up raw waiters, if any */
	  wake_up_interruptible(&dev->raw_waitq);
	  /* asynchronous notifications */
	  if (dev->raw_asyncq != NULL)
	    kill_fasync(dev->raw_asyncq, SIGIO);
	  
	  return;
	}
      }
    }
  } /* if (dev->grabber) */
  
  /* wake up raw waiters, if any */
  wake_up_interruptible(&dev->raw_waitq);
  /* asynchronous notifications */
  if (dev->raw_asyncq != NULL)
    kill_fasync(dev->raw_asyncq, SIGIO);
  
  /* wakeup waiters, if any */
  for (client=dev->clients; client!=NULL; client=client->next) {
    if (client->sid == sid) {
      wake_up_interruptible(&client->waitq);
      if (client->asyncq != NULL)
	kill_fasync(client->asyncq, SIGIO);
    }
  }
}

/*
 * device operations
 */

static inline int
ts_raw_device (const struct inode *inode)
{
  /*
   * modifies: nothing
   * effects:  returns TRUE iff inode is the "raw" ts device
   */
  return (MINOR(inode->i_rdev) == TS_RAW_MINOR);
}

static inline int
ts_grab_device (const struct inode *inode)
{
  return (MINOR(inode->i_rdev) == TS_GRAB_MINOR);
}

static int
ts_open (struct inode *inode, struct file * file)
{
  ts_dev *dev = &ts_global;
  ts_client *client;
  unsigned int minor=MINOR(inode->i_rdev);
  int id;

  IF_DEBUG( printk(TS_NAME ": ts_open\n") );

  if (minor == TS_BATTERY_MINOR  ||  minor == TS_ADC_MINOR) {
    if (itsy_major_rev() > 1)
      return (EACCES);
    else {
      file->private_data = NULL;
      MOD_INC_USE_COUNT;
      return (0);
    }
  }
  
  /* ustrokes not supported yet LSB */
  if (minor != TS_MINOR && minor != TS_RAW_MINOR && minor != TS_GRAB_MINOR)
    return (EACCES);
  
  /* create new client, fail if unable */
  if (minor == TS_GRAB_MINOR)
    id = 0;
  else
    id = dev->client_id_next++;
  if ((client = ts_client_create(id, itsy_session_from_task(current))) == NULL)
    return (-ENOMEM);

  /* skip over old events */
  client->buf_next = dev->buf_next;

  /* add to appropiate client list */
  if (minor == TS_GRAB_MINOR)
    dev->grabber = client;
  else if (minor == TS_RAW_MINOR)
    ts_client_insert(&dev->rawclients, client);
  else
    ts_client_insert(&dev->clients, client);

  /* attach client state to file */
  file->private_data = client;

  MOD_INC_USE_COUNT;

  /* everything OK */
  return (0);
}

static int ts_fasync(struct inode *inode, struct file *file, int mode);

static void
ts_release (struct inode *inode, struct file * file)
{
  ts_client *client = (ts_client *) file->private_data;
  ts_dev *dev = &ts_global;
  ts_area *area;
  unsigned int minor=MINOR(inode->i_rdev);

  IF_DEBUG( printk(TS_NAME ": ts_release\n") );

  /* remove form asynchronous notification queues */
  ts_fasync(inode, file, 0);

  if (client != NULL) {
  
    /* if client is the grabber, release the grab areas */
    if (client == dev->grabber) {
      while ((area=dev->areas) != NULL) {
	ts_area_remove(&dev->areas, area);
	ts_area_destroy(area);
      }
      dev->grabber = NULL;
    }
    else if (minor == TS_RAW_MINOR) {	/* remove from rawclients list */
      ts_client_remove(&dev->rawclients, client);
    }
    else {    				/* remove from client list */
      ts_client_remove(&dev->clients, client);
    }
    /* reclaim storage */
    ts_client_destroy(client);
    file->private_data = NULL;
  }
  
  MOD_DEC_USE_COUNT;
}

static int
ts_ioctl (struct inode *inode, struct file *file, uint cmd, ulong arg)
{
  ts_client *client = (ts_client *) file->private_data;
  ts_dev *dev = &ts_global;
  ts_calibration cal;
  ts_area *area;
  ts_sid_map *smp;
  unsigned int minor=MINOR(inode->i_rdev);
  unsigned long cur_jiffies = jiffies;
  int size, err;
  long v, temp;
  
  IF_DEBUG( printk(TS_NAME ": ts_ioctl: cmd=%d, arg=%lu\n", cmd, arg) );

  /* sanity checks */
  if (_IOC_TYPE(cmd) != TS_IOCTL_TYPE)
    return (-EINVAL);
  if (_IOC_NR(cmd) > TS_IOCTL_MAXNR)
    return (-EINVAL);

  /* verify transfer memory */
  size = _IOC_SIZE(cmd);
  if (_IOC_DIR(cmd) & _IOC_READ)
    if ((err = verify_area(VERIFY_WRITE, (void *) arg, size)) != 0)
      return (err);
  if (_IOC_DIR(cmd) & _IOC_WRITE)
    if ((err = verify_area(VERIFY_READ, (void *) arg, size)) != 0)
      return (err);

  /* ioctls for /dev/adc and dev/battery */
  if (minor == TS_BATTERY_MINOR  ||  minor == TS_ADC_MINOR) {
    int adc_input;
    ulong flags;

    switch (cmd) {
    case ADC_0_GET:
      adc_input = 0;
      break;
    case ADC_1_GET:
      adc_input = 1;
      break;
    case ADC_2_GET:
      adc_input = 2;
      break;
    case ADC_3_GET:
      adc_input = 3;
      break;
    default:
      return (-EINVAL);
    }
    
    while (cur_jiffies > dev->adc_times[adc_input]) { 
      save_flags_cli(flags);
      dev->adc_read_flags |= (1 << adc_input);
      if (!dev->pen_down) {
	ts_disable_codec_int();
	ts_start_measurements();
      }
      restore_flags(flags);
      interruptible_sleep_on(&dev->adc_waitq[adc_input]);
      if (current->signal & ~current->blocked) {
	/* abort if interrupted */
	return (-EINTR);
      }
    }
    put_user_long((long)dev->adc_values[adc_input], (void *)arg);
    return (0);
  }
  
  /* ioctls for touchscreen */
  
  /* dispatch based on command */
  switch (cmd) {
  case TS_SET_DEFAULT_CALIBRATION:		/* check for root priv? */
    memcpy_fromfs(&cal, (void *)arg, sizeof(ts_calibration));
    if (cal.xscale == 0  ||  cal.yscale == 0)
      return (-EINVAL);
    /*     if (!suser()) */
    /*       return (-EPERM); */
    /* Change current clients calibration if they are using the default one */
    {
      ts_client *p;

      if (dev->grabber != NULL  &&  
	  calibrations_are_equal(&dev->grabber->cal, &default_cal))
	memcpy(&dev->grabber->cal, &cal, sizeof(ts_calibration));
      
      for (p = dev->rawclients; p != NULL; p = p->next) {
	if (calibrations_are_equal(&p->cal, &default_cal))
	  memcpy(&p->cal, &cal, sizeof(ts_calibration));
      }

      for (p = dev->clients; p != NULL; p = p->next) {
	if (calibrations_are_equal(&p->cal, &default_cal))
	  memcpy(&p->cal, &cal, sizeof(ts_calibration));
      }
    }
    memcpy(&default_cal, &cal, sizeof(ts_calibration));
    break;
  case TS_GET_DEFAULT_CALIBRATION:
    memcpy_tofs((void *)arg, &default_cal, sizeof(ts_calibration));
    break;
  case TS_SET_CALIBRATION:
    memcpy_fromfs(&cal, (void *)arg, sizeof(ts_calibration));
    if (cal.xscale == 0  ||  cal.yscale == 0)
      return (-EINVAL);
    memcpy(&client->cal, &cal, sizeof(ts_calibration));
    break;
  case TS_GET_CALIBRATION:
    memcpy_tofs((void *)arg, &client->cal, sizeof(ts_calibration));
    break;
/*   case TS_SET_GRABBER: */
/*     if (dev->grabber != NULL) */
/*       return (-EPERM); */
/*     dev->grabber = client; */
/*     break; */
  case TS_GET_GRABBER:
    if (dev->grabber == NULL)
      put_user_long(0L, (void *) arg);
    else
      put_user_long(1L, (void *) arg);
    break;
  case TS_ADD_GRAB_AREA:
    /* only grabber can add grab areas */
    if (dev->grabber != client)
      return (-EACCES);

    /* create grab area and copy data from user */
    if ((area = ts_area_create(dev->area_id_next++)) == NULL)
      return (-ENOMEM);
    memcpy_fromfs(&area->garea, (void *)arg, sizeof(ts_grab_area));

    /* return grab area id to user */
    area->garea.id = area->id;
    memcpy_tofs((void *) arg, &area->garea, sizeof(int));

    /* transform from screen to raw coordinates */
    ts_cal_to_raw(&client->cal, &area->garea.p0, &area->garea.p0);
    ts_cal_to_raw(&client->cal, &area->garea.p1, &area->garea.p1);

    /* want p0.y < p1.y in raw coordinates */
    if (area->garea.p0.y > area->garea.p1.y) {
      temp = area->garea.p0.y;
      area->garea.p0.y = area->garea.p1.y;
      area->garea.p1.y = temp;
      temp = area->garea.p0.x;
      area->garea.p0.x = area->garea.p1.x;
      area->garea.p1.x = temp;
    }
      
    /* insert into grab area link list */
    ts_area_insert_in_order(&dev->areas, area);
    break;
  case TS_DELETE_GRAB_AREA:
    /* only grabber can delete grab areas */
    if (dev->grabber != client)
      return (-EACCES);

    /* find the specified area (by id) from list */
    area = ts_area_lookup(dev->areas, (int)arg);
    if (area == NULL)
      return (-EINVAL);

    ts_area_remove(&dev->areas, area);
    break;
  case TS_SET_RATE:
    v = arg;
    if (v > TS_MAX_RATE  ||  v <= 0)
      return (-EINVAL);
    if ((smp = ts_sid_map_lookup(tsSidList, client->sid)) == NULL) {
      /*       printk("ts_sid_map_lookup fails for sid:%d\n", client->sid); */
      return (-EINVAL);
    }
    smp->rate = v;
    smp->rate_delay = OS_TIMER_RATE / smp->rate;
    if (client->sid == dev->sm_active) {
      dev->ts_fg_rate_delay = smp->rate_delay;
    }
    break;
  case TS_GET_RATE:
    if ((smp = ts_sid_map_lookup(tsSidList, client->sid)) == NULL) {
      /*       printk("ts_sid_map_lookup fails for sid:%d\n", client->sid); */
      return (-EINVAL);
    }
    put_user_long((long)smp->rate, (void *)arg);
    break;
  case TS_SET_DEFAULT_RATE:
    if (!suser())
      return (-EINVAL);
    v = arg;
    if (v > TS_MAX_RATE  ||  v <= 0)
      return (-EINVAL);
    dev->ts_rate = v;
    dev->ts_rate_delay = OS_TIMER_RATE / dev->ts_rate;
    break;
  case TS_GET_DEFAULT_RATE:
    put_user_long((long)dev->ts_rate, (void *)arg);
    break;
    
  default:
    return (-EINVAL);
  }

  return (0);
}

static int
ts_read (struct inode * inode, struct file * file, char * buf, int count)
{
  const int event_bytes = sizeof(ts_event);
  ts_client *client = (ts_client *) file->private_data;
  ts_dev *dev = &ts_global;
  int total, err, buf_first, raw, grabber, sid, next;
  unsigned int minor=MINOR(inode->i_rdev);
  
  IF_DEBUG( printk(TS_NAME ": ts_read\n") );

  /* can't read battery or adc, only access is through ioctl */
  if (client == NULL || minor == TS_BATTERY_MINOR || minor == TS_ADC_MINOR)
    return (-EINVAL);
  
  /* can't send part of events */
  if (count < sizeof(ts_event))
    return 0;
  
  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_WRITE, (void *) buf, count)) != 0)
    return (err);

  /* initialize */
  sid = itsy_session_from_task(current);
  total = 0;
  raw = ts_raw_device(inode);
  grabber = (dev->grabber == client);

  /* determine first available buffered event */
  buf_first = MAX(0, dev->buf_next - dev->buf_events);
  
  /* skip over dropped events */
  if (client->buf_next < buf_first) {
    client->buf_next = buf_first;
    client->drop_count++;
  }

  /* repeat until we get at least one event */
  while (total == 0) {
    
    /* transfer integral number of matching events */
    for (next = client->buf_next;
	 (next < dev->buf_next) && (count >= event_bytes);
	 next++, client->buf_next++) {
      
      ts_event *evp, ev;

      evp = &dev->buf[next & dev->buf_mask];
      if (raw ||		  	/* raw client */
	  (!grabber && 			/* sid matches and event not grabbed */
	   evp->sid == sid &&
	   !(evp->common.flags & ITSY_EVENT_FLAG_GRABBED)) ||
	  (grabber &&			/* grabber and event grabbed */
	   (evp->common.flags & ITSY_EVENT_FLAG_GRABBED))
	  ) {
	/* go from raw to calibrated data */
	memcpy((void *)&ev, (void *)evp, sizeof(ts_event));
	ts_raw_to_cal(&client->cal, &evp->point, &ev.point);
	
	/* copyout matching event */
	memcpy_tofs((void *)buf, (void *)&ev, event_bytes);

	/* update counts */
	total += event_bytes;
	buf   += event_bytes;
	count -= event_bytes;
      }
    }
    
    /* there were no events */
    if (total == 0) {
      
      /* if non-blocking return now */
      if (file->f_flags & O_NONBLOCK)
	return (-EAGAIN);
      
      /* else block on appropiate wait queue */
      if (raw)
	interruptible_sleep_on(&dev->raw_waitq);
      else
	interruptible_sleep_on(&client->waitq);
      if (current->signal & ~current->blocked) {
	/* abort if interrupted */
	return (-EINTR);
      }
    }
    
  }  /* end of while loop, exit => total > 0 */

  IF_DEBUG( printk(TS_NAME ": ts_read returning, total=%d\n", total) );
	   
  return (total);
}
      
static int
ts_write (struct inode * inode, struct file * file, const char * buf, int count)
{
  const int event_bytes = sizeof(ts_event);
  ts_client *client = (ts_client *) file->private_data;
  ts_dev *dev = &ts_global;
  unsigned int minor=MINOR(inode->i_rdev);
  ts_event ev;
  int err, total=0;
  
  IF_DEBUG( printk("TS_NAME" ": ts_write\n") );

  /* can't read battery or adc, only access is through ioctl */
  if (client == NULL || minor == TS_BATTERY_MINOR || minor == TS_ADC_MINOR)
    return (-EINVAL);
  
  /* can't send part of events */
  if (count < sizeof(ts_event))
    return 0;
  
  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_READ, (void *) buf, count)) != 0)
    return (err);

  memcpy_fromfs((void *)&ev, (void *)buf, event_bytes);
  ev.common.event_class = ITSY_EVENT_CLASS_TOUCHSCREEN;
  ev.common.event_type  = ITSY_EVENT_TOUCHSCREEN;
  ev.common.size  = sizeof(ts_event);
  ev.common.flags = ev.common.flags | ITSY_EVENT_FLAG_SYNTHETIC;
  if (suser())
    ev.common.flags |= ITSY_EVENT_FLAG_SUSER;

  /* foreground sid? */
  if (ev.sid == ITSY_SESSION_ID_FOREGROUND)
    ev.sid = dev->sm_active;
  
  /* add it to event queue */
  ts_cal_to_raw(&client->cal, &ev.point, &ev.point);
  ts_dev_event (ev.sid, ev.pressure, ev.point.x, 
 		ev.point.y, ev.common.flags); 

  /* only do 1 event at a time? */
  total += sizeof(ts_event);
  
  return total;
}

static int
ts_fasync (struct inode *inode, struct file *file, int mode)
{
  ts_client *client = (ts_client *) file->private_data;
  ts_dev *dev = &ts_global;
  unsigned int minor = MINOR(inode->i_rdev);

  IF_DEBUG( printk(TS_NAME ": ts_fasync\n") );
  /* add client to appropiate notifications queue */
  if (minor == TS_RAW_MINOR)
    return fasync_helper(inode, file, mode, &dev->raw_asyncq);
  else if (client != NULL)
    return fasync_helper(inode, file, mode, &client->asyncq);
  else
    return (-EINVAL);
}

static int
ts_select (struct inode *inode, struct file *file, int mode,
	   select_table *table)
{
  ts_client *client = (ts_client *) file->private_data;
  ts_dev *dev = &ts_global;

  IF_DEBUG( printk(TS_NAME ": ts_select\n") );

  if (client == NULL)
    return (-EINVAL);
  
  /* handle select on input */
  if (mode == SEL_IN) {

    if (ts_raw_device(inode)) {
      /* looking for any event, regardless of event's sid */
      
      if (client->buf_next < dev->buf_next)
	return (1);

      /* wait for raw event */
      IF_DEBUG( printk(TS_NAME ": ts_select: wait (raw) ...\n") );
      select_wait(&dev->raw_waitq, table);
      return (0);
    }
    else {
      /* any sid-specific events buffered? */
      if (client->buf_next < dev->buf_next) {
	int buf_first, next;

	buf_first = MAX(0, dev->buf_next - dev->buf_events);
	buf_first = MAX(buf_first, client->buf_next);
	
	if (dev->grabber == client) {
	/* this is the grabber */
	  for (next = buf_first; next < dev->buf_next; next++)
	    if (dev->buf[next & dev->buf_mask].common.flags &
		ITSY_EVENT_FLAG_GRABBED) {
	      client->buf_next = next;
	      return (1);
	    }
	}
	else {
	  /* not the grabber */
	  for (next = buf_first; next < dev->buf_next; next++)
	    if (!(dev->buf[next & dev->buf_mask].common.flags &
		  ITSY_EVENT_FLAG_GRABBED)  && 
		(dev->buf[next & dev->buf_mask].sid == client->sid)) {
	      client->buf_next = next;
	      return (1);
	    }
	}
	client->buf_next = next;
      }
      
      IF_DEBUG( printk(TS_NAME ": ts_select: wait (sid %d)...\n", client->sid));

      select_wait(&client->waitq, table);
      return (0);
    } 
  }
  
  /* SEL_OUT, SEL_EX not supported */
  return (0);
}

/*
 * Power management functions
 *
 */


/* static int openFlag=0; */

static struct file_operations ts_fops = {
  NULL,		/* no special lseek */
  ts_read,
  ts_write,
  NULL,         /* no special readdir */
  ts_select,
  ts_ioctl,     
  NULL,         /* no special mmap */
  ts_open,
  ts_release,
  NULL,         /* no special fsync */
  ts_fasync,
  NULL,         /* no special check_media_change */
  NULL,         /* no special revalidate */
};

static void ts_session_create_cb (void *vp, int sid)
{
  ts_sid_map *smp;

  /*   printk("ts_session_create_cb - sid:%d ", sid); */
  if ((smp = ts_sid_map_lookup(tsSidList, sid)) == NULL) {
    if (tsSidFreeList != NULL) {
      smp = tsSidFreeList;
      tsSidFreeList = tsSidFreeList->next;
    }
    else {
      if ((smp = KOBJ_ALLOC(ts_sid_map)) == NULL) {
	/* 	printk(" FAILS\n"); */
	return;
      }
      KOBJ_INIT(smp, ts_sid_map);
    }
    /*     printk(" OK\n"); */
    smp->id = sid;
    smp->rate = ts_global.ts_rate;
    smp->rate_delay = ts_global.ts_rate_delay;
    ts_sid_map_insert(&tsSidList, smp);
  }
}

static void ts_session_destroy_cb (void *vp, int sid)
{
  ts_sid_map *smp;

  /*   printk("ts_session_destroy_cb - sid:%d", sid); */
  if ((smp = ts_sid_map_lookup(tsSidList, sid)) != NULL) {
    ts_sid_map_remove(&tsSidList, smp);
    smp->next = tsSidFreeList;
    tsSidFreeList = smp;
    /*     printk(" OK\n"); */
  }
  /*   else */
    /*     printk(" DOESN'T EXIST\n"); */
}

static void ts_session_switchto_cb (void *vp, int sid)
{
  /*
   * modifies: ts_global.sm_active
   * effect:   keeps track of foreground (active) sesion
   */
  ts_sid_map *smp;
  
  ts_global.sm_active = sid;
  if ((smp = ts_sid_map_lookup(tsSidList, sid)) != NULL) {
    ts_global.ts_fg_rate_delay = smp->rate_delay;
  }
  else {
    ts_global.ts_fg_rate_delay = ts_global.ts_rate_delay;
  }
}

static void ts_stop (void)
{
  /*
   * modifies: 	ts_global (global)
   * effects:	Stop processing touchscreen
   */
  
  /* Stop timer 1 */ 
  cleanup_timer1();
  
  /* disable mcp interrupts */
  mcp_common_disable_irq(ts_global.mcp_id);  
}

static int
ts_unregister (ts_dev *dev, char *msg)
{
  if (msg != NULL)
    printk(msg);

  ts_disable_all_int();
  unregister_chrdev(TS_MAJOR, TS_NAME);

  if (dev->mcp_id != MCP_COMMON_ID_INVALID) {
    MCP_DISABLE(dev->mcp_id);
    mcp_common_unregister(dev->mcp_id);
  }
  
  if (dev->timer1_alloc) 
    free_irq(TS_TIMER, NULL);

#ifdef CONFIG_POWERMGR
  if (dev->pm_id != POWERMGR_ID_INVALID)
    if (powermgr_unregister(dev->pm_id) == POWERMGR_FAILURE)
      printk("ERROR: Failed to unregister %s module from powermgr!!\n",
	     TS_NAME_VERBOSE);
#endif

  if (dev->sm_id != ITSY_SESSION_NOTIFY_ID_INVALID)
    if (itsy_session_notify_unregister(dev->sm_id) != 0)
      printk("ERROR: unable to unregister %s from session manager!!\n",
	     TS_NAME_VERBOSE);

  return -1;
}

static char *
ts_version (void)
{
  /*
   * modifies: nothing
   * effect:   returns the current cvs revision number.
   * storage:  return value is statically-allocated.
   */
  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = TS_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return (version);
}


#ifdef CONFIG_POWERMGR

/*
 * Power management functions
 *
 */

static int 
ts_suspend_check (void *unused, int idle)
{
  /*
   * modifies: nothing
   * effects: Always returns zero: i.e. willing to sleep immediately.
   */
  return 0;
}

static int 
ts_suspend (void *unused)
{
  /*
   * modifies: 	ts_global (global)
   * effects:	Stop timer
   *		Always returns zero (ignored)
   */
  cleanup_timer1();
  ts_disable_all_int();
  ts_pin_power_off();
  MCP_DISABLE(ts_global.mcp_id);
/*   mcp_common_disable_irq(ts_global.mcp_id); */

  return 1;
}

static int ts_resume (void *unused, int resume_flags)
{
  /*
   * modifies:	ts_global (global)
   * effects:	Resumes after wakeup from sleep mode.
   *		Always returns zero (ignored).
   */
  ts_dev *dev = &ts_global;
  
  MCP_ENABLE(dev->mcp_id);

/*   v1 = codec_read(mcp, 4); */
/*   if (v1) { */
/*     codec_write(mcp, 4, 0); */
/*     codec_write(mcp, 4, v1); */
/*   } */

  ts_enable_timer_int();
  ts_clear_codec_int();
  mcp_common_enable_irq(dev->mcp_id);
  ts_start_measurements();
  return 1;
}

static const powermgr_client ts_powermgr = {
  /* callback functions */
  ts_suspend_check,
  ts_suspend,
  ts_resume,

  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  TS_NAME_VERBOSE,

  /* power-consumption info */
  2000,
  0
};

#endif /* CONFIG_POWERMGR */

int
ts_init (void)
{
  int err;
  ts_dev *dev = &ts_global;

  IF_DEBUG( printk("TS: init_module\n") );

  if ((err = ts_dev_init(dev)) != 0)
    return err;

  /* 
   * register ts character device 
   */ 
  if ((err=register_chrdev(TS_MAJOR, TS_NAME, &ts_fops)) != 0) {
    printk("TS: unable to get major (%d) for %s device, err=%d\n", 
	   TS_MAJOR, TS_NAME, err);
    return err;
  }

  /*
   * register with mcp_common
   */
  if ((dev->mcp_id = mcp_common_register()) == MCP_COMMON_ID_INVALID) 
    return ts_unregister(dev, "TS: mcp_common_register fails!\n");

  /* turn on mcp */
  MCP_ENABLE(dev->mcp_id);

  IF_DEBUG( printk("ts module: major number: %d\n", TS_MAJOR) );

  /* 
   * request timer1
   */
  if (request_irq(TS_TIMER, ts_timer_int_handler, SA_SHIRQ, "ts - OST1", NULL) 
      != 0) 
    return ts_unregister(dev, 
			 "ERROR: ts driver could not allocate ts OST1 IRQ!\n");
  else
    dev->timer1_alloc = 1;
  
  /* 
   * request shared mcp irq
   */
  if (request_irq(TS_IRQ, ts_codec_int_handler, SA_INTERRUPT | SA_SHIRQ, 
		  TS_NAME, &ts_global)  !=  0) 
    return ts_unregister(dev, "ERROR: Could not allocate ts gpio IRQ!\n");

  mcp_common_enable_irq(dev->mcp_id);

  /* enable timer1 interrupts */
  ts_enable_timer_int();

  /* clear pending interrupts at codec? */
/*   if (v=codec_read(mcp, 4)) { */
/*     codec_write(mcp, 4, 0); */
/*     codec_write(mcp, 4, v); */
/*   } */

#ifdef CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &ts_proc_entry);
#endif
  
#ifdef CONFIG_POWERMGR
  if ((dev->pm_id = powermgr_register(&ts_powermgr)) == POWERMGR_ID_INVALID) {
    printk("Failed to register %s module for power management!!\n",
	   TS_NAME);
  }
#endif

  dev->sm_id = itsy_session_notify_register(NULL, "touchscreen", 
					    ts_session_create_cb,
					    ts_session_destroy_cb,
					    ts_session_switchto_cb, NULL);
  if (dev->sm_id == ITSY_SESSION_NOTIFY_ID_INVALID)
    printk("%s unable to register with session manager\n", TS_NAME);
  
  ts_start_measurements();

  /* log device registration */
  printk(TS_NAME_VERBOSE " version %s initialized\n", ts_version());
  
  return 0;

}

#if 0
long ts_init(long mem_start, long mem_end) {
  if (register_chrdev(TS_MAJOR, TS_NAME, &ts_fops))
    printk("unable to get major for ts device\n");
  
  return mem_start;
}
#endif

/*
 * module functions
 */
#ifdef MODULE

int 
init_module(void)
{
  return (ts_init());
}

void 
cleanup_module(void)
{
  int err;
  
  IF_DEBUG( printk("TS: cleanup_module\n") );

  ts_stop();
  ts_pin_power_off();
  MCP_DISABLE(ts_global.mcp_id);

  ts_unregister(&ts_global, NULL);

#ifdef CONFIG_PROC_FS
  /* unregister procfs entry */
  if ((err = proc_unregister(&proc_root, ts_proc_entry.low_ino)) != 0)
    IF_DEBUG( printk(TS_NAME ": cleanup_module proc_unregister(ts) %d\n", err));
#endif
  
  free_irq(TS_IRQ, &ts_global); 

}
#endif


