/*
 * linux/arch/arm/drivers/char/keyb_brutus.c
 *
 * Copyright (c) Compaq Computer Corporation, 1998
 *
 * Author: Larry Brakmo.
 * Based on code from initial SA1100 evaluation board sample software.
 *
 * Keyboard driver for Brutus ARM Linux.
 *
 * Note!!! This driver talks directly to the keyboard.
 *
 * Changelog:
 * 22-01-1998	lsb	Created by modifying keyb_rpc.c
 *
 */

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/ptrace.h>
#include <linux/signal.h>
#include <linux/timer.h>
#include <linux/random.h>
#include <linux/ctype.h>
#include <linux/config.h>

#include <asm/bitops.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>

extern void kbd_keyboardkey(unsigned int keycode, unsigned int up_flag);
extern void kbd_reset(void);
extern void kbd_setregs(struct pt_regs *regs);


/*
 * Area of brutus keyboard specific functions
 */

#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/io_reg.h>
#include <asm/proc/page.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/proc/pgtable.h>
#include <linux/itsy_console.h>
#include <asm/arch/sa1100.h>
#include <asm/arch/ostimers.h>
#include <asm/arch/mmap.h>
#include "kbd_kern.h"

#ifndef CONFIG_ITSY  /* change later to hard keyboard! */
#define INCLUDE_KBMAPS
#endif
#include "kbmap_brutus.h"

#define KBCTL_NODATA -1
#define KBCTL_AGAIN  -2
#define KBCTL_OLD_LEDS 0xffaef

static int key_fn_down=0;
static int key_numl_down=0;

#ifndef CONFIG_ITSY
static struct sspreg *ssp_base_v;
static struct scm_gpio_regs *gp_base_v;
static int *gpdr_v;
static int putcToKBCTL (unsigned char);
static void putsToKBCTL (char *, int);
static char lrc (char *, int);
static void setKBled (int);
static int checkInputKBCTL (void);
static void waitInputKBCTL (void);
static int getcFromKBCTL (void);
static int wgetcFromKBCTL (void);
static void initKBCTL (void);
#endif

void kbd_drv_setleds(unsigned int);
extern void lsb_flash (int);

#ifndef CONFIG_ITSY
static int putcToKBCTL (unsigned char c)
{
  int i;
  volatile struct sspreg *ssp;

  ssp  = ssp_base_v;
  
  /*
   * NOTE: on older versions of sa1100 you need to reset the ssp port
   *       every time (I assume new versions here). These functions
   *       were copied from code that came with early brutus boards.
   */
  
  while (((ssp->sssr) & SSPSR_M_TNF) == 0); /* check for space */
  ssp->ssdr = (c<<8) | c; /* put char */
  while (((ssp->sssr) & SSPSR_M_RNE) == 0); /* wait to get char back */
  i = ssp->ssdr;
  return i;
}

static void putsToKBCTL (char *m, int cnt)
{
  volatile struct scm_gpio_regs *gpr;
  int i,j,j2,x;

  gpr  = gp_base_v;
  
  gpr->gp_pinclr = 1<<23;
  for (j=0; j<3; j++)
    x = gpr->gp_pinlevel;
  gpr->gp_pinset = 1<<23;
  for (j=0; j<250*30; j++)
    x = gpr->gp_pinlevel;
  for (i=0; i<cnt; i++) {
    j2 = putcToKBCTL(m[i]);
    for (j=0; j<400; j++)
      x = gpr->gp_pinlevel;
  }
}

static char lrc (char *m, int l)
{
  int i,crc;
  
  crc = 0;			/* init crc */
  for(i=0; i<l; i++) { 		/* xor in all data */
    crc ^= *m++;
  }
   if ((crc & 0x80) == 0x80) {
     crc ^= 0xC0;
   }
  return crc;
}

static void setKBled (int i)
{
  char msg[20];
  
  msg[0]=0x1B; /* led on */
  msg[1]=0xa6;
  msg[2]=0;
  msg[3]=(char) (i & 1);
  msg[4]=0;
  msg[5]=0;
  msg[6]=0;
  msg[7]=0;
  msg[8]=lrc(msg,8);
  putsToKBCTL(msg, 9);
}

static int checkInputKBCTL (void)
{
  volatile int *gpdr;
  
  gpdr = gpdr_v; 		/* address of GPIO level register */
  return (EXT(*gpdr,25)); 	/* check KBCTL has data to send */
}

static void waitInputKBCTL (void)
{
  volatile int *gpdr;
  
  gpdr = gpdr_v; 		/* address of GPIO level register */
  while (EXT(*gpdr,25)); 
}


static int getcFromKBCTL (void)
{
  int x,y;
  volatile int *gpdr;
  static int lastc=-1;
  
  gpdr = gpdr_v; 		/* address of GPIO level register */
  if (EXT(*gpdr,25)) {
    return KBCTL_NODATA;
  }

  x = putcToKBCTL(0)&0xff; 	/* get char by sending one */
  if (x == lastc || x == 0 || x == 0xff)
    return KBCTL_NODATA;

  lastc = x;

  if (EXT(x,7)) {    		/* key up */
    x &= 0x7f;
    if (x == 0x21) {		/* fn key up */
      key_fn_down = 0;
      return KBCTL_AGAIN;
    }
    else if (key_fn_down) { 	/* this is a fn modified key */
      y = kbmapFN[x];
      if (y == KK_NUML) 
	return KBCTL_AGAIN;
    }
    else if (key_numl_down) 	/* this is a numlock modified key */
      y = kbmapNL[x];
    else
      y = kbmap[x];
    return (y | 0x80);
  } else {			/* key down */
    /* key down */
    if (x == 0x21) {		/* fn key down */
      key_fn_down = 1;
      return KBCTL_AGAIN;
    }
    else if (key_fn_down) {	/* this is a fn modified key */
      y = kbmapFN[x];
      if (y == KK_NUML) {	/* toggle local numlock */
	key_numl_down = !key_numl_down;
	kbd_drv_setleds(KBCTL_OLD_LEDS);
	return KBCTL_AGAIN;
      }
    }
    else if (key_numl_down)	/* this is a numlock modified key */
      y = kbmapNL[x];
    else
      y = kbmap[x];
    return y;
  }
}

static int wgetcFromKBCTL (void)
{
  waitInputKBCTL();
  return getcFromKBCTL();
}

static void initKBCTL (void)
{
  char msg[30];
  int  i, tmp;
  volatile int * ppar;
  volatile struct scm_gpio_regs *gpreg;
  volatile struct sspreg *ssp;

  /*
   * define the base for each structure 
   */
  ssp_base_v = (struct sspreg *) io_p2v(SSP_BASE_P);
  gp_base_v = (struct scm_gpio_regs *) io_p2v(GP_BASE_P);
  gpdr_v = (int *) io_p2v(GP_BASE_P);
  ppar = (int *) io_p2v(PPC_PIN_ASSIGN_P);
  gpreg  = gp_base_v;
  ssp  = ssp_base_v;

  tmp = gpreg->gp_pindir;
  tmp = (tmp | (0xD << 10) | 1<<23) & ~(0x2 << 10);
  gpreg->gp_pindir  = tmp; 	/* set dir for ssp */
  tmp = gpreg->gp_altfunc;
  tmp = tmp | (0xf << 10);
  gpreg->gp_altfunc = tmp;

  gpreg->gp_pinset = 1<<23; 	/* deassert the kbctl wakup pin */
  i = *ppar;
  *ppar = i | 1<<18; 		/* set alt function for spi interface  */
 
  ssp->sscr0 = 0;
  ssp->sscr1 = 0; 		/* no ints no loopback */
  ssp->sssr = 0; 		/* remove any rcv overrun errors */
  /* turn on SSP */
  ssp->sscr0 = (3<<SSPCR0_V_SCR) | (1<<SSPCR0_V_SSE) | (0<<SSPCR0_V_FRF) | ((8-1)<<SSPCR0_V_DSS);

  while (((ssp->sssr) & SSPSR_M_RNE) == 1) { /* drain any data already there */
    i = ssp->ssdr;
  }
  
/*   msetleds(3); */
  msg[0]=0x1B; /* led on */
  msg[1]=0xa7;
  msg[2]=0;
  msg[3]=3;
  msg[4]=lrc(msg, 4);
  putsToKBCTL(msg, 5);

  /* clear keyboard buffer */
  while (getcFromKBCTL() != KBCTL_NODATA) {
    ;
  }
  
}

/*
 * End of Brutus Functions
 */

#endif /* ! CONFIG_ITSY */
	  
#define VERSION 101

void kbd_keyboard_dev (char c)
{
  int v;

  if ((v=kbmapASCII[(int)(c & 0x7f)]) & 0x80) { /* shifted */
    v = (-v) & 0x7f;
/*     printk("/dev/kbd 1: got=0x%x, sending=LSFT,0x%x\n", c, v); */
    kbd_keyboardkey(KK_LSFT, 0);
    kbd_keyboardkey(v, 0);
    kbd_keyboardkey(v, 0x80);
    kbd_keyboardkey(KK_LSFT, 0x80); 
  }
  else if (c <= 0x1a) { /* control */
    if ((v=kbmapASCII[(int)c]) != KK_NONE) {
/*       printk("/dev/kbd 2: got=0x%x, sending=0x%x\n", c, v); */
      kbd_keyboardkey(v, 0);
      kbd_keyboardkey(v, 0x80);
    }
    else {
      v = kbmapASCII[c+0x60];
/*       printk("/dev/kbd 3: got=0x%x, sending=LCTL,0x%x\n", c, v); */
      kbd_keyboardkey(KK_LCTL, 0);
      kbd_keyboardkey(v, 0);
      kbd_keyboardkey(v, 0x80);
      kbd_keyboardkey(KK_LCTL, 0x80);
    }
  }
  else {
/*     printk("/dev/kbd 4: got=0x%x, sending=0x%x\n", c, v); */
    kbd_keyboardkey(v, 0);
    kbd_keyboardkey(v, 0x80);
  }
}
void kbd_keyboard_raw_dev (unsigned int keycode, unsigned int up_flag)
{
  kbd_keyboardkey(keycode, up_flag ? 0x80 : 0);
}

#ifndef CONFIG_ITSY
static void kbd_drv_key(unsigned int keycode, unsigned int up_flag)
{
	kbd_keyboardkey(keycode, up_flag ? 0x80 : 0);
}
#endif

static inline void kbd_drv_reset(void)
{
	key_fn_down = 0;
	key_numl_down = 0;
	kbd_reset();
}

#include "con_info.h"

void kbd_drv_setleds(unsigned int leds)
{
  int state=0;
  static int old_leds=0;

  if (leds == KBCTL_OLD_LEDS)
    leds = old_leds;
  else
    old_leds = leds;

  if (leds & (1 << VC_CAPSLOCK))
    state |= CON_INFO_KBSTATE_CAPS;
  if (leds & (1 << VC_SCROLLOCK))
    state |= CON_INFO_KBSTATE_SCROLLOCK;

  /* LSB Numlock is local! */
  if (key_numl_down)
    state |= CON_INFO_KBSTATE_NUML;

  con_info(CON_INFO_KBSTATE, state);
  
  return;
}

int ignore_kbd_irq = 0;
extern void count_fb_red(void);

#define VirtAdd(PhAdd) (io_p2v(PhAdd))
#include <asm/arch/Itsy.h>
#include <linux/sched.h>

#ifndef CONFIG_ITSY
static void kbd_drv_rx(int irq, void *dev_id, struct pt_regs *regs)
{
	kbd_setregs(regs);
	{
	  int i;
	  while (!ignore_kbd_irq  &&  (i=getcFromKBCTL()) != KBCTL_NODATA) {
	    if (i != KBCTL_AGAIN  &&  i != KK_NONE)
	      kbd_drv_key(i & 0x7f, i & 0x80);
	  }
	}
	reset_timer1();
	mark_bh(KEYBOARD_BH);
}
#endif

/* static void kbd_drv_tx(int irq, void *dev_id, struct pt_regs *regs) */
/* { */
/* } */

int kbd_drv_init (void)
{
	unsigned long flags;

	save_flags_cli (flags);
#ifndef CONFIG_ITSY
	initKBCTL();
	setup_timer1();
	if (request_irq (IRQ_OST1, kbd_drv_rx, SA_SHIRQ, "keyboard", NULL) != 0)
	  printk("Could not allocate keyboard receive IRQ!");
	/* LSB should use panic() above */
#endif
	
	restore_flags (flags);

	printk (KERN_INFO "Keyboard driver v%d.%02d\n", VERSION/100, VERSION%100);
	printk("**** Turning off console printing ****\n");

	/* Turn off console printk messages */
#ifndef CONFIG_ITSY
 	con_printable(0);
#endif
	return 0;
}
