/*
 * Itsy (SA1100+UCB1200) MCP Common Module 
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Carl Waldspurger
 *
 * $Log: mcp_common.c,v $
 * Revision 2.7  1999/12/23  01:28:00  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.6  1999/10/05  01:33:24  caw
 * Minor updates to support both V1 and V2 codec/gpio interfaces.
 *
 * Revision 2.5  1999/05/12  18:01:04  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.4.2.2  1999/05/12  17:11:46  caw
 * Fixed revision log comment.
 *
 * Revision 2.4.2.1  1999/04/27  21:28:38  caw
 * Added missing (un)mask_irq operations in mcp irq primitives.
 *
 * Revision 2.4  1999/04/14  01:08:51  caw
 * Bug fix: do not attempt to report /proc/codec state when mcp disabled.
 *
 * Revision 2.3  1999/04/01  01:25:24  caw
 * Updated mcp_clear_irq_prim() to disable and clear all codec interrupts.
 * Removed several old revision log entries.
 *
 * Revision 2.2  1999/03/31  03:01:55  caw
 * Defined new mcp_common_{enable,disable}_irq() operations.
 * Added procfs entries for "/proc/gpio" and "/proc/codec".
 * Updated "/proc/mcp", other minor improvments.
 *
 * Revision 2.1  1999/03/29  23:49:15  caw
 * Include newer "itsy_codec.h" instead of obsolete "mcp-sa1100.h" header.
 *
 * Revision 1.7  1998/03/28  00:17:48  caw
 * Added support for power management.
 *
 * Revision 1.3  1998/03/06  00:27:57  caw
 * Module debugged and working.  Reduced logging, fixed /proc/mcp formatting.
 *
 * Revision 1.1  1998/03/05  03:55:24  caw
 * Initial revision.
 *
 */

/*
 *
 * debugging 
 *
 */

#define	MCP_DEBUG			(1)
#define	MCP_DEBUG_VERBOSE		(0)

/*
 * includes
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/errno.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef	CONFIG_POWERMGR
#include <linux/powermgr.h>
#include <asm/arch/irq.h>
#endif

#include <asm/io_reg.h>

#include <linux/mcp_common.h> 
#include <linux/itsy_codec.h>
#include <linux/itsy_audio.h>
#include <linux/itsy_version.h>

/*
 * constants
 *
 */

/* names */
#define	MCP_NAME			"mcp-common"
#define	MCP_NAME_VERBOSE		"itsy mcp module"
#define	MCP_VERSION_STRING		"$Revision: 2.7 $"

/* codec */
#define	CODEC_ASD_NUMERATOR		(375000)

/*
 * types
 *
 */

typedef struct {
  /* memory-mapped registers */
  mcp_reg *mcp;
  gpio_reg *gpio;

  /* bitmasks */
  ulong id_mask;
  ulong enabled_mask;
  ulong irq_enabled_mask;

  /* mcp statistics */
  ulong register_count;
  ulong unregister_count;
  ulong enable_count;
  ulong disable_count;
  ulong prim_enable_count;
  ulong prim_disable_count;

  /* irq statistics */
  ulong irq_enable_count;
  ulong irq_disable_count;
  ulong irq_prim_enable_count;
  ulong irq_prim_disable_count;

#ifdef	CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
#endif
} mcp_state;

/*
 * globals
 *
 */

static mcp_state mcp_global;

#ifdef	CONFIG_PROC_FS
static int mcp_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry mcp_proc_entry = {
  0, 3, "mcp", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, mcp_procfs_info,
};

static int gpio_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry gpio_proc_entry = {
  0, 4, "gpio", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, gpio_procfs_info,
};

static int codec_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry codec_proc_entry = {
  0, 5, "codec", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, codec_procfs_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int mcp_procfs_info(char *buf, 
			   char **start,
			   off_t offset,
			   int length,
			   int unused)
{
  mcp_state *m = &mcp_global;
  mcp_reg *mcp = m->mcp;
  int len = 0;

  /* report MCP state */
  len += sprintf(buf + len, 
		 "vaddr    state    name  description\n"
		 "%08lx %08x MCCR0 mcp control reg\n"
		 "%08lx %08x MCSR  mcp status reg\n\n",
		 &mcp->mccr, mcp->mccr,
		 &mcp->mcsr, mcp->mcsr);

  /* format header */
  len += sprintf(buf + len,
		 "operation    requests  actual  bitmask\n"
		 "register               %6d  %08x\n"
		 "unregister             %6d\n"
		 "mcp-enable   %8d  %6d  %08x\n"
		 "mcp-disable  %8d  %6d\n"
		 "irq-enable   %8d  %6d  %08x\n"
		 "irq-disable  %8d  %6d\n",
		 m->register_count,  
		 m->id_mask,
		 m->unregister_count,
		 m->enable_count, 
		 m->prim_enable_count,
		 m->enabled_mask,
		 m->disable_count, 
		 m->prim_disable_count,
		 m->irq_enable_count,
		 m->irq_prim_enable_count,
		 m->irq_enabled_mask,
		 m->irq_disable_count,
		 m->irq_prim_disable_count);
		 
  return(len);
}

static int gpio_procfs_info(char *buf, 
			    char **start,
			    off_t offset,
			    int length,
			    int unused)
{
  mcp_state *m = &mcp_global;
  gpio_reg *gpio = m->gpio;

  int len = 0;

  /* report gpio state */
  len += sprintf(buf + len, 
		 "vaddr    state    name description\n"
		 "%08lx %08x GPLR pin level\n"
		 "%08lx %08x GPDR pin direction\n"
		 "%08lx %08x GRER rising  edge detect\n"
		 "%08lx %08x GFER falling edge detect\n"
		 "%08lx %08x GEDR edge detect status\n"
		 "%08lx %08x GAFR alternate function\n",
		 &gpio->pin_level,     gpio->pin_level,
		 &gpio->pin_direction, gpio->pin_direction,
		 &gpio->rising_edge,   gpio->rising_edge,
		 &gpio->falling_edge,  gpio->falling_edge,
		 &gpio->edge_detect,   gpio->edge_detect,
		 &gpio->alternate_fun, gpio->alternate_fun);

  return(len);
}

static int codec_procfs_info(char *buf, 
			     char **start,
			     off_t offset,
			     int length,
			     int unused)
{
  mcp_state *m = &mcp_global;
  mcp_reg *mcp = m->mcp;
  int len = 0;

  /* punt if codec not enabled */
  if (m->enabled_mask == 0)
    {
      len += sprintf(buf + len, "codec not enabled\n");
      return(len);
    }

  /* report codec state */
  len += sprintf(buf + len, 
		 "reg  state description\n"
		 "  0  %04x  io port data\n"
		 "  1  %04x  io port direction\n"
		 "  2  %04x  rising  edge intr enable\n"
		 "  3  %04x  falling edge intr enable\n"
		 "  4  %04x  interrupt status\n"
		 "  5  %04x  telecom control A\n"
		 "  6  %04x  telecom control B\n"
		 "  7  %04x  audio control A\n"
		 "  8  %04x  audio control B\n"
		 "  9  %04x  touchscreen control\n"
		 " 10  %04x  ADC control\n"
		 " 11  %04x  ADC data\n"
		 " 12  %04x  ID register\n"
		 " 13  %04x  mode register\n",
		 codec_read(mcp, 0),
		 codec_read(mcp, 1),
		 codec_read(mcp, 2),
		 codec_read(mcp, 3),
		 codec_read(mcp, 4),
		 codec_read(mcp, 5),
		 codec_read(mcp, 6),
		 codec_read(mcp, 7),
		 codec_read(mcp, 8),
		 codec_read(mcp, 9),
		 codec_read(mcp, 10),
		 codec_read(mcp, 11),
		 codec_read(mcp, 12),
		 codec_read(mcp, 13));

  return(len);
}
#endif	/* CONFIG_PROC_FS */

/*
 * mcp_state operations
 *
 */

static void mcp_state_init(mcp_state *m)
{
  /* initialize mcp state */
  memset((void *) m, 0, sizeof(mcp_state));

  /* initialize memory-mapped regions */
  m->mcp  = MCP_BASE_V;
  m->gpio = GPIO_BASE_V;

  /* initialize bitmasks */
  m->id_mask = 0;
  m->enabled_mask = 0;
  m->irq_enabled_mask = 0;

  /* initialize stats */
  m->register_count   = 0;
  m->unregister_count = 0;
  m->enable_count  = 0;
  m->disable_count = 0;
  m->prim_enable_count  = 0;
  m->prim_disable_count = 0;

  /* initialize irq stats */
  m->irq_enable_count  = 0;
  m->irq_disable_count = 0;
  m->irq_prim_enable_count  = 0;
  m->irq_prim_disable_count = 0;

#ifdef	CONFIG_POWERMGR
  /* initialize powermgr state */
  m->pm_id = POWERMGR_ID_INVALID;
#endif	/* CONFIG_POWERMGR */
}

static inline int mcp_id_valid(const mcp_state *m, mcp_common_id id)
{
  /*
   * modifies: nothing
   * effects:  Returns TRUE iff id is a valid and allocated for m.
   *
   */

  /* check id range */
  if ((id < 0) || (id > MCP_COMMON_ID_MAX))
    return(0);

  /* check id allocation */
  if ((m->id_mask & (1 << id)) == 0)
    return(0);

  /* valid id */
  return(1);
}

static inline void mcp_enable_prim(mcp_state *m)
{
  /* turn on the MCP */
  int asd = CODEC_ASD_NUMERATOR / AUDIO_RATE_DEFAULT;
  m->mcp->mccr = 0;
  m->mcp->mccr = MCCR_M_ADM | MCCR_M_MCE | (40 << MCCR_V_TSD) | asd;

  if (is_itsy_v1())
    {
      /* reset V1 I/O port directions, initialize output ports */
      codec_write(m->mcp, CODEC_REG_IO_DIRECTION, CODEC_ITSY_V1_IO_DIRECTIONS);
      codec_write(m->mcp, CODEC_REG_IO_DATA, CODEC_ITSY_V1_IO_DATA_INIT);
    }
  else
    {
      /* reset V2 I/O port directions, initialize output ports */
      codec_write(m->mcp, CODEC_REG_IO_DIRECTION, CODEC_ITSY_V2_IO_DIRECTIONS);
      codec_write(m->mcp, CODEC_REG_IO_DATA, CODEC_ITSY_V2_IO_DATA_INIT);
    }

  /* update stats */
  m->prim_enable_count++;

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": mcp_enable_prim\n");
}

static inline void mcp_disable_prim(mcp_state *m)
{
  /* turn off the MCP */
  m->mcp->mccr = 0;

  /* update stats */
  m->prim_disable_count++;

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": mcp_disable_prim\n");
}

static void mcp_clear_irq_prim(mcp_state *m)
{
  ushort status;

  /* disable all codec interrupts */
  codec_write(m->mcp, CODEC_REG_RISE_INT_ENABLE, 0);
  codec_write(m->mcp, CODEC_REG_FALL_INT_ENABLE, 0);

  /* clear codec interrupt status (0->1 transition) */
  status = codec_read(m->mcp, CODEC_REG_INT_STATUS);  
  codec_write(m->mcp, CODEC_REG_INT_STATUS, 0);
  codec_write(m->mcp, CODEC_REG_INT_STATUS, status);

  /* clear any pending codec interrupts */
  m->gpio->edge_detect |= (1 << CODEC_IRQ);
}

static inline void mcp_enable_irq_prim(mcp_state *m)
{
  /* clear any pending interrupts */
  mcp_clear_irq_prim(m);
  
  /* enable interrupt on rising-edge from codec */
  m->gpio->rising_edge  |=  (1 << CODEC_IRQ);
  m->gpio->falling_edge &= ~(1 << CODEC_IRQ);
  
  /* enable codec interrupts */
  unmask_irq(CODEC_IRQ);

  /* update stats */
  m->irq_prim_enable_count++;

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": mcp_enable_irq_prim\n");
}

static inline void mcp_disable_irq_prim(mcp_state *m)
{
  /* disable codec interrupts */
  mask_irq(CODEC_IRQ);

  /* disable interrupts from codec */
  m->gpio->rising_edge  &= ~(1 << CODEC_IRQ);
  m->gpio->falling_edge &= ~(1 << CODEC_IRQ);

  /* update stats */
  m->irq_prim_disable_count++;

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": mcp_disable_irq_prim\n");
}

/*
 * exported operations
 *
 */

mcp_common_id mcp_common_register(void)
{
  /*
   * modifies: mcp_global
   * effects:  Register as user of SA1100 MCP.  Returns a unique mcp id
   *           for use with future calls, or MCP_COMMON_ID_INVALID if
   *           unable to register.
   *
   */

  mcp_state *m = &mcp_global;
  mcp_common_id id;
  ulong flags, i;

  /* search bitmask for available id */
  id = MCP_COMMON_ID_INVALID;
  save_flags_cli(flags);
  for (i = 0; i <= MCP_COMMON_ID_MAX; i++)
    if ((m->id_mask & (1 << i)) == 0)
      {
	/* allocate id, clear enabled masks */
	m->id_mask |= (1 << i);
	m->enabled_mask &= ~(1 << i);
	m->irq_enabled_mask &= ~(1 << i);
	id = (mcp_common_id) i;
	break;
      }
  restore_flags(flags);

#ifdef	MODULE
  if (id != MCP_COMMON_ID_INVALID)
    {
      MOD_INC_USE_COUNT;
    }
#endif

  /* update stats */
  m->register_count++;

  return(id);
}

int mcp_common_unregister(mcp_common_id id)
{
  /*
   * modifies: mcp_global
   * effects:  Deregisters id as user of SA1100 MCP. 
   *           Returns 0 iff successful, otherwise returns error code.
   *
   */

  mcp_state *m = &mcp_global;
  ulong flags;
  int retval;

  /* sanity check */
  if ((id < 0) || (id > MCP_COMMON_ID_MAX))
    return(-EINVAL);

  /* search bitmask for id */
  save_flags_cli(flags);
  if (m->id_mask & (1 << id))
    {
      /* forcibly disable irq */
      if (m->irq_enabled_mask & (1 << id))
	(void) mcp_common_disable_irq(id);

      /* forcibly disable */
      if (m->enabled_mask & (1 << id))
	(void) mcp_common_disable(id);

      /* deallocate id */
      m->id_mask &= ~(1 << id);
      retval = 0;
    }
  else
    {
      /* id not allocated */
      retval = -EINVAL;
    }
  restore_flags(flags);

#ifdef	MODULE
  if (retval == 0)
    {
      MOD_DEC_USE_COUNT;
    }
#endif

  /* update stats */
  m->unregister_count++;

  return(retval);
}

int mcp_common_enable(mcp_common_id id)
{
  /*
   * modifies: mcp_global
   * effects:  Turns on SA1100 MCP; no effect if MCP is already on.
   *           Redundant calls with the same id are safely ignored.
   *           Returns 0 iff successful, otherwise returns error code.
   *
   */

  mcp_state *m = &mcp_global;
  ulong flags;

  /* sanity check */
  if (!mcp_id_valid(m, id))
    return(-EINVAL);

  save_flags_cli(flags);
  if ((m->enabled_mask & (1 << id)) == 0)
    {
      /* turn on if necessary */
      if (m->enabled_mask == 0)
	mcp_enable_prim(m);	
      m->enabled_mask |= (1 << id);
    }
  restore_flags(flags);

  /* update stats */
  m->enable_count++;

  /* successful */
  return(0);
}

int mcp_common_disable(mcp_common_id id)
{
  /*
   * modifies: mcp_global
   * effects:  Turns off SA1100 MCP; no effect if MCP is already off.
   *           Redundant calls with the same id are safely ignored.
   *           Returns 0 iff successful, otherwise returns error code.
   *
   */

  mcp_state *m = &mcp_global;
  ulong flags;

  /* sanity check */
  if (!mcp_id_valid(m, id))
    return(-EINVAL);

  save_flags_cli(flags);
  if (m->enabled_mask & (1 << id))
    {
      /* turn off if necessary */
      m->enabled_mask &= ~(1 << id);      
      if (m->enabled_mask == 0)
	mcp_disable_prim(m);
    }
  restore_flags(flags);

  /* update stats */
  m->disable_count++;

  /* successful */
  return(0);
}

int mcp_common_enable_irq(mcp_common_id id)
{
  /*
   * modifies: mcp_global
   * effects:  Enables shared interrupt on rising edge of Itsy GPIO 6.
   *           Redundant calls with the same id are safely ignored.
   *           Returns 0 iff successful, otherwise returns error code.
   *
   */

  mcp_state *m = &mcp_global;
  ulong flags;

  /* sanity check */
  if (!mcp_id_valid(m, id))
    return(-EINVAL);

  save_flags_cli(flags);
  if ((m->irq_enabled_mask & (1 << id)) == 0)
    {
      /* turn on if necessary */
      if (m->irq_enabled_mask == 0)
	mcp_enable_irq_prim(m);	
      m->irq_enabled_mask |= (1 << id);
    }
  restore_flags(flags);

  /* update stats */
  m->irq_enable_count++;

  /* successful */
  return(0);
}

int mcp_common_disable_irq(mcp_common_id id)
{
  /*
   * modifies: mcp_global
   * effects:  Disables interrupt on rising edge of Itsy GPIO 6.
   *           Redundant calls with the same id are safely ignored.
   *           Returns 0 iff successful, otherwise returns error code.
   *
   */

  mcp_state *m = &mcp_global;
  ulong flags;

  /* sanity check */
  if (!mcp_id_valid(m, id))
    return(-EINVAL);

  save_flags_cli(flags);
  if (m->irq_enabled_mask & (1 << id))
    {
      /* turn off if necessary */
      m->irq_enabled_mask &= ~(1 << id);      
      if (m->irq_enabled_mask == 0)
	mcp_disable_irq_prim(m);
    }
  restore_flags(flags);

  /* update stats */
  m->irq_disable_count++;

  /* successful */
  return(0);
}

/*
 * power management operations
 *
 */

#ifdef	CONFIG_POWERMGR
static int mcp_common_suspend_check(void *cookie, int idle_jiffies)
{
  /*
   * modifies: nothing
   * effects:  Always returns zero; i.e. willing to sleep immediately.
   *
   */

  return(0);
}

static int mcp_common_suspend(void *cookie)
{
  /*
   * modifies: nothing
   * effects:  Saves state in preparation for sleep mode.
   *           Always returns zero (ignored).
   *
   */

  mcp_state *m = &mcp_global;

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": mcp_common_suspend\n");

  /* disable codec interrupts, if necessary */
  if (m->irq_enabled_mask != 0)
    mcp_disable_irq_prim(m);

  /* turn off mcp, if necessary */
  if (m->enabled_mask != 0)
    mcp_disable_prim(m);

  /* return value ignored */
  return(0);
}

static int mcp_common_resume(void *cookie, int resume_flags)
{
  /*
   * modifies: mcp_global
   * effects:  Resumes mcp state after wakeup from sleep mode.
   *           Always returns zero (ignored).
   *
   */

  mcp_state *m = &mcp_global;

  /* restore mcp state */
  if (m->enabled_mask == 0)
    mcp_disable_prim(m);
  else
    mcp_enable_prim(m);

  /* restore irq setup */
  if (m->irq_enabled_mask == 0)
    mcp_disable_irq_prim(m);
  else
    mcp_enable_irq_prim(m);

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": mcp_common_resume (%lx)\n", m->enabled_mask);

  /* return value ignored */
  return(0);
}

static const powermgr_client mcp_powermgr = {
  /* callback functions */
  mcp_common_suspend_check,
  mcp_common_suspend,
  mcp_common_resume,
  
  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  MCP_NAME_VERBOSE,

  /* power-consumption info */
  0,
  0
};

#endif	/* CONFIG_POWERMGR */

/*
 * module operations
 *
 */

static struct symbol_table mcp_common_syms = {
#include <linux/symtab_begin.h>
        X(mcp_common_register),
        X(mcp_common_unregister),
	X(mcp_common_enable),
	X(mcp_common_disable),
	X(mcp_common_enable_irq),
	X(mcp_common_disable_irq),
#include <linux/symtab_end.h>
};

static char *mcp_version(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current cvs revision number.
   * storage:  Return value is statically-allocated.
   *
   */

  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = MCP_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return(version);
}

#ifdef	MODULE
int init_module(void)
{
  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": init_module\n");

  /* initialize global state */
  mcp_state_init(&mcp_global);

#ifdef	CONFIG_PROC_FS
  /* register procfs devices */
  proc_register_dynamic(&proc_root, &mcp_proc_entry);
  proc_register_dynamic(&proc_root, &gpio_proc_entry);
  proc_register_dynamic(&proc_root, &codec_proc_entry);
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_POWERMGR
  mcp_global.pm_id = powermgr_register(&mcp_powermgr);
  if (mcp_global.pm_id == POWERMGR_ID_INVALID)
    if (MCP_DEBUG)
      printk(MCP_NAME ": init_module: powermgr_register failed\n");
#endif	/* CONFIG_POWERMGR */

  /* export symbols */
  (void) register_symtab(&mcp_common_syms);

  /* log registration */
  printk(MCP_NAME_VERBOSE " version %s initalized\n", mcp_version());

  /* everything OK */
  return(0);
}

void cleanup_module(void)
{
  mcp_state *m = &mcp_global;
  int err;

  if (MCP_DEBUG_VERBOSE)
    printk(MCP_NAME ": cleanup_module\n");

#ifdef	CONFIG_PROC_FS
  /* unregister procfs entries */
  if ((err = proc_unregister(&proc_root, mcp_proc_entry.low_ino)) != 0)
    if (MCP_DEBUG)
      printk(MCP_NAME ": cleanup_module: proc_unregister(mcp) %d\n", err);
  if ((err = proc_unregister(&proc_root, gpio_proc_entry.low_ino)) != 0)
    if (MCP_DEBUG)
      printk(MCP_NAME ": cleanup_module: proc_unregister(gpio) %d\n", err);
  if ((err = proc_unregister(&proc_root, codec_proc_entry.low_ino)) != 0)
    if (MCP_DEBUG)
      printk(MCP_NAME ": cleanup_module: proc_unregister(codec) %d\n", err);
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_POWERMGR
  if ((err = powermgr_unregister(m->pm_id)) != POWERMGR_SUCCESS)
    if (MCP_DEBUG)
      printk(MCP_NAME ": cleanup_module: powermgr_unregister %d\n",err);
#endif	/* CONFIG_POWERMGR */

  /* sanity check */
  if ((m->id_mask != 0) || (m->enabled_mask != 0))
    printk(MCP_NAME ": cleanup_module: id=%lx, enabled=%lx\n",
	   m->id_mask,
	   m->enabled_mask);
  
  /* forcibly disable mcp */
  mcp_disable_prim(m);

  /* log unload */
  printk(MCP_NAME_VERBOSE " unloaded\n");
}
#endif	/* MODULE */
