/*
 * Itsy Powerbutton Device
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Authors: Deborah A. Wallach and Carl Waldspurger
 *
 * $Log: powerbutton.c,v $
 * Revision 2.7  2000/01/11  22:35:32  kerr
 * Support for deep sleep mode.
 *
 * Revision 2.6  1999/12/23  01:28:01  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.5  1999/12/11  04:45:36  mann
 * Power manager support for flash.
 * Powerbutton bugfix: don't try to go to sleep from interrupt level, as
 * drivers may be in an unclean state.
 *
 * Revision 2.4  1999/10/08 21:58:57  kerr
 * Include linux/version.h so dependencies are correct.
 *
 * Revision 2.3  1999/09/10  21:41:18  caw
 * Updated to reflect changes to itsy_event_header field names.
 *
 * Revision 2.2  1999/05/21 19:20:26  brakmo
 * Modified so releasing the button doesn't wake it up (used in gmanager)
 *
 * Revision 2.1  1999/05/12 18:01:05  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.0.2.3  1999/04/23  22:06:02  caw
 * Added support for POWERBUTTON_ENTER_SLEEP_MODE operation.
 * Expanded overview documentation.
 *
 * Revision 2.0.2.2  1999/04/23  20:04:21  caw
 * Updated to support both up and down events.  Other minor improvements.
 *
 * Revision 2.0.2.1  1999/04/21  01:24:30  caw
 * Substantial rewrite, with support for file I/O and button events.
 * Backwards-compatibility OK, but new features still experimental.
 *
 */

/*
 * Overview
 *
 * This driver supports interrupt-driven button input for the special
 * power button on the Itsy platform.  The original driver (written
 * completely by Debby Wallach) invoked sleep mode whenever this button
 * was pressed and released.  The current driver also supports standard
 * file I/O operations, plus an ioctl() to control whether or not sleep
 * mode should be entered.
 *
 * Note that the routine powerbutton_powermgr_register() is invoked
 * early, prior to the main powerbutton_init() initialization routine.
 * This is done in the powermgr initialization code to ensure that the
 * powerbutton device is registered in the appropriate order.  The global
 * variable pb_pm_ready is used as a guard to prevent any early callbacks
 * from the powermgr from accessing uninitialized data.
 *
 */
 
/*
 * debugging
 *
 */

#define	PB_DEBUG		(1)
#define	PB_DEBUG_VERBOSE	(0)

/*
 * includes
 *
 */

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <linux/types.h>
#include <linux/ctype.h>
#include <linux/config.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/irq-no.h>
#include <asm/hardware.h>
#include <asm/io_reg.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#ifdef	CONFIG_POWERMGR
#include <linux/powermgr.h>
#endif

#include <linux/itsy_session.h>
#include <linux/itsy_codec.h>
#include <linux/itsy_powerbutton.h>

#include "itsy_util.h"

/*
 * externals
 *
 */

extern void sa1100_sleep_mode_enter(void);
extern void sa1100_deep_sleep_mode_enter(void);

/*
 * constants
 *
 */

/* names */
#define	PB_NAME			"itsy-powerbutton"
#define	PB_NAME_VERBOSE		"itsy powerbutton driver"
#define	PB_VERSION_STRING	"$Revision: 2.7 $"

/* device number */
#define	PB_MAJOR		(120)

/* interrupt number */
#define	PB_IRQ			(IRQ_MAINPB)

/* buffer sizes */
#define	PB_BUFFER_LG_NEVENTS	(6)
#define	PB_BUFFER_NEVENTS	(1 << PB_BUFFER_LG_NEVENTS)
#define	PB_BUFFER_INDEX_MASK	(PB_BUFFER_NEVENTS - 1)

/*
 * types
 *
 */

typedef struct pb_client {
  /* linked list */
  struct pb_client *prev, *next;

  /* identity */
  int id;

  /* allow sleep mode? */
  int enable_sleep;

  /* buffer management */
  int buf_next;

  /* dropped events */
  int drop_count;
} pb_client;

typedef struct pb_dev {
  /* memory-mapped regs */
  gpio_reg *gpio;

  /* buffer management */
  struct wait_queue *waitq;
  itsy_buttons_event *buf;
  int buf_events;
  int buf_bytes;

  /* buffer indexing */
  int buf_mask;
  int buf_next;

  /* clients */
  pb_client *clients;
  int client_id_next;
  int nclients;

  /* statistics */
  ulong interrupt_count;
  ulong event_count;

#ifdef	CONFIG_POWERMGR
  /* power management */
  powermgr_id pm_id;
  int resuming;
#endif
} pb_dev;

/*
 * globals
 *
 */

/* early powermgr initialization (ugh!) */
static powermgr_id pb_pm_id = POWERMGR_ID_INVALID;
static int pb_pm_ready = 0;

static pb_dev pb_global;

#ifdef	CONFIG_PROC_FS
static int pb_procfs_info(char *, char **, off_t, int, int);
static struct proc_dir_entry pb_proc_entry = {
  0, 11, "powerbutton", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, pb_procfs_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * instantiate list operations
 *
 */

INSTANTIATE_LIST_INSERT(pb_client)
INSTANTIATE_LIST_REMOVE(pb_client)

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int pb_procfs_info(char *buf, 
			  char **start,
			  off_t offset,
			  int length,
			  int unused)
{
  pb_dev *dev = &pb_global;
  int len = 0;

  /* format debugging info */
  len += sprintf(buf + len,
		 "debugging:\n"
		 "  %5d pb_pm_ready\n",
		 pb_pm_ready);

  /* format current state */
  len += sprintf(buf + len,
		 "statistics:\n"
		 "  %5ld interrupts\n"
		 "  %5ld events\n"
		 "  %5d clients\n",
		 dev->interrupt_count,
		 dev->event_count,
		 dev->nclients);

  /* format client info */
  if (dev->clients != NULL)
    {
      pb_client *c;

      len += sprintf(buf + len, "clients:\n");
      for (c = dev->clients; c != NULL; c = c->next)
	len += sprintf(buf + len,
		       "  %5d %s\n",
		       c->id, 
		       (c->enable_sleep ? "yes" : "no"));
    }

  return(len);  
}
#endif	/* CONFIG_PROC_FS */

/*
 * memory-mapped register operations
 *
 */

static inline void pb_gpio_clear(gpio_reg *gpio)
{
  /* clear pending interrupts, if any */
  gpio->edge_detect |= (1 << PB_IRQ);
}

static inline void pb_gpio_setup(gpio_reg *gpio)
{
  /* clear interrupts */
  pb_gpio_clear(gpio);
  
  /* setup interrupts on both button up and down */
  gpio->rising_edge  |= (1 << PB_IRQ);
  gpio->falling_edge |= (1 << PB_IRQ);    
}

static inline void pb_gpio_suspend_setup(gpio_reg *gpio)
{
  /* clear interrupts */
  pb_gpio_clear(gpio);
  
  /* setup interrupts only on button down */
  /*   gpio->rising_edge  |= (1 << PB_IRQ);   */
  gpio->rising_edge  &= ~(1 << PB_IRQ);  

  gpio->falling_edge |= (1 << PB_IRQ);      
  /*   gpio->falling_edge &= ~(1 << PB_IRQ); */
}

static inline void pb_enable_wakeup(void)
{
  /* enable power button wakeup */
  *((uint *) PWER_V) |= (1 << PB_IRQ);
}

/*
 * pb_client operations
 *
 */

static pb_client *pb_client_create(int id)
{
  /*
   * modifies: nothing
   * effects:  Returns a new, initialized pb_client, or
   *	       NULL if unable to allocate storage.
   */

  pb_client *client;

  /* allocate storage, fail if unable */
  if ((client = KOBJ_ALLOC(pb_client)) == NULL)
    return(NULL);
  KOBJ_INIT(client, pb_client);

  /* set identity */
  client->id = id;

  /* enable sleep mode by default */
  client->enable_sleep = 1;

  /* everything OK */
  return(client);
}

static void pb_client_destroy(pb_client *client)
{
  /*
   * modifies: client
   * effects:  Reclaims all storage associated with client.
   *
   */

  kfree(client);
}

#ifdef	CONFIG_POWERMGR
/*
 * power management operations
 *
 */

static int powerbutton_suspend_check(void *ignore, int idle_jiffies)
{
  /*
   * modifies: nothing
   * effects:  Always returns zero; i.e. willing to sleep immediately.
   *
   */

  return(0);
}

/* n.b. also called externally from wakeup_from_sleep_mode() */
int powerbutton_suspend(void *ignore)
{
  pb_dev *dev = &pb_global;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_suspend\n");

  if (!pb_pm_ready)
    return(0);

  pb_gpio_suspend_setup(dev->gpio);
  pb_enable_wakeup();
  disable_irq(PB_IRQ);

  /* return value ignored */
  return(0);
}

static int powerbutton_resume(void *ignore, int resume_flags)
{
  pb_dev *dev = &pb_global;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_resume\n");

  if (!pb_pm_ready)
    return(0);

  /* set flag */
  dev->resuming = 1;

  pb_gpio_setup(dev->gpio);
  enable_irq(PB_IRQ);

  /* return value ignored */
  return(0);
}

static const powermgr_client pb_powermgr = {
  /* callback functions */
  powerbutton_suspend_check,
  powerbutton_suspend,
  powerbutton_resume,
  
  /* uninterpreted client data */
  NULL,

  /* identity */
  POWERMGR_ID_INVALID,
  PB_NAME_VERBOSE,

  /* power-consumption info */
  0,
  0
};
#endif	/* CONFIG_POWERMGR */

/*
 * pb_dev operations
 *
 */

static int pb_dev_init(pb_dev *dev)
{
  /*
   * modifies: dev
   * effects:  Initializes powerbutton device dev.
   *	       Returns 0 iff successful, otherwise error code.
   *
   */

  /* clear everything */
  KOBJ_INIT(dev, pb_dev);
  
  /* allocate buffer, fail if unable */
  dev->buf_events = PB_BUFFER_NEVENTS;
  dev->buf_bytes  = dev->buf_events * sizeof(itsy_buttons_event);
  dev->buf = (itsy_buttons_event *) kmalloc(dev->buf_bytes, GFP_KERNEL);
  if (dev->buf == NULL)
    return(-ENOMEM);

  /* buffer indexing */
  dev->buf_mask = PB_BUFFER_INDEX_MASK;
  dev->buf_next = 0;
  
  /* initialize gpio */
  dev->gpio = GPIO_BASE_V;

#ifdef	CONFIG_POWERMGR
  /* initialize powermgr state */
  dev->pm_id = POWERMGR_ID_INVALID;
  dev->resuming = 0;
#endif	/* CONFIG_POWERMGR */

  /* everything OK */
  return(0);
}

static void pb_dev_event(pb_dev *dev, ushort state)
{
  /*
   * modifies: buf
   * effects:  Creates event with specified state.  Event is added to the
   *           buffer associated with dev.  Wakes up any clients that may
   *           be blocked waiting for the event.
   *
   */

  itsy_buttons_event *next;

  /* use next slot in circular buffer */
  next = &dev->buf[dev->buf_next & dev->buf_mask];
  
  /* generic session event */
  do_gettimeofday(&next->common.timestamp);

  next->common.event_class = ITSY_EVENT_CLASS_BUTTONS;
  next->common.event_type  = ITSY_EVENT_BUTTONS_STATE;
  next->common.size  = sizeof(itsy_buttons_event);
  next->common.flags = ITSY_EVENT_FLAG_NONE;

  /* specific buttons event */
  next->sid = ITSY_SESSION_ID_NONE;
  next->state = state;
  
  /* advance buffer index */
  dev->buf_next++;

  /* wakeup waiters, if any */
  wake_up_interruptible(&dev->waitq);

  /* update statistics */
  dev->event_count++;
}

/*
 * interrupt processing 
 *
 */

static int pb_sleep_enabled(pb_dev *dev)
{
  pb_client *c;

  /* disallowed if disabled by any client */
  for (c = dev->clients; c != NULL; c = c->next)
    if (!c->enable_sleep)
      return(0);

  /* enabled */
  return(1);
}

static void pb_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  pb_dev *dev = &pb_global;
  gpio_reg *gpio = dev->gpio;
  uint button_up;

  /* determine button state */
  button_up = gpio->pin_level & (1 << PB_IRQ);

  /* clear edge detect register */
  pb_gpio_clear(gpio);

  /* buffer event */
  pb_dev_event(dev, button_up ? 0 : BUTTON_POWER);

  /* update statistics */
  dev->interrupt_count++;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_interrupt %ld, up=%u\n", 
	   dev->interrupt_count,
	   button_up);

#ifdef	CONFIG_POWERMGR
  /* on button up, enter sleep mode if enabled */
  if (button_up && pb_sleep_enabled(dev))
    {
      /* enter sleep mode, unless resuming */
      if (dev->resuming) {
	dev->resuming = 0;
      } else {
	/* Wait until we are no longer at interrupt level  --mann */
	static struct tq_struct tq;
	tq.routine = (void (*)(void*)) sa1100_sleep_mode_enter;
	tq.data = NULL;
	queue_task(&tq, &tq_scheduler);
      }
    }
#endif	/* CONFIG_POWERMGR */
}

static void pb_hardware_init(pb_dev *dev)
{
  pb_gpio_setup(dev->gpio);
  pb_enable_wakeup();
}

/*
 * file operations
 *
 */

#ifdef	CONFIG_POWERMGR
static int pb_enter_sleep_mode(void)
{
  /* must be root to force sleep */
  if (!suser())
    return(-EPERM);

  /* enter sleep mode */
  sa1100_sleep_mode_enter();
  return(0);
}

static int pb_enter_deep_sleep_mode(void)
{
  /* must be root to force sleep */
  if (!suser())
    return(-EPERM);

  /* enter sleep mode */
  sa1100_deep_sleep_mode_enter();
  return(0);
}
#endif	/* CONFIG_POWERMGR */

static int pb_dev_get_sleep(pb_dev *dev, int *enabled)
{
  /* copyout reply value */
  put_user(pb_sleep_enabled(dev), enabled);
  return(0);
}

static int pb_client_get_sleep(pb_client *client, int *enabled)
{
  /* copyout reply value */
  put_user(client->enable_sleep, enabled);
  return(0);
}

static int pb_client_set_sleep(pb_client *client, int enable)
{
  /* update client flag */
  client->enable_sleep = enable;
  return(0);
}

static int pb_ioctl(struct inode *inode,
		    struct file *file,
		    uint command,
		    ulong arg)
{
  pb_client *client = (pb_client *) file->private_data;
  pb_dev *dev = &pb_global;
  int size, err;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_ioctl: cmd=%d, arg=%lu\n", command, arg);

  /* sanity checks */
  if (_IOC_TYPE(command) != POWERBUTTON_IOCTL_TYPE)
    return(-EINVAL);
  if (_IOC_NR(command) > POWERBUTTON_IOCTL_MAXNR)
    return(-EINVAL);

  /* verify transfer memory */
  size = _IOC_SIZE(command);
  if (_IOC_DIR(command) & _IOC_READ)
    if ((err = verify_area(VERIFY_WRITE, (void *) arg, size)) != 0)
      return(err);
  if (_IOC_DIR(command) & _IOC_WRITE)
    if ((err = verify_area(VERIFY_READ,  (void *) arg, size)) != 0)
      return(err);

  /* dispatch based on command */
  switch (command) {
  case POWERBUTTON_GET_SLEEP_ALL:
    return(pb_dev_get_sleep(dev, (int *) arg));
  case POWERBUTTON_GET_SLEEP:
    return(pb_client_get_sleep(client, (int *) arg));
  case POWERBUTTON_SET_SLEEP:
    return(pb_client_set_sleep(client, (int) arg));
#ifdef	CONFIG_POWERMGR
  case POWERBUTTON_ENTER_SLEEP_MODE:
    return(pb_enter_sleep_mode());
  case POWERBUTTON_ENTER_DEEP_SLEEP_MODE:
    return(pb_enter_deep_sleep_mode());
#endif	/* CONFIG_POWERMGR */
  }

  /* invalid command */
  return(-EINVAL);
}

static int pb_read(struct inode *inode, 
		   struct file *file,
		   char *buffer,
		   int count)
{
  /*
   * modifies: buffer
   * effects:  Copies available powerbutton events, if any, to buffer.
   *           Returns the number of bytes copied.
   *
   */

  const int event_bytes = sizeof(itsy_buttons_event);

  pb_client *client = (pb_client *) file->private_data;
  pb_dev *dev = &pb_global;
  int total, buf_first, err;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_read\n");

  /* verify user-space buffer validity */
  if ((err = verify_area(VERIFY_WRITE, (void *) buffer, count)) != 0)
    return(err);

  /* initialize */
  total = 0;

  /* determine first available buffered event */
  buf_first = MAX(0, dev->buf_next - dev->buf_events);

  /* skip over dropped events */
  if (client->buf_next < buf_first)
    {
      client->buf_next = buf_first;
      client->drop_count++;
    }

  /* transfer integral number of events */
  for (;
       (client->buf_next < dev->buf_next) && (count >= event_bytes);
       client->buf_next++)
    {
      itsy_buttons_event *event;
      
      /* examine next buffered event */
      event = &dev->buf[client->buf_next & dev->buf_mask];

      /* copyout event */
      memcpy_tofs((void *) buffer, (const void *) event, event_bytes);

      /* update counts */
      total  += event_bytes;
      buffer += event_bytes;
      count  -= event_bytes;
    }

  return(total);
}

static int pb_select(struct inode *inode,
		     struct file *file,
		     int mode,
		     select_table *table)
{
  pb_client *client = (pb_client *) file->private_data;
  pb_dev *dev = &pb_global;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_select\n");

  /* handle select on input */
  if (mode == SEL_IN)
    {
      /* any events buffered? */
      if (client->buf_next < dev->buf_next)
	return(1);

      /* wait for event */
      if (PB_DEBUG_VERBOSE)
	printk(PB_NAME ": pb_select: wait...\n");
      select_wait(&dev->waitq, table);
      return(0);
    }

  /* SEL_OUT, SEL_EX not supported */
  return(0);
}

static int pb_open(struct inode *inode, struct file *file)
{
  pb_dev *dev = &pb_global;
  pb_client *client;
  int id;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_open\n");

  /* add new client, fail if unable */
  id = dev->client_id_next++;
  if ((client = pb_client_create(id)) == NULL)
    return(-ENOMEM);

  /* skip over old events */
  client->buf_next = dev->buf_next;

  /* add to client list */
  pb_client_insert(&dev->clients, client);
  dev->nclients++;

  /* attach client state to file */
  file->private_data = client;

  /* everything OK */
  return(0);
}

static void pb_release(struct inode *inode, struct file *file)
{
  pb_client *client = (pb_client *) file->private_data;
  pb_dev *dev = &pb_global;

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": pb_release\n");

  /* remove from client list */
  pb_client_remove(&dev->clients, client);
  dev->nclients--;

  /* reclaim storage */
  pb_client_destroy(client);
  file->private_data = NULL;
}

static struct file_operations pb_fops =
{
  NULL,			/* no special lseek */
  pb_read,
  NULL,			/* no special write */
  NULL,			/* no special readdir */
  pb_select,
  pb_ioctl,
  NULL,			/* no special mmap */
  pb_open,
  pb_release,
  NULL,			/* no special fsync */
  NULL,			/* no special fasync */
  NULL,			/* no special check_media_change */
  NULL			/* no special revalidate */
};

/*
 * initialization
 *
 */

static char *pb_version(void)
{
  /*
   * modifies: nothing
   * effects:  Returns the current cvs revision number.
   * storage:  Return value is statically-allocated.
   *
   */

  static char version[8];
  char *revision;
  int i = 0;

  /* extract version number from cvs string */
  for (revision = PB_VERSION_STRING; *revision != '\0'; revision++)
    if (isdigit((int) *revision) || (*revision == '.'))
      version[i++] = *revision;
  version[i] = '\0';

  return(version);
}

void powerbutton_powermgr_register(void)
{
#ifdef	CONFIG_POWERMGR
  /* register with power manager */
  pb_pm_id = powermgr_register(&pb_powermgr);
#endif	/* CONFIG_POWERMGR */
}

int powerbutton_init(void)
{
  static int initialized = 0;

  pb_dev *dev = &pb_global;
  int err;

  /* initialize only once */
  if (initialized++)
    return(-EPERM);

  if (PB_DEBUG_VERBOSE)
    printk(PB_NAME ": powerbutton_init\n");

  /* initialize global state */
  pb_dev_init(dev);

  /* register device */
  if ((err = register_chrdev(PB_MAJOR, PB_NAME, &pb_fops)) != 0)
    {
      /* log error and fail */
      printk(PB_NAME ": unable to register major device %d\n", PB_MAJOR);
      return(err);
    }

#ifdef	CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &pb_proc_entry);
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_POWERMGR
  /* use result of earlier power manager registration */
  dev->pm_id = pb_pm_id;
  if (dev->pm_id == POWERMGR_ID_INVALID)
    if (PB_DEBUG)
      printk(PB_NAME ": unable to register with power manager\n");
#endif	/* CONFIG_POWERMGR */

  /* initialize shared IRQ */
  if ((err = request_irq(PB_IRQ,
			 pb_interrupt,
			 0 /* XXX SA_INTERRUPT */,
			 PB_NAME,
			 dev)) != 0)
    {
      /* log error and fail */
      printk(PB_NAME ": request_irq failed: %d\n", err);
      return(err);
    }

  /* initialize hardware */
  pb_hardware_init(dev);

  /* allow powermgr operations */
  pb_pm_ready = 1;

  /* log device registration */
  printk(PB_NAME_VERBOSE " version %s initialized\n", pb_version());

  /* everything OK */
  return(0);
}
