/*
 * Rock 'n' Scroll driver for Itsy V1.5 and V2.x with ADXL202 directly 
 * connected to the StrongARM.
 *
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * $Log: rocknscroll.c,v $
 * Revision 2.7  1999/12/23  01:28:01  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.6  1999/10/20  23:47:30  bartlett
 * Enable required interrupts on first open and on resume
 *
 * Revision 2.5  1999/10/09 01:56:19  bartlett
 * Added support for Itsy V2
 *
 * Revision 2.4  1999/10/08 21:58:58  kerr
 * Include linux/version.h so dependencies are correct.
 *
 * Revision 2.3  1999/06/12  02:57:59  bartlett
 * Fixed free counter clear in interrupt handler
 *
 * Revision 2.2  1999/05/21  23:41:07  bartlett
 * Add select and process info
 *
 *
 * Revision 2.1  1999/05/18  01:37:48  bartlett
 * Initial version of driver for Rock 'n' Scroll
 *
 */

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/itsy_session.h>
#include <linux/itsy_version.h>
#include <linux/itsy_rocknscroll.h>
#include <linux/config.h>

#include <asm/io.h>
#include <asm/irq.h>

#include <asm/irq-no.h>
#include <asm/hardware.h>
#include <asm/io_reg.h>

#ifdef CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#include <linux/powermgr.h>

powermgr_client  pm_client;
powermgr_id  pm_id; 

/* Device information */

#define ROCKNSCROLL_MAJOR 45
#define ROCKNSCROLL_NAME "rocknscroll"

static int gpio_xaxis;
static int gpio_yaxis;

static int  uniqueid;

static int  opencnt = 0;

#ifdef CONFIG_PROC_FS
static int rocknscroll_procfs_info( char *, char **, off_t, int, int );

static struct proc_dir_entry  rocknscroll_proc_entry = {
   0, 11, "rocknscroll", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, 
   rocknscroll_procfs_info,
};
#endif /* CONFIG_PROC_FS */

/* Sample buffer definitions */

static struct ADXL_SAMPLE  sample_buffers[ 2 ];

static volatile int  interrupt_sample = 1;
static int  driver_sample = 0;

static struct wait_queue  *waitq = NULL;

static int  sm_id = ITSY_SESSION_NOTIFY_ID_INVALID, /* Session management */
            sm_active = ITSY_SESSION_ID_INVALID;

static void rocknscroll_sessions_register( void );

static void rocknscroll_sessions_unregister( void );

static void rocknscroll_powermgr_register( void );

static void rocknscroll_powermgr_unregister( void );

/* Supported driver operations */

static int rocknscroll_read( struct inode *, struct file *, char *, int );

static int rocknscroll_select( struct inode *, struct file *, int, 
			       select_table * );

static int rocknscroll_ioctl( struct inode *, struct file *, unsigned int,
			      unsigned long );

static int rocknscroll_open( struct inode *, struct file * );

static void rocknscroll_release( struct inode *, struct file * );

struct file_operations rocknscroll_fops = {
   NULL,
   rocknscroll_read,
   NULL,
   NULL,
   rocknscroll_select,
   rocknscroll_ioctl,
   NULL,
   rocknscroll_open,
   rocknscroll_release
};

/*
 * On each interrupt, the current value of the OS Timer Count Register (driven
 * off the 3.6864-MHz clock) and the GPIO pins is saved.  Buffer overflows
 * are noted.
 *
 */

static void rocknscroll_interrupt( int irq, void *dev_id, 
				   struct pt_regs *regs ) {
   int  time,
	gpios,
	free;
   struct ADXL_SAMPLE  *sample = &sample_buffers[ interrupt_sample ];

   /* Clear edge detect interrupts, read time and gpio values. */
   outl( (1<<gpio_xaxis | 1<<gpio_yaxis), GEDR_V );
   time = inl( OSCR_V );
   gpios = inl( GPLR_V );

   if ((free = sample->free) >= ADXL_SAMPLE_COUNT) {
      /* Buffer is full, so empty it and set overflow */
      sample->overflow = 1;
      free = sample->free = 0;
   }
   sample->samples[ free ].time = time;
   sample->samples[ free ].gpios = gpios;
   sample->free = sample->free+1;

   /* Wake up any waiters */
   wake_up_interruptible( &waitq );
}

/*
 * Point interrupt handler to the next buffer, and point the driver to the
 * previous interrupt buffer.
 *
 */

static void next_sample_buffer( int overflow ) {
   sample_buffers[ driver_sample ].free = 0;
   sample_buffers[ driver_sample ].overflow = overflow;
   interrupt_sample = 1-interrupt_sample;
   driver_sample = 1-driver_sample;
}

/*
 * Open the device.
 *
 */

static int rocknscroll_open( struct inode *inode, struct file *file ) {
   if (opencnt == 0) {
      /* First open:  initialize interrupt data structures */
      next_sample_buffer( 1 );

      /* Clear pending edge detect interrupts on each signal */
      outl( (1<<gpio_xaxis | 1<<gpio_yaxis), GEDR_V );

      /* Request rising and falling edge interrupts on each signal */
      outl( (inl( GRER_V ) | 1<<gpio_xaxis | 1<<gpio_yaxis), GRER_V );
      outl( (inl( GFER_V ) | 1<<gpio_xaxis | 1<<gpio_yaxis), GFER_V );

      /* Make sure required interrupt(s) are enabled */
      enable_irq( (gpio_xaxis < 11) ? gpio_xaxis : IRQ_GPIO11_27 );
      enable_irq( (gpio_yaxis < 11) ? gpio_yaxis : IRQ_GPIO11_27 );
   }
   opencnt = opencnt+1;
   MOD_INC_USE_COUNT;
   return  0;
}

/*
 * Close the device.
 *
 */

static void rocknscroll_release( struct inode *inode, struct file *file ) {
   opencnt = opencnt-1;
   if (opencnt == 0) {
      /* Last close:  clear rising & falling edge interrupts on each signal */
      outl( (inl( GRER_V ) & ~(1<<gpio_xaxis | 1<<gpio_yaxis)), GRER_V );
      outl( (inl( GFER_V ) & ~(1<<gpio_xaxis | 1<<gpio_yaxis)), GFER_V );
   }
   MOD_DEC_USE_COUNT;
}

/*
 * Retrieve the contents of the current sample buffer when this task is in
 * the foreground session.  Otherwise return 0 bytes.
 *
 */

static int rocknscroll_read( struct inode *inode, struct file *file, 
			     char *userbuf, int usercount ) {
   int  available;

   if (sm_active != ITSY_SESSION_ID_INVALID  &&
       sm_active != itsy_session_from_task( current )) {
      /* Read by a background task returns 0 bytes */
      return 0;
   }
   next_sample_buffer( 0 );
   available = sizeof( int )*2+
	       sizeof( int )*2*sample_buffers[ driver_sample ].free;
   if  (available < usercount) usercount = available;
   memcpy_tofs( userbuf, &sample_buffers[ driver_sample ], usercount );
   return usercount;
}

/*
 * Check to see if there is any buffered samples for the foreground session.
 *
 */

static int rocknscroll_select( struct inode *inode, struct file *file,
                               int mode, select_table *table) {
   if (mode == SEL_IN) {
      /* Select on input is the only mode supported */
      if ((sm_active != ITSY_SESSION_ID_INVALID  &&
           sm_active != itsy_session_from_task( current )) ||
          sample_buffers[ interrupt_sample ].free == 0) {
         /* No data available if background task or buffer empty */
         select_wait( &waitq, table );
         return 0;
      } else {
         /* Buffered samples for foreground task */
         return 1;
      }
   }
   /* SEL_OUT and SEL_EX are not supported */
   return 0;
}

/*
 * Device specific control information.
 *
 */

static int rocknscroll_ioctl( struct inode *inode, struct file *file,
			      unsigned int cmd, unsigned long arg ) {
   int  err = 0, size = _IOC_SIZE( cmd );

   /* Verify that the command is for this device. */
   if (_IOC_TYPE( cmd ) != ROCKNSCROLL_IOCTL_TYPE) return -EINVAL;
   if (_IOC_NR( cmd ) > ROCKNSCROLL_IOCTL_MAXNR) return -EINVAL;

   /* Check direction and size of data transfers */
   if (_IOC_DIR( cmd ) & _IOC_READ)
      err = verify_area( VERIFY_WRITE, (void *)arg, size );
   else if (_IOC_DIR( cmd ) & _IOC_WRITE)
      err = verify_area( VERIFY_READ, (void *)arg, size );
   if (err) return err;

   switch (cmd) {
      case ROCKNSCROLL_GET_ADXL_INFO:
	 memcpy_tofs( (void *)arg, &adxl_info[ itsy_major_rev() ], size );
	 break;

      default:
	 return -EINVAL;
   }
   return 0;
}

/*
 * Module initialization and cleanup.
 *
 */ 

int init_module( void ) {
   int  err;

   /* Register the device driver */
   if ((err = register_chrdev( ROCKNSCROLL_MAJOR, ROCKNSCROLL_NAME, 
			       &rocknscroll_fops )) != 0) {
      printk( "RnS: unable to get major device = %i, err = %i\n",
	      ROCKNSCROLL_MAJOR, err );
      return err;
   }

   /* Initialize interrupt data structures */
   next_sample_buffer( 1 );

   /* Use hardware version to figure out GPIOs */
   gpio_xaxis = adxl_info[ itsy_major_rev() ].gpio_xaxis;
   gpio_yaxis = adxl_info[ itsy_major_rev() ].gpio_yaxis;

   /* Get the IRQ(s) associated with edge detection */
   if (itsy_major_rev() == 2) {
      /* Get the IRQs associated with edge detection for each axis */
      if (request_irq( gpio_xaxis, rocknscroll_interrupt, 
		       SA_INTERRUPT, ROCKNSCROLL_NAME, 
		       &uniqueid) != 0) {
         printk( "RnS: unable to allocate IRQ_GPIO%i\n", gpio_xaxis );
         unregister_chrdev( ROCKNSCROLL_MAJOR, ROCKNSCROLL_NAME );
         return -1;
      }
      if (request_irq( gpio_yaxis, rocknscroll_interrupt, 
		       SA_INTERRUPT, ROCKNSCROLL_NAME, 
		       &uniqueid) != 0) {
         printk( "RnS: unable to allocate IRQ_GPIO%i\n", gpio_yaxis );
         free_irq( gpio_xaxis, &uniqueid );
         unregister_chrdev( ROCKNSCROLL_MAJOR, ROCKNSCROLL_NAME );
         return -1;
      }
   } else {
      /* Get the IRQ associated with edge detection for GPIOs 11-27 */
      if (request_irq( IRQ_GPIO11_27, rocknscroll_interrupt, 
		       SA_INTERRUPT | SA_SHIRQ, ROCKNSCROLL_NAME, 
		       &uniqueid) != 0) {
         printk( "RnS: unable to allocate IRQ_GPIO11_27\n" );
         unregister_chrdev( ROCKNSCROLL_MAJOR, ROCKNSCROLL_NAME );
         return -1;
      }
   }

   /* Initialize power and session management */
   rocknscroll_sessions_register();
   rocknscroll_powermgr_register();

#ifdef  CONFIG_PROC_FS
   /* Register procfs device */
   proc_register_dynamic( &proc_root, &rocknscroll_proc_entry);
#endif  /* CONFIG_PROC_FS */

   return 0;
}

void cleanup_module( void ) {
   int  err;

   if ((err = unregister_chrdev( ROCKNSCROLL_MAJOR, ROCKNSCROLL_NAME )) != 0) {
      printk( "RnS: unregister_chrdev returns: %i\n", err );
   }
   if (itsy_major_rev() == 2) {
      free_irq( gpio_xaxis, &uniqueid );
      free_irq( gpio_yaxis, &uniqueid );
   } else {
      free_irq( IRQ_GPIO11_27, &uniqueid );
   }
   rocknscroll_sessions_unregister();
   rocknscroll_powermgr_unregister();

#ifdef  CONFIG_PROC_FS
   /* Unregister procfs device */
   proc_unregister( &proc_root, rocknscroll_proc_entry.low_ino );
#endif  /* CONFIG_PROC_FS */
}

/*
 * Session management functions
 *
 */

static void rocknscroll_session_destroy( void *private, int sid ) {
   if (sid == sm_active) {
      next_sample_buffer( 1 );
      sm_active = itsy_session_get_foreground();
   }
}

static void rocknscroll_session_switchto( void *private, int sid ) {
   next_sample_buffer( 1 );
   sm_active = sid;
}

static void rocknscroll_sessions_register() {
   sm_id = itsy_session_notify_register( NULL, ROCKNSCROLL_NAME, 
					 NULL,
					 rocknscroll_session_destroy,
					 rocknscroll_session_switchto, 
					 NULL );
   sm_active = itsy_session_get_foreground();
}

static void rocknscroll_sessions_unregister() {
   if (itsy_session_notify_unregister( sm_id ) != 0) {
      printk( "RnS: unable to unregister with session manager\n" );
   }
}

/* 
 * Power management functions
 *
 */

static int rocknscroll_suspend_check( void *unused, int idle_jiffies ) {
   return 0;
}

static int rocknscroll_suspend( void *ignore ) {
   return 1;
}

static int rocknscroll_resume( void *ignore, int resume_flags ) {
   /* Initialize sample buffer */
   next_sample_buffer( 1 );
   if (opencnt != 0) {
      /* Clear pending edge detect interrupts on each signal */
      outl( (1<<gpio_xaxis | 1<<gpio_yaxis), GEDR_V );

      /* Request rising and falling edge interrupts on each signal */
      outl( (inl( GRER_V ) | 1<<gpio_xaxis | 1<<gpio_yaxis), GRER_V );
      outl( (inl( GFER_V ) | 1<<gpio_xaxis | 1<<gpio_yaxis), GFER_V );

      /* Make sure required interrupt(s) are enabled */
      enable_irq( (gpio_xaxis < 11) ? gpio_xaxis : IRQ_GPIO11_27 );
      enable_irq( (gpio_yaxis < 11) ? gpio_yaxis : IRQ_GPIO11_27 );
   }
   return 1; 
}

static void rocknscroll_powermgr_register( void ) {
   pm_client.suspend_check = rocknscroll_suspend_check;
   pm_client.suspend = rocknscroll_suspend;
   pm_client.resume = rocknscroll_resume;
   pm_client.cookie = NULL;
   pm_client.name = ROCKNSCROLL_NAME;
   pm_client.uamps_idle = 0;
   pm_client.uamps_sleep = 0;
   pm_id = powermgr_register( &pm_client );
   if (pm_id == POWERMGR_ID_INVALID) {
      printk( "Failed to register %s module for power management!!!\n", 
	      pm_client.name );
  }
}

static void rocknscroll_powermgr_unregister( void ) {
   if (powermgr_unregister( pm_id ) == POWERMGR_FAILURE) {
      printk( "Failed to unregister %s module for power management!!!\n",
	      pm_client.name );
   }
}

/*
 * Process information
 *
 */

#ifdef  CONFIG_PROC_FS
static int  rocknscroll_procfs_info( char *buf, char **start, off_t offset,
                                     int length, int unused ) {
   int len = 0;

   /* Format device information */
   len += sprintf( buf+len, 
		   "X-axis: %2i   Y-axis: %2i    itsy_major_rev: %i\n", 
		   gpio_xaxis, gpio_yaxis, itsy_major_rev() );
   return(len);
}
#endif  /* CONFIG_PROC_FS */
