/*
 * SA1100 Real Time Clock
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Deborah A. Wallach
 * 03-30-98
 *
 * Save and restore time using RTC during sleep mode
 *	
 */

#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/io_reg.h>
#include <linux/mm.h>
#include <linux/powermgr.h>

#include <asm/irq.h>

/* power management functions */
static long			clock_cmos_diff;
static int			got_clock_diff = 0;

#define RTSR_STATUS_MASK (3)
#define RTSR_ENABLE_MASK (0xc)
#define RTSR_CURRENT_ENABLED ((inl(RTSR_V) & RTSR_ENABLE_MASK))
#define RTSR_ALARM_ENABLE (4)
#define RTSR_1HZ_ENABLE   (0x8)
#define RTSR_ALARM_STATUS (1)
#define RTSR_1HZ_STATUS   (2)

static int rtc_suspend_check(void *unused, int idle_jiffies)
{
  return(0);
}

unsigned long get_cmos_time(void)
{
  return(inl(RCNR_V));
}

void set_cmos_time(unsigned int seconds)
{
  outl(seconds, RCNR_V);
}

static int rtc_suspend(void *cookie)
{
  clock_cmos_diff = -get_cmos_time();
  clock_cmos_diff += CURRENT_TIME;
  got_clock_diff = 1;

  return(1); /* YYY - values ignored for now */
}

static int rtc_resume(void *cookie, int resume_flags)
{
  if (got_clock_diff)	{ /* If don't know time zone, can't set clock */
    CURRENT_TIME = get_cmos_time() + clock_cmos_diff;
  }

  outl(RTSR_CURRENT_ENABLED & RTSR_1HZ_ENABLE, RTSR_V); /* clear RTC alarm enable */
  return(1); /* YYY - values ignored for now */
}

static void rtc_powermgr_register(void)
{
  powermgr_client client;
  powermgr_id rtc_powermgr_id; 

  client.suspend_check = rtc_suspend_check;
  client.suspend = rtc_suspend;
  client.resume = rtc_resume;
  client.cookie = 0;
  client.name = "RTC";
  client.uamps_idle = 0;
  client.uamps_sleep = 0;
  rtc_powermgr_id = powermgr_register(&client);
  if (rtc_powermgr_id == POWERMGR_ID_INVALID) {
    printk("Failed to register %s module for power management!!!\n", client.name);
    return;
  }
}

void rtc_init(void)
{
  /* YYY RTC clock should really be set up better than this */
  outl(0x7fff, RTTR_V); /* YYY fake trim to pretend crystal is at optimal speed */
                        /* 0x7fff = 32768 Hz - 1 */

  /* clear rtc status register; turn on enables */
  outl(RTSR_ALARM_ENABLE | RTSR_1HZ_ENABLE, RTSR_V); 
  outl(inl(RCNR_V)+2, RTAR_V); /* set up alarm so will definitely get alarm bit set */

  rtc_powermgr_register();
}

void rtc_set_alarm(int seconds)
{
  int new_rtsr_value = RTSR_CURRENT_ENABLED | RTSR_ALARM_ENABLE; /* set RTC alarm enable on */
  outl(inl(RCNR_V)+seconds, RTAR_V);
  if (seconds != 0) {
    /* clear RTC alarm status bit */
    new_rtsr_value |= RTSR_ALARM_STATUS;
  }
  outl(new_rtsr_value, RTSR_V);
  outl((inl(PWER_V)|(1<<IRQ_RTCAlrm)), PWER_V);
}

void rtc_clear_alarm(void)
{
  /* clear RTC alarm status bit and enable */
  outl(((RTSR_CURRENT_ENABLED & RTSR_1HZ_ENABLE) | RTSR_ALARM_STATUS), RTSR_V);
  outl((inl(PWER_V)&~(1<<IRQ_RTCAlrm)), PWER_V);
}

void rtc_disable_alarm(void)
{
  /* disable alarm but don't clear RTC alarm status bit */
  outl((RTSR_CURRENT_ENABLED & RTSR_1HZ_ENABLE), RTSR_V);
  outl((inl(PWER_V)&~(1<<IRQ_RTCAlrm)), PWER_V);
}

int rtc_alarm_happened(void)
{
  return(
	 (inl(RTSR_V)&1) &&                /* the alarm went off   and */
	 (inl(PWER_V)&(1<<IRQ_RTCAlrm))    /* we were watching for it */
	 );
}

void rtc_set_1hz_detect(void)
{
  /* clear 1hz detect status bit; set 1hz detect enable on */
  outl((RTSR_CURRENT_ENABLED | RTSR_1HZ_ENABLE | RTSR_1HZ_STATUS), RTSR_V);
  outl((inl(PWER_V)|(1<<IRQ_RTC1Hz)), PWER_V);
}

void rtc_enable_1hz_detect(void)
{
  outl((inl(PWER_V)|(1<<IRQ_RTC1Hz)), PWER_V);
}

void rtc_clear_1hz_detect(void)
{
  /* clear 1hz detect status bit and enable */
  outl(((RTSR_CURRENT_ENABLED & RTSR_ALARM_ENABLE) | RTSR_1HZ_STATUS), RTSR_V);
  outl((inl(PWER_V)&~(1<<IRQ_RTC1Hz)), PWER_V);
}

void rtc_disable_1hz_detect(void)
{
  outl((inl(PWER_V)&~(1<<IRQ_RTC1Hz)), PWER_V);
}

int rtc_1hz_detected(void)
{
  return(
	 (inl(RTSR_V)&2) &&                /* the alarm went off   and */
	 (inl(PWER_V)&(1<<IRQ_RTC1Hz))    /* we were watching for it */
	 );
}

