/*
 * linux/arch/arm/drivers/char/serialecho.c
 *
 * Serial echoing for kernel console messages.
 */
#include <linux/config.h>
#if defined (CONFIG_ARCH_A5K) || defined (CONFIG_ARCH_RPC) || defined(CONFIG_ARCH_EBSA)

#include <asm/io.h>

#include <linux/serial_reg.h>

extern int serial_echo_init (int base);
extern int serial_echo_print (const char *s);

/*
 * this defines the address for the port to which printk echoing is done
 *  when CONFIG_SERIAL_ECHO is defined
 */
#ifndef SERIAL_ECHO_PORT
#define SERIAL_ECHO_PORT	0x3f8	/* internal serial port */
#endif

#ifndef SERIAL_ECHO_DIVISOR
#define SERIAL_ECHO_DIVISOR	12	/* 9600 baud */
#endif

static int serial_echo_port = 0;

#define serial_echo_outb(v,a) outb((v),(a)+serial_echo_port)
#define serial_echo_inb(a)    inb((a)+serial_echo_port)

#define BOTH_EMPTY (UART_LSR_TEMT | UART_LSR_THRE)

/* wait for the transmitter & holding register to empty */
#define WAIT_FOR_XMITR \
 do { \
	lsr = serial_echo_inb (UART_LSR); \
 } while ((lsr & BOTH_EMPTY) != BOTH_EMPTY)

/*
 * These two functions abstract the actual communications with the
 * debug port.  This is so we can change the underlying communications
 * mechanism without modifying the rest of the code.
 */
int serial_echo_print (const char *s)
{
	int lsr, ier;
	int i;

	if (!serial_echo_port)
		return 0;

	/*
	 * First save the IER then disable interrupts
	 */
	ier = serial_echo_inb (UART_IER);
	serial_echo_outb (0x00, UART_IER);

	/*
	 * Now do each character
	 */
	for (i = 0; *s; i++, s++) {
		WAIT_FOR_XMITR;

		/* send the character out. */
		serial_echo_outb (*s, UART_TX);

		/* if a LF, also do CR... */
		if (*s == 10) {
			WAIT_FOR_XMITR;
			serial_echo_outb (13, UART_TX);
		}
	}

	/*
	 * Finally, wait for transmitter & holding register to empty
	 *  and restore the IER.
	 */
	WAIT_FOR_XMITR;
	serial_echo_outb (ier, UART_IER);

	return 0;
}

int serial_echo_init (int base)
{
	int comstat, hi, lo;

	if (base != 0x3f8 && base != 0x2f8) {
		serial_echo_port = 0;
		return 0;
	} else
		serial_echo_port = base;

	/*
	 * Read the Divisor Latch
	 */
	comstat = serial_echo_inb (UART_LCR);
	serial_echo_outb (comstat | UART_LCR_DLAB, UART_LCR);
	hi = serial_echo_inb (UART_DLM);
	lo = serial_echo_inb (UART_DLL);
	serial_echo_outb (comstat, UART_LCR);

	/*
	 * now do a hardwired init
	 */
	serial_echo_outb (0x03, UART_LCR); /* No parity, 8 bits, 1 stop */
	serial_echo_outb (0x83, UART_LCR); /* Access divisor latch */
	serial_echo_outb (SERIAL_ECHO_DIVISOR >> 8, UART_DLM);
	serial_echo_outb (SERIAL_ECHO_DIVISOR & 0xff, UART_DLL);
	serial_echo_outb (0x03, UART_LCR); /* Done with divisor */

	/* Prior to disabling interrupts, read the LSR and RBR
	 * registers
	 */
	comstat = serial_echo_inb (UART_LSR); /* COM? LSR */
	comstat = serial_echo_inb (UART_RX);  /* COM? RBR */
	serial_echo_outb (0x00, UART_IER);    /* Disable all interrupts */

	return 0;
}

#elif defined(CONFIG_ARCH_BRUTUS)

/*
 *  Code below is:
 *
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Deborah A. Wallach
 */

#include <asm/io.h>

#include <linux/serial_reg.h>
#include <asm/io_reg.h>
#include <asm/arch/serial.h>

#include "serial-include.c"

extern int serial_echo_init (unsigned int base);
extern int serial_echo_print (const char *s);

/*
 * this defines the address for the port to which printk echoing is done
 *  when CONFIG_SERIAL_ECHO is defined
 */
#define SERIAL_ECHO_PORT        0

/* for debugging: */
#ifndef SERIAL_ECHO_PORT
#define SERIAL_ECHO_PORT	UART_SERIAL1_BASE_V	/* internal serial port */  
#endif

#ifndef SERIAL_ECHO_DIVISOR
#define SERIAL_ECHO_DIVISOR	0x17	/* 9600 baud */
#endif

static int serial_echo_port = 0;

#define serial_echo_outl(v,a) outl((v),(a)+serial_echo_port)
#define serial_echo_inl(a)    inl((a)+serial_echo_port)

/* wait for the transmitter & holding register to empty */
#define WAIT_FOR_XMITR \
 do { \
	lsr = serial_echo_inl (UTSR1); \
 } while ((lsr & UTSR1_TNF) != UTSR1_TNF)

/*
 * These two functions abstract the actual communications with the
 * debug port.  This is so we can change the underlying communications
 * mechanism without modifying the rest of the code.
 */
int serial_echo_print (const char *s)
{
	int lsr, ier;
	int i;

	if (!serial_echo_port)
		return 0;

	/*
	 * First save the IER then disable interrupts
	 */
	ier = serial_echo_inl (UTCR3);
	serial_echo_outl (UTCR3_TXE, UTCR3);

	/*
	 * Now do each character
	 */
	for (i = 0; *s; i++, s++) {
		WAIT_FOR_XMITR;

		/* send the character out. */
		serial_echo_outl (*s, UART_TX);

		/* if a LF, also do CR... */
		if (*s == 10) {
			WAIT_FOR_XMITR;
			serial_echo_outl (13, UART_TX);
		}
	}

	/*
	 * Finally, wait for transmitter & holding register to empty
	 *  and restore the IER.
	 */
	WAIT_FOR_XMITR;
	serial_echo_outl (ier, UTCR3);

	return 0;
}

int serial_echo_init (unsigned int base)
{
	if (base != (unsigned int)UART_SERIAL1_BASE_V && 
	    base != (unsigned int)UART_SERIAL3_BASE_V) {
		serial_echo_port = 0;
		return 0;
	} else {
		serial_echo_port = base;

#ifndef CONFIG_ITSY
		if(serial_echo_port == (unsigned int)UART_SERIAL1_BASE_V) {
		  itsy_setup_uart1_alt();
		}
#else /* CONFIG_ITSY */
		if(serial_echo_port == (unsigned int)UART_SERIAL1_BASE_V) {
		  if (is_itsy_v1()) {
		    itsy_setup_uart1();
		  }
		  else if (is_itsy_v2()) {
		    itsy_turn_mb_uart_hardware_on();
		    itsy_setup_uart1_alt();
		  }
		}
		if(serial_echo_port == (unsigned int)UART_SERIAL3_BASE_V) {
		  if (is_itsy_v1()) {
		    itsy_turn_mb_uart_hardware_on();
		  }
		}
#endif /* CONFIG_ITSY */
        }

	/*
	 * now do a hardwired init
	 */
	serial_echo_outl (0x08, UTCR0); /* No parity, 8 bits, 1 stop */
	serial_echo_outl (SERIAL_ECHO_DIVISOR >> 8, UTCR1);
	serial_echo_outl (SERIAL_ECHO_DIVISOR & 0xff, UTCR2);

	serial_echo_outl (0, UTCR3);            /* Disable all interrupts */
	serial_echo_outl (UTCR3_TXE, UTCR3);    /* Turn on transmitter */
	/* clear out any errors by writing the line status register */
	serial_echo_outl(0xff, UTSR0) ;

	return 0;
}

#endif



