/*
 * linux/arch/arm/drivers/scsi/acornscsi.c
 *
 *  Acorn SCSI 3 driver
 *  By R.M.King.
 *
 * Abandoned using the Select and Transfer command since there were
 * some nasty races between our software and the target devices that
 * were not easy to solve, and the device errata said that there were
 * some nasty bugs in that command...
 */
#define DEBUG_NO_WRITE	1
#define DEBUG_QUEUES	2
#define DEBUG_DMA	4
#define DEBUG_ABORT	8
#define DEBUG_DISCON	16
#define DEBUG_CONNECT	32
#define DEBUG_PHASES	64
#define DEBUG_WRITE	128
#define DEBUG_LINK	256
#define DEBUG_MESSAGES	512
#define DEBUG_RESET	1024

/* DRIVER CONFIGURATION
 *
 * SCSI-II Tagged queue support.
 *
 * I don't have any SCSI devices that support it, so it is totally untested
 * (except to make sure that it doesn't interfere with any non-tagging
 * devices).  It is not fully implemented either - what happens when a
 * tagging device reconnects???
 *
 * You can tell if you have a device that supports tagged queueing my
 * cating /proc/scsi/acornscsi/0 and see if the SCSI revision is reported
 * as '2 TAG'.
 */
#undef SCSI2_TAG

/*
 * SCSI-II Linked command support.
 *
 * The higher level code doesn't support linked commands yet.
 */
#undef SCSI2_LINK

/*
 * SCSI-II Synchronous transfer support.
 *
 * Tried and tested...
 *
 * SDTR_SIZE	  - maximum number of un-acknowledged bytes (0 = off, 12 = max)
 * SDTR_PERIOD	  - period of REQ signal (min=228, max=1020)
 * DEFAULT_PERIOD - default REQ period.
 */
#define SCSI2_SYNC
#define SDTR_SIZE	12
#define SDTR_PERIOD	250
#define DEFAULT_PERIOD	250

/*
 * Debugging information
 *
 * Please see debug list above
 * DEBUG_TARGET	  - is defined to the target number if you want to debug
 *		    a specific target. [only recon/write/dma].
 */
#define DEBUG (DEBUG_ABORT)
/*#define DEBUG_TARGET 2*/

/*
 * Select timeout time (in 10ms units)
 *
 * This is the timeout used between the start of selection and the WD33C93
 * chip deciding that the device isn't responding.
 */
#define TIMEOUT_TIME 10

/*
 * Define this if you want to have verbose explaination of SCSI
 * status/messages.
 */
#undef CONFIG_ACORNSCSI_CONSTANTS

/*
 * Define this if you want to use the on board DMAC
 */
#define USE_DMAC

/*
 * List of devices that the driver will recognise
 */
#define ACORNSCSI_LIST { MANU_ACORN, PROD_ACORN_SCSI }

/*
 *====================================================================================
 */

#ifdef DEBUG_TARGET
#define DBG(cmd,xxx...) \
  if (cmd->target == DEBUG_TARGET) { \
    xxx; \
  }
#else
#define DBG(cmd,xxx...) xxx
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/ioport.h>
#include <asm/bitops.h>
#include <asm/delay.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/ecard.h>

#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "acornscsi.h"
#include "constants.h"

#define VER_MAJOR 1
#define VER_MINOR 9
#define VER_PATCH 6

#ifdef SCSI2_TAG
#define AS_MSG_TAG 1
#else
#define AS_MSG_TAG 0
#endif

#ifdef SCSI2_SYNC
#define AS_MSG_SYNC 1
#else
#define AS_MSG_SYNC 0
#endif

/*
 * DMAC setup parameters
 */ 
#define INIT_DEVCON0	(DEVCON0_RQL | DEVCON0_EXW)
#define INIT_DEVCON1	(DEVCON1_BHLD)
#define INIT_SBICDMA	(CTRL_DMABURST)
#define DMAC_READ	(MODECON_READ)
#define DMAC_WRITE	(MODECON_WRITE)

/*
 * 1 for MESSAGE_REJECTs, 1 for IDENTIFYs, and one spare just in case of an ABORT.
 */
#define AS_MAX_MSGS (AS_MSG_TAG + AS_MSG_SYNC + 3)

#ifdef SCSI2_LINK
#error SCSI2 LINKed commands not supported (yet)!
#endif

/*
 * This is used to dump the previous states of the SBIC
 */
static unsigned char status_array[9][64];
static unsigned int status_ptr[9];

static AS_Msg outgoing_messages[AS_MAX_MSGS];
static struct Scsi_Host *first_instance;
static Scsi_Host_Template *the_template;
static struct proc_dir_entry proc_scsi_acornscsi = {
	PROC_SCSI_EATA, 9, "acornscsi", S_IFDIR | S_IRUGO | S_IXUGO, 2
};

static void del_all_msgs (AS_Host *hostdata);
static void acornscsi_done (struct Scsi_Host *instance, struct as_cmnd *ascmd);

static inline
void sbic_write (const struct Scsi_Host *instance, int reg, int value)
{
    unsigned long port = instance->io_port + 0x800;
    outb (reg, port);
    outw (value, port + 1);
}

static inline
void sbic_writenext (const struct Scsi_Host *instance, int value)
{
    outw(value, instance->io_port + 0x801);
}

static inline
int sbic_read (const struct Scsi_Host *instance, int reg)
{
    unsigned long port = instance->io_port + 0x800;
    if(reg == ASR)
    	   return inw(port) & 0xff;
    outb(reg, port);
    return inw(port + 1) & 0xff;
}

static inline
int sbic_readnext (const struct Scsi_Host *instance)
{
    return inw(instance->io_port + 0x801) & 0xff;
}

static inline
void page_write (const struct Scsi_Host *instance, int value)
{
    outb(value, POD_SPACE(instance->io_port) + 0xc00);
}

static inline
int dmac_read (const struct Scsi_Host *instance, int reg)
{
    return inb (instance->io_port + 0xc00 + reg);
}

static inline
void dmac_write (const struct Scsi_Host *instance, int reg, int value)
{
    outb (value, instance->io_port + 0xc00 + reg);
}

static inline
void dmac_clearintr (const struct Scsi_Host *instance)
{
    outb (0, POD_SPACE(instance->io_port) + 0x800);
}

static inline
int read_irqstat (const struct Scsi_Host *instance)
{
    return inb (POD_SPACE(instance->io_port) + 0x800);
}

static inline
void sbic_reset (const struct Scsi_Host *instance)
{
    sbic_write (instance, OWNID, OWNID_EAF | instance->this_id);
    sbic_write (instance, CMND, CMND_RESET);
    sbic_write (instance, SOURCEID, SOURCEID_ER | SOURCEID_DSP);
}

static
void acornscsi_resetcard (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *) instance->hostdata;
    unsigned long target_jiffies, flags;
    
    hostdata->page_reg = 0x80;
    page_write (instance, 0x80); /* pulls reset line active */
    target_jiffies = jiffies + 3;
    del_all_msgs (hostdata);

    save_flags (flags);
    sti ();

    while (jiffies < target_jiffies) barrier(); /* wait 3 cs.  SCSI standard says 25ms. */

    restore_flags (flags);

    hostdata->page_reg = 0x00;
    page_write (instance, 0x00);

    udelay (5);

    /* wait for hard reset interrupt */
    while (!(sbic_read (instance, ASR) & ASR_INT));
    /* clear pending interrupt */
    sbic_read (instance, SSR);

    hostdata->page_reg = 0x40;
    page_write (instance, 0x40);

    /* setup sbic - WD33C93A */
    sbic_reset (instance);

    /* setup dmac - uPC71071 */
    dmac_write (instance, INIT, 0);
    dmac_write (instance, INIT, INIT_8BIT);
    dmac_write (instance, CHANNEL, CHANNEL_0);
    dmac_write (instance, DEVCON0, INIT_DEVCON0);
    dmac_write (instance, DEVCON1, INIT_DEVCON1);
}

/*=============================================================================================
 * utility routines
 */
#ifdef CONFIG_ACORNSCSI_CONSTANTS
static char *acornscsi_interrupttype[] = {
  "reset",	  "successful",	 "paused/aborted", "*undefined*",
  "terminated",	  "*undefined*", "*undefined*",	   "*undefined*",
  "service-reqd", "*undefined*", "*undefined*",	   "*undefined*",
  "*undefined*",  "*undefined*", "*undefined*",	   "*undefined*"
};

static signed char acornscsi_map[] = {
  0,  1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1,  2, -1, -1,  -1, -1,  3, -1,   4,  5,  6,  7,   8,  9, 10, 11,
 12, 13, 14, -1,  -1, -1, -1, -1,  15, 16, 17, 18,  19, 20, 21, 22,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 23, 24, 25, 26,  27, -1, -1, 28,  29, 30, 31, 32,  33, 34, 35, 36,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 37, 38, -1, -1,  -1, 39, -1, -1,  40, 41, 42, 43,  44, 45, 46, 47
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,
 -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1,  -1, -1, -1, -1
};	

static char *acornscsi_interruptcode[] = {
	/* 0 */
	"reset - normal mode - disconnected",		/* 00 */
	"reset - advanced mode - disconnected",		/* 01 */

	/* 2 */
	"select command completed",			/* 11 */
	"select-and-transfer completed",		/* 16 */
	"transfer-info: data-out",			/* 18 */
	"transfer-info: data-in",			/* 19 */
	"transfer-info: command",			/* 1A */
	"transfer-info: status",			/* 1B */
	"transfer-info: unspec-out",			/* 1C */
	"transfer-info: unspec-in",			/* 1D */
	"transfer-info: message-out",			/* 1E */
	"transfer-info: message-in",			/* 1F */

	/* 12 */
	"transfer-info: message-in, /ACK asserted",	/* 20 */
	"save-data-pointer",				/* 21 */
	"select/reselect command aborted",		/* 22 */
	"transfer-cmd: data out",			/* 28 */
	"transfer-cmd: data-in",			/* 29 */
	"transfer-cmd: command",			/* 2A */
	"transfer-cmd: status",				/* 2B */
	"transfer-cmd: unspec-out",			/* 2C */
	"transfer-cmd: unspec-in",			/* 2D */
	"transfer-cmd: message-out",			/* 2E */
	"transfer-cmd: message-in",			/* 2F */

	/* 23 */
	"invalid command",				/* 40 */
	"unexpected disconnect",			/* 41 */
	"select timeout",				/* 42 */
	"parity error, /ATN not asserted",		/* 43 */
	"parity error, /ATN asserted",			/* 44 */
	"incorrect status byte",			/* 47 */
	"unexpected phase: data out",			/* 48 */
	"unexpected phase: data-in",			/* 49 */
	"unexpected phase: command",			/* 4A */
	"unexpected phase: status",			/* 4B */
	"unexpected phase: unspec-out",			/* 4C */
	"unexpected phase: unspec-in",			/* 4D */
	"unexpected phase: message-out",		/* 4E */
	"unexpected phase: message-in",			/* 4F */

	/* 37 */
	"reselected - no id message",			/* 80 */
	"reselected in advanced mode",			/* 81 */
	"disconnect occurred",				/* 85 */
	"new phase: data out",				/* 88 */
	"new phase: data-in",				/* 89 */
	"new phase: command",				/* 8A */
	"new phase: status",				/* 8B */
	"new phase: unspec-out",			/* 8C */
	"new phase: unspec-in",				/* 8D */
	"new phase: message-out",			/* 8E */
	"new phase: message-in",			/* 8F */
};

static
void print_scsi_status (unsigned int ssr)
{
    if (acornscsi_map[ssr] != -1)
	printk ("%X:%X (%s %s)", ssr >> 4, ssr & 0x0f,
		acornscsi_interrupttype[(ssr >> 4)],
		acornscsi_interruptcode[acornscsi_map[ssr]]);
    else
	printk ("%X:%X (*unknown*)", ssr >> 4, ssr & 0x0f);    
}    
#endif

static
void print_sbic_status (int asr, int ssr, int cmdphase)
{
#ifdef CONFIG_ACORNSCSI_CONSTANTS
    printk ("sbic status: %s %s %s %s %s %s ",
        asr & ASR_INT ? "INT" : "int",
        asr & ASR_LCI ? "LCI" : "lci",
        asr & ASR_BSY ? "BSY" : "bsy",
        asr & ASR_CIP ? "CIP" : "cip",
    	asr & ASR_PE  ? "PE"  : "pe",
    	asr & ASR_DBR ? "DBR" : "dbr");
    printk ("scsi status: ");
    print_scsi_status (ssr);
    printk (" phase %02X\n", cmdphase);
#else
    printk ("sbic status: %02X scsi status: %X:%X phase %02X\n",
        asr, (ssr & 0xf0)>>4, ssr & 0x0f, cmdphase);
#endif
}

static
void acornscsi_printstatus (struct Scsi_Host *instance, int target)
{
    do {
	int statptr;

	printk ("%c: ", target == 8 ? 'H' : ('0' + target));
	statptr = status_ptr[target] - 22;

	if (statptr < 0)
	    statptr += 64;
	for (; statptr != status_ptr[target]; statptr = (statptr + 2) & 63) {
#ifdef CONFIG_ACORNSCSI_CONSTANTS
	    printk ("[DPh %02X, SST ", status_array[target][statptr + 1]);
	    print_scsi_status (status_array[target][statptr]);
	    printk ("]\n");
#else
	    printk ("[%02X:%02X] ", status_array[target][statptr+1],
			status_array[target][statptr]);
#endif
	}
	printk ("\n");
	if (target == 8)
	    break;
	target = 8;
    } while (1);
}

static
void dmac_dump (struct Scsi_Host *instance, struct as_cmnd *ascmd)
{
	unsigned int mode, addr, len;

	mode = dmac_read (instance, MODECON);
	addr = dmac_read (instance, TXADRHI) << 16 |
	       dmac_read (instance, TXADRMD) << 8 |
	       dmac_read (instance, TXADRLO);
	len  = dmac_read (instance, TXCNTHI) << 8 |
	       dmac_read (instance, TXCNTLO);

	printk ("scsi%d: DMAC %02x @%06x +%04x msk %02x, ",
		instance->host_no, mode, addr, len,
		dmac_read (instance, MASKREG));

	printk ("OL %lx CL %lx, ", ascmd->origlen, ascmd->dmaclen);
	printk ("BH @%p +%04x, ", ascmd->cmd->SCp.ptr,
		ascmd->cmd->SCp.this_residual);
	printk ("BUF%d @%04x +%04x\n", ascmd->buffer_no,
		ascmd->bufaddr[ascmd->buffer_no],
		ascmd->buflen[ascmd->buffer_no]);
}

static inline
int acornscsi_cmdtype (int command)
{
    switch (command) {
    	case WRITE_6:
    	case WRITE_10:
    	case WRITE_12:
    	    return CMD_WRITE;
    	case READ_6:
    	case READ_10:
    	case READ_12:
    	    return CMD_READ;
    	default:
    	    return CMD_MISC;
    }
}

static
int acornscsi_writedirection (int command)
{
    switch (command) {
	case CHANGE_DEFINITION:
	case COMPARE:
	case COPY:
	case COPY_VERIFY:
	case LOG_SELECT:
	case MODE_SELECT:
	case MODE_SELECT_10:
	case SEND_DIAGNOSTIC:
	case WRITE_BUFFER:
	case FORMAT_UNIT:
	case REASSIGN_BLOCKS:
	case RESERVE:
	case SEARCH_EQUAL:
	case SEARCH_HIGH:
	case SEARCH_LOW:
	case WRITE_6:
	case WRITE_10:
	case WRITE_VERIFY:
	case UPDATE_BLOCK:
	case WRITE_LONG:
	case WRITE_SAME:
	case SEARCH_HIGH_12:
	case SEARCH_EQUAL_12:
	case SEARCH_LOW_12:
	case WRITE_12:
	case WRITE_VERIFY_12:
	case SET_WINDOW:
	case MEDIUM_SCAN:
	case SEND_VOLUME_TAG:
	case 0xea:
	    return 1;
	case TEST_UNIT_READY:
	default:
	    return 0;
    }
}

static struct sync_xfer_tbl {
    unsigned int period_ns;
    unsigned char reg_value;
} sync_xfer_table[] = {
    {   1, 0x20 },
    { 252, 0x20 },
    { 376, 0x30 },
    { 500, 0x40 },
    { 624, 0x50 },
    { 752, 0x60 },
    { 876, 0x70 },
    {1000, 0x00 },
    {   0,    0 }
};

static inline
int round_period (unsigned int period)
{
    int i;

    for (i = 1; sync_xfer_table[i].period_ns; i++) {
	if ((period <= sync_xfer_table[i].period_ns) &&
	    (period > sync_xfer_table[i - 1].period_ns))
	    return i;
    }
    return 7;
}

static unsigned char calc_sync_xfer (unsigned int period, unsigned int offset)
{
    return sync_xfer_table[round_period(period)].reg_value |
		((offset < SDTR_SIZE) ? offset : SDTR_SIZE);
}

static inline
unsigned short dmac_length (unsigned long len)
{
    if (len < 4096)
	return len;
    else
	return 4096;
}

static inline
unsigned long sbic_readlength (struct Scsi_Host *instance)
{
    unsigned long length;

    length = sbic_read (instance, TRANSCNTH) << 16;
    length |= sbic_readnext (instance) << 8;
    length |= sbic_readnext (instance);

    return length;
}

static
void add_msg (AS_Host *hostdata, unsigned char length, ...)
{
    va_list ap;
    AS_Msg **m;
    int i, j;

    va_start (ap, length);

    for (i = 0; i < AS_MAX_MSGS; i++)
	if (!set_bit (0, &outgoing_messages[i].used)) {
	    outgoing_messages[i].next = NULL;
	    outgoing_messages[i].len = length;
	    for (j = 0; j < length; j++)
		outgoing_messages[i].vals[j] = va_arg (ap, unsigned char);
	    for (m = &hostdata->out_msg; *m; m = &(*m)->next);
		*m = &outgoing_messages[i];
	    va_end (ap);
	    return;
	}
    printk ("acornscsi: out of outgoing messages\n");
}

static
void del_all_msgs (AS_Host *hostdata)
{
    AS_Msg **m, *n;
    for (m = &hostdata->out_msg; (n = *m) != NULL; m = &n->next) {
    	(*m) = NULL;
    	n->used = 0;
    }
}

static inline
void dmac_setup (struct Scsi_Host *instance, unsigned int mode,
		unsigned long address, unsigned int length)
{
    dmac_write (instance, MASKREG, MASK_ON);
    if (length) {
	length -= 1;
	dmac_write (instance, TXCNTLO, length);
	dmac_write (instance, TXCNTHI, length >> 8);
	dmac_write (instance, TXADRLO, address);
	dmac_write (instance, TXADRMD, address >> 8);
	dmac_write (instance, TXADRHI, address >> 16);
	dmac_write (instance, MODECON, mode);
	dmac_write (instance, MASKREG, MASK_OFF);
    } else {
	dmac_write (instance, TXCNTLO, 0);
	dmac_write (instance, TXCNTHI, 0);
    }
}

/*
 * Function : acornscsi_readtobh (struct Scsi_Host *instance,
 *			AS_Host *hostdata,
 *			struct as_cmd *ascmd, short residual)
 * Purpose  : transfer a block of data from the scsi card to memory
 * Params   : instance  - host data
 *	    : hostdata	- host specific data
 *	    : ascmd	- currently executing command
 *	    : residual	- residual to transfer
 */
static
void acornscsi_readtobh (struct Scsi_Host *instance,
		AS_Host *hostdata,
		struct as_cmnd *ascmd, short residual)
{
    extern void __acornscsi_in (int port, char *buf, int len);
    Scsi_Cmnd *SCpnt = ascmd->cmd;
    unsigned long buffer_length;
    unsigned int page, offset = 0;
    register unsigned long ptrlen;
    register unsigned char *ptr;

    buffer_length = ascmd->buflen[ascmd->buffer_no] - residual;
    page = (ascmd->bufaddr[ascmd->buffer_no] >> 12) & 0x3f;
    page_write (instance, hostdata->page_reg | page);

#if (DEBUG & DEBUG_DMA)
    DBG(SCpnt,printk (KERN_DEBUG "scsi%d id%d: readtobh: (%04lX %04lX)\n        ",
		    instance->host_no, SCpnt->target, (page << 12) | offset,
		    buffer_length));
#endif
    ptr = SCpnt->SCp.ptr;
    ptrlen = SCpnt->SCp.this_residual;

    while (buffer_length > 0) {
	int this_length;

	if (buffer_length + offset > 4096)
	    this_length = 4096 - offset;
	else
	    this_length = buffer_length;

	if (this_length > ptrlen)
	    this_length = ptrlen;

#if (DEBUG & DEBUG_DMA)
	DBG(SCpnt,printk ("[%p %04lX %04X] ", ptr, (page << 12) | offset, this_length));
#endif

	__acornscsi_in (instance->io_port + (offset >> 1), ptr, this_length);

	offset += this_length;
	ptr += this_length;
	ptrlen -= this_length;
	buffer_length -= this_length;

	if (!ptrlen) {
	    if (SCpnt->SCp.buffers_residual) {
		SCpnt->SCp.buffer++;
		SCpnt->SCp.buffers_residual--;
		ptr = (unsigned char *)SCpnt->SCp.buffer->address;
		ptrlen = SCpnt->SCp.buffer->length;
	    } else {
		ptr = NULL;
		ptrlen = 0;
		break;
	    }
	}
    }

    SCpnt->SCp.ptr = ptr;
    SCpnt->SCp.this_residual = ptrlen;
#if (DEBUG & DEBUG_DMA)
    DBG(SCpnt,printk ("\n"));
#endif
}

/*
 * Function : acornscsi_writebuffer (struct Scsi_Host *instance, char *buffer,
 *			unsigned long boardaddr, int len)
 * Purpose  : transfer a block of data from memory to the scsi card
 * Params   : instance  - host specific data
 *          : buffer    - address to put data
 *          : boardaddr - address of data on scsi card
 *          : len       - length of data to transfer
 */
static
void acornscsi_writefrombh (struct Scsi_Host *instance,
		AS_Host *hostdata,
		struct as_cmnd *ascmd, int buffer)
{
    extern void __acornscsi_out (int port, char *buf, int len);
    Scsi_Cmnd *SCpnt = ascmd->cmd;
    unsigned int page, offset = 0;
    register unsigned long ptrlen;
    register unsigned char *ptr;

    page = (ascmd->bufaddr[buffer] >> 12) & 0x3f;

#if (DEBUG & (DEBUG_DMA|DEBUG_WRITE))
	DBG(SCpnt,printk (KERN_DEBUG "scsi%d id%d: writefrombh: (%04X): ",
		instance->host_no, SCpnt->target, page << 12));
#endif
    ptr = SCpnt->SCp.ptr;
    if (!ptr) {
	ascmd->buflen[buffer] = 0;
#if (DEBUG & (DEBUG_DMA|DEBUG_WRITE))
	DBG(SCpnt,printk ("\n"));
#endif
	return;
    }
    ptrlen = SCpnt->SCp.this_residual;
    page_write (instance, hostdata->page_reg | page);

    while (offset < 4096) {
	int this_length;

	this_length = 4096 - offset;

	if (this_length > ptrlen)
	    this_length = ptrlen;

#if (DEBUG & (DEBUG_DMA|DEBUG_WRITE))
	DBG(SCpnt,printk ("[%p %04X] ", ptr, this_length));
#endif
	__acornscsi_out (instance->io_port + (offset >> 1), ptr, this_length);

	offset += this_length;
	ptr += this_length;
	ptrlen -= this_length;

	if (!ptrlen) {
	    if (SCpnt->SCp.buffers_residual) {
		SCpnt->SCp.buffer++;
		SCpnt->SCp.buffers_residual--;
		ptr = (unsigned char *)SCpnt->SCp.buffer->address;
		ptrlen = SCpnt->SCp.buffer->length;
	    } else {
		ptr = NULL;
		ptrlen = 0;
		break;
	    }
	}
    }

    SCpnt->SCp.ptr = ptr;
    SCpnt->SCp.this_residual = ptrlen;
#if (DEBUG & (DEBUG_DMA|DEBUG_WRITE))
    DBG(SCpnt,printk ("\n"));
#endif
    ascmd->buflen[buffer] = offset;
}    

/*
 * Function : acornscsi_handletransfer (struct Scsi_Host *instance);
 * Purpose  : transfers a block of data from the on-board buffer
 *            to the scatter list or buffer and updates the dmac
 * Params   : instance - the device to handle
 */
static
int acornscsi_handletransfer (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    struct as_cmnd *ascmd = hostdata->connected;
    unsigned char buffer;
    
    dmac_write (instance, MASKREG, MASK_ON);
    dmac_clearintr (instance);

    if (!ascmd || (hostdata->phase != PHASE_DATAIN && hostdata->phase != PHASE_DATAOUT)) {
    	printk (KERN_WARNING "scsi%d: unexpected dmac interrupt\n",
		instance->host_no);
    	return 0;
    }

    switch (hostdata->phase) {
    case PHASE_DATAIN:
	buffer = ascmd->buffer_no;
	ascmd->dmaclen -= ascmd->buflen[buffer];
	buffer ^= 1;
	ascmd->buflen[buffer] = dmac_length (ascmd->dmaclen);
	dmac_setup (instance, DMAC_READ, ascmd->bufaddr[buffer], ascmd->buflen[buffer]);
	acornscsi_readtobh (instance, hostdata, ascmd, 0);
	ascmd->buffer_no = buffer;
	return ascmd->dmaclen ? 1 : 0;

    case PHASE_DATAOUT:
	buffer = ascmd->buffer_no ? 0 : 1;
	ascmd->dmaclen -= ascmd->buflen[buffer ^ 1];
#if (DEBUG & DEBUG_WRITE)
	DBG(ascmd->cmd,printk (KERN_DEBUG "di%d: sl %06lX dl %06lX da %04X dl %04X\n",
		buffer, ascmd->sbiclen, ascmd->dmaclen, ascmd->bufaddr[buffer],
		ascmd->buflen[buffer]));
#endif
	if (ascmd->buflen[buffer] == 0) {
	    int asr, count = 0;
	    do {
		asr = sbic_read (instance, ASR);
		if (asr & ASR_DBR) {
		    (void)sbic_read (instance, DATA);
		    count ++;
		}
	    } while (!(asr & ASR_INT));
	    printk ("scsi%d: %d extra bytes transferred - sbic len %02X%02X%02X\n", instance->host_no, count,
		sbic_read(instance, TRANSCNTH), sbic_read(instance, TRANSCNTM), sbic_read(instance, TRANSCNTL));
	    return 0;
	} else {
	    dmac_setup (instance, DMAC_WRITE, ascmd->bufaddr[buffer], ascmd->buflen[buffer]);
	    ascmd->buffer_no = buffer;
	    acornscsi_writefrombh (instance, hostdata, ascmd, buffer ^ 1);
	    return ascmd->dmaclen ? 1 : 0;
	}
    }
    return 0;
}

static
void acornscsi_sendcommand (struct Scsi_Host *instance,
			AS_Host *hostdata)
{
    struct as_dev *dev;
    struct as_cmnd *asprev, *ascmd;
    int i;

    if (hostdata->connected)
	return;
    cli ();

    if (hostdata->aborting) {
	if (!(sbic_read (instance, ASR) & (ASR_INT|ASR_BSY|ASR_CIP))) {
	    hostdata->phase = PHASE_CONNECTING_FOR_ABORT;
	    sbic_write (instance, DESTID, hostdata->aborting->cmd->target);
	    sbic_write (instance, CMND, CMND_SELWITHATN);
#if (DEBUG & (DEBUG_QUEUES|DEBUG_ABORT|DEBUG_CONNECT))
	    printk (KERN_DEBUG "scsi%d id%d: restarting abortion\n", instance->host_no,
	    		hostdata->aborting->cmd->target);
#endif
	}
	goto not_busy;
    }

    if (hostdata->connecting) {
	if (!(sbic_read (instance, ASR) & (ASR_INT|ASR_BSY|ASR_CIP))) {
	    hostdata->phase = PHASE_CONNECTING;
	    sbic_write (instance, DESTID, hostdata->connecting->cmd->target);
	    sbic_write (instance, CMND, CMND_SELWITHATN);
#if (DEBUG & (DEBUG_QUEUES|DEBUG_CONNECT))
	    printk (KERN_DEBUG "scsi%d id%d: restarting selection\n", instance->host_no,
	    		hostdata->connecting->cmd->target);
#endif
	}
	goto not_busy;
    }

    for (i = 0, dev = hostdata->dev; i < 8; i++, dev++) {
	for (ascmd = dev->issue_head, asprev = NULL; ascmd; asprev = ascmd, ascmd = ascmd->next) {
	    if (!ascmd->used || !ascmd->cmd || !ascmd->cmd->scsi_done) {
	        printk (KERN_WARNING "scsi%d: invalid %s found in issue queue\n",
			instance->host_no, ascmd->used ? ascmd->cmd ? "scsi_done" : "cmd" : "used");
		if (asprev)
		    asprev = ascmd->next;
		else
		    dev->issue_head = ascmd->next;
		if (ascmd == dev->issue_tail)
		    dev->issue_tail = asprev;
		continue;
	    }

	    if (ascmd->cmd->device->tagged_queue == 0) {
		if (set_bit (ascmd->cmd->lun, &dev->busy))
		    continue;
	    }
#ifdef SCSI2_TAG
	    else {
		/*
		 * SCSI-II Tagged support.  If the command is not a request sense, then
		 * we allocate a tag to it.
		 */
		if (ascmd->cmd->cmnd[0] != REQUEST_SENSE) {
		    if (ascmd->cmd->device->current_tag == 0) {
			printk (KERN_NOTICE "scsi%d id%d: enabling tagged queueing.\n",
			    ascmd->cmd->host->host_no, ascmd->cmd->target);
			ascmd->cmd->device->current_tag = 1; /* enable tagging */
		    }
		    ascmd->cmd->tag = ascmd->cmd->device->current_tag++;
		    if (ascmd->cmd->device->current_tag == 0) /* MUST not become zero */
			ascmd->cmd->device->current_tag += 1;
		}
	    }
#endif

	    hostdata->connecting = ascmd;

	    if (asprev)
	        asprev->next = ascmd->next;
	    else
	        dev->issue_head = ascmd->next;

	    if (ascmd == dev->issue_tail)
	        dev->issue_tail = asprev;

	    hostdata->stats.removes += 1;

	    if (!(sbic_read (instance, ASR) & (ASR_INT|ASR_BSY|ASR_CIP))) {
		sti ();
	    	hostdata->phase = PHASE_CONNECTING;
		sbic_write (instance, DESTID, hostdata->connecting->cmd->target);
		sbic_write (instance, CMND, CMND_SELWITHATN);
#if (DEBUG & (DEBUG_QUEUES|DEBUG_CONNECT))
		printk (KERN_DEBUG "scsi%d id%d: starting selection\n",
			instance->host_no, hostdata->connecting->cmd->target);
#endif
	    }

	    sti ();

	    switch (acornscsi_cmdtype (ascmd->cmd->cmnd[0])) {
	    case CMD_WRITE:
		hostdata->stats.writes += 1;
		break;
	    case CMD_READ:
	        hostdata->stats.reads += 1;
	        break;
	    case CMD_MISC:
	        hostdata->stats.miscs += 1;
	        break;
	    }
	    return;
	}
    }

not_busy:
    sti ();
}

static inline
void acornscsi_startsendcommand (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;

    if (set_bit (0, (void *)&hostdata->busy) == 0)
	acornscsi_sendcommand (instance, hostdata);
    hostdata->busy = 0;
}

static
void acornscsi_done (struct Scsi_Host *instance, struct as_cmnd *ascmd)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    Scsi_Cmnd *SCpnt = ascmd->cmd;

    if (!SCpnt)
	panic ("scsi%d: null command in acornscsi_done", instance->host_no);

    /*
     * In theory, this should not happen, but just in case it does.
     * This problem does actually seem to happen randomly.  Maybe we ought to
     * set DID_ERROR status?  But then what about partial transfers during
     * detection of devices etc?
     */
    if (SCpnt->SCp.ptr && !(SCpnt->result & 0xffff0000) &&
    			    ascmd->buffer_initialised) {
	printk ("scsi%d: incomplete data transfer detected\n",
		instance->host_no);
	dmac_dump (instance, ascmd);
	acornscsi_printstatus (instance, SCpnt->target);
	if (SCpnt->SCp.this_residual == ascmd->buflen[ascmd->buffer_no])
		SCpnt->result |= DID_ERROR << 16;
    }

    sbic_write (instance, SOURCEID, SOURCEID_ER | SOURCEID_DSP);
    dmac_write (instance, MASKREG, MASK_ON);
    dmac_clearintr (instance);

    ascmd->cmd = NULL;
    ascmd->next = NULL;
    ascmd->origlen = 0;
    ascmd->sbiclen = 0;
    ascmd->dmaclen = 0;
#ifdef FIX_64K
    ascmd->transferred = 0;
#endif
    ascmd->buffer_initialised = 0;
    ascmd->buffer_no = 0;
    ascmd->saved = 0;
    ascmd->buflen[0] = 0;
    ascmd->buflen[1] = 0;
    ascmd->sdmacaddress = 0;
    ascmd->sdmaclen = 0;
    ascmd->scmdptr = 0;
    ascmd->used = 0;

    hostdata->dev[SCpnt->target].busy &= ~(1 << SCpnt->lun);

    if (!SCpnt->scsi_done)
	panic ("scsi%d: null scsi_done function in acornscsi_done", instance->host_no);
    SCpnt->scsi_done (SCpnt);
}

static
void acornscsi_savepointers (struct Scsi_Host *instance,
				AS_Host *hostdata)
{
    struct as_cmnd *ascmd = hostdata->connected;
    unsigned long transferred, buffer_offset;
    unsigned char buffer_no;

    if (!ascmd->buffer_initialised)
	return;

    dmac_write (instance, MASKREG, MASK_ON);
#ifdef FIX_64K
    ascmd->sbiclen = hostdata->tmpsbiclen + ascmd->transferred;
    ascmd->transferred = 0;
#else
    ascmd->sbiclen = hostdata->tmpsbiclen;
#endif

    if (!ascmd->sbiclen) {
	ascmd->sdmacaddress = 0;
	ascmd->sdmaclen = 0;
	ascmd->saved = 1;
	return;
    }
    transferred = ascmd->origlen - ascmd->sbiclen;

#if (DEBUG & DEBUG_DISCON)
    DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: dis ol %06lX sl %06lX dl %06lX",
		instance->host_no, ascmd->cmd->target, ascmd->origlen, ascmd->sbiclen,
		ascmd->dmaclen));
#endif

    buffer_no = (transferred >> 12) & 1;
    buffer_offset = transferred & 4095;
    if (buffer_no != ascmd->buffer_no)
	printk (" buffer_no mismatch! - %d != %d", ascmd->buffer_no, buffer_no);
    ascmd->sdmacaddress = ascmd->bufaddr[buffer_no] + buffer_offset;
    ascmd->sdmaclen = ascmd->buflen[buffer_no] - buffer_offset;
    ascmd->saved = 1;

#if (DEBUG & DEBUG_DISCON)
    DBG(ascmd->cmd,printk (" da %04X dl %04X\n", ascmd->sdmacaddress, ascmd->sdmaclen));
#endif
}

/*
 * Function : acornscsi_disconnect (struct Scsi_Host *instance)
 * Purpose  : disconnect the current command & put onto disconnected queue
 * Params   : instance - host data
 *
 * Notes:
 * We appear to have a race condition between this exiting, sending the next command,
 * and a device on the bus reselecting us...
 */
static inline
void acornscsi_disconnect (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    struct as_dev *dev;
    struct as_cmnd *ascmd;
    unsigned long flags;

    dev = &hostdata->dev[hostdata->connected->cmd->target];

    dmac_write (instance, MASKREG, MASK_ON);

    save_flags_cli (flags);
    ascmd = hostdata->connected;
    hostdata->connected = NULL;

    ascmd->next = dev->disconnected;
    dev->disconnected = ascmd;
    restore_flags (flags);

#if (DEBUG & (DEBUG_QUEUES|DEBUG_DISCON))
    DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: command for lun %d tag %d"
		" moved to disconnected queue\n", instance->host_no,
		ascmd->cmd->target, ascmd->cmd->lun, ascmd->cmd->tag));
#endif
}

/*
 * Function : acornscsi_reconnect (struct Scsi_Host *instance)
 * Purpose  : restore a previously saved command state
 * Params   : instance - host data
 * Remarks  : SCSI spec says:
 *		'The set of active pointers is restored from the set
 *		 of saved pointers upon reconnection of the I/O process'
 */
static inline
int acornscsi_reconnect (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    struct as_dev *dev;
    struct as_cmnd *ascmd, **asprev;
    int sourceid, sourcelun, sourcetag = 0;

    sourceid = sbic_read (instance, SOURCEID);

    if (hostdata->connected) {
	printk (KERN_ERR "scsi%d id%d: reconnecting while connected!?!  Loosing command\n",
	    instance->host_no, hostdata->connected->cmd->target);
	hostdata->connected = NULL;
    }

    if (!(sourceid & 8)) {
	printk (KERN_WARNING "scsi%d: invalid source id after reselection - chip/device fault?\n",
	    instance->host_no);
	return 0;
    }

    sourceid &= 7;
    sourcelun = sbic_read (instance, DATA);
    sourcelun &= 7;

    dev = &hostdata->dev[sourceid];

    for (asprev = &dev->disconnected; *asprev; asprev = &(*asprev)->next) {
    	ascmd = *asprev;
	if (ascmd->cmd->target != sourceid)
	    printk (KERN_WARNING "scsi%d: command on wrong disconnected queue %d instead of %d\n",
		instance->host_no, ascmd->cmd->target, sourceid);
	if (ascmd->cmd->lun == sourcelun && ascmd->cmd->tag == sourcetag) {
	    *asprev = ascmd->next;
	    ascmd->next = NULL;
	    hostdata->connected = ascmd;
#if (DEBUG & (DEBUG_QUEUES|DEBUG_DISCON))
	DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: command for lun %d tag %d reconnected\n",
		instance->host_no, ascmd->cmd->target, ascmd->cmd->lun, ascmd->cmd->tag));
#endif
	    hostdata->cmdptr = ascmd->scmdptr;
	    return 1;
	}
    }
    printk (KERN_WARNING "scsi%d id%d: lun %d tag %d not in disconnect_queue.\n",
	instance->host_no, sourceid, sourcelun, sourcetag);
    return 0;
}

/*
 * Function : acornscsi_forcedisconnect (struct Scsi_Host *instance)
 * Purpose  : force a disconnect from the bus
 * Params   : instance - hostdata
 */
static inline
void acornscsi_forcedisconnect (struct Scsi_Host *instance)
{
    sbic_write (instance, CMND, CMND_DISCONNECT);
}

/*
 * Function : acornscsi_startabort (struct Scsi_Host *instance)
 * Purpose  : start aborting a connected command
 * Params   : instance - hostdata
 */
static inline
void acornscsi_startabort (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    add_msg (hostdata, 1, ABORT);
    sbic_write (instance, CMND, CMND_ASSERTATN);
    acornscsi_printstatus (instance, 8);
}

/* ==========================================================================================
 * Interrupt routines.
 */
static
int acornscsi_sbicintr (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    struct as_cmnd *ascmd;
    int asr, ssr;

    asr = sbic_read (instance, ASR);
    if (!(asr & ASR_INT))
	return INTR_CONTINUE;

    ssr = sbic_read (instance, SSR);
#if (DEBUG & (DEBUG_DMA|DEBUG_RECON|DEBUG_PHASES|DEBUG_WRITE))
    print_sbic_status(asr, ssr, sbic_read (instance, CMNDPHASE));
#endif
    ADD_STATUS(8, ssr);
    ADD_STATUS(8, hostdata->phase);

    ascmd = hostdata->connected;
    if (ascmd) {
	ADD_STATUS(ascmd->cmd->target, ssr);
	ADD_STATUS(ascmd->cmd->target, hostdata->phase);
    }

    /*
     * This is just one big switch statement
     */
    switch (ssr) {
/* ---------------------- Reset -------------------------------------------------------- */
    case 0x00: /* reset state - not advanced */
	printk (KERN_ERR "scsi%d: reset in standard mode but wanted advanced mode.\n",
	    instance->host_no);
    case 0x01: /* reset state - advanced */
	sbic_write (instance, CTRL, INIT_SBICDMA | CTRL_IDI);
	sbic_write (instance, TIMEOUT, TIMEOUT_TIME);
	sbic_write (instance, SYNCHTRANSFER, SYNCHTRANSFER_2DBA);
	return INTR_CONTINUE;

/* ---------------------- Select ------------------------------------------------------- */
    case 0x11: /* select command ok */
        switch (hostdata->phase) {
	case PHASE_CONNECTING:
	    if (hostdata->connecting) {
		ascmd = hostdata->connected = hostdata->connecting;
		hostdata->connecting = NULL;
		hostdata->phase = PHASE_CONNECTED;
		sbic_write (instance, SYNCHTRANSFER, hostdata->dev[ascmd->cmd->target].sync_xfer);
		add_msg (hostdata, 1,
			IDENTIFY(hostdata->dev[ascmd->cmd->target].disconnect_ok,
			ascmd->cmd->lun));
#ifdef SCSI2_TAG
		if (ascmd->cmd->device->tagged_queue) {
		    if (ascmd->cmd->cmnd[0] == REQUEST_SENSE)
			add_msg (hostdata, 2, HEAD_OF_QUEUE_TAG, ascmd->cmd->tag);
		    else
			add_msg (hostdata, 2, SIMPLE_QUEUE_TAG, ascmd->cmd->tag);
		}
#endif
#ifdef SCSI2_SYNC
		if (hostdata->dev[ascmd->cmd->target].sync_stat == SDTR_UNSET) {
		    hostdata->dev[ascmd->cmd->target].sync_stat = SDTR_SENT;
		    add_msg (hostdata, 5,
			EXTENDED_MESSAGE,
			3,
			EXTENDED_SDTR,
			SDTR_PERIOD / 4,
			SDTR_SIZE);
		}
#endif
		hostdata->cmdptr = 0;
#if (DEBUG & (DEBUG_QUEUES|DEBUG_CONNECT))
		printk (KERN_DEBUG "scsi%d id%d: command connected\n",
		    instance->host_no, hostdata->connected->cmd->target);
#endif
		return INTR_CONNECTED;
	    } else {
		printk (KERN_WARNING "scsi%d: select ok but no connecting command\n",
		    instance->host_no);
		goto abort_connect;
	    }

	case PHASE_CONNECTING_FOR_ABORT:
	    if (hostdata->aborting) {
		hostdata->connected = hostdata->aborting;
		hostdata->aborting = NULL;
		add_msg (hostdata, 1, ABORT);
#if (DEBUG & (DEBUG_QUEUES|DEBUG_CONNECT))
		printk (KERN_DEBUG "scsi%d id%d: abort command connected\n",
		    instance->host_no, hostdata->connected->cmd->target);
#endif
		return INTR_CONNECTED;
	    } else {
		printk (KERN_WARNING "scsi%d: select ok but no connecting aborting command\n",
		    instance->host_no);
		goto abort_connect;
	    }

	default:
	    printk (KERN_WARNING "scsi%d: select connected to target with invalid phase (%X)\n",
		instance->host_no, hostdata->phase);
	abort_connect:
	    /* At this point we have lost the command totally (and hence can't return an
	     * error until the upper code causes an abort.
	     */
	    acornscsi_printstatus (instance, 8);
	    acornscsi_forcedisconnect (instance);
	    hostdata->phase = PHASE_IDLE;
	    return INTR_NOTCONNECTED;
	}

    case 0x42:		/* select timeout */
	switch (hostdata->phase) {
	case PHASE_CONNECTING:
	    ascmd = hostdata->connecting;
	    hostdata->connecting = NULL;
	    break;

	case PHASE_CONNECTING_FOR_ABORT:
	    ascmd = hostdata->aborting;
	    hostdata->aborting = NULL;
	    break;

	default:
	    printk (KERN_WARNING "scsi%d id%d: select timeout in wrong phase\n",
	    	instance->host_no, hostdata->connecting->cmd->target);
	    acornscsi_printstatus (instance, 8);
	    return INTR_NOTCONNECTED;
	}
	ascmd->cmd->result = DID_NO_CONNECT << 16;
	hostdata->phase = PHASE_IDLE;
	acornscsi_done (instance, ascmd);
	return INTR_NOTCONNECTED;

    case 0x81: /* reselection occurred */
	/* for this, we need to add extra code to allow us to keep track of whether this
	 * is SCSI 2 tagged or not.  I am not sure how this part is supposed to work.
	 *
	 * For the moment, I will assume that it works the same way that we tell the
	 * device, ie. SELECT->MESSAGE(IDENTIFY)->MESSAGE(SIMPLE_TAG)->MESSAGE(TAGID)
	 */
	if (acornscsi_reconnect (instance)) {
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    hostdata->phase = PHASE_RECONNECTED;
	} else {
	    sbic_write (instance, CMND, CMND_ASSERTATN);
	    hostdata->phase = PHASE_ABORTED;
	    add_msg (hostdata, 1, ABORT);
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	}
	return INTR_CONNECTED;
	
    case 0x41:		/* unexpected disconnect aborted command */
	if (!ascmd)
	    return INTR_NOTCONNECTED;
	printk (KERN_WARNING "scsi%d id%d: unexpected disconnect\n",
		instance->host_no, ascmd->cmd->target);
#if (DEBUG & DEBUG_ABORT)
	acornscsi_printstatus (instance, 8);
#endif
	hostdata->connected = NULL;
	ascmd->cmd->result = DID_ABORT << 16;
	acornscsi_done (instance, ascmd);
	hostdata->phase = PHASE_IDLE;
	return INTR_NOTCONNECTED;

    case 0x85: /* disconnection occurred */
	if (ascmd == NULL) {
	    printk (KERN_WARNING "scsi%d: disconnect with no connected command\n",
	    	instance->host_no);
	    return INTR_NOTCONNECTED;
	}
	switch (hostdata->phase) {
	case PHASE_FINISHED:
	    hostdata->connected = NULL;
	    hostdata->stats.fins += 1;
	    hostdata->phase = PHASE_IDLE;
	    ascmd->cmd->result = ascmd->cmd->SCp.Status | (ascmd->cmd->SCp.Message << 8);
	    acornscsi_done (instance, ascmd);
	    return INTR_NOTCONNECTED;

	case PHASE_DISCONNECTING:
	    acornscsi_disconnect (instance);
	    hostdata->phase = PHASE_IDLE;
	    return INTR_NOTCONNECTED;

	case PHASE_ABORTED:
	    hostdata->connected = NULL;
	    hostdata->stats.fins += 1;
	    hostdata->phase = PHASE_IDLE;
	    ascmd->cmd->result = DID_ABORT << 16;
	    acornscsi_done (instance, ascmd);
	    return INTR_NOTCONNECTED;

	case PHASE_IDLE:
	    break;

	default:
	    printk (KERN_WARNING "scsi%d id%d: invalid phase in disconnect (%X)\n",
		instance->host_no, ascmd->cmd->target, hostdata->phase);
	    acornscsi_printstatus (instance, 8);
	}
    	return INTR_NOTCONNECTED;

/* ---------------------- data out ----------------------------------------------------- */
    case 0x88: /* connect - data out (only after reconnection) */
    case 0x18: /* data transfer completed - data out */
    {
	int bufinit = ascmd->buffer_initialised;
	if (!bufinit) {
#if (DEBUG & (DEBUG_DISCON|DEBUG_WRITE))
	    DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: do init\n",
		instance->host_no, ascmd->cmd->target));
#endif
	    ascmd->buffer_no = 0;
	    ascmd->buffer_initialised = 1;
	    acornscsi_writefrombh (instance, hostdata, ascmd, 0);
	    ascmd->sdmacaddress = ascmd->bufaddr[0];
	    ascmd->sdmaclen = ascmd->buflen[0];
	} else {
#ifdef FIX_64K
	    if (ascmd->transferred) {
		sbic_write (instance, TRANSCNTH, 0);
		sbic_writenext (instance, 128);
		sbic_writenext (instance, 0);
	    } else
#endif
		   {
#if (DEBUG & (DEBUG_DISCON|DEBUG_WRITE))
		DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: do rcnt\n",
			instance->host_no, ascmd->cmd->target));
#endif
		if (!ascmd->saved)
		    printk (KERN_WARNING "scsi%d id%d: no saved data or reconnect\n",
			instance->host_no, ascmd->cmd->target);
		ascmd->saved = 0;
	    }
	}
	sbic_write (instance, SYNCHTRANSFER, hostdata->dev[ascmd->cmd->target].sync_xfer);
#ifdef FIX_64K
	ascmd->transferred = 0;
	if (ascmd->sbiclen >= 32768) {
		ascmd->transferred = 32768;
		sbic_writenext (instance, (ascmd->sbiclen - 32768) >> 16);
		sbic_writenext (instance, (ascmd->sbiclen - 32768) >> 8);
		sbic_writenext (instance, (ascmd->sbiclen - 32768));
	} else
#endif
	{
		sbic_writenext (instance, ascmd->sbiclen >> 16);
		sbic_writenext (instance, ascmd->sbiclen >> 8);
		sbic_writenext (instance, ascmd->sbiclen);
	}
#if (DEBUG & (DEBUG_DISCON|DEBUG_WRITE))
	DBG(ascmd->cmd,printk (" sl %06lX dl %06lX da %04X dl %04X\n",
			ascmd->sbiclen, ascmd->dmaclen,
			ascmd->sdmacaddress, ascmd->sdmaclen));
#endif
	dmac_setup (instance, DMAC_WRITE, ascmd->sdmacaddress, ascmd->sdmaclen);
	sbic_write (instance, CMND, CMND_XFERINFO);
	if (!bufinit)
	    acornscsi_writefrombh (instance, hostdata, ascmd, 1);
	hostdata->phase = PHASE_DATAOUT;
	return INTR_CONNECTED;
    }

/* ---------------------- data in ------------------------------------------------------ */
    case 0x89: /* connect - data in [only after reconnection] */
    case 0x19: /* transfer completed - data in */
    {
	int bufinit = ascmd->buffer_initialised;
	if (!bufinit) {
#if (DEBUG & DEBUG_DISCON)
	    DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: di init:",
		instance->host_no, ascmd->cmd->target));
#endif
	    ascmd->buffer_no = 0;
	    ascmd->buffer_initialised = 1;
	    ascmd->buflen[0] = dmac_length (ascmd->dmaclen);
	    ascmd->sdmacaddress = ascmd->bufaddr[0];
	    ascmd->sdmaclen = ascmd->buflen[0];
	} else {
#if (DEBUG & DEBUG_DISCON)
	    DBG(ascmd->cmd,printk (KERN_DEBUG "scsi%d id%d: di rcnt:",
		instance->host_no, ascmd->cmd->target));
#endif
	    if (!ascmd->saved)
		printk (KERN_WARNING "scsi%d id%d: no saved data for reconnect\n",
			instance->host_no, ascmd->cmd->target);
	    ascmd->saved = 0;
	}
	sbic_write (instance, SYNCHTRANSFER, hostdata->dev[ascmd->cmd->target].sync_xfer);
#ifdef FIX_64K
	ascmd->transferred = 0;
#endif
	sbic_writenext (instance, ascmd->sbiclen >> 16);
	sbic_writenext (instance, ascmd->sbiclen >> 8);
	sbic_writenext (instance, ascmd->sbiclen);
#if (DEBUG & DEBUG_DISCON)
	DBG(ascmd->cmd,printk (" sl %06lX da %04X dl %04X\n", ascmd->sbiclen,
				ascmd->sdmacaddress, ascmd->sdmaclen));
#endif
	dmac_setup (instance, DMAC_READ, ascmd->sdmacaddress, ascmd->sdmaclen);
	sbic_write (instance, CMND, CMND_XFERINFO);
	hostdata->phase = PHASE_DATAIN;
	return INTR_CONNECTED;
    }

/* ---------------------- command out -------------------------------------------------- */
    case 0x1a: /* transfer completed - command info next */
    case 0x4a:
    case 0x8a: /* connection - command phase out */
	if (hostdata->phase == PHASE_CONNECTED) {
	    Scsi_Cmnd *SCpnt = hostdata->connected->cmd;
	    int i = hostdata->cmdptr;

	    if (i >= SCpnt->cmd_len) {
	    	add_msg (hostdata, 1, ABORT);
		sbic_write (instance, CMND, CMND_XFERINFO | CMND_SBT);
		sbic_write (instance, CMND, CMND_ASSERTATN);
		sbic_write (instance, DATA, 0);
		hostdata->phase = PHASE_ABORTED;
		return INTR_CONNECTED;
	    }

	    sbic_write (instance, TRANSCNTH, 0);
	    sbic_writenext (instance, 0);
	    sbic_writenext (instance, SCpnt->cmd_len - i);
	    sbic_write (instance, CMND, CMND_XFERINFO);
	    do {
		asr = sbic_read (instance, ASR);
		if (asr & ASR_DBR) {
		    sbic_write (instance, DATA, SCpnt->cmnd[i++]);
		    if (i >= SCpnt->cmd_len) {
			hostdata->phase = PHASE_COMMANDSENT;
			break;
		    }
	        }
	    } while (!(asr & ASR_INT));
	    hostdata->cmdptr = i;
	    return INTR_CONNECTED;
	} else {
	    printk (KERN_WARNING "scsi%d id%d: invalid phase for command (%X)\n",
		instance->host_no, ascmd->cmd->target, hostdata->phase);
	    acornscsi_startabort (instance);
	    return INTR_CONNECTED;
	}

/* ---------------------- status in ---------------------------------------------------- */
    case 0x1b: /* transfer completed - status phase in */
    case 0x8b: /* connection - status phase in [only after reconnection] */
	sbic_write (instance, CMND, CMND_XFERINFO|CMND_SBT);
	while ((sbic_read (instance, ASR) & ASR_DBR) == 0);
	ascmd->cmd->SCp.Status = sbic_read (instance, DATA);
	hostdata->phase = PHASE_STATUSIN;
	return INTR_CONNECTED;

    case 0x4b: { /* unexpected status info */
	unsigned short residual;

	dmac_write (instance, MASKREG, MASK_ON);
	residual = ((dmac_read (instance, TXCNTHI) << 8) | dmac_read (instance, TXCNTLO)) + 1;

	sbic_write (instance, TRANSCNTH, 0);
	sbic_writenext (instance, 0);
	sbic_writenext (instance, 0);

	if (hostdata->phase == PHASE_DATAIN)
	    acornscsi_readtobh (instance, hostdata, ascmd, residual);

	sbic_write (instance, CMND, CMND_XFERINFO|CMND_SBT);
	while ((sbic_read (instance, ASR) & ASR_DBR) == 0);
	hostdata->connected->cmd->SCp.Status = sbic_read (instance, DATA);
	hostdata->phase = PHASE_STATUSIN;

	return INTR_CONNECTED;
	}

/* ---------------------- message out -------------------------------------------------- */
    case 0x1e:
    case 0x2e:
    case 0x4e:
    case 0x8e: /* connection - message out phase [only after reconnection] */
	if (!hostdata->out_msg) {
	    sbic_write (instance, CMND, CMND_XFERINFO | CMND_SBT);
	    while ((sbic_read (instance, ASR) & ASR_DBR) == 0);
	    sbic_write (instance, DATA, NOP);
	    hostdata->last_message = NOP;
	    return INTR_CONNECTED;
	} else {
	    int setup = 0;
	    AS_Msg *m;

	    for (m = hostdata->out_msg; m; m = hostdata->out_msg) {
	    	int i, asr;

		hostdata->out_msg = m->next;

#if (DEBUG & DEBUG_MESSAGES)
		printk (KERN_DEBUG "scsi%d id%d: sending ",
				instance->host_no, ascmd->cmd->target);
		print_msg (m->vals);
		printk ("\n");
#endif
		hostdata->last_message = m->vals[0];
		if (hostdata->last_message == EXTENDED_MESSAGE)
		    hostdata->last_message |= m->vals[2] << 8;

		if (!setup) {
		    AS_Msg *p;
		    int msglen = 0;

		    setup = 1;

		    for (p = m; p; p = p->next)
		    msglen += p->len;

		    if (msglen < 2)
			sbic_write (instance, CMND, CMND_XFERINFO | CMND_SBT);
		    else {
			sbic_write (instance, TRANSCNTH, 0);
			sbic_writenext (instance, 0);
			sbic_writenext (instance, msglen);
			sbic_write (instance, CMND, CMND_XFERINFO);
		    }
		}

		i = 0;
		do {
		    asr = sbic_read (instance, ASR);
		    if (asr & ASR_DBR) {
			sbic_write (instance, DATA, m->vals[i++]);
			if (i >= m->len)
			    break;
		    }
		} while (!(asr & ASR_INT));
		m->len = 0;
		m->next = NULL;
		m->used = 0;
		if (asr & ASR_INT)
		    break;
	    }
	}
	return INTR_CONNECTED;

/* ---------------------- message in --------------------------------------------------- */
    case 0x1f: /* transfer completed - message in phase */
    case 0x4f: /* transfer terminated - message in phase */
    case 0x8f: /* connection - message in phase [only after reconnection] [???] */
	dmac_write (instance, MASKREG, MASK_ON);
	hostdata->tmpsbiclen = sbic_readlength (instance); /* needed in case of save pointers */
	hostdata->in_msgptr = 0;
	do {
	    sbic_write (instance, CMND, CMND_XFERINFO|CMND_SBT);
	    while ((sbic_read (instance, ASR) & ASR_DBR) == 0);
	    hostdata->in_msg[hostdata->in_msgptr++] = sbic_read (instance, DATA);
	    while ((sbic_read (instance, ASR) & ASR_INT) == 0);	/* wait for msg-in-xfer-paused */
	    sbic_read (instance, SSR);
	    if (hostdata->in_msg[0] != EXTENDED_MESSAGE ||
		(hostdata->in_msgptr > 1 && hostdata->in_msgptr > 1 + hostdata->in_msg[1]))
		break;
	    sbic_write (instance, CMND, CMND_NEGATEACK);	/* negate-ack to continue */
	    while ((sbic_read (instance, ASR) & ASR_INT) == 0);	/* wait for next msg-in */
	    sbic_read (instance, SSR);
	} while (1);
	hostdata->connected->cmd->SCp.Message = hostdata->in_msg[0];
#if (DEBUG & DEBUG_MESSAGES)
printk (KERN_DEBUG "scsi%d id%d: ", instance->host_no, ascmd->cmd->target);
print_msg (hostdata->in_msg);
printk ("\n");
#endif
	switch (hostdata->in_msg[0]) {
	case COMMAND_COMPLETE:
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    hostdata->phase = PHASE_FINISHED;
	    return INTR_CONNECTED;

	case SAVE_POINTERS:
	    /*
	     * Save the command, data & status pointers
	     */
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    ascmd->scmdptr = hostdata->cmdptr;
	    if (ascmd->buffer_initialised)
		acornscsi_savepointers (instance, hostdata);
	    break;

	case RESTORE_POINTERS:
	    /*
	     * Restore the command, data & status pointers
	     */
	    hostdata->cmdptr = ascmd->scmdptr;
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    break;

	case DISCONNECT:
	    del_all_msgs (hostdata);
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    /* If the device has not issued a 'save pointers' message, do it.
	     * ANSI standard says: (Section SCSI-2 Rev. 10c Sect 6.4.2)
	     * 'On those occasions when an error or exception condition occurs
	     *  and the target elects to repeat the information transfer, the
	     *  target may repeat the transfer either issuing a RESTORE POINTERS
	     *  message or by disconnecting without issuing a SAVE POINTERS
	     *  message.  When reconnection is completed, the most recent
	     *  saved pointer values are restored.'
	     *
	     * If a device disconnects before it transfers any data, it may or
	     * may not issue a save pointers.  To make this work correctly,
	     * we have to save more pointers in the code...
	     */
	    hostdata->stats.disconnects ++;
	    if (ascmd->buffer_initialised && !ascmd->saved)
		acornscsi_savepointers (instance, hostdata);
	    hostdata->phase = PHASE_DISCONNECTING;
	    return INTR_CONNECTED;

	case MESSAGE_REJECT:
	    if (hostdata->out_msg)
		sbic_write (instance, CMND, CMND_ASSERTATN);
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    switch (hostdata->last_message) {
#ifdef SCSI2_TAG
	    case HEAD_OF_QUEUE_TAG:
	    case ORDERED_QUEUE_TAG:
	    case SIMPLE_QUEUE_TAG:
		ascmd->cmd->device->tagged_queue = 0;
		set_bit (ascmd->cmd->lun, &hostdata->dev[ascmd->cmd->target].busy);
		printk (KERN_NOTICE "scsi%d id%d: disabling tagged queueing\n",
		    instance->host_no, ascmd->cmd->target);
		break;
#endif
	    case EXTENDED_MESSAGE | (EXTENDED_SDTR << 8):
		printk (KERN_NOTICE "scsi%d id%d: Using asynchronous transfer\n",
			instance->host_no, ascmd->cmd->target);
		hostdata->dev[ascmd->cmd->target].sync_xfer = SYNCHTRANSFER_2DBA;
		hostdata->dev[ascmd->cmd->target].sync_stat = SDTR_DONE;
		sbic_write (instance, SYNCHTRANSFER, hostdata->dev[ascmd->cmd->target].sync_xfer);
		break;
		
	    default:
	        break;
	    }
	    break;

	case SIMPLE_QUEUE_TAG:
	    /* tag queue reconnect... hostdata->in_msg[1] = queue tag */
	    sbic_write (instance, CMND, CMND_ASSERTATN);
#ifdef SCSI2_TAG
	    if (!hostdata->in_msg[1] == 0) {
		add_msg (hostdata, 1, ABORT_TAG);
		hostdata->phase = PHASE_ABORTED;
	    } else
#endif
		add_msg (hostdata, 1, MESSAGE_REJECT);
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    break;

	case EXTENDED_MESSAGE:
	    switch (hostdata->in_msg[2]) {
	    case EXTENDED_SDTR:
		if (hostdata->dev[ascmd->cmd->target].sync_stat == SDTR_SENT) {
		    hostdata->dev[ascmd->cmd->target].sync_stat = SDTR_DONE;
		    sbic_write (instance, CMND, CMND_NEGATEACK);
		    printk (KERN_NOTICE "scsi%d id%d: Using synchronous transfer, offset %d, %d ns\n",
			instance->host_no, ascmd->cmd->target, hostdata->in_msg[4],
			hostdata->in_msg[3] * 4);
		    hostdata->dev[ascmd->cmd->target].sync_xfer =
			calc_sync_xfer (hostdata->in_msg[3] * 4, hostdata->in_msg[4]);
		} else {
		    unsigned char period, length;
		    sbic_write (instance, CMND, CMND_ASSERTATN);
#ifdef SCSI2_SYNC
		    period = max (hostdata->in_msg[3], DEFAULT_PERIOD / 4);
		    length = min (hostdata->in_msg[4], 12);
		    add_msg (hostdata, 5,
				EXTENDED_MESSAGE,
				3,
				EXTENDED_SDTR,
				period,
				length);
		    sbic_write (instance, CMND, CMND_NEGATEACK);
		    hostdata->dev[ascmd->cmd->target].sync_xfer =
			calc_sync_xfer (period * 4, length);
#else
		    add_msg (hostdata, 1, MESSAGE_REJECT);
		    sbic_write (instance, CMND, CMND_NEGATEACK);
		    hostdata->dev[ascmd->cmd->target].sync_xfer = SYNCHTRANSFER_2DBA;
#endif
		}
		sbic_write (instance, SYNCHTRANSFER, hostdata->dev[ascmd->cmd->target].sync_xfer);
		break;

	    case EXTENDED_WDTR:
		sbic_write (instance, CMND, CMND_ASSERTATN);
		add_msg (hostdata, 4,
				EXTENDED_MESSAGE,
				2,
				EXTENDED_WDTR,
				0);	/* 8-bit only */
		sbic_write (instance, CMND, CMND_NEGATEACK);
		break;

	    default:
		sbic_write (instance, CMND, CMND_ASSERTATN);
		add_msg (hostdata, 1, MESSAGE_REJECT);
		sbic_write (instance, CMND, CMND_NEGATEACK);
		break;
	    }
	    break;

#ifdef SCSI2_LINK
	case LINKED_CMD_COMPLETE:
	case LINKED_FLG_CMD_COMPLETE:
	    /*
	     * We don't support linked commands yet
	     */
	    if (0) {
#if (DEBUG & DEBUG_LINKED)
		printk (KERN_DEBUG "scsi%d id%d: lun %d tag %d linked command complete\n",
			instance->host_no, ascmd->connected->target, ascmd->connected->tag);
#endif
		/*
		 * A linked command should only terminate with one of these messages
		 * if there are more linked commands available.
		 */
		if (!ascmd->cmd->next_link) {
		    sbic_write (instance, CMND, CMND_ASSERTATN);
		    printk (KERN_WARNING "scsi%d id%d: lun %d tag %d linked command complete, but no next_link\n",
			instance->host_no, ascmd->connected->target, ascmd->connected->tag);
		    add_msg (hostdata, 1, ABORT);
		}
		sbic_write (instance, CMND, CMND_NEGATEACK);

		ascmd->cmd->next_link->tag = ascmd->cmd->tag;
		/* scsi_done */
		hostdata->connected->cmd = ascmd->cmd->next_link;
		break;
	    }
#endif

	default: /* reject message */
	    sbic_write (instance, CMND, CMND_ASSERTATN);
	    add_msg (hostdata, 1, MESSAGE_REJECT);
	    sbic_write (instance, CMND, CMND_NEGATEACK);
	    break;
	}
	return INTR_CONNECTED;

    default:
	printk (KERN_WARNING "scsi%d: unknown interrupt number %02X\n", instance->host_no, ssr);
	acornscsi_printstatus (instance, 8);
	return INTR_CONTINUE;
    }
}

static
void acornscsi_intr (int irq, void *dev_id, struct pt_regs *regs)
{
    struct Scsi_Host *instance = (struct Scsi_Host *)dev_id;
    int need_command, need_command_done;

    do {
	need_command_done = 0;
	need_command = 0;

	if (read_irqstat(instance) & 2 && acornscsi_handletransfer(instance)) {
	    /* setup for next buffer & transfer current buffer */
	    need_command_done = 1;
	    continue;
	}

	if (read_irqstat(instance) & 8)
	    switch (acornscsi_sbicintr (instance)) {
	    case INTR_NOTCONNECTED:
		need_command = 1;
		break;
	    case INTR_CONNECTED:
		need_command_done = 1;
		break;
	    }
	if (need_command) {
	    acornscsi_startsendcommand (instance);
	    need_command_done = 1;
	}
    }
    while(need_command_done);
}

/*=============================================================================================
 * Interfaces between interrupt handler and rest of scsi code
 */

/*
 * Function : acornscsi_queuecmd (Scsi_Cmnd *cmd, void (*done)(Scsi_Cmnd *))
 * Purpose  : queues a SCSI command
 * Params   : cmd  - SCSI command
 *            done - function called on completion, with pointer to command descriptor
 * Returns  : 0, or < 0 on error.
 */
int acornscsi_queuecmd (Scsi_Cmnd *cmd, void (*done)(Scsi_Cmnd *))
{
    AS_Host *hostdata = (AS_Host *)cmd->host->hostdata;
    struct as_cmnd *as_cmd;
    int i;
#if (DEBUG & DEBUG_QUEUES)
    const char *where;
#endif

    if (!done) {
    	/* there should be some way of rejecting errors like this without panicing... */
	panic ("scsi%d: queuecommand called with NULL done function [cmd=%p]",
		cmd->host->host_no, cmd);
	return -EINVAL;
    }

#if (DEBUG & DEBUG_NO_WRITE)
    if (acornscsi_cmdtype (cmd->cmnd[0]) == CMD_WRITE) {
	printk (KERN_DEBUG "scsi%d id%d: WRITE attempted with NO_WRITE flag set\n",
	    cmd->host->host_no, cmd->target);
	cmd->result = DID_ERROR << 16;
	done (cmd);
	return 0;
    }
#endif

    cmd->host_scribble = NULL;
    cmd->scsi_done = done;
    cmd->result = 0;
    cmd->SCpdirection = acornscsi_writedirection (cmd->cmnd[0]);

    if (cmd->use_sg) {
	cmd->SCp.buffer = (struct scatterlist *) cmd->buffer;
	cmd->SCp.buffers_residual = cmd->use_sg - 1;
	cmd->SCp.ptr = (char *) cmd->SCp.buffer->address;
	cmd->SCp.this_residual = cmd->SCp.buffer->length;
    } else {
	cmd->SCp.buffer = NULL;
	cmd->SCp.buffers_residual = 0;
	cmd->SCp.ptr = (char *) cmd->request_buffer;
	cmd->SCp.this_residual = cmd->request_bufflen;
    }

    cmd->tag = 0;

    for (i = 0, as_cmd = hostdata->queuedcommands; i < CAN_QUEUE; i++, as_cmd ++) {
	if (set_bit (0, (void *)&as_cmd->used) == 0) {
	    if (as_cmd->cmd) {
		printk (KERN_ERR "scsi%d: free command found wasn't free\n",
			cmd->host->host_no);
		continue;
	    }
	    as_cmd->cmd = cmd;
	    as_cmd->buffer_no = 0;
	    as_cmd->buffer_initialised = 0;
	    as_cmd->saved = 0;
	    as_cmd->scmdptr = 0;
	    as_cmd->origlen = as_cmd->dmaclen = as_cmd->sbiclen = cmd->request_bufflen;
#ifdef FIX_64K
	    as_cmd->transferred = 0;
#endif

	    if (cmd->cmnd[0] == REQUEST_SENSE) {
		cli ();
		as_cmd->next = hostdata->dev[cmd->target].issue_head;
		hostdata->dev[cmd->target].issue_head = as_cmd;
		sti ();
#if (DEBUG & DEBUG_QUEUES)
		where = "head";
#endif
	    } else {
		cli ();
		as_cmd->next = NULL;
		if (hostdata->dev[cmd->target].issue_head == NULL)
		    hostdata->dev[cmd->target].issue_head = as_cmd;
		if (hostdata->dev[cmd->target].issue_tail)
		    hostdata->dev[cmd->target].issue_tail->next = as_cmd;
		hostdata->dev[cmd->target].issue_tail = as_cmd;
		sti ();
#if (DEBUG & DEBUG_QUEUES)
		where = "tail";
#endif
	    }
	    hostdata->stats.queues += 1;
#if (DEBUG & DEBUG_QUEUES)
	    printk (KERN_DEBUG "scsi%d: command put on %s of queue for target %d\n",
			cmd->host->host_no, where, cmd->target);
#endif
	    acornscsi_startsendcommand (cmd->host);
	    return 0;
	}
    }

    printk (KERN_CRIT "scsi%d: could not find a free slot for command?\n", cmd->host->host_no);

    cmd->result = DID_ERROR << 16;
    done (cmd);
    return -1;
}

int acornscsi_abort (Scsi_Cmnd * cmd)
{
    struct Scsi_Host *instance = cmd->host;
    AS_Host *hostdata = (AS_Host *) instance->hostdata;
    struct as_cmnd *ascmd, *asprev;

    hostdata->stats.aborts += 1;

    printk (KERN_WARNING "acornscsi_abort: ");
    print_sbic_status(sbic_read (instance, ASR), sbic_read (instance, SSR),
			sbic_read (instance, CMNDPHASE));
    acornscsi_printstatus (instance, cmd->target);

    /*
     * Check to see if the command hasn't been issued.  If so, just remove it from the queue.
     */
    for (ascmd = hostdata->dev[cmd->target].issue_head, asprev = NULL; ascmd;
		asprev = ascmd, ascmd = ascmd->next) {

	if (ascmd->cmd == cmd) {
	    if (ascmd == hostdata->dev[cmd->target].issue_tail)
		hostdata->dev[cmd->target].issue_tail = asprev;
	    if (asprev)
	        asprev->next = ascmd->next;
	    else
	        hostdata->dev[cmd->target].issue_head = ascmd->next;
            ascmd->next = NULL;
	    sti();

	    ascmd->cmd->result = DID_ABORT << 16;
#if (DEBUG & DEBUG_ABORT)
	    printk (KERN_DEBUG "scsi%d: abort removed command from issue queue\n", instance->host_no);
#endif
	    acornscsi_done (instance, ascmd);
	    return SCSI_ABORT_SUCCESS;
	}
    }

    /*
     * If the command is connected, try to abort it
     */
    if (hostdata->connected && hostdata->connected->cmd == cmd) {
    	int asr;
	sti();
	ascmd = hostdata->connected;
#if (DEBUG & DEBUG_ABORT)
	printk (KERN_DEBUG "scsi%d: aborting connected command\n", instance->host_no);
#endif
	ascmd->cmd->result = DID_ABORT << 16;
#if 0
	/* Hmm... doesn't seem to work.. */
	add_msg (hostdata, 1, ABORT);
	sbic_write (instance, CMND, CMND_ASSERTATN);
	return SCSI_ABORT_PENDING;
#else
	/*
	 * Very dirty abort - initiator disconnects from bus
	 * WD33C93 doesn't like asserting ATN...
	 */
	hostdata->connected = NULL;
	acornscsi_done (instance, ascmd);
	sbic_write (instance, CMND, CMND_DISCONNECT);
	/*
	 * Now flush fifo
	 */
	do {
	    asr = sbic_read(instance, ASR);
	    if (asr & ASR_DBR)
		(void)sbic_read(instance, DATA);
	} while (!(asr & ASR_INT));
	return SCSI_ABORT_SUCCESS;
#endif
    }

    /*
     * If the command is disconnected, and there are no connected commands, reconnect
     * and send abort message
     */
    for (ascmd = hostdata->dev[cmd->target].disconnected, asprev = NULL; asprev;
		asprev = ascmd, ascmd = ascmd->next) {
	if (ascmd->cmd == cmd) {
#if (DEBUG & DEBUG_ABORT)
	    printk (KERN_DEBUG "scsi%d: aborting disconnected command\n", instance->host_no);
#endif
	    if (asprev)
		asprev->next = ascmd->next;
	    else
	        hostdata->dev[cmd->target].disconnected = ascmd->next;
	    ascmd->next = NULL;

	    hostdata->aborting = ascmd;
	    sti ();

	    acornscsi_startsendcommand (instance);
	    return SCSI_ABORT_PENDING;
	}
    }

    sti();
    printk (KERN_WARNING "scsi%d: SCSI command probably completed successfully before abortion\n",
			instance->host_no);
while (1);
    return SCSI_ABORT_NOT_RUNNING;
}

int acornscsi_reset (Scsi_Cmnd * SCpnt, unsigned int reset_flags)
{
    struct Scsi_Host *instance = SCpnt->host;
    AS_Host *hostdata = (AS_Host *)instance->hostdata;
    int device_reset = -1;
    
    hostdata->stats.resets += 1;

#if (DEBUG & DEBUG_RESET)
    printk (KERN_WARNING "acornscsi_reset\n");
    print_sbic_status(sbic_read (instance, ASR), sbic_read (instance, SSR), sbic_read (instance, CMNDPHASE));
    acornscsi_printstatus (instance, SCpnt->target);
#endif
#if 0
    /*
     * Work out what to do.  We attempt to find out where the command is only
     * if we are just trying to reset a device.
     */
    if (!reset_flags & (SCSI_RESET_SUGGEST_BUS_RESET|SCSI_RESET_SUGGEST_HOST_RESET)) {
	struct as_cmnd *ascmd;
	/*
	 * If it is on the connecting list it should time out.
	 * Maybe the HW has gone wrong? (let it timeout for a while longer).
	 */
	if (hostdata->connecting && hostdata->connecting->cmd == SCpnt)
	    device_reset = -1;
	else
	/*
	 * If it is on the aborting list, we really want to
	 * reset the device.
	 */
	if (hostdata->aborting && hostdata->aborting->cmd == SCpnt)
	    device_reset = SCpnt->target;
	else
	/*
	 * If it is on the connected list, as above.
	 */
	if (hostdata->connected && hostdata->connected->cmd == SCpnt)
	    device_reset = -1;
	else {
	/*
	 * Check the disconnected queue for the device.
	 */
	    for (ascmd = hostdata->dev[SCpnt->target].disconnected; ascmd; ascmd = ascmd->next)
		if (ascmd->cmd == SCpnt) {
		    device_reset = SCpnt->target;
		    break;
		}
	    if (!ascmd)
		printk (KERN_ERR "scsi%d: can't find command to reset (dev %d, lun %d)\n",
			instance->host_no, SCpnt->target, SCpnt->lun);
	}
    }
#endif
    if (device_reset == -1) {
	struct as_cmnd *ascmd;
	int i;

	/*
	 * do hard reset.  This resets all devices on this host, and so we
	 * must set the reset status on all commands.
	 */
	acornscsi_resetcard (instance);

	if (hostdata->connecting) {
	    ascmd = hostdata->connecting;
	    hostdata->connecting = NULL;
	    ascmd->cmd->result = DID_RESET << 16;
	    if (SCpnt == ascmd->cmd)
		SCpnt = NULL;
	    acornscsi_done (instance, ascmd);
	}
	if (hostdata->aborting) {
	    ascmd = hostdata->aborting;
	    hostdata->aborting = NULL;
	    ascmd->cmd->result = DID_RESET << 16;
	    if (SCpnt == ascmd->cmd)
		SCpnt = NULL;
	    acornscsi_done (instance, ascmd);
	}
	if (hostdata->connected) {
	    ascmd = hostdata->connected;
	    hostdata->connected = NULL;
	    ascmd->cmd->result = DID_RESET << 16;
	    if (SCpnt == ascmd->cmd)
		SCpnt = NULL;
	    acornscsi_done (instance, ascmd);
	}

	for (i = 0; i < 8; i++) {
	    struct as_cmnd *as_next;

	    ascmd = hostdata->dev[i].disconnected;
	    hostdata->dev[i].disconnected = NULL;
	    hostdata->dev[i].sync_stat = SDTR_UNSET;
	    hostdata->dev[i].sync_xfer = SYNCHTRANSFER_2DBA;
	    hostdata->dev[i].busy = 0;

	    for (; ascmd; ascmd = as_next) {
		ascmd->cmd->result = DID_RESET << 16;
		as_next = ascmd->next;
		if (SCpnt == ascmd->cmd)
		    SCpnt = NULL;
		acornscsi_done (instance, ascmd);
	    }
	}

	if (SCpnt) {
	    SCpnt->result = DID_RESET << 16;
	    SCpnt->scsi_done (SCpnt);
	}
	return SCSI_RESET_BUS_RESET | SCSI_RESET_HOST_RESET | SCSI_RESET_SUCCESS;
    } else {
	struct as_cmnd *ascmd, *as_next;

	/*
	 * Reset specific device ???
	 */
	if (hostdata->aborting && hostdata->aborting->cmd == SCpnt) {
	    ascmd = hostdata->aborting;
	    hostdata->aborting = hostdata->aborting->next;
	    ascmd->cmd->result = DID_RESET << 16;
	    acornscsi_done (instance, ascmd);
	}
	if (hostdata->connected && hostdata->connected->cmd == SCpnt) {
	    ascmd = hostdata->connected;
	    hostdata->connected = NULL;
	    ascmd->cmd->result = DID_RESET << 16;
	    acornscsi_done (instance, ascmd);
	}

	for (ascmd = hostdata->dev[device_reset].disconnected; ascmd; ascmd = as_next) {
	    ascmd->cmd->result = DID_RESET << 16;
	    as_next = ascmd->next;
	    acornscsi_done (instance, ascmd);
	}
	hostdata->dev[device_reset].disconnected = NULL;
	hostdata->dev[device_reset].sync_stat = SDTR_UNSET;
	hostdata->dev[device_reset].sync_xfer = SYNCHTRANSFER_2DBA;

	return SCSI_RESET_SUCCESS;
    }
}

/*==============================================================================================
 * initialisation & miscellaneous support
 */
static struct expansion_card *ecs[MAX_ECARDS];

static
void acornscsi_init (struct Scsi_Host *instance)
{
    AS_Host *hostdata = (AS_Host *) instance->hostdata;
    int i;

    if (!the_template) {
	the_template = instance->hostt;
	first_instance = instance;
    }

    hostdata->connecting = NULL;
    hostdata->aborting = NULL;
    hostdata->connected = NULL;
    hostdata->busy = 0;
    hostdata->phase = 0;
    hostdata->stats.queues	= 0;
    hostdata->stats.removes	= 0;
    hostdata->stats.fins	= 0;
    hostdata->stats.reads	= 0;
    hostdata->stats.writes	= 0;
    hostdata->stats.aborts	= 0;
    hostdata->stats.resets	= 0;
    hostdata->stats.disconnects = 0;

    for (i = 0; i < CAN_QUEUE; i++) {
	hostdata->queuedcommands[i].used       = 0;
	hostdata->queuedcommands[i].cmd        = NULL;
	hostdata->queuedcommands[i].next       = NULL;
	hostdata->queuedcommands[i].bufaddr[0] = i * 0x2000;
	hostdata->queuedcommands[i].bufaddr[1] = i * 0x2000 + 0x1000;
	hostdata->queuedcommands[i].buflen[0]  = 0;
	hostdata->queuedcommands[i].buflen[1]  = 0;
    }

    for (i = 0; i < 8; i++) {
	hostdata->dev[i].issue_head	= NULL;
	hostdata->dev[i].issue_tail	= NULL;
	hostdata->dev[i].disconnected	= NULL;
	hostdata->dev[i].busy		= 0;
	hostdata->dev[i].sync_stat	= SDTR_UNSET;
	hostdata->dev[i].sync_xfer	= SYNCHTRANSFER_2DBA;
	hostdata->dev[i].disconnect_ok  = 1;
    }

    acornscsi_resetcard (instance);
}

int acornscsi_detect(Scsi_Host_Template * tpnt)
{
    static const card_ids acornscsi_cids[] = { ACORNSCSI_LIST, { 0xffff, 0xffff } };
    int i, count = 0;
    struct Scsi_Host *instance;

    tpnt->proc_dir = &proc_scsi_acornscsi;

    for (i = 0; i < MAX_ECARDS; i++)
    	ecs[i] = NULL;

    ecard_startfind ();

    while(1) {
	ecs[count] = ecard_find(0, acornscsi_cids);
	if (!ecs[count])
	    break;

	if (ecs[count]->irq == 0xff) {
	    printk ("scsi: WD33C93 does not have IRQ enabled - ignoring\n");
	    continue;
	}

	instance = scsi_register (tpnt, sizeof(AS_Host));
	instance->io_port = ecard_address (ecs[count], ECARD_MEMC, 0);
	instance->irq = ecs[count]->irq;

	request_region (instance->io_port, 2048, "acornscsi(ram)");
	request_region (instance->io_port + 0x800, 2, "acornscsi(sbic)");
	request_region (instance->io_port + 0xc00, 256, "acornscsi(dmac)");
	request_region (POD_SPACE(instance->io_port) + 0x800, 1, "acornscsi(intr)");
	request_region (POD_SPACE(instance->io_port) + 0xc00, 1, "acornscsi(page)");

	if (request_irq(instance->irq, acornscsi_intr, SA_INTERRUPT, "acornscsi", instance)) {
	    printk("scsi%d: IRQ%d not free, interrupts disabled\n",
		instance->host_no, instance->irq);
	}

	ecard_claim(ecs[count]); /* Must claim here - card produces irq on reset */
	acornscsi_init(instance);

	++count;
    }
    return count;
}

const char *acornscsi_info(struct Scsi_Host *host)
{
    static char string[100], *p;

    p = string;
    
    p += sprintf (string, "%s at port %X irq %d v%d.%d.%d",
		host->hostt->name, host->io_port, host->irq,
		VER_MAJOR, VER_MINOR, VER_PATCH);
#ifdef SCSI2_SYNC
    p += sprintf (p, " SYNC");
#endif
#ifdef SCSI2_TAG
    p += sprintf (p, " TAG");
#endif
#ifdef SCSI2_LINK
    p += sprintf (p, " LINK");
#endif

    return string;
}

int acornscsi_proc_info(char *buffer, char **start, off_t offset,
			int length, int host_no, int inout)
{
    int len, pos, begin, first;
    struct Scsi_Host *host = scsi_hostlist;
    AS_Host *hostdata;
    Scsi_Device *scd;

    while (host) {
	if (host->host_no == host_no)
	    break;
	host = host->next;
    }

    hostdata  = (AS_Host *) host->hostdata;
    
    if (inout == 1)
	return -EINVAL;

    begin = 0;
    pos = sprintf(buffer,
			"Acorn SCSI driver version %d.%d.%d\n", VER_MAJOR, VER_MINOR, VER_PATCH);

    len = sprintf(buffer+pos,
    			"Address: %08X          IRQ : %d\n"
    			"SBIC   : WD33C93A          DMAC: uPC71071\n\n"
			"Statistics:\n", host->io_port, host->irq);
    pos += len;
    
    len = sprintf(buffer+pos,
			"Queued commands: %-10ld    Issued commands: %-10ld\n"
			"Done commands  : %-10ld    Reads          : %-10ld\n"
			"Writes         : %-10ld    Others         : %-10ld\n"
			"Disconnects    : %-10ld    Aborts         : %-10ld\n"
			"Resets         : %-10ld\n",
				hostdata->stats.queues,
				hostdata->stats.removes,
				hostdata->stats.fins,
				hostdata->stats.reads,
				hostdata->stats.writes,
				hostdata->stats.miscs,
				hostdata->stats.disconnects,
				hostdata->stats.aborts,
				hostdata->stats.resets);
    pos += len;

    scd = scsi_devices;

    first = 1;
    
    while (scd) {
	if (scd->host == host) {
    	    if (first) {
		len = sprintf(buffer+pos, "\nAttached devices:\n");
		pos += len;
		first = 0;
	    }

	    proc_print_scsidevice (scd, buffer, &len, pos);
	    pos += len;
	    pos += sprintf (buffer+pos, "Extensions: ");
	    if (scd->tagged_supported)
		pos += sprintf (buffer+pos, "TAG %sabled [%d] ",
			scd->tagged_queue ? "en" : "dis", scd->current_tag);
	    pos += sprintf (buffer+pos, "\n");

	    if (pos + begin < offset) {
	    	begin += pos;
	    	pos = 0;
	    }
	    if (pos + begin > offset + length)
	    	break;
	}
	scd = scd->next;
    }

    if (first) {
	len = sprintf(buffer+pos,
 			"Attached devices: none\n");
 	pos += len;
    }
    
    *start = buffer + (offset - begin);
    pos -= offset - begin;
    if (pos > length)
	pos = length;

    return pos;
}

#ifdef MODULE

Scsi_Host_Template driver_template = ACORNSCSI_3;

#define SCSI_CLEANUP \
	{ int i; for (i = 0; i < MAX_ECARDS; i++) if (ecs[i]) ecard_release(ecs[i]); }

#include "scsi_module.c"
#endif
