/*
 * Simple Power Manager
 * Copyright (c) Compaq Computer Corporation, 1998, 1999
 *
 * Author: Carl Waldspurger
 *
 * $Log: powermgr.c,v $
 * Revision 2.4  2000/01/19  18:25:23  kerr
 * More copyright conformance.
 *
 * Revision 2.3  1999/12/23  01:28:09  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.2  1999/05/12  18:01:06  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.1.2.1  1999/04/21  01:26:16  caw
 * Changed powerbutton initialization to powerbutton_powermgr_register().
 *
 * Revision 2.1  1998/11/25  01:28:02  kerr
 * Added clock speed change capability (with loadable module ioctl interface).
 * Does not depend on sleep mode for implementation.
 *
 * Revision 2.0  1998/06/24  21:45:22  kramer
 * Major Bump: Move all version numbers up to 2.0 for external release.
 * -drew
 *
 * Revision 1.8  1998/06/16  19:19:47  kramer
 * Added copyright information
 *
 * Revision 1.7  1998/04/14  02:25:43  caw
 * Implemented updated suspend_check() interfaces.  Added support for
 * automatic timeout-based suspends based on kernel timers.
 *
 * Revision 1.6  1998/03/28  00:14:48  caw
 * Export symbols to allow calls from modules.  Minor cleanup.
 *
 * Revision 1.5  1998/03/27  02:01:11  caw
 * Added missing include of "config.h", fixing missing procfs behavior.
 *
 * Revision 1.4  1998/03/26  01:01:55  kerr
 * Power management using the official interface.  It works!
 *
 * Revision 1.3  1998/03/26  00:34:26  caw
 * Updated interface. Changed suspend/resume ordering.
 *
 * Revision 1.2  1998/03/21  02:04:38  caw
 * Fixed procfs problem.  Most functionality still untested.
 *
 * Revision 1.1  1998/03/21  02:02:18  caw
 * Initial revision.
 *
 */

/*
 * TODO
 *
 *   - should kmalloc/copy client name fields
 *   - unregisters can cause impl not to meet resume/suspend ordering spec
 *     fix by using list of clients instead of array
 *
 */

/*
 * debugging 
 *
 */

#define	POWERMGR_DEBUG			(1)
#define	POWERMGR_DEBUG_VERBOSE		(0)

/*
 * includes
 *
 */


#include <linux/config.h>
#include <linux/module.h>

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/timer.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

#include <linux/powermgr.h> 

/*
 * constants
 *
 */

#define	POWERMGR_NAME			"powermgr"
#define	POWERMGR_NAME_VERBOSE		"Power Manager"
#define	POWERMGR_VERSION_STRING		"$Revision: 2.4 $"

#define	POWERMGR_NCLIENTS		(32)

/*
 * types
 *
 */

typedef struct {
  /* clients */
  powermgr_client clients[POWERMGR_NCLIENTS];

  /* periodic timer */
  struct timer_list timer;
  int check_jiffies;
  int idle_jiffies;

  /* statistics */
  ulong suspend_check_count;
  ulong suspend_count;
  ulong resume_count;
} powermgr_state;

/*
 * globals
 *
 */

static powermgr_state powermgr_global;

#ifdef	CONFIG_PROC_FS
static int powermgr_get_info(char *, char **, off_t, int, int);
static struct proc_dir_entry powermgr_proc_entry = {
  0, 8, "powermgr", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, powermgr_get_info,
};
#endif	/* CONFIG_PROC_FS */


/*
 * exported symbols
 *
 */

static struct symbol_table powermgr_syms = {
#include <linux/symtab_begin.h>
        X(powermgr_suspend_check),
        X(powermgr_suspend),
        X(powermgr_resume),
        X(powermgr_register),
        X(powermgr_unregister),
#include <linux/symtab_end.h>
};

/*
 * powermgr_client operations
 *
 */

static inline int powermgr_client_valid(const powermgr_client *c)
{
  /*
   * modifies: nothing
   * effects:  Returns TRUE iff c is a valid client.
   *
   */

  return(c->id != POWERMGR_ID_INVALID);
}

static inline void powermgr_client_copy(powermgr_client *to, 
					const powermgr_client *from)
{
  /*
   * modifies: to
   * effects:  Copies all fields of from into to.
   *
   */

  memcpy((void *) to, (void *) from, sizeof(powermgr_client));
}

/*
 * procfs operations
 *
 */

#ifdef	CONFIG_PROC_FS
static int powermgr_get_info(char *buf, 
			     char **start,
			     off_t offset,
			     int length,
			     int unused)
{
  powermgr_state *m = &powermgr_global;
  int len, i;

  /* format stats */
  len = sprintf(buf,
		"suspend-checks: %8lu\n"
		"suspends:       %8lu\n"
		"restores:       %8lu\n",
		m->suspend_check_count,
		m->suspend_count,
		m->resume_count);

  for (i = 0; i < POWERMGR_NCLIENTS; i++)
    {
      powermgr_client *c = &m->clients[i];
      if (powermgr_client_valid(c))
	len += sprintf(buf + len, 
		       "client %2d: %s\n",
		       c->id,
		       (c->name == NULL) ? "" : c->name);
    }

  return(len);
}
#endif	/* CONFIG_PROC_FS */

/*
 * powermgr_state operations
 *
 */

static void powermgr_state_init(powermgr_state *m)
{
  int i;

  /* initialize */
  memset((void *) m, 0, sizeof(powermgr_state));

  /* invalidate clients */
  for (i = 0; i < POWERMGR_NCLIENTS; i++)
    m->clients[i].id = POWERMGR_ID_INVALID;

  /* initialize stats */
  m->suspend_check_count = 0;
  m->suspend_count = 0;
  m->resume_count = 0;
}

/*
 * exported client operations
 *
 */

powermgr_id powermgr_register(const powermgr_client *client)
{
  /*
   * modifies: powermgr_global
   * effects:  Register as a power manager client.  The power manager
   *           will communicate via the client-specified callbacks.
   *           Returns a unique identifier for use with future calls,
   *           or POWERMGR_ID_INVALID if unable to register.
   * storage:  Data is copied from client, so it is safe to deallocate
   *           client after registering.
   * caveats:  The "id" field specified by client is ignored.
   *
   */

  powermgr_state *m = &powermgr_global;
  int i;

  /* search for free slot */
  for (i = 0; i < POWERMGR_NCLIENTS; i++)
    {
      powermgr_client *c = &m->clients[i];
      if (!powermgr_client_valid(c))
	{
	  /* copy client into available slot, update id */
	  powermgr_client_copy(c, client);
	  c->id = i;
	  return(c->id);
	}
    }

  /* no free slots */
  return(POWERMGR_ID_INVALID);
}

extern int powermgr_unregister(powermgr_id id)
{
  /*
   * modifies: powermgr_global
   * effects:  Deregisters id as a power manager client.
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

  powermgr_state *m = &powermgr_global;

  /* fail if invalid id */
  if ((id < 0) || (id > POWERMGR_ID_MAX))
    return(POWERMGR_FAILURE);

  /* fail if no currently registered client with id */
  if (!powermgr_client_valid(&m->clients[id]))
    return(POWERMGR_FAILURE);
  
  /* invalidate client */
  m->clients[id].id = POWERMGR_ID_INVALID;
  return(POWERMGR_SUCCESS);
}

/*
 * exported power management operations
 *
 */

int powermgr_suspend_check(int idle_jiffies, int *retry_jiffies)
{
  /*
   * modifies: retry_jiffies, powermgr_global
   * effects:  Returns TRUE iff all registered clients are willing
   *           to be suspended.  Clients should be willing to 
   *           suspend after an inactivity period of idle_jiffies.
   *           Otherwise returns FALSE and sets retry_jiffies to a 
   *           reasonable time delay in jiffies, after which another
   *           check may return TRUE.
   *
   */

  powermgr_state *m = &powermgr_global;
  int i, suspendible, delay;

  /* initialize */
  suspendible = 1;
  delay = 0;

  /* query each valid client */
  for (i = 0; i < POWERMGR_NCLIENTS; i++)
    {
      powermgr_client *c = &m->clients[i];
      if (powermgr_client_valid(c))
	if (c->suspend_check != NULL)
	  {
	    int check = (*(c->suspend_check))(c->cookie, idle_jiffies);
	    if (check != 0)
	      {
		/* client doesn't want to suspend */
		suspendible = 0;
		if (check > delay) 
		  delay = check;
	      }
	  }
    }

  /* update stats */
  m->suspend_check_count++;

  /* set retry time delay (measured in jiffies) */
  *retry_jiffies = delay;

  return(suspendible);
}

int powermgr_suspend(void)
{
  /*
   * modifies: powermgr_global
   * effects:  Forcibly suspends all registered clients, in the reverse order
   *           of registration (i.e. first to register suspends last).
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

  powermgr_state *m = &powermgr_global;
  int i;

  /* suspend each valid client, in reverse registration order */
  for (i = POWERMGR_NCLIENTS - 1; i >= 0; i--)
    {
      powermgr_client *c = &m->clients[i];
      if (powermgr_client_valid(c))
	if (c->suspend != NULL)
	  (void) (*(c->suspend))(c->cookie);
    }

  /* update stats */
  m->suspend_count++;

  /* always successful */
  return(POWERMGR_SUCCESS);
}

int powermgr_resume(int resume_flags)
{
  /*
   * modifies: powermgr_global
   * effects:  Resumes all registered clients, in the order of registration
   *           (i.e. first to register resumes first).
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

  powermgr_state *m = &powermgr_global;
  int i;

  /* resume each valid client, in registration order */
  for (i = 0; i < POWERMGR_NCLIENTS; i++)
    {
      powermgr_client *c = &m->clients[i];
      if (powermgr_client_valid(c))
	if (c->resume != NULL)
	  (void) (*(c->resume))(c->cookie, resume_flags);
    }

  /* update stats */
  m->resume_count++;

  /* always successful */
  return(POWERMGR_SUCCESS);
}

int powermgr_init(void)
{
  /*
   * modifies: powermgr_global
   * effects:  Initializes power manager.
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

  static int initialized = 0;

  /* initialize only once */
  if (!initialized)
    {
      extern void powerbutton_powermgr_register(void);
      extern void general_powermgr_register(void);

      /* initialize global state */
      initialized = 1;
      powermgr_state_init(&powermgr_global);

      /* register low-level clients */
      general_powermgr_register();
      powerbutton_powermgr_register();
      
      /* everything OK */
      return(POWERMGR_SUCCESS);
    }

  /* already initialized */
  return(POWERMGR_FAILURE);
}

void powermgr_init_ksyms(void)
{
  /*
   * effects:  Exports powermgr kernel symbols for modules.
   * caveats:  Separate function due to boot order restrictions.
   *
   */

#ifdef	CONFIG_MODULES
  /* export symbols */
  (void) register_symtab(&powermgr_syms);
#endif	/* CONFIG_MODULES */
}

void powermgr_init_procfs(void)
{
  /*
   * effects:  Registers /proc/powermgr device.
   * caveats:  Separate function due to boot order restrictions.
   *
   */

#ifdef	CONFIG_PROC_FS
  /* register procfs entry */
  proc_register_dynamic(&proc_root, &powermgr_proc_entry);
#endif	/* CONFIG_PROC_FS */
}

static void powermgr_suspend_timeout(unsigned long ignore)
{
  /*
   * requires: powermgr_global.timer initialized
   * modifies: powermgr_global
   * effects:  Invoked by kernel timer mechanism.
   *
   */

  extern void sa1100_sleep_mode_enter(void);
  
  powermgr_state *m = &powermgr_global;
  int retry_jiffies = 0;
  
  /* suspend if possible, otherwise wait retry_jiffies */
  if (powermgr_suspend_check(m->idle_jiffies, &retry_jiffies))
    sa1100_sleep_mode_enter();

  /* restart timer */
  if (retry_jiffies < m->check_jiffies)
    retry_jiffies = m->check_jiffies;
  m->timer.expires = jiffies + retry_jiffies;
  add_timer(&m->timer);  
}

void powermgr_start_timer(int idle_jiffies, int check_jiffies)
{
  /*
   * modifies: powermgr_global
   * effects:  Starts periodic timer to initiate suspends.  Clients are
   *           recommended to suspend after idle_jiffies of inactivity.
   *           Suspend checks will be made at most once every check_jiffies.
   * 
   */

  powermgr_state *m = &powermgr_global;

  /* save timeouts */
  m->check_jiffies = check_jiffies;
  m->idle_jiffies = idle_jiffies;

  /* initialize timer */
  init_timer(&m->timer);
  m->timer.function = powermgr_suspend_timeout;
  m->timer.data = 0;
  m->timer.expires = jiffies + check_jiffies;

  /* start timer */
  add_timer(&m->timer);
}
