/*
 * SA1100 Idle Mode Statistics
 * Copyright (c) Compaq Computer Corporation, 1998
 *
 * Author: Carl Waldspurger
 *
 * $Log: sa1100-idle-mode.c,v $
 * Revision 2.2  1999/12/23  01:28:09  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.1  1998/11/25  01:28:03  kerr
 * Added clock speed change capability (with loadable module ioctl interface).
 * Does not depend on sleep mode for implementation.
 *
 * Revision 2.0  1998/06/24  21:45:27  kramer
 * Major Bump: Move all version numbers up to 2.0 for external release.
 * -drew
 *
 * Revision 1.5  1998/06/16  19:19:48  kramer
 * Added copyright information
 *
 * Revision 1.4  1998/04/13  22:48:14  kerr
 * Fixed timing bug (reenable clock switching whenever leave idle mode).
 * No idea why this is necessary.
 *
 * Revision 1.3  1998/02/12  04:06:01  caw
 * Improved formatting.  Output header, average idle time in usec.
 *
 * Revision 1.2  1998/02/03  03:12:46  caw
 * Improved statistics output formatting.
 *
 * Revision 1.1  1998/02/03  02:57:45  caw
 * Initial revision.
 *
 */

/*
 * includes
 * 
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/config.h>

#include <asm/io.h>
#include <asm/io_reg.h>

#ifdef	CONFIG_PROC_FS
#include <linux/stat.h>
#include <linux/proc_fs.h>
#endif

int clock_switching_enabled = 1;  /* on by default */

/*
 * constants
 *
 */

#define	OSTIMER_HZ		(3686400)

#define	NIRQS			(32)
#define	NSPECIAL		(1)
#define	SPECIAL_NO_IRQ		(32)
#define	NSTATS			(NIRQS + NSPECIAL)

/*
 * types
 *
 */

typedef unsigned long long uint64;

typedef struct {
  uint64 duration;
  ulong  count;
} idle_mode_stats;

/*
 * globals
 *
 */

static ulong enter_count, enter_time, enter_flag;
static idle_mode_stats stats[NSTATS];
static ulong init_flag = 0;

#ifdef	CONFIG_PROC_FS
static int idle_get_info(char *, char **, off_t, int, int);
static struct proc_dir_entry idle_proc_entry = {
  0, 4, "idle", S_IFREG | S_IRUGO, 1, 0, 0, 0, NULL, idle_get_info,
};
#endif	/* CONFIG_PROC_FS */

/*
 * operations
 *
 */

#ifdef	CONFIG_PROC_FS
static inline ulong ticks_to_usec(ulong ticks)
{
  uint64 cvt = ((uint64) ticks * 1000000) / OSTIMER_HZ;
  return((ulong) cvt);
}

static int idle_format_stats(char *buf, const idle_mode_stats *stats)
{
  ulong mticks, avg_ticks, avg_usec;

  mticks = (stats->duration >> 20);
  avg_ticks = (ulong) (stats->duration / stats->count);
  avg_usec  = ticks_to_usec(avg_ticks);

  return(sprintf(buf, 
		 "%8lu %10lu%c %8lu %6lu\n",
		 stats->count,
		 (mticks > 0) ? mticks  : (ulong) stats->duration,
		 (mticks > 0) ? 'm' : ' ',
		 avg_ticks,
		 avg_usec));
}

static int idle_get_info(char *buf,
			 char **start,
			 off_t offset,
			 int length, 
			 int unused)
{
  idle_mode_stats total;
  int len, i;

  /* initialize */
  total.duration = 0;
  total.count = 0;

  /* format header */
  len = sprintf(buf, "event  wakeups  idle-ticks  average   usec\n" );

  /* format per-irq stats */
  for (i = 0; i < NSTATS; i++)
    if (stats[i].count > 0)
      {
	if (i == SPECIAL_NO_IRQ)
	  len += sprintf(buf + len, "NO-irq");
	else
	  len += sprintf(buf + len, "irq-%2d", i);
	
	len += idle_format_stats(buf + len, &stats[i]);

	/* accumulate totals */
	total.count += stats[i].count;
	total.duration += stats[i].duration;
      }

  /* format totals */
  if (total.count > 0)
    {
      len += sprintf(buf + len, "totals");
      len += idle_format_stats(buf + len, &total);
    }

  return(len);
}
#endif	/* CONFIG_PROC_FS */

#ifdef	CONFIG_MODULES

#include <linux/module.h>
#include <linux/version.h>

static struct symbol_table idle_syms = {
#include <linux/symtab_begin.h>
        X(clock_switching_enabled),
#include <linux/symtab_end.h>
};
#endif

void sa1100_idle_mode_init(void)
{
  int i;

  /* reset only once */
  if (init_flag) return;
  init_flag = 1;

#if	0
  /* XXX debugging */
  printk("** sa1100_idle_mode_init()\n");
#endif

  /* initialize globals */
  for (i = 0; i < NSTATS; i++)
    {
      stats[i].count    = 0;
      stats[i].duration = 0;
    }
  enter_count = 0;
  enter_time  = 0;
  enter_flag  = 0;

#ifdef	CONFIG_PROC_FS
  /* register procfs device */
  proc_register_dynamic(&proc_root, &idle_proc_entry);
#endif	/* CONFIG_PROC_FS */
#ifdef	CONFIG_MODULES
  (void) register_symtab(&idle_syms);
#endif	/* CONFIG_MODULES */
}

void sa1100_idle_mode_leave(int irq)
{
  if (enter_flag)
    {
      ulong now, duration;

      /* reset flag */
      enter_flag = 0;

      /* measure idle duration */
      now = inl(OSCR_V);  
      if (now < enter_time)
	duration = enter_time - now;
      else
	duration = now - enter_time;

      /* update stats */
      stats[irq].count++;
      stats[irq].duration += (uint64) duration;
    }
}

void sa1100_idle_mode_enter(void)
{
  extern void sa1100_idle_mode(void);

  if (!init_flag)
    sa1100_idle_mode_init();

  /* capture current time */
  enter_count++;
  enter_flag = 1;
  enter_time = inl(OSCR_V);

  /* enter idle mode */
  sa1100_idle_mode();

  if (clock_switching_enabled) 
      sa1100_idle_mode_finish();

  /* tally non-interrupt wakeups */
  if (enter_flag)
    sa1100_idle_mode_leave(SPECIAL_NO_IRQ);
}

