/*
 *  linux/arch/arm/kernel/time.c
 *
 *  Copyright (C) 1991, 1992, 1995  Linus Torvalds
 *  Modifications for ARM processor (C) 1994, 1995, 1996 Russell King
 *
 * This file contains the ARM-specific time handling details:
 * reading the RTC at bootup, etc...
 *
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/config.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/timex.h>
#include <asm/irq-no.h>
#include <asm/hardware.h>

#ifndef BCD_TO_BIN
#define BCD_TO_BIN(val) ((val)=((val)&15) + ((val)>>4)*10)
#endif

#ifndef BIN_TO_BCD
#define BIN_TO_BCD(val) ((val)=(((val)/10)<<4) + (val)%10)
#endif

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static inline unsigned long mktime(unsigned int year, unsigned int mon,
	unsigned int day, unsigned int hour,
	unsigned int min, unsigned int sec)
{
	if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
		mon += 12;	/* Puts Feb last since it has leap day */
		year -= 1;
	}
	return (((
		    (unsigned long)(year/4 - year/100 + year/400 + 367*mon/12 + day) +
		      year*365 - 719499
		    )*24 + hour /* now have hours */
		   )*60 + min /* now have minutes */
		  )*60 + sec; /* finally seconds */
}

#include <asm/arch/io_reg.h>
#include <asm/arch/time.h>

extern int setup_arm_irq(int, struct irqaction *);

static unsigned long do_gettimeoffset(void)
{
	return gettimeoffset ();
}

void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;

	save_flags_cli (flags);
	*tv = xtime;
	tv->tv_usec += do_gettimeoffset();
	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
	restore_flags(flags);
}

void do_settimeofday(struct timeval *tv)
{
  void set_cmos_time(unsigned int seconds);
	cli ();
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_state = TIME_BAD;
	time_maxerror = 0x70000000;
	time_esterror = 0x70000000;
	set_cmos_time(tv->tv_sec);
	sti ();
}
#if 0
/*
 * In order to set the CMOS clock precisely, set_rtc_mmss has to be
 * called 500 ms after the second nowtime has started, because when
 * nowtime is written into the registers of the CMOS clock, it will
 * jump to the next second precisely 500 ms later. Check the Motorola
 * MC146818A or Dallas DS12887 data sheet for details.
 */
int set_rtc_mmss(unsigned long nowtime)
{
	int retval = 0;
	return retval;
}

/* last time the cmos clock got updated */
static long last_rtc_update = 0;
#endif
/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick.
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	if (reset_timer ())
		do_timer(regs);
#if 0
	/* If we have an externally synchronized linux clock, then update
	 * CMOS clock accordingly every ~11 minutes.  Set_rtc_mmss() has to be
	 * called as close as possible to 500 ms before the new second starts.
	 */
	if (time_state != TIME_BAD && xtime.tv_sec > last_rtc_update + 660 &&
	    xtime.tv_usec > 50000 - (tick >> 1) &&
	    xtime.tv_usec < 50000 + (tick >> 1)) {
		if (set_rtc_mmss(xtime.tv_sec) == 0)
			last_rtc_update = xtime.tv_sec;
		else
			last_rtc_update = xtime.tv_sec - 600; /* do it again in 60 s */
	}
#endif
}

static struct irqaction irqtimer0 = { timer_interrupt, 0, 0, "timer", NULL, NULL};

void time_init(void)
{
#ifdef CONFIG_POWERMGR
  static void time_powermgr_register(void);
#endif /* CONFIG_POWERMGR */

  setup_timer();
        xtime.tv_sec = get_cmos_time();
	xtime.tv_usec = 0;

	setup_arm_irq(IRQ_TIMER0, &irqtimer0);
#ifdef CONFIG_POWERMGR
  time_powermgr_register();
  rtc_init();
#endif /* CONFIG_POWERMGR */
}

#ifdef CONFIG_POWERMGR

/* power management functions */
#include <linux/powermgr.h>

static unsigned long time_regs[1];  /* storage of timer 0 registers during sleep */

static int time_suspend_check(void *unused, int idle_jiffies)
{
  return(0);
}

static int time_suspend(void *cookie)
{
  /* we are assuming that no match is happening 
   * ie we could lose a match here
   */
  /* XXX should we be turning off enable of interrupts? */
  unsigned long *time_regs = (unsigned long *)cookie;
  time_regs[0] = inl(OIER_V) & 0x1;  /* match on timer 0 */
  disable_irq(IRQ_OST0);
  return(1); /* YYY - values ignored for now */
}

static int time_resume(void *cookie, int resume_flags)
{
  unsigned long *time_regs = (unsigned long *)cookie;
  reset_timer();
  outl(time_regs[0], OIER_V);
  enable_irq(IRQ_OST0);
  return(1); /* YYY - values ignored for now */
}

static void time_powermgr_register(void)
{
  powermgr_client client;
  powermgr_id time_powermgr_id; 

  client.suspend_check = time_suspend_check;
  client.suspend = time_suspend;
  client.resume = time_resume;
  client.cookie = time_regs;
  client.name = "Time modules (OS Timer 0)";
  client.uamps_idle = 0; /* YYY ? */
  client.uamps_sleep = 0; /* YYY ? */
  time_powermgr_id = powermgr_register(&client);
  if (time_powermgr_id == POWERMGR_ID_INVALID) {
    printk("Failed to register %s module for power management!!!\n", client.name);
    return;
  }
}

#endif /* CONFIG_POWERMGR */
