/*
 *  linux/arch/arm/mm/fault.c
 *
 *  Copyright (C) 1995  Linus Torvalds
 *  Modifications for ARM processor (c) 1995, 1996 Russell King
 */

#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/unaligned.h>
#include <asm/proc/unaligned.h>


#define FAULT_CODE_READ		0x02
#define FAULT_CODE_USER		0x01

extern void die_if_kernel(char *msg, struct pt_regs *regs, unsigned int err, unsigned int ret);

static void
page_fault (unsigned long addr, int mode, struct pt_regs *regs)
{
    struct task_struct *tsk = current;
    struct mm_struct *mm = tsk->mm;
    struct vm_area_struct *vma;

    down(&mm->mmap_sem);
    vma = find_vma (mm, addr);
    if (!vma)
	goto bad_area;
    if (addr >= vma->vm_start)
	goto good_area;
    if (!(vma->vm_flags & VM_GROWSDOWN) || expand_stack (vma, addr))
	goto bad_area;

    /*
     * Ok, we have a good vm_area for this memory access, so
     * we can handle it..
     */
good_area:
    if (mode & FAULT_CODE_READ) { /* read? */
	if (!(vma->vm_flags & (VM_READ|VM_EXEC)))
	    goto bad_area;
    } else {
	if (!(vma->vm_flags & VM_WRITE))
	    goto bad_area;
    }
    handle_mm_fault (vma, addr, !(mode & FAULT_CODE_READ));
    up(&mm->mmap_sem);
    return;

    /*
     * Something tried to access memory that isn't in our memory map..
     * Fix it, but check if it's kernel or user first..
     */
bad_area:
    up(&mm->mmap_sem);
    if (mode & FAULT_CODE_USER) {
	tsk->tss.error_code = mode;
	tsk->tss.trap_no = 14;
#ifdef DEBUG
	printk ("SEGV: %s (%d): %p", current->comm, current->pid, (void *)addr);
	show_regs (regs);
	c_backtrace (regs->ARM_fp, regs->ARM_cpsr);
#endif
	force_sig(SIGSEGV, tsk);
	return;
    } else {
	/*
	 * Oops. The kernel tried to access some bad page. We'll have to
	 * terminate things with extreme prejudice.
	 */
    	pgd_t *pgd;
	if (addr < PAGE_SIZE)
	    printk (KERN_ALERT "Unable to handle kernel NULL pointer dereference");
	else
	    printk (KERN_ALERT "Unable to handle kernel paging request");
	printk (" at virtual address %08lx\n", addr);
	printk (KERN_ALERT "current->tss.memmap = %08lX\n", tsk->tss.memmap);
	pgd = pgd_offset (mm, addr);
	printk (KERN_ALERT "*pgd = %08lx", pgd_val (*pgd));
	if (!pgd_none (*pgd)) {
		pmd_t *pmd;
		pmd = pmd_offset (pgd, addr);
		printk (", *pmd = %08lx", pmd_val (*pmd));
		if (!pmd_none (*pmd))
			printk (", *pte = %08lx", pte_val (*pte_offset (pmd, addr)));
	}
	printk ("\n");
	die_if_kernel ("Oops", regs, mode, SIGKILL);
	do_exit (SIGKILL);
    }
}

/*
 * Handle a data abort.  Note that we have to handle a range of addresses
 * on ARM2/3 for ldm.  If both pages are zero-mapped, then we have to force
 * a copy-on-write
 */
asmlinkage void
do_DataAbort (unsigned long addr, int fsr, int error_code, struct pt_regs *regs)
{
	if (user_mode(regs))
		error_code |= FAULT_CODE_USER;

#define DIE(signr,nam)\
		force_sig(signr, current);\
		die_if_kernel(nam, regs, fsr, signr);\
		break;

	switch (fsr & 15) {
	case 2:
		DIE(SIGKILL, "Terminal exception")
	case 0:
		DIE(SIGSEGV, "Vector exception")
	case 1:
	case 3:
	  //		DIE(SIGBUS, "Alignment exception")
		do_alignfixup(addr,fsr,error_code,regs);
		  break;
	case 12:
	case 14:
		DIE(SIGBUS, "External abort on translation")
	case 9:
	case 11:
		DIE(SIGSEGV, "Domain fault")
	case 13:/* permission fault on section */
#ifndef DEBUG
		show_regs(regs);
		{
			unsigned int i, j, a;
			a = regs->ARM_sp;
			for (j = 0; j < 10; j++) {
				printk ("%08x: ", a);
				for (i = 0; i < 8; i += 1, a += 4)
					printk ("%08lx ", *(unsigned long *)a);
				printk ("\n");
			}
		}
#endif
		DIE(SIGSEGV, "Permission fault")

	case 15:/* permission fault on page */
	case 5:	/* page-table entry descriptor fault */
	case 7:	/* first-level descriptor fault */
		page_fault (addr, error_code, regs);
		break;
	case 4:
	case 6:
		DIE(SIGBUS, "External abort on linefetch")
	case 8:
	case 10:
		DIE(SIGBUS, "External abort on non-linefetch")
	}
}

asmlinkage int
do_PrefetchAbort (unsigned long addr, struct pt_regs *regs)
{
#if 0
	/* does this still apply ? */
	if (the memc mapping for this page exists - can check now...) {
		printk ("Page in, but got abort (undefined instruction?)\n");
		return 0;
	}
#endif
	page_fault (addr, FAULT_CODE_USER|FAULT_CODE_READ, regs);
	return 1;
}

// Code to do alignment fixups for ARM..
// NOTE: we really should restrict this to mode SVC_32 as
// this is really expensive...

#include <linux/sysctl.h>

/*
 * Possible values = 0 - align, 1 - align & report, 2 - use ROR (no handler)
 */
int arm_mm_alignment_enabled = 2;

int proc_do_arm_mm_alignment_handler(ctl_table *ctl,
				     int write, struct file *filp,
				     void *buffer, size_t *lenp)
{
  int val = arm_mm_alignment_enabled;
  int retv;

  retv = proc_dointvec(ctl, write, filp, buffer, lenp);

  if ( write ) {
    if ( arm_mm_alignment_enabled != val ) {
      int old_flags;
      int new_flags;
      asm("mrc	p15, 0, %0, c1, c0" : "=r" (old_flags));
      new_flags = old_flags;

      printk(KERN_INFO "sysctl: setting arm_mm_alignment_enabled to %d\n",  arm_mm_alignment_enabled);
      switch (arm_mm_alignment_enabled ) {
      case 0:
      case 1:
	/*
	 * Enable alignment - 
	 */
	new_flags = old_flags | (0x2);
	break;
      case 2:
	/*
	 * disable alignment
	 */
	new_flags = old_flags & (~0x2);
	break;
      default:
	/*
	 * Improper specification - don't change
	 */
	retv = val;
	break;
      }
      if ( new_flags != old_flags ) { 
	asm("mcr	p15, 0, %0, c1, c0" : : "r" (new_flags));
      }
    }
  }

  return retv;
}


STATS_MISALIGNED Tmisalignstat_stats={0,0,0,0,0,0,0};

void misalign_init()
{
  printk("-ARMLinux 16/32 bit Alignment Handler v0.9 - sanm@corelcomputer.com\n");

  // There actually *is* other stuff that should be done here... but...
	return;
}

// Macros to determine if the access should be done in user or kernel space.

#define get_checked(ptr) \
  (regs->uregs[15] < 0xc0000000) ? get_user((ptr)) : get_unaligned((ptr))

#define put_checked(ptr,val) \
  (regs->uregs[15] < 0xc0000000) ? \
  put_user((ptr),(val)) : \
  put_unaligned((ptr),(val))

asmlinkage int
do_alignfixup(unsigned long addr,int fsr,int error_code,struct pt_regs *regs)
{
	unsigned int UI_inst;      	// Instruction that caused fault  
      	unsigned int Rn;		// The base register
       	unsigned int Rd;		// Destination register
        unsigned int Rm;	        // Offset register
     	unsigned int UI_effectiveaddr;
      	unsigned int UI_offsetdelta;	// Addressing mode independant delta value
	unsigned int *pc;
	char *mode;

	if (user_mode(regs)) {
	  Tmisalignstat_stats.UI_user++;
	  mode = "user";
	}
	else {
	  Tmisalignstat_stats.UI_sys++;
	  mode = "system";
	}

	// We will need to +8 to the regs->pc so that any ldr/str using
	// r15(pc) will do the correct math


	pc = regs->uregs[15];
        UI_inst = *(unsigned int*) pc;
	regs->uregs[15]+=8;

	if ( arm_mm_alignment_enabled == 1) {
	  /*
	   * Report alignment fault
	   */
	  printk(KERN_INFO "Alignment fault in process %d at pc 0x%0x,in %s-mode: inst = 0x%0x\n",
		 current -> pid, pc, mode, UI_inst);
	}

	// pick off bits 27,26 and 25......
	switch((char)CODING_BITS(UI_inst))
		{
		case 0x2:
		case 0x3:
			{

			Tmisalignstat_stats.UI_word++;
			// Step 1 - Calculate EA
			Rn=RN_BITS(UI_inst);
			Rd=RD_BITS(UI_inst);

			// Step 2  - Check if immediate offset / register offset / scaled reg offset
			if (IMMED_BIT(UI_inst))			// Offset is an immediate value
				{
				UI_offsetdelta=OFFSET_BITS(UI_inst);
				}
			else					// Offset is a register
				{
				Rm=RM_BITS(UI_inst);    
				if (IS_SHIFT(UI_inst))		// Offset register gets shifted (scaled)
					{
					unsigned int UI_shiftval;	// Immediate shift value
					unsigned int UI_shiftcode;	// Shifting operation

					UI_shiftcode=SHIFT_TYPE(UI_inst);
					UI_shiftval=SHIFT_BITS(UI_inst);
					switch (UI_shiftcode)
						{
						case SHIFT_LSL:
							UI_offsetdelta=regs->uregs[Rm] << UI_shiftval;
							break;
						case SHIFT_LSR:
							UI_offsetdelta=(unsigned int)(regs->uregs[Rm]) >> UI_shiftval;
							break;
						case SHIFT_ASR:
							UI_offsetdelta=(int)(regs->uregs[Rm]) >> UI_shiftval;
							break;
						case SHIFT_RORRRX:
							if (UI_shiftval==0)
								UI_offsetdelta=((unsigned int)(regs->uregs[Rm]) >>1) 
									| (regs->uregs[16] & CC_C_BIT ? 0x80000000:0);
							else
								UI_offsetdelta=((unsigned int)(regs->uregs[Rm]) >> 
									UI_shiftval) | ((unsigned int)(regs->uregs[Rm]) << 
									(32-UI_shiftval));
							break;
						default:
							do_glap("Undefined or unsupported shifting operation occurred",regs);
							return(0);
						}
					}
				else
					{
					UI_offsetdelta=regs->uregs[Rm];
					}
				}

			// Step 3 - Do necessary pre-indexing and create the load address
			if (PREINDEX_BIT(UI_inst))
				{
				if (ADDOFFSET_BIT(UI_inst))
					UI_effectiveaddr=regs->uregs[Rn]+UI_offsetdelta;
				else
					UI_effectiveaddr=regs->uregs[Rn]-UI_offsetdelta;
				}
			else // We're post-indexing... just use the base as EA
				UI_effectiveaddr=regs->uregs[Rn];

			// Step 4 - Determine type of access (load/store)
			if (LOAD_BIT(UI_inst))
				regs->uregs[Rd]=get_checked((int *)UI_effectiveaddr);
			else
				put_checked((regs->uregs[Rd]),(int *)UI_effectiveaddr);

			// Step 5 - Do necessary post-indexing (and possible base update)
			if (!PREINDEX_BIT(UI_inst))
				{
				if (ADDOFFSET_BIT(UI_inst))
					UI_effectiveaddr=regs->uregs[Rn]+UI_offsetdelta;
				else
					UI_effectiveaddr=regs->uregs[Rn]-UI_offsetdelta;
				regs->uregs[Rn]=UI_effectiveaddr; // Unconditional base update
				}
			else
				if (BASEUPDATE_BIT(UI_inst))
					{
					regs->uregs[Rn]=UI_effectiveaddr;
					}
			}
			break;
		case 0x4:
			{
			int I_rn_scratch;
			int I_registerlist;
 			int I_regnum;
			int I_regcount = 0;

			Tmisalignstat_stats.UI_multiple++;
			Rn = RN_BITS(UI_inst);
			I_rn_scratch=regs->uregs[Rn]; // Address to load/store
			I_registerlist=(UI_inst & 0xffff);

			// step 1: how many registers do we have to do?
			for (I_regnum=0;I_regnum<16;I_regnum++)
				if ( (1<<I_regnum) & I_registerlist)
					I_regcount++;

			// step 2: for decrementing ops, we need to skip down
			if (!ADDOFFSET_BIT(UI_inst))
				I_rn_scratch -= 4*I_regcount;
			// precompensate for preinc or postdec
			if (ADDOFFSET_BIT(UI_inst)==0 && PREINDEX_BIT(UI_inst)==0)
				I_rn_scratch += 4;
			if (ADDOFFSET_BIT(UI_inst) && PREINDEX_BIT(UI_inst))
				I_rn_scratch += 4;

			for (I_regnum=0;I_regnum<16;I_regnum++) 
				{
				if ( (1<<I_regnum) & I_registerlist) 
					{
					// check if load/store
					if (LOAD_BIT(UI_inst))
						regs->uregs[I_regnum]=get_checked((int *)I_rn_scratch); // load
					else
						put_checked((regs->uregs[I_regnum]),(int *)I_rn_scratch);// store
                                        I_rn_scratch+=4;
					}
				}
			// step 3: compensate the address and writeback
			if (BASEUPDATE_BIT(UI_inst)) 
				{
				if (PREINDEX_BIT(UI_inst) && !ADDOFFSET_BIT(UI_inst))
					I_rn_scratch -= 4*I_regcount;
				else if (PREINDEX_BIT(UI_inst))
					I_rn_scratch -= 4;
				else if (!ADDOFFSET_BIT(UI_inst))
					I_rn_scratch -= 4 + 4*I_regcount;
				regs->uregs[Rn]=I_rn_scratch;
				}
			}
			// check S bit (with pc flagged)
			if (UI_inst & 0x400000)
				{ // s bit is set
				if (user_mode(regs))
					{
					do_glap("Illegal user mode use of S bit in misaligned 32-bit multiple load/store",regs);
					return 0;
					}
				do_glap("Misaligned 32-bit load/store multiple with unimplimented S bit trap",regs);
				return 0;
				}
			break;
		case 0x0://halfwords
			// check for swap word instruction
			if ((UI_inst & 0x0ff00ff0) == 0x01000090) 
				{
				Tmisalignstat_stats.UI_swap++;
				Rn=RN_BITS(UI_inst);
				Rd=RD_BITS(UI_inst);
				Rm=RM_BITS(UI_inst);
				// NOTE: we can move the bytes around, but we
				// can't really ensure the lock aspect. Oh well.
				regs->uregs[Rd] = get_checked((int*)regs->uregs[Rn]);
				put_checked(regs->uregs[Rm], (int*)regs->uregs[Rn]);
				break;
				}
			else if (((UI_inst & 0x0e000090) == 0x00000090) // chk for half ldr/str
					&& ((UI_inst & 0x60) != 0)) 
				{
				Tmisalignstat_stats.UI_short++;
				Rn=RN_BITS(UI_inst);
				Rd=RD_BITS(UI_inst);
				Rm=RM_BITS(UI_inst);
				// step 1: form the delta
				if ((UI_inst & 0x00400000) == 0x00400000) // immed offset
					UI_offsetdelta = ((UI_inst & 0xf00) >> 4) + (UI_inst & 0xf);
				else  // register offset
					UI_offsetdelta = regs->uregs[Rm];
				// step 2: form the effective address
				UI_effectiveaddr = regs->uregs[Rn];
				if (PREINDEX_BIT(UI_inst)) 
					{
					if (ADDOFFSET_BIT(UI_inst))
						UI_effectiveaddr += UI_offsetdelta;
					else
						UI_effectiveaddr -= UI_offsetdelta;
					}
				// step 3: do the load/store 
				// NOTE: if bits d6,d5 are 00, this insn is a multiply
				// if the bits are 10, this is signed byte op
				if (LOAD_BIT(UI_inst))
					regs->uregs[Rd] = get_checked((unsigned short*)UI_effectiveaddr);
				else
					put_checked(regs->uregs[Rd],(unsigned short*)UI_effectiveaddr);
				if ((UI_inst & 0x40) == 0x40) // signed halfword
					regs->uregs[Rd] = (long) ((short) regs->uregs[Rd]);
				// step 4: do the writeback
				if (!(PREINDEX_BIT(UI_inst))) 
					{
					if (ADDOFFSET_BIT(UI_inst))
						UI_effectiveaddr += UI_offsetdelta;
					else
						UI_effectiveaddr -= UI_offsetdelta;
					regs->uregs[Rn] = UI_effectiveaddr;
					} 
				else
					if (BASEUPDATE_BIT(UI_inst))
						regs->uregs[Rn] = UI_effectiveaddr;
				break;
				}
		default:
				Tmisalignstat_stats.UI_unsupported++;
				do_glap("Undefined or unsupported misaligned access fault occurred",regs);
				break;
		}
	// Done with fixup.. time to go; yes, I know it's the same location <g>
	regs->uregs[15]-=8; // Make up for the fact that we added 8 to pc
	regs->ARM_pc+=4;     // Do not want to re-execute fault instruction

	return(0);
}


void do_glap(char *glap_msg,struct pt_regs *regs)
{
  struct task_struct *tsk = current;
  unsigned long *Ip_pc;

  // Since we're going to get the hell out of dodge,
  // fix things up for display (and return)
  regs->uregs[15]-=8;

  Ip_pc=instruction_pointer(regs);

  printk("[GLAP!!] %s\n",glap_msg);
  printk("         Inst Ptr.   :%08x\n",Ip_pc);
  printk("         Instruction :%08x\n",*Ip_pc);
  printk("         Called from :%08x\n",(unsigned long)regs->ARM_lr);
  if (!user_mode(regs))
    show_regs(regs); // SEGV handler will do userland for us
  printk("Please email sanm@corelcomputer.com with above details\n");
  if (user_mode(regs))
    {
    printk("          (Process Operations Halted)\n");
    force_sig(SIGSEGV, tsk);
    return;
    }
  printk("            (Kernel Operations Halted)\n");
  while(1);
}

