/*
 *  linux/fs/adfs/inode.c
 *
 * Copyright (C) 1997 Russell King
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/adfs_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/mm.h>

/*
 * Inode numbers:
 *  bit 30 - 16		FragID of parent object
 *  bit 15		0			1
 *  bit 14 - 0		FragID of object	Offset into parent FragID
 */
#define parent_frag_id(x)	(((x) >> 16) & 0x7fff)
#define fragment_id(x)		((((x) & 1<<15) ? (x)>>16 : (x)) & 0x7fff)
#define offset(x)		(((x) & 1<<15) ? (x) ? ((x) & 0xff) - 1 : 0 : 0)
#define object_id(x)		(((x) & 1<<15) ? ((x)>>8 & 0x7fff00) | ((x) & 0xff) : \
					((x)<<8 & 0x7fff00))

#define frag_id(x)		(((x) >> 8) & 0x7fff)
#define off(x)			((x) & 0xff)

int adfs_inode_validate (struct inode *inode)
{
	struct super_block *sb;
	unsigned long max_frag_id;

	sb = inode->i_sb;

	max_frag_id = sb->u.adfs_sb.s_map_size * sb->u.adfs_sb.s_ids_per_zone;

	return inode->i_ino == -1 || fragment_id(inode->i_ino) >= max_frag_id ||
		    parent_frag_id(inode->i_ino) >= max_frag_id;
}

unsigned long adfs_inode_generate (unsigned long parent_id, unsigned long obj_id)
{
	if (!parent_id || !obj_id)
		return -1;

	if (frag_id(parent_id) != ADFS_ROOT_FRAG && off(parent_id) != 1) {
		printk ("adfs_inode_generate: can't generate for %lX and %lX\n", parent_id, obj_id);
		return -1;	/* can't do this! */
	}

	if (frag_id(parent_id) == frag_id(obj_id))
		return frag_id(parent_id) << 16 | off(obj_id) | 1 << 15;

	if (off(obj_id) > 1) {
		printk ("adfs_inode_generate: can't generate for %lX and %lX\n", parent_id, obj_id);
		return -1;	/* can't do this! */
	}

	return frag_id(parent_id) << 16 | frag_id(obj_id);
}

unsigned long adfs_inode_objid (struct inode *inode)
{
	unsigned long objid;

	if (adfs_inode_validate (inode)) {
		adfs_error (inode->i_sb, "adfs_inode_objid",
				"bad inode number: %lu", inode->i_ino);
		return 0;
	}

	if (inode->i_ino == inode->i_sb->u.adfs_sb.s_root)
		return inode->i_sb->u.adfs_sb.s_dr->root;
	objid = object_id(inode->i_ino);
	return objid + (S_ISDIR(inode->i_mode) ? 1 : 0);;
}

int adfs_bmap (struct inode *inode, int block)
{
	struct super_block *sb = inode->i_sb;

	if (adfs_inode_validate (inode)) {
		adfs_error (sb, "adfs_bmap",
				"bad inode number: %lu", inode->i_ino);
		return 0;
	}

	if (fragment_id(inode->i_ino) == ADFS_ROOT_FRAG)
		return sb->u.adfs_sb.s_map_block + offset(inode->i_ino) + block;
	return adfs_map_lookup (sb, fragment_id(inode->i_ino), offset (inode->i_ino) + block);
}

int adfs_parent_bmap (struct inode *inode, int block)
{
	struct super_block *sb = inode->i_sb;

	if (adfs_inode_validate (inode)) {
		adfs_error (sb, "adfs_parent_bmap",
				"bad inode number: %lu", inode->i_ino);
		return 0;
	}

	if (parent_frag_id (inode->i_ino) == ADFS_ROOT_FRAG)
		return sb->u.adfs_sb.s_map_block + offset(sb->u.adfs_sb.s_root) + block;
	return adfs_map_lookup (sb, parent_frag_id (inode->i_ino), block);
}

static int adfs_atts2mode (unsigned char mode)
{
	int omode = 0;
	if (mode & ADFS_NDA_DIRECTORY)
		omode |= S_IFDIR|S_IRUSR|S_IXUSR|S_IXGRP|S_IXOTH;
	else
		omode |= S_IFREG;
	if (mode & ADFS_NDA_OWNER_READ)
		omode |= S_IRUSR;
	if (mode & ADFS_NDA_OWNER_WRITE)
		omode |= S_IWUSR;
	if (mode & ADFS_NDA_PUBLIC_READ)
		omode |= S_IRGRP | S_IROTH;
	if (mode & ADFS_NDA_PUBLIC_WRITE)
		omode |= S_IWGRP | S_IWOTH;
	return omode;
}

void adfs_read_inode (struct inode *inode)
{
	struct super_block *sb;
	struct buffer_head *bh[4];
	struct adfs_idir_entry ide;
	int buffers;

	sb = inode->i_sb;
	inode->i_uid = 0;
	inode->i_gid = 0;
	inode->i_version = ++event;

	if (adfs_inode_validate (inode)) {
		adfs_error (sb, "adfs_read_inode",
			    "bad inode number: %lu", inode->i_ino);
		goto bad;
	}

	if (fragment_id(inode->i_ino) == ADFS_ROOT_FRAG) {
		/* root dir */
		inode->i_mode	 = S_IRWXUGO | S_IFDIR;
		inode->i_nlink	 = 2;
		inode->i_size	 = ADFS_NEWDIR_SIZE;
		inode->i_blksize = PAGE_SIZE;
		inode->i_blocks  = inode->i_size / sb->s_blocksize;
		inode->i_mtime   =
		inode->i_atime   =
		inode->i_ctime   = 0;
	} else {
		if (!(buffers = adfs_dir_read_parent (inode, bh)))
			goto bad;
		if (adfs_dir_check (inode, bh, buffers, NULL)) {
			adfs_dir_free (bh, buffers);
			goto bad;
		}

		if (!adfs_dir_find_entry (sb, bh, buffers, object_id (inode->i_ino), &ide)) {
			adfs_dir_free (bh, buffers);
			goto bad;
		}
		adfs_dir_free (bh, buffers);
		inode->i_mode	 = adfs_atts2mode (ide.mode);
		inode->i_nlink	 = 2;
		inode->i_size    = ide.size;
		inode->i_blksize = PAGE_SIZE;
		inode->i_blocks	 = (inode->i_size + sb->s_blocksize - 1) >> sb->s_blocksize_bits;
		inode->i_mtime	 =
		inode->i_atime   =
		inode->i_ctime   = 0;
	}

	if (S_ISDIR(inode->i_mode))
		inode->i_op	 = &adfs_dir_inode_operations;
	else if (S_ISREG(inode->i_mode))
		inode->i_op	 = &adfs_file_inode_operations;
	return;

bad:
	inode->i_mode	 = 0;
	inode->i_nlink	 = 1;
	inode->i_size    = 0;
	inode->i_blksize = 0;
	inode->i_blocks  = 0;
	inode->i_mtime   =
	inode->i_atime   =
	inode->i_ctime   = 0;
	inode->i_op	 = NULL;
}
