/*
 * SA1100 Clock Speed Device Driver Header File
 * Copyright (c) Compaq Computer Corporation, 1998
 *
 * Author: Deborah A. Wallach
 *
 */

#ifndef	clkspeed_sa1100_h
#define	clkspeed_sa1100_h

/*
 * The processor clock can run at one of eleven clock levels, from
 * 59.0 to 206.4Mhz (in approx 15Mhz increments). The code refers
 * to the hardware index (0..10) rather than the actual clock speed. See
 * Table 8.2 in the SA1100 manual concerning the clock configuration
 * for more information.
 *
 * In addition, you can enable/disable "clock switching". When swtiching
 * is enabled, the CPU runs at 2x the DRAM clock speed until the processor
 * needs to wait for a memory operation, at which time it runs at the DRAM
 * clock speed until the operation is completed.  With clock switching 
 * disabled, the CPU always runs at the DRAM clock speed (e.g. from 30-103MHz).
 *
 * When the clock speed is changed, individual devices can (and should)
 * be suspended during the speed change. 
 *
 * You should only use the following IOCTL's:
 *
 * Getting information:
 *   CLKSPEED_GET_CLOCK -- returns current clock speed index, or -1
 *			on failure.
 *
 *   CLKSPEED_GET_SWITCHING 
 *			-1 -- error
 *			 0 -- switching is disabled
 *			 1 -- switching is enabled
 *
 * Setting information
 *   CALL_IOCTL_CLKSPEED_SET_SPEED -- this macro should be used
 *   to set the clock speed and the state of "clock switching".
 *   You must also indicate if devices should suspend during the
 *   clock speed change. Normally, you will always specify TRUE
 *   for both switching and suspend.
 *   Parameters:
 *     fd -- the file descriptor of the clock device (/dev/clkspeed)
 *     speed -- the desired new speed (from 0 to 10, inclusive)
 *     switching -- enable (1) or disable (0) switching after the speed change
 *     suspend -- if 1, suspend devices during the switch; if 0, do not
 *   Typical use:  retval = CALL_IOCTL_CLKSPEED_SET_SPEED(clockfd, newspeed, 1, 1);
 *
 */
   
/*
 *  This is the only definition you should use. 
 */

#define CALL_IOCTL_CLKSPEED_SET_SPEED(fd, clock, switching, suspend) (ioctl(fd, CLKSPEED_SET_SPEED, (clock | _CLKSPEED_MAKE_SWITCHING_PARAM(switching) | _CLKSPEED_MAKE_SUSPEND_PARAM(suspend))))

/* driver ioctl type */
#define	CLKSPEED_IOCTL_TYPE	'y'

/* ioctl commands */
#define	CLKSPEED_SET_CLOCK          _IOW(CLKSPEED_IOCTL_TYPE,  0, uint)
#define	CLKSPEED_GET_CLOCK	    _IOW(CLKSPEED_IOCTL_TYPE,  1, uint)
#define	CLKSPEED_ENABLE_SWITCHING   _IOW(CLKSPEED_IOCTL_TYPE,  2, uint)
#define	CLKSPEED_DISABLE_SWITCHING  _IOW(CLKSPEED_IOCTL_TYPE,  3, uint)
#define	CLKSPEED_GET_SWITCHING	    _IOW(CLKSPEED_IOCTL_TYPE,  4, uint)
#define	CLKSPEED_SET_SPEED          _IOW(CLKSPEED_IOCTL_TYPE,  5, uint)
#define	CLKSPEED_PRINT_PARAMS	    _IOW(CLKSPEED_IOCTL_TYPE,  6, uint)
#define	CLKSPEED_TIME_IT	    _IOW(CLKSPEED_IOCTL_TYPE,  7, uint)

#define _CLKSPEED_SWITCHING_FIELD (29)
#define _CLKSPEED_SWITCHING_MASK  (1<<_CLKSPEED_SWITCHING_FIELD)
#define _CLKSPEED_MAKE_SWITCHING_PARAM(switching) (switching<<_CLKSPEED_SWITCHING_FIELD)
#define _CLKSPEED_GET_SWITCHING_PARAM(arg) ((arg&_CLKSPEED_SWITCHING_MASK)>>_CLKSPEED_SWITCHING_FIELD)

#define _CLKSPEED_SUSPEND_FIELD (30)
#define _CLKSPEED_SUSPEND_MASK  (1<<_CLKSPEED_SUSPEND_FIELD)
#define _CLKSPEED_MAKE_SUSPEND_PARAM(suspend) (suspend<<_CLKSPEED_SUSPEND_FIELD)
#define _CLKSPEED_GET_SUSPEND_PARAM(arg) ((arg&_CLKSPEED_SUSPEND_MASK)>>_CLKSPEED_SUSPEND_FIELD)

#define _CLKSPEED_GET_BASE_PARAM(arg) (((unsigned)arg<<(32-_CLKSPEED_SWITCHING_FIELD))>>(32-_CLKSPEED_SWITCHING_FIELD))


#endif	/* clkspeed_sa1100_h */
