/*
 * flinfo.h
 * Flash device identification, programming, and erasing primitives
 *
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * Author: Tim Mann
 *
 * $Log: flinfo.h,v $
 * Revision 2.4  1999/12/23  01:28:21  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.3  1999/11/23  05:00:59  mann
 * Multiple region support.
 *
 * Revision 2.2  1999/10/09  00:22:10  mann
 * Changes to support 32Mbit flash parts.
 * Changes to support /proc/kcore_elf
 *
 * Revision 2.1  1999/04/29  02:49:50  mann
 * Itsy FTL (flash translation layer) support, initial working version.
 *
 */

/* Essential type information about a flash chip */

/* Maximum number of sub-banks */
#define FLMEM_MAXSUBBNK 4
/* Maximum number of special short sectors */
#define FLMEM_MAXSPSCT 16

/* Programming and erasing primitives.  All addresses, both flash
   addresses and RAM data buffer addresses, must be 16-bit aligned for
   16-bit devices, 32-bit aligned for 32-bit devices.  This condition
   is not checked.  The primitives return error codes from
   itsymon-error.h. */

/* Program device.  Must not be called while erase is in progress on
   this bank.  If erase is suspended, must not program the sector
   being erased.  These conditions are not checked.  Returns when
   programming is complete. */
typedef unsigned int
(*FlMemPrgmAlg)(volatile void *FlMemBsAdd,        /* Sub-bank base addr */
		volatile void *FlMemAdd,          /* First addr to program */
		void *DataAdd,                    /* Data to program in */
		unsigned int DataSz);             /* Data size in bytes */

/* Start a sector erase.  Must not be called when erase is in progress
   or suspended on this bank; this condition is not checked.  Upon
   return, erase may still be in progress. */
typedef unsigned int
(*FlMemSctErsAlg)(volatile void *FlMemBsAdd,      /* Sub-bank base address */
		  volatile void *FlMemSctAdd);    /* First address to erase */

/* Check if sector erase is in progress.  Must not be called when
   erase is suspended.  Returns -1 if in progress, Success (0) if
   done, error code if last erase failed.  In the error case, the
   chips are reset; thus the error code will appear only on the first
   call after the failure. */
typedef int
(*FlMemSctErsStat)(volatile void *FlMemBsAdd,     /* Sub-bank base addr */
		   volatile void *FlMemSctAdd);   /* First address to erase */

/* Suspend the current erase on the specified bank.  Must be called
   only when FlMemSctErsAlg has started an erase and FlMemSctErsStat
   has not yet indicated the erase is complete.  Returns Success if
   the erase has been suspended successfully or -1 if it was already
   finished.  May return an error if the current erase has failed; as
   with FlMemSctErsStat, the chips are reset in this case and thus the
   error code will not appear again, so be sure to check for it.  */
typedef int
(*FlMemSctErsSusAlg)(volatile void *FlMemBsAdd,   /* Sub-bank base addr */
		     volatile void *FlMemSctAdd); /* First address to erase */

/* Resume a suspended erase.  No-op if no erase is suspended. */
typedef unsigned int
(*FlMemSctErsResAlg)(volatile void *FlMemBsAdd,   /* Sub-bank base addr */
		     volatile void *FlMemSctAdd); /* First address to erase */

typedef struct {
  FlMemPrgmAlg prgmAlg;
  FlMemSctErsAlg sctErsAlg;
  FlMemSctErsStat sctErsStat;
  FlMemSctErsSusAlg sctErsSusAlg;
  FlMemSctErsResAlg sctErsResAlg;

  /* Number of this bank */
  unsigned int stMemBnkNb;

  /* Base virtual address for each sub-bank */
  unsigned int numSubBnk;
  void *(subBnkAdd[FLMEM_MAXSUBBNK]);

  /* Organization: devsPerWord x bitsPerDev = bitsPerWord.
     bitsPerWord can be 16 or 32. */
  unsigned int devsPerWord;
  unsigned int bitsPerDev;
  unsigned int bitsPerWord;

  /* Alignment mask for low-order bits:
     1 if bank requires 16-bit alignment (devsPerWord * bitsPerDev = 16);
     3 if 32-bit (devsPerWord * bitsPerDev = 32). */
  unsigned int dataWdthMsk;

  /* Size populated in this bank, in bytes */
  unsigned int size;

  /* Size of a normal sector, in bytes */
  unsigned int nlSctSz;

  /* Virtual addresses and sizes of special short sectors, in order.
     These sectors occupy the space of one or two normal sector(s).  They must
     all be erased individually. */
  unsigned int numSpSct;
  unsigned int spSctAdd[FLMEM_MAXSPSCT];
  unsigned int spSctSz[FLMEM_MAXSPSCT];
  
} FlMemInfo;

/*
 * PUBLIC PROCEDURE: FlMemIdentify
 *
 * Purpose
 *    Identify the type of flash chips in a given bank and their organization,
 *    and return information necessary to erase or program them.
 *
 * Input
 *    StMemBnkNb        Static memory bank number.
 *    FlMemBsAdd	Flash memory virtual base address.
 *
 * Output
 *    FlMemIdentify	Set to 0 ("Success") if the chip type was identified
 *              	 or to the corresponding error number if an error occurred.
 *    fi                Filled in if success.                 
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.  (Under Linux, these variables are
 *    not global, but their values can be retrieved with FlMemGetError.)
 */
unsigned int
FlMemIdentify(int StMemBnkNb, unsigned int FlMemBsAdd, FlMemInfo *fi);

/* !!Note: this business of storing the error info in a global (not
 * even one per bank) is pretty ugly; maybe should be cleaned up.
 */

/*
 * PUBLIC PROCEDURE: FlMemGetError
 *
 * Purpose
 *    Return error information associated with the last call to FlMemPrgm()
 *    or FlMemIdentify().
 *
 * Input
 *    None.
 *
 * Output
 *    status           A monitor error code; see itsy-mon.h.
 *    addr             Address where last error occurred, if status != Success.
 *    data             Data associated with last error, if status != Success.
 */
void FlMemGetError(int *status, void **addr, unsigned int *data);

/* The following is an entry point to the monitor's programming
 * routine that is also used by the character driver.
 *
 * Note: !!Maybe the character driver should be cleaned up to use
 * FlMemIdentify and call through the function pointers it returns
 * instead.  FlMemPrgm could then be split out into a separate,
 * monitor-only source file.  
 */

/*
 * PUBLIC PROCEDURE: FlMemPrgm
 *
 * Purpose
 *    The procedure "FlMemPrgm" programs (i.e., erases and re-writes) a given
 *    data buffer into a flash memory. Only the flash-memory sectors storing
 *    the new data are modified.
 *
 * Input
 *    FlMemAdd  	Virtual address, in the flash memory, where to write
 *              	the new data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *    Merge     	When set to 0 ("FALSE"), the data previously stored
 *              	in the flash-memory sectors to be modified are
 *              	ignored and un-written locations are left blank
 *              	(i.e., all bits set to 1). Otherwise ("TRUE"), the
 *              	new data is merged with the previously stored data
 *              	before being written into the flash memory.
 *
 * Output
 *    FlMemPrgm 	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.  (Under Linux, the
 *    variables are not global, but their values can be retrieved with 
 *    FlMemPrgm_GetError.)
 */
#ifdef	__linux__
unsigned int FlMemPrgm (
   unsigned long	flash_base_virtual,  /* virtual base of this bank */
   unsigned long        flash_base_physical, /* physical base of this bank */
   void        	        *FlMemAdd,
   void                 *DataAdd,
   unsigned int 	DataSz,
   int          	Merge);
#endif	/* __linux__ */


