/*
 * DS2437 Smart Battery Monitor Commands
 * Copyright (c) Compaq Computer Corporation, 1999
 * Platform: Itsy V2.0 with SA-1100 processor
 * Author: Jennifer Anderson
 */

#ifndef	ITSY_BATTMON_H
#define	ITSY_BATTMON_H

/* 
 * Issues commands to the DS2437 Smart Battery Monitor on Itsy V2.0.
 * The battery monitor commands are all accessed via the followings IOCTL's:
 *
 * BATTMON_DETECT
 *     Detect the battery monitor and return its 64b ROM code.
 *
 * BATTMON_GET_TEMP
 *     Initiate a temperature conversion and return the temperature.
 *     The temperature value is in units of 1/32 degrees C, and  
 *     the valid range is -55C to 125 C.  Temperature conversion
 *     can take up to 1 second, and the process is put to sleep
 *     while the conversion takes place.
 * 
 * BATTMON_GET_VOLTAGE    
 *     Initiate a voltage conversion and return the Vdd voltage.
 *     The voltage value is in units of 10mV, and the valid range
 *     voltages is 0V to 10.23V.  Voltage conversion can take 2 ms,
 *     and the process is put to sleep for 10ms (minimum timer
 *     resolution) while the conversion takes place.
 *
 * BATTMON_GET_VAD_VOLTAGE
 *     Initiate a voltage conversion and return the Vad voltage.
 *     The voltage value is in units of 10mV, and the valid range
 *     voltages is 0V to 10.23V.  Voltage conversion can take 2 ms,
 *     and the process is put to sleep for 10ms (minimum timer
 *     resolution) while the conversion takes place.
 *
 * BATTMON_READ_CURRENT 
 *     Return the value of the current register.  The current is in
 *     units of 1/205 C, where full capacity is 1C (for Itsy, C =
 *     500 mA).  The valid range of current values is -2.0C to +2.0C.
 *     The current register is updated 32 times per second.
 *     
 * BATTMON_READ_NET_CURRENT           
 *     Return the value of the net current (ICA) register.  The ICA
 *     maintains a net accumulated total of current flowing into and
 *     out of the battery; therefore the reading in this register is
 *     an indication of the remaining capacity in the battery.  The
 *     current is in units of .01C, where full battery capacity is 1C.
 *     Thus ICA=100 represents a fully charged battery and ICA=0
 *     represents fully discharged.  The valid range of current values
 *     is 0C to +2.55C (for Itsy, C = 500 mA).
 *     
 * BATTMON_WRITE_NET_CURRENT          
 *     Set the value of the net current (ICA) register.
 *
 * BATTMON_READ_CHARGING_CURRENT      
 *     Return the value of the charging current (CCA) register.  The
 *     CCA represents the total charging current the battery has
 *     encountered in its lifetime.  It is only updated when the
 *     current is positive, i.e., the battery is being charged.  The
 *     CCA registers are non-volatile, so the information will
 *     accumulate over the lifetime of the battery pack.  The current
 *     is in units of .32C, where full battery capacity is 1C.  The
 *     valid range is 0C to 20971C in .32 increments.
 *
 * BATTMON_WRITE_CHARGING_CURRENT     
 *     Set the value of the charging current (CCA) register.
 *
 * BATTMON_READ_DISCHARGING_CURRENT   
 *     Return the value of the discharging current (DCA) register.
 *     The DCA represents the total discharging current the battery
 *     has encountered in its lifetime.  It is only updated when the
 *     current is negative, i.e., the battery is being discharged.
 *     The DCA registers are non-volatile, so the information will
 *     accumulate over the lifetime of the battery pack.  The current
 *     is in units of .32C, where full battery capacity is 1C.  The
 *     valid range is 0C to 20971C in .32 increments.
 *
 * BATTMON_WRITE_DISCHARGING_CURRENT  
 *     Set the value of the discharging current (DCA) register.
 *
 * BATTMON_READ_RTC                   
 *     Return the value of the real-time-clock (RTC) registers.  The
 *     timestamp is in units of 1 sec.
 *
 * BATTMON_WRITE_RTC                  
 *     Set the value of the real-time-clock (RTC) registers.
 *
 * BATTMON_READ_DISCONNECT_TIME       
 *     Return the value of the disconnect-timestamp registers.  The
 *     disconnect-timestamp is written to by the DS2437 whenever it
 *     senses that the DQ line has been low for more than 1 second.
 *     This condition signals that the battery pack has been removed
 *     from the system.  The timestamp is in units of 1 sec.
 *
 * BATTMON_READ_EOC_TIME
 *     Return the value of the end-of-charge (EOC) registers.  The
 *     end-of-charge timestamp is written by the DS2437 whenever it
 *     senses that charging is finished (when the current changes
 *     direction).  The timestamp is in units of 1 sec.
 *
 */

#include <linux/types.h>

/* brakmo */
#ifndef BATTMON_DRIVER
typedef __u8 Byte;
typedef __u16 Word16;
typedef __u32 Word32;
#endif

typedef __u64 Word64;

/* driver ioctl type, use 'd' for DS2437 */
#define	BATTMON_IOCTL_TYPE	'd'

/* ioctl commands */
#define	BATTMON_DETECT                     _IOR(BATTMON_IOCTL_TYPE,  0, Word64)

#define BATTMON_GET_TEMP                   _IOR(BATTMON_IOCTL_TYPE,  1, int)
#define BATTMON_GET_VOLTAGE                _IOR(BATTMON_IOCTL_TYPE,  2, int)
#define BATTMON_GET_VAD_VOLTAGE            _IOR(BATTMON_IOCTL_TYPE,  3, int)

#define BATTMON_READ_CURRENT               _IOR(BATTMON_IOCTL_TYPE,  4, int)
#define BATTMON_READ_NET_CURRENT           _IOR(BATTMON_IOCTL_TYPE,  5, Byte)
#define	BATTMON_WRITE_NET_CURRENT          _IOW(BATTMON_IOCTL_TYPE,  6, Byte)
#define BATTMON_READ_CHARGING_CURRENT      _IOR(BATTMON_IOCTL_TYPE,  7, Word16)
#define BATTMON_WRITE_CHARGING_CURRENT     _IOW(BATTMON_IOCTL_TYPE,  8, Word16)
#define BATTMON_READ_DISCHARGING_CURRENT   _IOR(BATTMON_IOCTL_TYPE,  9, Word16)
#define BATTMON_WRITE_DISCHARGING_CURRENT  _IOW(BATTMON_IOCTL_TYPE, 10, Word16)

#define BATTMON_READ_RTC                   _IOR(BATTMON_IOCTL_TYPE, 11, Word32)
#define BATTMON_WRITE_RTC                  _IOW(BATTMON_IOCTL_TYPE, 12, Word32)
#define BATTMON_READ_DISCONNECT_TIME       _IOR(BATTMON_IOCTL_TYPE, 13, Word32)
#define BATTMON_READ_EOC_TIME              _IOR(BATTMON_IOCTL_TYPE, 14, Word32)
#define BATTMON_SET_CHARGE_RATE		   _IO(BATTMON_IOCTL_TYPE, 15) 
#define BATTMON_GET_CHARGE_RATE		   _IOR(BATTMON_IOCTL_TYPE, 16, int)
#define BATTMON_IOCTL_MAXNR                17

#endif	/* ITSY_BATTMON_H */
