/*
 * Itsy Session Manager Interface
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * Author: Carl Waldspurger
 *
 * $Log: itsy_session.h,v $
 * Revision 2.11  1999/12/23  01:28:28  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.10  1999/09/10  21:41:46  caw
 * Updated itsy_event_header field names to avoid conflicts with C++.
 *
 * Revision 2.9  1999/06/10  20:17:23  caw
 * Added session manager request definitions.
 *
 * Revision 2.8  1999/05/20  22:41:06  caw
 * Updated to reflect new "override" option to itsy_session_set_name().
 *
 * Revision 2.7  1999/05/12 18:01:11  caw
 * Included changes from itsy_sessions branch.
 *
 * Revision 2.6.2.4  1999/04/29  23:56:47  caw
 * Defined SESSION_REAP ioctl.
 *
 * Revision 2.6.2.3  1999/04/22  16:53:58  caw
 * Export more general itsy_session_signal() operation.
 * Added ioctls to terminate/kill session tasks.
 *
 * Revision 2.6.2.2  1999/04/22  00:20:05  caw
 * Defined kernel and ioctl interfaces to stop/continue session tasks.
 *
 * Revision 2.6.2.1  1999/04/15 22:08:22  caw
 * Use ITSY_SESSION_ID_NONE for all tasks if CONFIG_ITSY_SESSIONS undefined.
 * Added more documentation.
 *
 * Revision 2.6  1999/04/13  19:42:57  brakmo
 * Added a flag, foreground session id and register respectively.
 *
 * Revision 2.5  1999/03/31  23:56:57  caw
 * Defined itsy_session_from_task() to support future generalizations.
 *
 * Revision 2.4  1999/02/04 22:46:59  caw
 * Defined additional session event types.
 *
 * Revision 2.3  1999/02/03  18:37:39  caw
 * Expanded and updated ioctl interface.
 *
 * Revision 2.2  1999/02/02  03:05:16  caw
 * Updated ioctl interface.  Other minor improvements.
 *
 * Revision 2.1  1999/02/02  01:08:19  caw
 * Initial revision.
 *
 */

#ifndef	ITSY_SESSION_H
#define	ITSY_SESSION_H

/*
 * includes
 *
 */

#include <linux/itsy_event.h>

/*
 * constants
 *
 */

/* session event types */
#define	ITSY_EVENT_SESSION_FOREGROUND		(1)
#define	ITSY_EVENT_SESSION_BACKGROUND		(2)
#define	ITSY_EVENT_SESSION_CREATED		(3)
#define	ITSY_EVENT_SESSION_DESTROYED		(4)
#define	ITSY_EVENT_SESSION_NAME			(5)
#define	ITSY_EVENT_SESSION_DATA			(6)
#define	ITSY_EVENT_SESSION_MANAGER_REQUEST	(7)

/* special sids */
#define	ITSY_SESSION_ID_INVALID		(-1)
#define ITSY_SESSION_ID_FOREGROUND	(0)
#define ITSY_SESSION_ID_NONE		(0)

/* sizes */
#define	ITSY_SESSION_NAME_LEN		(16)

/*
 * types
 *
 */

typedef struct {
  int sender;
  int operation;
  int data;
} itsy_session_request;

typedef struct {
  itsy_event_header common;
  int sid;
  itsy_session_request request;
} itsy_session_event;

typedef char itsy_session_name[ITSY_SESSION_NAME_LEN];

typedef struct {
  int sid;
  itsy_session_name name;
} itsy_session_info;

typedef struct {
  int  sid;
  void *data;
  int  ndata;
  int  ntotal;
} itsy_session_data;

typedef struct {
  int *sids;
  int nsids;
  int ntotal;
} itsy_session_ids;

/*
 * ioctl interface
 *
 */

/* driver ioctl type */
#define	SESSION_IOCTL_TYPE	'x'
#define	SESSION_IOCTL_MAXNR	(14)

/* ioctl commands */
#define	SESSION_GET_ID		_IOR(SESSION_IOCTL_TYPE,  0, int)
#define	SESSION_GET_FOREGROUND	_IOR(SESSION_IOCTL_TYPE,  1, int)
#define	SESSION_SET_FOREGROUND 	_IO(SESSION_IOCTL_TYPE,   2)
#define	SESSION_GET_INFO	_IOWR(SESSION_IOCTL_TYPE, 3, itsy_session_info)
#define	SESSION_SET_INFO	_IOW(SESSION_IOCTL_TYPE,  4, itsy_session_info)
#define	SESSION_GET_DATA	_IOWR(SESSION_IOCTL_TYPE, 5, itsy_session_data)
#define	SESSION_SET_DATA	_IOW(SESSION_IOCTL_TYPE,  6, itsy_session_data)
#define	SESSION_SET_MANAGER	_IO(SESSION_IOCTL_TYPE,   7)
#define	SESSION_GET_IDS		_IOWR(SESSION_IOCTL_TYPE, 8, itsy_session_ids)
#define	SESSION_SIG_SUSPEND	_IO(SESSION_IOCTL_TYPE,   9)
#define	SESSION_SIG_CONTINUE	_IO(SESSION_IOCTL_TYPE,  10)
#define	SESSION_SIG_TERMINATE	_IO(SESSION_IOCTL_TYPE,  11)
#define	SESSION_SIG_KILL	_IO(SESSION_IOCTL_TYPE,  12)
#define	SESSION_REAP		_IO(SESSION_IOCTL_TYPE,  13)
#define	SESSION_MANAGER_REQUEST	_IOW(SESSION_IOCTL_TYPE, 14, itsy_session_request)

/*
 * kernel interface
 *
 */

#ifdef	__KERNEL__

/*
 * session macros
 *
 */

static inline int itsy_session_from_task(const struct task_struct *t)
{
  /*
   * modifies: nothing
   * effects:  Returns the current session associated with task t.
   *           Drivers should use this operation instead of directly
   *	       using the task's session field, since an Itsy session
   *           may not always match a Linux session.
   */

#ifdef	CONFIG_ITSY_SESSIONS
  /* use linux session id */
  return(t->session);
#else
  /* no session id */
  return(ITSY_SESSION_ID_NONE);
#endif	/* CONFIG_ITSY_SESSIONS */
}

/*
 * session interface 
 *
 */

/* primitive ops, not for general use by drivers */
extern int itsy_session_init(void);
extern int itsy_session_create(int sid, const char *name);
extern int itsy_session_destroy(int sid);

extern int itsy_session_get_foreground(void);
  /*
   * modifies: nothing
   * effects:  Returns the session id of the current foreground session,
   *           or ITSY_SESSION_ID_NONE if none exists.
   *
   */

extern int itsy_session_set_foreground(int sid);
  /*
   * modifies: nothing
   * effects:  Sets the foreground session to the specified sid.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

extern const char *itsy_session_get_name(int sid);
  /*
   * modifies: nothing
   * effects:  Returns the session name associated with sid, or
   *           NULL if no such association exists.
   *
   */ 

extern int itsy_session_set_name(int sid, const char *name, int override);
  /*
   * modifies: nothing
   * effects:  Sets the name associated with specified session.
   *           Fails if a name has already been set with override,
   *           unless the override flag is again specified.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

extern int itsy_session_signal(int sid, ulong sig);
  /*
   * modifies: nothing
   * effects:  Sends signal sig to all tasks associated with sid.
   *           Returns count of tasks successfully signalled.
   *
   */

/*
 * session notification interface 
 *
 */

/* invalid id */
#define	ITSY_SESSION_NOTIFY_ID_INVALID	(-1)

/* callback prototypes */

typedef void  (*session_notify_callback)(void *private, int sid);
typedef char* (*session_query_callback) (void *private, int sid);

extern int itsy_session_notify_register(void *private, 
					const char *name,
					session_notify_callback create,
					session_notify_callback destroy,
					session_notify_callback switchto,
					session_query_callback  query);
  /*
   * modifies: nothing
   * effects:  Register as a session notification client.  Session
   *           notifications occur via the client-specified callbacks.
   *           Returns a unique identifier for use with future calls,
   *           or ITSY_SESSION_NOTIFY_ID_INVALID if unable to register.
   *
   */

extern int itsy_session_notify_unregister(int id);
  /*
   * modifies: nothing
   * effects:  Deregisters id as a session notification client.
   *           Returns 0 iff successful, otherwise error code.
   *
   */

#endif	/* __KERNEL__ */

#endif	/* ITSY_SESSION_H */
