/*
 * Power Management Interface
 * Copyright (c) Compaq Computer Corporation, 1998
 *
 * Author: Carl Waldspurger
 *
 * $Log: powermgr.h,v $
 * Revision 2.1  1999/12/23  01:28:30  kerr
 * Updated Copyright information for entire source tree.
 *
 * Revision 2.0  1998/06/25  18:41:45  kramer
 * Major Bump: Update all files to version 2.0 for external release
 * -drew
 *
 * Revision 1.5  1998/06/16  19:20:09  kramer
 * Added copyright information
 *
 * Revision 1.4  1998/04/14  02:23:32  caw
 * Updated suspend_check() interfaces.
 *
 * Revision 1.3  1998/03/28  00:15:41  caw
 * Defined powermgr_init_ksyms().
 *
 * Revision 1.2  1998/03/26  00:35:25  caw
 * Updated interface.
 *
 * Revision 1.1  1998/03/20  22:59:48  caw
 * Initial revision.
 *
 */

#ifndef	powermgr_h
#define	powermgr_h

/*
 * constants
 *
 */

/* ids */
#define	POWERMGR_ID_INVALID	(-1)
#define	POWERMGR_ID_MAX		(31)

/* status */
#define	POWERMGR_SUCCESS	(0)
#define	POWERMGR_FAILURE	(-1)

/* retry delay */
#define	POWERMGR_RETRY_UNKNOWN	(-1)

/* resume flags */
#define	POWERMGR_RESUME_FROM_SLEEP		(1 << 0)
#define	POWERMGR_RESUME_FROM_BATTERY_FAULT	(1 << 1)
#define	POWERMGR_RESUME_FROM_VDD_FAULT		(1 << 2)

/*
 * types
 *
 */

/* client identifier */
typedef int powermgr_id;

/*
 * client callback functions
 *
 */

typedef int (*powermgr_suspend_check_callback_fun)(void *cookie, int idle);
typedef int (*powermgr_suspend_callback_fun)(void *cookie);
typedef int (*powermgr_resume_callback_fun)(void *cookie, int resume_flags);

/*
 * int client_suspend_check(void *cookie, int idle_jiffies);
 *   modifies: cookie
 *   effects:  Checks if client is willing to be suspended.  Suspending
 *             after an inactive period of idle_jiffies is recommended.
 *             Returns 0 iff suspendible, otherwise returns a retry delay
 *             measured in jiffies (or POWERMGR_RETRY_UNKNOWN).
 *
 * int client_suspend(void *cookie);
 *   modifies: cookie
 *   effects:  Saves state in preparation for imminent suspension.
 *             Return value is currently ignored.
 *
 * int client_resume(void *cookie, int resume_flags);
 *   modifies: cookie
 *   effects:  Restores state after wakeup from suspension.  The cause of
 *             the suspension is encoded by the bitmask resume_flags
 *             (e.g. POWERMGR_RESUME_FROM_SLEEP, _FROM_BATTERY_FAULT, etc).
 *             Return value is currently ignored.
 *
 */

typedef struct {
  /* callback functions */
  powermgr_suspend_check_callback_fun suspend_check;
  powermgr_suspend_callback_fun suspend;
  powermgr_resume_callback_fun  resume;

  /* uninterpreted client data */
  void *cookie;

  /* identity */
  powermgr_id id;
  char *name;

  /* power-consumption info */
  int uamps_idle;
  int uamps_sleep;
} powermgr_client;

/*
 * client operations
 *
 */

extern powermgr_id powermgr_register(const powermgr_client *client);
  /*
   * effects:  Register as a power manager client.  The power manager
   *           will communicate via the client-specified callbacks.
   *           Returns a unique identifier for use with future calls,
   *           or POWERMGR_ID_INVALID if unable to register.
   * storage:  Data is copied from client, so it is safe to deallocate
   *           client after registering.
   * caveats:  The "id" field specified by client is ignored.
   *
   */

extern int powermgr_unregister(powermgr_id id);
  /*
   * effects:  Deregisters id as a power manager client.
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

/*
 * power management operations
 *
 */

extern int powermgr_init(void);
  /*
   * effects:  Initializes power manager.
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

extern void powermgr_init_ksyms(void);
  /*
   * effects:  Exports powermgr kernel symbols for modules.
   * caveats:  Separate function due to boot order restrictions.
   *
   */

extern void powermgr_init_procfs(void);
  /*
   * effects:  Registers /proc/powermgr device.
   * caveats:  Separate function due to boot order restrictions.
   *
   */

extern int powermgr_suspend_check(int idle_jiffies, int *retry_jiffies);
  /*
   * modifies: retry_jiffies
   * effects:  Returns TRUE iff all registered clients are willing
   *           to be suspended.  Clients should be willing to 
   *           suspend after an inactivity period of idle_jiffies.
   *           Otherwise returns FALSE and sets retry_jiffies to a 
   *           reasonable time delay in jiffies, after which another
   *           check may return TRUE.
   *
   */

extern int powermgr_suspend(void);
  /*
   * effects:  Forcibly suspends all registered clients, in the reverse order
   *           of registration (i.e. first to register suspends last).
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

extern int powermgr_resume(int resume_flags);
  /*
   * effects:  Resumes all registered clients, in the order of registration
   *           (i.e. first to register resumes first).
   *           Returns POWERMGR_SUCCESS or POWERMGR_FAILURE.
   *
   */

extern void powermgr_start_timer(int idle_jiffies, int check_jiffies);
  /*
   * effects:  Starts periodic timer to initiate suspends.  Clients are
   *           recommended to suspend after idle_jiffies of inactivity.
   *           Suspend checks will be made at most once every check_jiffies.
   * 
   */

#endif	/* powermgr_h */
