/*
 * Flash Support
 * Ported from Itsy Monitor => Linux
 *
 * Copyright (c) Compaq Computer Corporation, 1999
 *
 * $Log: flprgm.c,v $
 * Revision 2.9  2000/01/19  18:25:14  kerr
 * More copyright conformance.
 *
 * Revision 2.8  1999/11/23  04:53:58  mann
 * Rewrote char flash driver to be fully compatible with block ftl driver.
 * Both now live in arch/arm/drivers/block/ftl_itsy.c.
 * The flash.c file in this directory still contains low-level support
 * and the /proc/flash implementation.
 *
 * Removed some of the linux code from flprgm.c that is no longer used after
 * the rewrite.
 *
 * Revision 2.7  1999/10/23  02:50:11  mann
 * Bug fix: On Am29DLxxx chips, erase suspend, erase resume, and unlock bypass
 * reset must all be directed to an address in the correct internal bank.
 *
 * Revision 2.6  1999/10/22 21:55:34  mann
 * Fixed a bug in FlMemIdentify for Am29DLxxxT (top boot block) flash chips.
 *
 * Revision 2.5  1999/10/09 00:17:22  mann
 * Added support for 32Mbit flash devices.
 * Added /dev/flash[0-3]o (minor devices 4-7), which write without erasing
 * first.
 *
 * Revision 2.4  1999/05/01  04:09:03  mann
 * Switched from __get_free_pages to vmalloc, which allows us to allocate
 * more than 128KB.  The limit was a problem with super daughtercard flash
 * banks, which have a 256KB erase sector size.
 *
 * Revision 2.3  1999/04/29 02:49:26  mann
 * Itsy FTL (flash translation layer) support, initial working version.
 *
 * Revision 2.2  1999/01/15 01:37:57  caw
 * Updated Itsy port to reflect latest Itsy monitor code, which supports
 * additional flash memory types and automatically handles unaligned accesses.
 * Also changed conditional compilation from "ITSY_LINUX" => "__linux__".
 *
 * Revision 2.1  1998/07/09  23:11:55  caw
 * Substantial modifications for interrupt-driven flash programming.
 * Currently blocks for sector erase and polls for word programming.
 * Dynamically allocate all transient buffers.  Improved /proc/flash support.
 *
 * Revision 2.0  1998/06/24  21:42:50  kramer
 * Major Bump: Move all version numbers up to 2.0 for external release.
 * -drew
 *
 * Revision 1.2  1998/06/24  02:25:00  caw
 * Statically allocate DataBuf in FlMemPrgm() to avoid kernel stack problems.
 *
 * Revision 1.1  1998/06/22  20:58:06  caw
 * Initial port of Itsy monitor code to Linux.
 *
 */

/*
 *	FILE    	flprgm.c
 *
 *	System  	Itsy versions 1.0/1.1, 1.5, 1.6, 1.7, and 2.0/2.1
 *	        	(StrongARM SA-1100)
 *	Language	C
 *	Purpose 	Flash-memory programming procedures for the Itsy
 *	        	computer.
 */

#ifdef	__linux__

/* linux includes */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/flinfo.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/arch/mmap.h>
#include <asm/arch/ARMv4.h>
#include <asm/arch/SA-1100.h>
#include <asm/arch/F29xxx.h>
#include <asm/arch/Am29xxx.h>
#include <asm/arch/itsymon-error.h>

#else

/* monitor includes */
#include <ARMv4.h>
#include <SA-1100.h>
#include <F29xxx.h>
#include <Am29xxx.h>
#include <error.h>
#include <misc2.h>
#include <flprgm.h>
#include <flinfo.h>
#include <itsymon.h>

#endif	/* __linux__ */

/*
 * GLOBAL DEFINITIONS
 */

/* Always busy-wait.  WAIT_ERASE is never used by Linux, others are. */
#define	WAIT_PROGRAM(vaddr)
#define	WAIT_ERASE(vaddr)
#define	WAIT_SUSPEND(vaddr)

#define FlMemCir2x8(Data, Nb) \
                	((Word16) F29_8Data (Data) << (Nb)*ByteWdth)
#define FlMemCir4x8(Data, Nb) \
                	((Word32) F29_8Data (Data) << (Nb)*ByteWdth)
#define FlMemCir1x16(Data, Nb) \
                	((Word16) F29_16Data (Data) << (Nb)*2*ByteWdth)
#define FlMemCir2x16(Data, Nb) \
                	((Word32) F29_16Data (Data) << (Nb)*2*ByteWdth)

#define FlMem2x8(Data)	(FlMemCir2x8 (Data, 0) + FlMemCir2x8 (Data, 1))
#define FlMem4x8(Data)	(FlMemCir4x8 (Data, 0) + FlMemCir4x8 (Data, 1) + \
                	 FlMemCir4x8 (Data, 2) + FlMemCir4x8 (Data, 3))
#define FlMem1x16(Data)	FlMemCir1x16 (Data, 0)
#define FlMem2x16(Data)	(FlMemCir2x16 (Data, 0) + FlMemCir2x16 (Data, 1))

#ifdef	__linux__
/*
 * Itsy Linux Modifications
 *
 */

/* make itsymon "globals" local */
static unsigned int ErrStatus;
static Address ErrAddress;
static Word ErrData;

void FlMemGetError(int *status,
		   Address *addr,
		   Word *data)
{
  /*
   * modifies: status, addr, data
   * effects:  Sets status, addr, and data to the error information
   *           associated with the last call to FlMemPrgm().
   *
   */

  *status = ErrStatus;
  *addr   = ErrAddress;
  *data   = ErrData;
}
#endif	/* __linux__ */


/*
 * PRIVATE PROCEDURE: A29_SctErs2x8
 *
 * Purpose
 *    The procedure "A29_SctErs2x8" starts the erase of a sector in a
 *    2 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErs2x8	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int A29_SctErs2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem2x8 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem2x8 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_8Add (A29_ErsAdd)] = FlMem2x8 (A29_ErsCmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem2x8 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem2x8 (A29_Unlck2Cmd) ;
   *FlMemSctAdd = FlMem2x8 (A29_SctErsCmd) ;
   return (ErrStatus = Success);
}

/*
 * PRIVATE PROCEDURE: A29_SctErsSus2x8
 *
 * Purpose
 *    The procedure "A29_SctErs2x8" suspends the erase of a sector in a
 *    2 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsSus2x8	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 * Notes
 *    Assumes that an erase was in progress or just completed on the specified
 *    sector; if instead the sector contains data, the implementation below 
 *    may loop forever.
 */

static unsigned int A29_SctErsSus2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16               FlMemVal;

   /* Issue command */
   *FlMemSctAdd = FlMem2x8 (A29_ErsSusCmd) ;

   /* Wait for erasing to stop on all chips */
   for (;;) {
      /* Read: either status or erased data */
      FlMemVal = *FlMemSctAdd;
      if ((FlMemVal & FlMemCir2x8 (A29_ErrTim + A29_BsyInvData, 0)) != 0 &&
	  (FlMemVal & FlMemCir2x8 (A29_ErrTim + A29_BsyInvData, 1)) != 0) {
	 /* All chips are either failed, done, or suspended */
	 break;
      }
      /* Else at least one chip is still erasing */
      WAIT_SUSPEND((void *) FlMemSctAdd);
   }

   /* Reread to be sure all bits are valid */
   FlMemVal = *FlMemSctAdd;
   if (((FlMemVal & FlMemCir2x8 (A29_BsyInvData, 0)) ==
	FlMemCir2x8 (A29_BsyInvData, 0)) &&
       ((FlMemVal & FlMemCir2x8 (A29_BsyInvData, 1)) ==
	FlMemCir2x8 (A29_BsyInvData, 1))) {
      /* All chips either done or suspended */
      if (FlMemVal == (Word16) -1) {
	 return (ErrStatus = -1);  /* done */
      } else {
	 return (ErrStatus = Success);  /* suspended */
      }
   }

   /* At least one erase failed */
   *FlMemBsAdd = FlMem2x8 (A29_RstCmd) ;
   ErrAddress = (Address) FlMemSctAdd;
   ErrData = -1;
   return (ErrStatus = FlSctErs2x8Err);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsRes2x8
 *
 * Purpose
 *    The procedure "A29_SctErs2x8" resumes the erase of a sector in a
 *    2 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsRes2x8	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int A29_SctErsRes2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   *FlMemSctAdd = FlMem2x8 (A29_ErsResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsStat2x8
 *
 * Purpose
 *    The procedure "A29_SctErs2x8" tests the status of a sector erase in a 
 *    2 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsStat2x8	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, to the corresponding error
 *			number if an error occurred, or to -1 if the erase
 *			is still in progress.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int A29_SctErsStat2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16       	FlMemVal ;

   FlMemVal = *FlMemSctAdd;
   if ((FlMemVal & FlMemCir2x8 (A29_ErrTim + A29_BsyInvData, 0)) != 0 &&
       (FlMemVal & FlMemCir2x8 (A29_ErrTim + A29_BsyInvData, 1)) != 0) {
      /* All chips are either failed or done */
      /* (caller guarantees they are not suspended) */

      /* Reread to be sure all bits are valid */
      if (*FlMemSctAdd == (Word16) -1) {
	 return (ErrStatus = Success) ;
      } else {
         *FlMemBsAdd = FlMem2x8 (A29_RstCmd) ;
         ErrAddress = (Address) FlMemSctAdd ;
         return (ErrStatus = FlSctErs2x8Err) ;
      }
   } else {
      /* At least one chip is still erasing */
      return (ErrStatus = -1);
   }
}


/*
 * PRIVATE PROCEDURE: A29_SctErs4x8
 *
 * Purpose
 *    The procedure "A29_SctErs4x8" starts the erase of a sector in a
 *    4 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErs4x8	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int A29_SctErs4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem4x8 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem4x8 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_8Add (A29_ErsAdd)] = FlMem4x8 (A29_ErsCmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem4x8 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem4x8 (A29_Unlck2Cmd) ;
   *FlMemSctAdd = FlMem4x8 (A29_SctErsCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: A29_SctErsSus4x8
 *
 * Purpose
 *    The procedure "A29_SctErs4x8" suspends the erase of a sector in a
 *    4 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsSus4x8	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 * Notes
 *    Assumes that an erase was in progress or just completed on the specified
 *    sector; if instead the sector contains data, the implementation below 
 *    may loop forever.
 */

static unsigned int A29_SctErsSus4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32               FlMemVal;

   /* Issue command */
   *FlMemSctAdd = FlMem4x8 (A29_ErsSusCmd) ;

   /* Wait for erasing to stop on all chips */
   for (;;) {
      /* Read: either status or erased data */
      FlMemVal = *FlMemSctAdd;
      if ((FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 0)) != 0 &&
	  (FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 1)) != 0 &&
          (FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 2)) != 0 &&
	  (FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 3)) != 0) {
	 /* All chips are either failed, done, or suspended */
	 break;
      }
      /* Else at least one chip is still erasing */
      WAIT_SUSPEND((void *) FlMemSctAdd);
   }

   /* Reread to be sure all bits are valid */
   FlMemVal = *FlMemSctAdd;
   if (((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 0)) ==
	FlMemCir4x8 (A29_BsyInvData, 0)) &&
       ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 1)) ==
	FlMemCir4x8 (A29_BsyInvData, 1)) &&
       ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 2)) ==
	FlMemCir4x8 (A29_BsyInvData, 2)) &&
       ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 3)) ==
	FlMemCir4x8 (A29_BsyInvData, 3))) {
      /* All chips either done or suspended */
      if (FlMemVal == (Word32) -1) {
	 return (ErrStatus = -1);  /* done */
      } else {
	 return (ErrStatus = Success);  /* suspended */
      }
   }

   /* At least one erase failed */
   *FlMemBsAdd = FlMem4x8 (A29_RstCmd) ;
   ErrAddress = (Address) FlMemSctAdd;
   ErrData = -1;
   return (ErrStatus = FlSctErs4x8Err);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsRes4x8
 *
 * Purpose
 *    The procedure "A29_SctErs4x8" resumes the erase of a sector in a
 *    4 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsRes4x8	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int A29_SctErsRes4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   *FlMemSctAdd = FlMem4x8 (A29_ErsResCmd) ;
   return (ErrStatus = Success);
}

/*
 * PRIVATE PROCEDURE: A29_SctErsStat4x8
 *
 * Purpose
 *    The procedure "A29_SctErsStat4x8" tests the status of a sector erase in a
 *    4 x 8-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsStat4x8	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error 
 *                      number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int A29_SctErsStat4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32       	FlMemVal ;

   FlMemVal = *FlMemSctAdd;
   if ((FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 0)) != 0 &&
       (FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 1)) != 0 &&
       (FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 2)) != 0 &&
       (FlMemVal & FlMemCir4x8 (A29_ErrTim + A29_BsyInvData, 3)) != 0) {
      /* All chips are either failed or done */
      /* (caller guarantees they are not suspended) */

      /* Reread to be sure all bits are valid */
      if (*FlMemSctAdd == (Word32) -1) {
	 return (ErrStatus = Success) ;
      } else {
         *FlMemBsAdd = FlMem4x8 (A29_RstCmd) ;
         ErrAddress = (Address) FlMemSctAdd ;
         return (ErrStatus = FlSctErs4x8Err) ;
      }
   } else {
      /* At least one chip is still erasing */
      return (ErrStatus = -1);
   }
}


/*
 * PRIVATE PROCEDURE: A29_SctErs1x16
 *
 * Purpose
 *    The procedure "A29_SctErs1x16" starts the erase of a sector in a
 *    1 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErs1x16	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int A29_SctErs1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem1x16 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem1x16 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_16Add (A29_ErsAdd)] = FlMem1x16 (A29_ErsCmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem1x16 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem1x16 (A29_Unlck2Cmd) ;
   *FlMemSctAdd = FlMem1x16 (A29_SctErsCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: A29_SctErsSus1x16
 *
 * Purpose
 *    The procedure "A29_SctErs1x16" suspends the erase of a sector in a
 *    1 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsSus1x16	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 * Notes
 *    Assumes that an erase was in progress or just completed on the specified
 *    sector; if instead the sector contains data, the implementation below 
 *    may loop forever.
 */

static unsigned int A29_SctErsSus1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16               FlMemVal;

   /* Issue command */
   *FlMemSctAdd = FlMem1x16 (A29_ErsSusCmd) ;

   /* Wait for erasing to stop on all chips */
   for (;;) {
      /* Read: either status or erased data */
      FlMemVal = *FlMemSctAdd;
      if ((FlMemVal & FlMemCir1x16 (A29_ErrTim + A29_BsyInvData, 0)) != 0) {
	 /* All chips are either failed, done, or suspended */
	 break;
      }
      /* Else at least one chip is still erasing */
      WAIT_SUSPEND((void *) FlMemSctAdd);
   }

   /* Reread to be sure all bits are valid */
   FlMemVal = *FlMemSctAdd;
   if (((FlMemVal & FlMemCir1x16 (A29_BsyInvData, 0)) ==
	FlMemCir1x16 (A29_BsyInvData, 0))) {
      /* All chips either done or suspended */
      if (FlMemVal == (Word16) -1) {
	 return (ErrStatus = -1);  /* done */
      } else {
	 return (ErrStatus = Success);  /* suspended */
      }
   }

   /* At least one erase failed */
   *FlMemBsAdd = FlMem1x16 (A29_RstCmd) ;
   ErrAddress = (Address) FlMemSctAdd;
   ErrData = -1;
   return (ErrStatus = FlSctErs1x16Err);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsRes1x16
 *
 * Purpose
 *    The procedure "A29_SctErs1x16" resumes the erase of a sector in a
 *    1 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsRes1x16	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int A29_SctErsRes1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   *FlMemSctAdd = FlMem1x16 (A29_ErsResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsStat1x16
 *
 * Purpose
 *    The procedure "A29_SctErsStat1x16" tests the status of a sector erase in a
 *    1 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErs1x16	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error
 *                      number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int A29_SctErsStat1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16 FlMemVal;

   FlMemVal = *FlMemSctAdd;
   if ((FlMemVal & FlMemCir1x16 (A29_ErrTim + A29_BsyInvData, 0)) != 0) {
      /* All chips are either failed or done */
      /* (caller guarantees they are not suspended) */

      /* Reread to be sure all bits are valid */
      if (*FlMemSctAdd == (Word16) -1) {
	 return (ErrStatus = Success) ;
      } else {
         *FlMemBsAdd = FlMem1x16 (A29_RstCmd) ;
         ErrAddress = (Address) FlMemSctAdd ;
         return (ErrStatus = FlSctErs1x16Err) ;
      }
   } else {
      /* At least one chip is still erasing */
      return (ErrStatus = -1);
   }
}


/*
 * PRIVATE PROCEDURE: A29_SctErs2x16
 *
 * Purpose
 *    The procedure "A29_SctErs2x16" starts the erase of a sector in a
 *    2 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErs2x16	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int A29_SctErs2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem2x16 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem2x16 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_16Add (A29_ErsAdd)] = FlMem2x16 (A29_ErsCmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem2x16 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem2x16 (A29_Unlck2Cmd) ;
   *FlMemSctAdd = FlMem2x16 (A29_SctErsCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: A29_SctErsSus2x16
 *
 * Purpose
 *    The procedure "A29_SctErs2x16" suspends the erase of a sector in a
 *    2 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsSus2x16 Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 * Notes
 *    Assumes that an erase was in progress or just completed on the specified
 *    sector; if instead the sector contains data, the implementation below 
 *    may loop forever.
 */

static unsigned int A29_SctErsSus2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32               FlMemVal;

   /* Issue command */
   *FlMemSctAdd = FlMem2x16 (A29_ErsSusCmd) ;

   /* Wait for erasing to stop on all chips */
   for (;;) {
      /* Read: either status or erased data */
      FlMemVal = *FlMemSctAdd;
      if ((FlMemVal & FlMemCir2x16 (A29_ErrTim + A29_BsyInvData, 0)) != 0 &&
	  (FlMemVal & FlMemCir2x16 (A29_ErrTim + A29_BsyInvData, 1)) != 0) {
	 /* All chips are either failed, done, or suspended */
	 break;
      }
      /* Else at least one chip is still erasing */
      WAIT_SUSPEND((void *) FlMemSctAdd);
   }

   /* Reread to be sure all bits are valid */
   FlMemVal = *FlMemSctAdd;
   if (((FlMemVal & FlMemCir2x16 (A29_BsyInvData, 0)) ==
	FlMemCir2x16 (A29_BsyInvData, 0)) &&
       ((FlMemVal & FlMemCir2x16 (A29_BsyInvData, 1)) ==
	FlMemCir2x16 (A29_BsyInvData, 1))) {
      /* All chips either done or suspended */
      if (FlMemVal == (Word32) -1) {
	 return (ErrStatus = -1);  /* done */
      } else {
	 return (ErrStatus = Success);  /* suspended */
      }
   }

   /* At least one erase failed */
   *FlMemBsAdd = FlMem2x16 (A29_RstCmd) ;
   ErrAddress = (Address) FlMemSctAdd;
   ErrData = -1;
   return (ErrStatus = FlSctErs2x16Err);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsRes2x16
 *
 * Purpose
 *    The procedure "A29_SctErs2x16" resumes the erase of a sector in a
 *    2 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErsRes2x16	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int A29_SctErsRes2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   *FlMemSctAdd = FlMem2x16 (A29_ErsResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: A29_SctErsStat2x16
 *
 * Purpose
 *    The procedure "A29_SctErsStat2x16" tests the status of a sector erase in a
 *    2 x 16-bit AMD Am29xxxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    A29_SctErs2x16	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int A29_SctErsStat2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32       	FlMemVal ;

   FlMemVal = *FlMemSctAdd;
   if ((FlMemVal & FlMemCir2x16 (A29_ErrTim + A29_BsyInvData, 0)) != 0 &&
       (FlMemVal & FlMemCir2x16 (A29_ErrTim + A29_BsyInvData, 1)) != 0) {
      /* All chips are either failed or done */
      /* (caller guarantees they are not suspended) */

      /* Reread to be sure all bits are valid */
      if (*FlMemSctAdd == (Word32) -1) {
	 return (ErrStatus = Success) ;
      } else {
         *FlMemBsAdd = FlMem2x16 (A29_RstCmd) ;
         ErrAddress = (Address) FlMemSctAdd ;
         return (ErrStatus = FlSctErs2x16Err) ;
      }
   } else {
      /* At least one chip is still erasing */
      return (ErrStatus = -1);
   }
}


/*
 * PRIVATE PROCEDURE: A29_Prgm2x8
 *
 * Purpose
 *    The procedure "A29_Prgm2x8" programs (i.e., writes) a given data
 *    buffer into a 2 x 8-bit AMD Am29xxxxx flash memory, using the standard
 *    programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm2x8	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemAdd,
   Word16       	*DataAdd,
   unsigned int 	DataSz)
{
   Word16       	Data, FlMemVal ;

   while ((int) (DataSz -= sizeof (Word16)) >= 0)
   {
      FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem2x8 (A29_Unlck1Cmd) ;
      FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem2x8 (A29_Unlck2Cmd) ;
      FlMemBsAdd [F29_8Add (A29_PrgmAdd)] = FlMem2x8 (A29_PrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!((((FlMemVal = *FlMemAdd) & FlMemCir2x8 (A29_BsyInvData, 0)) ==
                (Data & FlMemCir2x8 (A29_BsyInvData, 0)) ||
                FlMemVal & FlMemCir2x8 (A29_ErrTim, 0)) &&
               ((FlMemVal & FlMemCir2x8 (A29_BsyInvData, 1)) ==
                (Data & FlMemCir2x8 (A29_BsyInvData, 1)) ||
                FlMemVal & FlMemCir2x8 (A29_ErrTim, 1))))
         WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem2x8 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm2x8Err) ;
      }
   }
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm4x8
 *
 * Purpose
 *    The procedure "A29_Prgm4x8" programs (i.e., writes) a given data
 *    buffer into a 4 x 8-bit AMD Am29xxxxx flash memory, using the standard
 *    programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm4x8	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemAdd,
   Word32       	*DataAdd,
   unsigned int 	DataSz)
{
   Word32       	Data, FlMemVal ;

   while ((int) (DataSz -= sizeof (Word32)) >= 0)
   {
      FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem4x8 (A29_Unlck1Cmd) ;
      FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem4x8 (A29_Unlck2Cmd) ;
      FlMemBsAdd [F29_8Add (A29_PrgmAdd)] = FlMem4x8 (A29_PrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!((((FlMemVal = *FlMemAdd) & FlMemCir4x8 (A29_BsyInvData, 0)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 0)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 0)) &&
               ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 1)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 1)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 1)) &&
               ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 2)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 2)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 2)) &&
               ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 3)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 3)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 3))))
         WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem4x8 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm4x8Err) ;
      }
   }
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm1x16
 *
 * Purpose
 *    The procedure "A29_Prgm1x16" programs (i.e., writes) a given data
 *    buffer into a 1 x 16-bit AMD Am29xxxxx flash memory, using the standard
 *    programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm1x16	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemAdd,
   Word16       	*DataAdd,
   unsigned int 	DataSz)
{
   Word16       	Data, FlMemVal ;

   while ((int) (DataSz -= sizeof (Word16)) >= 0)
   {
      FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem1x16 (A29_Unlck1Cmd) ;
      FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem1x16 (A29_Unlck2Cmd) ;
      FlMemBsAdd [F29_16Add (A29_PrgmAdd)] = FlMem1x16 (A29_PrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!(((FlMemVal = *FlMemAdd) & FlMemCir1x16 (A29_BsyInvData, 0)) ==
               (Data & FlMemCir1x16 (A29_BsyInvData, 0)) ||
               FlMemVal & FlMemCir1x16 (A29_ErrTim, 0)))
	WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem1x16 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm1x16Err) ;
      }
   }
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm2x16
 *
 * Purpose
 *    The procedure "A29_Prgm2x16" programs (i.e., writes) a given data
 *    buffer into a 2 x 16-bit AMD Am29xxxxx flash memory, using the standard
 *    programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm2x16	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemAdd,
   Word32       	*DataAdd,
   unsigned int 	DataSz)
{
   Word32       	Data, FlMemVal ;

   while ((int) (DataSz -= sizeof (Word32)) >= 0)
   {
      FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem2x16 (A29_Unlck1Cmd) ;
      FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem2x16 (A29_Unlck2Cmd) ;
      FlMemBsAdd [F29_16Add (A29_PrgmAdd)] = FlMem2x16 (A29_PrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!((((FlMemVal = *FlMemAdd) & FlMemCir2x16 (A29_BsyInvData, 0)) ==
                (Data & FlMemCir2x16 (A29_BsyInvData, 0)) ||
                FlMemVal & FlMemCir2x16 (A29_ErrTim, 0)) &&
               ((FlMemVal & FlMemCir2x16 (A29_BsyInvData, 1)) ==
                (Data & FlMemCir2x16 (A29_BsyInvData, 1)) ||
                FlMemVal & FlMemCir2x16 (A29_ErrTim, 1))))
	WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem2x16 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm2x16Err) ;
      }
   }
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm2x8Byp
 *
 * Purpose
 *    The procedure "A29_Prgm2x8Byp" programs (i.e., writes) a given data
 *    buffer into a 2 x 8-bit AMD Am29xxxxx flash memory, using the "unlock
 *    bypass" programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm2x8Byp	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm2x8Byp (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemAdd,
   Word16       	*DataAdd,
   unsigned int 	DataSz)
{
   Word16       	Data, FlMemVal ;

   FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem2x8 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem2x8 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_8Add (A29_BypAdd)] = FlMem2x8 (A29_BypCmd) ;
   while ((int) (DataSz -= sizeof (Word16)) >= 0)
   {
      *FlMemBsAdd = FlMem2x8 (A29_BypPrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!((((FlMemVal = *FlMemAdd) & FlMemCir2x8 (A29_BsyInvData, 0)) ==
                (Data & FlMemCir2x8 (A29_BsyInvData, 0)) ||
                FlMemVal & FlMemCir2x8 (A29_ErrTim, 0)) &&
               ((FlMemVal & FlMemCir2x8 (A29_BsyInvData, 1)) ==
                (Data & FlMemCir2x8 (A29_BsyInvData, 1)) ||
                FlMemVal & FlMemCir2x8 (A29_ErrTim, 1))))
         WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem2x8 (A29_BypRst1Cmd) ;
         *FlMemBsAdd = FlMem2x8 (A29_BypRst2Cmd) ;
         *FlMemBsAdd = FlMem2x8 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm2x8Err) ;
      }
   }
   *FlMemBsAdd = FlMem2x8 (A29_BypRst1Cmd) ;
   *FlMemBsAdd = FlMem2x8 (A29_BypRst2Cmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm4x8Byp
 *
 * Purpose
 *    The procedure "A29_Prgm4x8Byp" programs (i.e., writes) a given data
 *    buffer into a 4 x 8-bit AMD Am29xxxxx flash memory, using the "unlock
 *    bypass" programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm4x8Byp	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm4x8Byp (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemAdd,
   Word32       	*DataAdd,
   unsigned int 	DataSz)
{
   Word32       	Data, FlMemVal ;

   FlMemBsAdd [F29_8Add (A29_Unlck1Add)] = FlMem4x8 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_8Add (A29_Unlck2Add)] = FlMem4x8 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_8Add (A29_BypAdd)] = FlMem4x8 (A29_BypCmd) ;
   while ((int) (DataSz -= sizeof (Word32)) >= 0)
   {
      *FlMemBsAdd = FlMem4x8 (A29_BypPrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!((((FlMemVal = *FlMemAdd) & FlMemCir4x8 (A29_BsyInvData, 0)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 0)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 0)) &&
               ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 1)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 1)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 1)) &&
               ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 2)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 2)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 2)) &&
               ((FlMemVal & FlMemCir4x8 (A29_BsyInvData, 3)) ==
                (Data & FlMemCir4x8 (A29_BsyInvData, 3)) ||
                FlMemVal & FlMemCir4x8 (A29_ErrTim, 3))))
         WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem4x8 (A29_BypRst1Cmd) ;
         *FlMemBsAdd = FlMem4x8 (A29_BypRst2Cmd) ;
         *FlMemBsAdd = FlMem4x8 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm4x8Err) ;
      }
   }
   *FlMemBsAdd = FlMem4x8 (A29_BypRst1Cmd) ;
   *FlMemBsAdd = FlMem4x8 (A29_BypRst2Cmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm1x16Byp
 *
 * Purpose
 *    The procedure "A29_Prgm1x16Byp" programs (i.e., writes) a given data
 *    buffer into a 1 x 16-bit AMD Am29xxxxx flash memory, using the "unlock
 *    bypass" programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm1x16Byp	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm1x16Byp (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemAdd,
   Word16       	*DataAdd,
   unsigned int 	DataSz)
{
   Word16       	Data, FlMemVal ;

   FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem1x16 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem1x16 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_16Add (A29_BypAdd)] = FlMem1x16 (A29_BypCmd) ;
   while ((int) (DataSz -= sizeof (Word16)) >= 0)
   {
      *FlMemBsAdd = FlMem1x16 (A29_BypPrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!(((FlMemVal = *FlMemAdd) & FlMemCir1x16 (A29_BsyInvData, 0)) ==
               (Data & FlMemCir1x16 (A29_BsyInvData, 0)) ||
               FlMemVal & FlMemCir1x16 (A29_ErrTim, 0)))
	WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem1x16 (A29_BypRst1Cmd) ;
         *FlMemBsAdd = FlMem1x16 (A29_BypRst2Cmd) ;
         *FlMemBsAdd = FlMem1x16 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm1x16Err) ;
      }
   }
   *FlMemBsAdd = FlMem1x16 (A29_BypRst1Cmd) ;
   *FlMemBsAdd = FlMem1x16 (A29_BypRst2Cmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: A29_Prgm2x16Byp
 *
 * Purpose
 *    The procedure "A29_Prgm2x16Byp" programs (i.e., writes) a given data
 *    buffer into a 2 x 16-bit AMD Am29xxxxx flash memory, using the "unlock
 *    bypass" programming algorithm.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    A29_Prgm2x16Byp	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

static unsigned int A29_Prgm2x16Byp (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemAdd,
   Word32       	*DataAdd,
   unsigned int 	DataSz)
{
   Word32       	Data, FlMemVal ;

   FlMemBsAdd [F29_16Add (A29_Unlck1Add)] = FlMem2x16 (A29_Unlck1Cmd) ;
   FlMemBsAdd [F29_16Add (A29_Unlck2Add)] = FlMem2x16 (A29_Unlck2Cmd) ;
   FlMemBsAdd [F29_16Add (A29_BypAdd)] = FlMem2x16 (A29_BypCmd) ;
   while ((int) (DataSz -= sizeof (Word32)) >= 0)
   {
      *FlMemBsAdd = FlMem2x16 (A29_BypPrgmCmd) ;
      *FlMemAdd = (Data = *(DataAdd++)) ;
      while (!((((FlMemVal = *FlMemAdd) & FlMemCir2x16 (A29_BsyInvData, 0)) ==
                (Data & FlMemCir2x16 (A29_BsyInvData, 0)) ||
                FlMemVal & FlMemCir2x16 (A29_ErrTim, 0)) &&
               ((FlMemVal & FlMemCir2x16 (A29_BsyInvData, 1)) ==
                (Data & FlMemCir2x16 (A29_BsyInvData, 1)) ||
                FlMemVal & FlMemCir2x16 (A29_ErrTim, 1))))
	WAIT_PROGRAM((void *) FlMemAdd);
      if (*(FlMemAdd++) != Data)
      {
         *FlMemBsAdd = FlMem2x16 (A29_BypRst1Cmd) ;
         *FlMemBsAdd = FlMem2x16 (A29_BypRst2Cmd) ;
         *FlMemBsAdd = FlMem2x16 (A29_RstCmd) ;
         ErrAddress = (Address) &(FlMemAdd [-1]) ;
         ErrData = Data ;
         return (ErrStatus = FlPrgm2x16Err) ;
      }
   }
   *FlMemBsAdd = FlMem2x16 (A29_BypRst1Cmd) ;
   *FlMemBsAdd = FlMem2x16 (A29_BypRst2Cmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: F29_SctErs2x8
 *
 * Purpose
 *    The procedure "F29_SctErs2x8" starts the erase of a sector in a
 *    2 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs2x8	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_SctErs2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem2x8 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem2x8 (F29_SctErsCmd) ;
   *FlMemSctAdd = FlMem2x8 (F29_CnfrmCmd) ;
   *FlMemBsAdd = FlMem2x8 (F29_StatRdCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: F29_SctErsSus2x8
 *
 * Purpose
 *    The procedure "F29_SctErsSus2x8" suspends the erase of a sector in a
 *    2 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsSus2x8	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsSus2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16		Status;

   *FlMemBsAdd = FlMem2x8 (F29_SusCmd) ;
   Status = *FlMemBsAdd & FlMem2x8 (F29_Sus) ;
   if (Status) 
      return (ErrStatus = Success); /* suspended */
   else
      return (ErrStatus = -1); /* complete */
}


/*
 * PRIVATE PROCEDURE: F29_SctErsRes2x8
 *
 * Purpose
 *    The procedure "F29_SctErsRes2x8" resumes the erase of a sector in a
 *    2 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsRes2x8	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsRes2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem2x8 (F29_ResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: F29_SctErsStat2x8
 *
 * Purpose
 *    The procedure "F29_SctErsStat2x8" tests the status of a sector erase in a
 *    2 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs2x8	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int F29_SctErsStat2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16       	Status ;

   if (!((Status = *FlMemBsAdd) & FlMemCir2x8 (F29_Rdy, 0) &&
            Status & FlMemCir2x8 (F29_Rdy, 1)))
      return (ErrStatus = -1);
   Status = *FlMemBsAdd & FlMem2x8 (F29_ErsErr + F29_Sus + F29_Rdy) ;
   *FlMemBsAdd = FlMem2x8 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem2x8 (F29_MemRdCmd) ;
   if (Status == FlMem2x8 (F29_Rdy))
      return (ErrStatus = Success) ;
   else
   {
      ErrAddress = (Address) FlMemSctAdd ;
      return (ErrStatus = FlSctErs2x8Err) ;
   }
}


/*
 * PRIVATE PROCEDURE: F29_SctErs4x8
 *
 * Purpose
 *    The procedure "F29_SctErs4x8" starts the erase of a sector in a
 *    4 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs4x8	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_SctErs4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem4x8 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem4x8 (F29_SctErsCmd) ;
   *FlMemSctAdd = FlMem4x8 (F29_CnfrmCmd) ;
   *FlMemBsAdd = FlMem4x8 (F29_StatRdCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: F29_SctErsSus4x8
 *
 * Purpose
 *    The procedure "F29_SctErsSus4x8" suspends the erase of a sector in a
 *    4 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsSus4x8	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsSus4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32		Status;

   *FlMemBsAdd = FlMem4x8 (F29_SusCmd) ;
   Status = *FlMemBsAdd & FlMem4x8 (F29_Sus) ;
   if (Status) 
      return (ErrStatus = Success); /* suspended */
   else
      return (ErrStatus = -1); /* complete */
}


/*
 * PRIVATE PROCEDURE: F29_SctErsRes4x8
 *
 * Purpose
 *    The procedure "F29_SctErsRes4x8" resumes the erase of a sector in a
 *    4 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsRes4x8	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsRes4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem4x8 (F29_ResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: F29_SctErsStat4x8
 *
 * Purpose
 *    The procedure "F29_SctErsStat4x8" tests the status of a sector erase in a
 *    4 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs4x8	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int F29_SctErsStat4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32       	Status ;

   if (!((Status = *FlMemBsAdd) & FlMemCir4x8 (F29_Rdy, 0) &&
            Status & FlMemCir4x8 (F29_Rdy, 1) &&
            Status & FlMemCir4x8 (F29_Rdy, 2) &&
            Status & FlMemCir4x8 (F29_Rdy, 3)))
      return (ErrStatus = -1);
   Status = *FlMemBsAdd & FlMem4x8 (F29_ErsErr + F29_Sus + F29_Rdy) ;
   *FlMemBsAdd = FlMem4x8 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem4x8 (F29_MemRdCmd) ;
   if (Status == FlMem4x8 (F29_Rdy))
      return (ErrStatus = Success) ;
   else
   {
      ErrAddress = (Address) FlMemSctAdd ;
      return (ErrStatus = FlSctErs4x8Err) ;
   }
}


/*
 * PRIVATE PROCEDURE: F29_SctErs1x16
 *
 * Purpose
 *    The procedure "F29_SctErs1x16" starts the erase of a sector in a
 *    1 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs1x16	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_SctErs1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem1x16 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem1x16 (F29_SctErsCmd) ;
   *FlMemSctAdd = FlMem1x16 (F29_CnfrmCmd) ;
   *FlMemBsAdd = FlMem1x16 (F29_StatRdCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: F29_SctErsSus1x16
 *
 * Purpose
 *    The procedure "F29_SctErsSus1x16" suspends the erase of a sector in a
 *    1 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsSus1x16	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsSus1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16		Status;

   *FlMemBsAdd = FlMem1x16 (F29_SusCmd) ;
   Status = *FlMemBsAdd & FlMem1x16 (F29_Sus) ;
   if (Status) 
      return (ErrStatus = Success); /* suspended */
   else
      return (ErrStatus = -1); /* complete */
}


/*
 * PRIVATE PROCEDURE: F29_SctErsRes1x16
 *
 * Purpose
 *    The procedure "F29_SctErsRes1x16" resumes the erase of a sector in a
 *    1 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsRes1x16	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsRes1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem1x16 (F29_ResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: F29_SctErsStat1x16
 *
 * Purpose
 *    The procedure "F29_SctErsStat1x16" tests the status of a sector erase in a
 *    1 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs1x16	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int F29_SctErsStat1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemSctAdd)
{
   Word16       	Status ;

   if (!(*FlMemBsAdd & FlMemCir1x16 (F29_Rdy, 0)))
     return (ErrStatus = -1);
   Status = *FlMemBsAdd & FlMem1x16 (F29_ErsErr + F29_Sus + F29_Rdy) ;
   *FlMemBsAdd = FlMem1x16 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem1x16 (F29_MemRdCmd) ;
   if (Status == FlMem1x16 (F29_Rdy))
      return (ErrStatus = Success) ;
   else
   {
      ErrAddress = (Address) FlMemSctAdd ;
      return (ErrStatus = FlSctErs1x16Err) ;
   }
}


/*
 * PRIVATE PROCEDURE: F29_SctErs2x16
 *
 * Purpose
 *    The procedure "F29_SctErs2x16" starts the erase of a sector in a
 *    2 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs2x16	Set to 0 ("Success") if the erasing operation was
 *              	started successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_SctErs2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem2x16 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem2x16 (F29_SctErsCmd) ;
   *FlMemSctAdd = FlMem2x16 (F29_CnfrmCmd) ;
   *FlMemBsAdd = FlMem2x16 (F29_StatRdCmd) ;
   return (ErrStatus = Success) ;
}

/*
 * PRIVATE PROCEDURE: F29_SctErsSus2x16
 *
 * Purpose
 *    The procedure "F29_SctErsSus2x16" suspends the erase of a sector in a
 *    2 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsSus2x16	Set to 0 ("Success") if the erasing operation was
 *              	suspended successfully, to -1 if it was already
 *                      complete, or to the corresponding error number 
 *                      if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsSus2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32		Status;

   *FlMemBsAdd = FlMem2x16 (F29_SusCmd) ;
   Status = *FlMemBsAdd & FlMem2x16 (F29_Sus) ;
   if (Status) 
      return (ErrStatus = Success); /* suspended */
   else
      return (ErrStatus = -1); /* complete */
}


/*
 * PRIVATE PROCEDURE: F29_SctErsRes2x16
 *
 * Purpose
 *    The procedure "F29_SctErsRes2x16" resumes the erase of a sector in a
 *    2 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErsRes2x16	Set to 0 ("Success") if the erasing operation was
 *              	resumed successfully, or to the corresponding error
 *			number if an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 *
 */

static unsigned int F29_SctErsRes2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   *FlMemBsAdd = FlMem2x16 (F29_ResCmd) ;
   return (ErrStatus = Success);
}


/*
 * PRIVATE PROCEDURE: F29_SctErsStat2x16
 *
 * Purpose
 *    The procedure "F29_SctErsStat2x16" tests the status of a sector erase in a
 *    2 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemSctAdd	Flash-memory sector address.
 *
 * Output
 *    F29_SctErs2x16	Set to 0 ("Success") if the erasing operation was
 *              	completed successfully, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static int F29_SctErsStat2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemSctAdd)
{
   Word32       	Status ;

   if (!((Status = *FlMemBsAdd) & FlMemCir2x16 (F29_Rdy, 0) &&
            Status & FlMemCir2x16 (F29_Rdy, 1)))
     return (ErrStatus = -1);
   Status = *FlMemBsAdd & FlMem2x16 (F29_ErsErr + F29_Sus + F29_Rdy) ;
   *FlMemBsAdd = FlMem2x16 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem2x16 (F29_MemRdCmd) ;
   if (Status == FlMem2x16 (F29_Rdy))
      return (ErrStatus = Success) ;
   else
   {
      ErrAddress = (Address) FlMemSctAdd ;
      return (ErrStatus = FlSctErs2x16Err) ;
   }
}


/*
 * PRIVATE PROCEDURE: F29_Prgm2x8
 *
 * Purpose
 *    The procedure "F29_Prgm2x8" programs (i.e., writes) a given data
 *    buffer into a 2 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash
 *    memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    F29_Prgm2x8	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_Prgm2x8 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemAdd,
   Word16       	*DataAdd,
   unsigned int 	DataSz)
{
   int          	PgCnt ;
   Word16       	*FlMemPtr, Status ;

   if (DataSz < sizeof (Word16))
      return (ErrStatus = Success) ;
   *FlMemBsAdd = FlMem2x8 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem2x8 (F29_PgPrgmCmd) ;
   FlMemPtr = (Word16 *) ((unsigned int) FlMemAdd & ~(2*F29_PgSz - 1)) ;
   for (PgCnt = 2*F29_PgSz/sizeof (Word16) ; FlMemPtr < FlMemAdd ; PgCnt--)
      *(FlMemPtr++) = (Word16) -1 ;
   while (TRUE)
   {
      for ( ; PgCnt > 0 && DataSz >= sizeof (Word16) ; PgCnt--)
      {
         *(FlMemPtr++) = *(DataAdd++) ;
         DataSz -= sizeof (Word16) ;
      }
      for ( ; PgCnt > 0 ; PgCnt--)
         *(FlMemPtr++) = (Word16) -1 ;
      *FlMemBsAdd = FlMem2x8 (F29_StatRdCmd) ;
      while (!((Status = *FlMemBsAdd) & FlMemCir2x8 (F29_Rdy, 0) &&
               Status & FlMemCir2x8 (F29_Rdy, 1)))
         WAIT_PROGRAM((void *) FlMemAdd);
      Status = *FlMemBsAdd &
               FlMem2x8 (F29_SctPrgmErr + F29_PrgmErr + F29_Sus + F29_Rdy) ;
      *FlMemBsAdd = FlMem2x8 (F29_StatClrCmd) ;
      if (Status != FlMem2x8 (F29_Rdy))
      {
         *FlMemBsAdd = FlMem2x8 (F29_MemRdCmd) ;
         ErrAddress = (Address) ((unsigned int) FlMemPtr - 2*F29_PgSz) ;
         return (ErrStatus = FlPgPrgm2x8Err) ;
      }
      if (DataSz >= sizeof (Word16))
      {
         *FlMemBsAdd = FlMem2x8 (F29_PgPrgmCmd) ;
         PgCnt = 2*F29_PgSz/sizeof (Word16) ;
      }
      else
         break ;
   }
   *FlMemBsAdd = FlMem2x8 (F29_MemRdCmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: F29_Prgm4x8
 *
 * Purpose
 *    The procedure "F29_Prgm4x8" programs (i.e., writes) a given data
 *    buffer into a 4 x 8-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash
 *    memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    F29_Prgm4x8	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_Prgm4x8 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemAdd,
   Word32       	*DataAdd,
   unsigned int 	DataSz)
{
   int          	PgCnt ;
   Word32       	*FlMemPtr, Status ;

   if (DataSz < sizeof (Word32))
      return (ErrStatus = Success) ;
   *FlMemBsAdd = FlMem4x8 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem4x8 (F29_PgPrgmCmd) ;
   FlMemPtr = (Word32 *) ((unsigned int) FlMemAdd & ~(4*F29_PgSz - 1)) ;
   for (PgCnt = 4*F29_PgSz/sizeof (Word32) ; FlMemPtr < FlMemAdd ; PgCnt--)
      *(FlMemPtr++) = (Word32) -1 ;
   while (TRUE)
   {
      for ( ; PgCnt > 0 && DataSz >= sizeof (Word32) ; PgCnt--)
      {
         *(FlMemPtr++) = *(DataAdd++) ;
         DataSz -= sizeof (Word32) ;
      }
      for ( ; PgCnt > 0 ; PgCnt--)
         *(FlMemPtr++) = (Word32) -1 ;
      *FlMemBsAdd = FlMem4x8 (F29_StatRdCmd) ;
      while (!((Status = *FlMemBsAdd) & FlMemCir4x8 (F29_Rdy, 0) &&
               Status & FlMemCir4x8 (F29_Rdy, 1) &&
               Status & FlMemCir4x8 (F29_Rdy, 2) &&
               Status & FlMemCir4x8 (F29_Rdy, 3)))
         WAIT_PROGRAM((void *) FlMemAdd);
      Status = *FlMemBsAdd &
               FlMem4x8 (F29_SctPrgmErr + F29_PrgmErr + F29_Sus + F29_Rdy) ;
      *FlMemBsAdd = FlMem4x8 (F29_StatClrCmd) ;
      if (Status != FlMem4x8 (F29_Rdy))
      {
         *FlMemBsAdd = FlMem4x8 (F29_MemRdCmd) ;
         ErrAddress = (Address) ((unsigned int) FlMemPtr - 4*F29_PgSz) ;
         return (ErrStatus = FlPgPrgm4x8Err) ;
      }
      if (DataSz >= sizeof (Word32))
      {
         *FlMemBsAdd = FlMem4x8 (F29_PgPrgmCmd) ;
         PgCnt = 4*F29_PgSz/sizeof (Word32) ;
      }
      else
         break ;
   }
   *FlMemBsAdd = FlMem4x8 (F29_MemRdCmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: F29_Prgm1x16
 *
 * Purpose
 *    The procedure "F29_Prgm1x16" programs (i.e., writes) a given data
 *    buffer into a 1 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash
 *    memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    F29_Prgm1x16	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_Prgm1x16 (
   volatile Word16	*FlMemBsAdd,
   volatile Word16	*FlMemAdd,
   Word16       	*DataAdd,
   unsigned int 	DataSz)
{
   int          	PgCnt ;
   Word16       	*FlMemPtr, Status ;

   if (DataSz < sizeof (Word16))
      return (ErrStatus = Success) ;
   *FlMemBsAdd = FlMem1x16 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem1x16 (F29_PgPrgmCmd) ;
   FlMemPtr = (Word16 *) ((unsigned int) FlMemAdd & ~(F29_PgSz - 1)) ;
   for (PgCnt = F29_PgSz/sizeof (Word16) ; FlMemPtr < FlMemAdd ; PgCnt--)
      *(FlMemPtr++) = (Word16) -1 ;
   while (TRUE)
   {
      for ( ; PgCnt > 0 && DataSz >= sizeof (Word16) ; PgCnt--)
      {
         *(FlMemPtr++) = *(DataAdd++) ;
         DataSz -= sizeof (Word16) ;
      }
      for ( ; PgCnt > 0 ; PgCnt--)
         *(FlMemPtr++) = (Word16) -1 ;
      *FlMemBsAdd = FlMem1x16 (F29_StatRdCmd) ;
      while (!(*FlMemBsAdd & FlMemCir1x16 (F29_Rdy, 0)))
	WAIT_PROGRAM((void *) FlMemAdd);
      Status = *FlMemBsAdd &
               FlMem1x16 (F29_SctPrgmErr + F29_PrgmErr + F29_Sus + F29_Rdy) ;
      *FlMemBsAdd = FlMem1x16 (F29_StatClrCmd) ;
      if (Status != FlMem1x16 (F29_Rdy))
      {
         *FlMemBsAdd = FlMem1x16 (F29_MemRdCmd) ;
         ErrAddress = (Address) ((unsigned int) FlMemPtr - F29_PgSz) ;
         return (ErrStatus = FlPgPrgm1x16Err) ;
      }
      if (DataSz >= sizeof (Word16))
      {
         *FlMemBsAdd = FlMem1x16 (F29_PgPrgmCmd) ;
         PgCnt = F29_PgSz/sizeof (Word16) ;
      }
      else
         break ;
   }
   *FlMemBsAdd = FlMem1x16 (F29_MemRdCmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: F29_Prgm2x16
 *
 * Purpose
 *    The procedure "F29_Prgm2x16" programs (i.e., writes) a given data
 *    buffer into a 2 x 16-bit Hitachi HN29WXxxx or Motorola M29Fxxx flash
 *    memory.
 *
 * Input
 *    FlMemBsAdd	Flash-memory base address.
 *    FlMemAdd  	Address, in the flash memory, where to write the new
 *              	data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *
 * Output
 *    F29_Prgm2x16	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int F29_Prgm2x16 (
   volatile Word32	*FlMemBsAdd,
   volatile Word32	*FlMemAdd,
   Word32       	*DataAdd,
   unsigned int 	DataSz)
{
   int          	PgCnt ;
   Word32       	*FlMemPtr, Status ;

   if (DataSz < sizeof (Word32))
      return (ErrStatus = Success) ;
   *FlMemBsAdd = FlMem2x16 (F29_StatClrCmd) ;
   *FlMemBsAdd = FlMem2x16 (F29_PgPrgmCmd) ;
   FlMemPtr = (Word32 *) ((unsigned int) FlMemAdd & ~(2*F29_PgSz - 1)) ;
   for (PgCnt = 2*F29_PgSz/sizeof (Word32) ; FlMemPtr < FlMemAdd ; PgCnt--)
      *(FlMemPtr++) = (Word32) -1 ;
   while (TRUE)
   {
      for ( ; PgCnt > 0 && DataSz >= sizeof (Word32) ; PgCnt--)
      {
         *(FlMemPtr++) = *(DataAdd++) ;
         DataSz -= sizeof (Word32) ;
      }
      for ( ; PgCnt > 0 ; PgCnt--)
         *(FlMemPtr++) = (Word32) -1 ;
      *FlMemBsAdd = FlMem2x16 (F29_StatRdCmd) ;
      while (!((Status = *FlMemBsAdd) & FlMemCir2x16 (F29_Rdy, 0) &&
               Status & FlMemCir2x16 (F29_Rdy, 1)))
	WAIT_PROGRAM((void *) FlMemAdd);
      Status = *FlMemBsAdd &
               FlMem2x16 (F29_SctPrgmErr + F29_PrgmErr + F29_Sus + F29_Rdy) ;
      *FlMemBsAdd = FlMem2x16 (F29_StatClrCmd) ;
      if (Status != FlMem2x16 (F29_Rdy))
      {
         *FlMemBsAdd = FlMem2x16 (F29_MemRdCmd) ;
         ErrAddress = (Address) ((unsigned int) FlMemPtr - 2*F29_PgSz) ;
         return (ErrStatus = FlPgPrgm2x16Err) ;
      }
      if (DataSz >= sizeof (Word32))
      {
         *FlMemBsAdd = FlMem2x16 (F29_PgPrgmCmd) ;
         PgCnt = 2*F29_PgSz/sizeof (Word32) ;
      }
      else
         break ;
   }
   *FlMemBsAdd = FlMem2x16 (F29_MemRdCmd) ;
   return (ErrStatus = Success) ;
}


/*
 * PRIVATE PROCEDURE: SubBnkIdentify
 *
 * Purpose
 *    The procedure "SubBnkIdentify" identifies the type of flash memory chips
 *    and their organization in a given sub-bank, and returns the information
 *    necessary to erase or program them.
 *
 * Input
 *    StMemBnkNb        Static memory bank number.
 *    FlMemBsAdd	Flash memory virtual base address.
 *    fi        	Pointer to an empty flash memory information
 *              	structure to be filled in.
 *    ManufIDPtr	Pointer to an unsigned integer where to store the
 *              	manufacturer identification value, or 0 (NULL) if
 *              	this value should not be saved.
 *    DevIDPtr  	Pointer to an unsigned integer where to store the
 *              	device identification value, or 0 (NULL) if this
 *              	value should not be saved.
 *
 * Output
 *    SubBnkIdentify	Set to 0 ("Success") if the flash memory was
 *              	successfully identified or to the corresponding error
 *              	number if an error occurred.
 *    fi        	Partially filled in if success.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

static unsigned int SubBnkIdentify (
   int          	StMemBnkNb,
   unsigned int		FlMemBsAdd,
   FlMemInfo    	*fi,
   unsigned int		*ManufIDPtr,
   unsigned int		*DevIDPtr)
{
   int spSct;          /* to iterate over all special sectors */
   Word	StMemCtrl;     /* relevant bits from MSC for this device */
   Word32 ManufID32,   /* manufacturer id */
	  DevID32;     /* device id */
   Word16 ManufID16,   /* manufacturer id */
	  DevID16;     /* device id */
   Word32 FlMemData32; /* temp var */
   Word16 FlMemData16; /* temp var */

   fi->numSubBnk = 1 ;
   fi->subBnkAdd [0] = (void *) FlMemBsAdd;
   /* Query processor's static memory control register */
   if ((StMemCtrl = FExtr (MSC [StMemBnkNb/2], MSC_Bnk (StMemBnkNb)) &
                    (FMsk (MSC_RT) + MSC_RBW)) ==
       MSC_NonBrst + MSC_32BitStMem)
   {
      /* 32-bit */
      fi->bitsPerWord = 32;
      fi->dataWdthMsk = sizeof (Word32) - 1 ;

      /* Query flash chip itself */
      *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (A29_RstCmd) ;
      *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (F29_MemRdCmd) ;
      FlMemData32 = ((volatile Word32 *) FlMemBsAdd) [F29_ManufIDAdd] ;
      ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_Unlck1Add)] =
         FlMem4x8 (A29_Unlck1Cmd) ;
      ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_Unlck2Add)] =
         FlMem4x8 (A29_Unlck2Cmd) ;
      ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_AutoSelAdd)] =
         FlMem4x8 (F29_IDCmd) ;
      ManufID32 = ((volatile Word32 *) FlMemBsAdd) [F29_ManufIDAdd] ;
      if (ManufID32 == FlMemData32)
      {
         ((volatile Word32 *) FlMemBsAdd) [F29_16Add (A29_Unlck1Add)] =
            FlMem2x16 (A29_Unlck1Cmd) ;
         ((volatile Word32 *) FlMemBsAdd) [F29_16Add (A29_Unlck2Add)] =
            FlMem2x16 (A29_Unlck2Cmd) ;
         ((volatile Word32 *) FlMemBsAdd) [F29_16Add (A29_AutoSelAdd)] =
            FlMem4x8 (F29_IDCmd) ;
         ManufID32 = ((volatile Word32 *) FlMemBsAdd) [F29_ManufIDAdd] ;
      }
      switch (ManufID32)
      {
         case FlMem4x8 (A29_AMD):
            ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_Unlck1Add)] =
               FlMem4x8 (A29_Unlck1Cmd) ;
            ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_Unlck2Add)] =
               FlMem4x8 (A29_Unlck2Cmd) ;
            ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_AutoSelAdd)] =
               FlMem4x8 (A29_AutoSelCmd) ;
            DevID32 =
               ((volatile Word32 *) FlMemBsAdd) [F29_8Add (A29_DevIDAdd)] ;
            switch (DevID32)
            {
               case FlMem4x8 (A29_Am29LV160B):
                  fi->size = 4*A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 4*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29LV160T):
                  fi->size = 4*A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_Am29LV160Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                4*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                4*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29LV800B):
                  fi->size = 4*A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 4*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8 ;
                  break ;
               case FlMem4x8 (A29_Am29LV800T):
                  fi->size = 4*A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_Am29LV800Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                4*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                4*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8 ;
                  break ;
               case FlMem4x8 (A29_Am29LV400B):
                  fi->size = 4*A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 4*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8 ;
                  break ;
               case FlMem4x8 (A29_Am29LV400T):
                  fi->size = 4*A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_Am29LV400Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                4*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                4*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8 ;
                  break ;
               case FlMem4x8 (A29_Am29LV200B):
                  fi->size = 4*A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 4*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8 ;
                  break ;
               case FlMem4x8 (A29_Am29LV200T):
                  fi->size = 4*A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                4*(A29_Am29LV200Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                4*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                4*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                4*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8 ;
                  break ;
               case FlMem4x8 (A29_Am29DL322B):
                  fi->size = 4*A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           4*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL322T):
                  fi->size = 4*A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       4*(A29_Am29DL322Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL323B):
                  fi->size = 4*A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           4*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL323T):
                  fi->size = 4*A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       4*(A29_Am29DL323Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL162B):
                  fi->size = 4*A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           4*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL162T):
                  fi->size = 4*A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       4*(A29_Am29DL162Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL163B):
                  fi->size = 4*A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           4*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL163T):
                  fi->size = 4*A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       4*(A29_Am29DL163Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL800B):
                  fi->size = 4*A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 4*(A29_NlSctSz/4) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/2) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL800T):
                  fi->size = 4*A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                   4*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2 -
                                      A29_NlSctSz/4) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL400B):
                  fi->size = 4*A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 4*(A29_NlSctSz/4) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   4*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/2) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               case FlMem4x8 (A29_Am29DL400T):
                  fi->size = 4*A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                   4*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2 -
                                      A29_NlSctSz/4) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm4x8Byp ;
                  break ;
               default:
                  *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (A29_RstCmd) ;
                  ErrData = DevID32 ;
                  return (ErrStatus = FlUnkwnDev32Err) ;
            }
            *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (A29_RstCmd) ;
            fi->devsPerWord = 4 ;
            fi->sctErsAlg = (FlMemSctErsAlg) A29_SctErs4x8 ;
            fi->sctErsSusAlg = (FlMemSctErsSusAlg) A29_SctErsSus4x8 ;
            fi->sctErsResAlg = (FlMemSctErsResAlg) A29_SctErsRes4x8 ;
            fi->sctErsStat = (FlMemSctErsStat) A29_SctErsStat4x8 ;
            break ;
         case FlMem2x16 (A29_AMD):
            DevID32 =
               ((volatile Word32 *) FlMemBsAdd) [F29_16Add (A29_DevIDAdd)] ;
            switch (DevID32)
            {
               case FlMem2x16 (A29_Am29LV160B):
                  fi->size = 2*A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29LV160T):
                  fi->size = 2*A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29LV800B):
                  fi->size = 2*A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16 ;
                  break ;
               case FlMem2x16 (A29_Am29LV800T):
                  fi->size = 2*A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16 ;
                  break ;
               case FlMem2x16 (A29_Am29LV400B):
                  fi->size = 2*A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16 ;
                  break ;
               case FlMem2x16 (A29_Am29LV400T):
                  fi->size = 2*A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16 ;
                  break ;
               case FlMem2x16 (A29_Am29LV200B):
                  fi->size = 2*A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16 ;
                  break ;
               case FlMem2x16 (A29_Am29LV200T):
                  fi->size = 2*A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16 ;
                  break ;
               case FlMem2x16 (A29_Am29DL322B):
                  fi->size = 2*A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL322T):
                  fi->size = 2*A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL322Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL323B):
                  fi->size = 2*A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL323T):
                  fi->size = 2*A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL323Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL162B):
                  fi->size = 2*A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL162T):
                  fi->size = 2*A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL162Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL163B):
                  fi->size = 2*A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL163T):
                  fi->size = 2*A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL163Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL800B):
                  fi->size = 2*A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*(A29_NlSctSz/4) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/2) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL800T):
                  fi->size = 2*A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2 -
                                      A29_NlSctSz/4) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL400B):
                  fi->size = 2*A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*(A29_NlSctSz/4) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/2) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               case FlMem2x16 (A29_Am29DL400T):
                  fi->size = 2*A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2 -
                                      A29_NlSctSz/4) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x16Byp ;
                  break ;
               default:
                  *((volatile Word32 *) FlMemBsAdd) = FlMem2x16 (A29_RstCmd) ;
                  ErrData = DevID32 ;
                  return (ErrStatus = FlUnkwnDev32Err) ;
            }
            *((volatile Word32 *) FlMemBsAdd) = FlMem2x16 (A29_RstCmd) ;
            fi->devsPerWord = 2 ;
            fi->sctErsAlg = (FlMemSctErsAlg) A29_SctErs2x16 ;
            fi->sctErsSusAlg = (FlMemSctErsSusAlg) A29_SctErsSus2x16 ;
            fi->sctErsResAlg = (FlMemSctErsResAlg) A29_SctErsRes2x16 ;
            fi->sctErsStat = (FlMemSctErsStat) A29_SctErsStat2x16 ;
            break ;
         case FlMem4x8 (F29_Hitachi):
            DevID32 =
               ((volatile Word32 *) FlMemBsAdd) [F29_8Add (F29_DevIDAdd)] ;
            if (DevID32 == ManufID32)
            {
               DevID32 =
                  ((volatile Word32 *) FlMemBsAdd) [F29_16Add (F29_DevIDAdd)] ;
               fi->devsPerWord = 2 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm2x16 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs2x16 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus2x16 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes2x16 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat2x16 ;
            }
            else
            {
               fi->devsPerWord = 4 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm4x8 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs4x8 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus4x8 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes4x8 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat4x8 ;
            }
            switch (DevID32)
            {
               case FlMem4x8 (F29_HN29WB800):
                  fi->size = fi->devsPerWord*F29_HN29W800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + fi->devsPerWord*F29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz +
                                        F29_SpSct2Sz) ;
                  break ;
               case FlMem4x8 (F29_HN29WT800):
                  fi->size = fi->devsPerWord*F29_HN29W800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz -
                                        F29_SpSct3Sz) ;
                  break ;
               default:
                  *((volatile Word32 *) FlMemBsAdd) =
                     FlMem4x8 (F29_MemRdCmd) ;
                  ErrData = DevID32 ;
                  return (ErrStatus = FlUnkwnDev32Err) ;
            }
            *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (F29_MemRdCmd) ;
            break ;
         case FlMem4x8 (F29_Motorola):
            DevID32 =
               ((volatile Word32 *) FlMemBsAdd) [F29_8Add (F29_DevIDAdd)] ;
            if (DevID32 == ManufID32)
            {
               DevID32 =
                  ((volatile Word32 *) FlMemBsAdd) [F29_16Add (F29_DevIDAdd)] ;
               fi->devsPerWord = 2 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm2x16 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs2x16 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus2x16 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes2x16 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat2x16 ;
            }
            else
            {
               fi->devsPerWord = 4 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm4x8 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs4x8 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus4x8 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes4x8 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat4x8 ;
            }
            switch (DevID32)
            {
               case FlMem4x8 (F29_M29F800A3B):
                  fi->size = fi->devsPerWord*F29_M29F800A3Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + fi->devsPerWord*F29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz +
                                        F29_SpSct2Sz) ;
                  break ;
               case FlMem4x8 (F29_M29F800A3U):
                  fi->size = fi->devsPerWord*F29_M29F800A3Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz -
                                        F29_SpSct3Sz) ;
                  break ;
               default:
                  *((volatile Word32 *) FlMemBsAdd) =
                     FlMem4x8 (F29_MemRdCmd) ;
                  ErrData = DevID32 ;
                  return (ErrStatus = FlUnkwnDev32Err) ;
            }
            *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (F29_MemRdCmd) ;
            break ;
         default:
            *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (A29_RstCmd) ;
            *((volatile Word32 *) FlMemBsAdd) = FlMem4x8 (F29_MemRdCmd) ;
            ErrData = ManufID32 ;
            return (ErrStatus = FlUnkwnMnf32Err) ;
      }
      if (ManufIDPtr != NULL)
         *ManufIDPtr = (unsigned int) ManufID32 ;
      if (DevIDPtr != NULL)
         *DevIDPtr = (unsigned int) DevID32 ;
   }
   else if (StMemCtrl == MSC_NonBrst + MSC_16BitStMem)
   {
      /* 16-bit */
      fi->bitsPerWord = 16;
      fi->dataWdthMsk = sizeof (Word16) - 1 ;

      /* Query flash chip itself */
      *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (A29_RstCmd) ;
      *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (F29_MemRdCmd) ;
      FlMemData16 = ((volatile Word16 *) FlMemBsAdd) [F29_ManufIDAdd] ;
      ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_Unlck1Add)] =
         FlMem2x8 (A29_Unlck1Cmd) ;
      ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_Unlck2Add)] =
         FlMem2x8 (A29_Unlck2Cmd) ;
      ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_AutoSelAdd)] =
         FlMem2x8 (F29_IDCmd) ;
      ManufID16 = ((volatile Word16 *) FlMemBsAdd) [F29_ManufIDAdd] ;
      if (ManufID16 == FlMemData16)
      {
         ((volatile Word16 *) FlMemBsAdd) [F29_16Add (A29_Unlck1Add)] =
            FlMem1x16 (A29_Unlck1Cmd) ;
         ((volatile Word16 *) FlMemBsAdd) [F29_16Add (A29_Unlck2Add)] =
            FlMem1x16 (A29_Unlck2Cmd) ;
         ((volatile Word16 *) FlMemBsAdd) [F29_16Add (A29_AutoSelAdd)] =
            FlMem2x8 (F29_IDCmd) ;
         ManufID16 = ((volatile Word16 *) FlMemBsAdd) [F29_ManufIDAdd] ;
      }
      switch (ManufID16)
      {
         case FlMem2x8 (A29_AMD):
            ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_Unlck1Add)] =
               FlMem2x8 (A29_Unlck1Cmd) ;
            ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_Unlck2Add)] =
               FlMem2x8 (A29_Unlck2Cmd) ;
            ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_AutoSelAdd)] =
               FlMem2x8 (A29_AutoSelCmd) ;
            DevID16 =
               ((volatile Word16 *) FlMemBsAdd) [F29_8Add (A29_DevIDAdd)] ;
            switch (DevID16)
            {
               case FlMem2x8 (A29_Am29LV160B):
                  fi->size = 2*A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29LV160T):
                  fi->size = 2*A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV160Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29LV800B):
                  fi->size = 2*A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8 ;
                  break ;
               case FlMem2x8 (A29_Am29LV800T):
                  fi->size = 2*A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV800Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8 ;
                  break ;
               case FlMem2x8 (A29_Am29LV400B):
                  fi->size = 2*A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8 ;
                  break ;
               case FlMem2x8 (A29_Am29LV400T):
                  fi->size = 2*A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV400Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8 ;
                  break ;
               case FlMem2x8 (A29_Am29LV200B):
                  fi->size = 2*A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_SpSct0Sz + A29_SpSct1Sz +
                                   A29_SpSct2Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8 ;
                  break ;
               case FlMem2x8 (A29_Am29LV200T):
                  fi->size = 2*A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                2*(A29_Am29LV200Sz - A29_SpSct0Sz -
                                   A29_SpSct1Sz - A29_SpSct2Sz -
                                   A29_SpSct3Sz) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8 ;
                  break ;
               case FlMem2x8 (A29_Am29DL322B):
                  fi->size = 2*A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL322T):
                  fi->size = 2*A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL322Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL323B):
                  fi->size = 2*A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL323T):
                  fi->size = 2*A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL323Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL162B):
                  fi->size = 2*A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL162T):
                  fi->size = 2*A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL162Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL163B):
                  fi->size = 2*A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           2*spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL163T):
                  fi->size = 2*A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       2*(A29_Am29DL163Sz -
			  (fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL800B):
                  fi->size = 2*A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*(A29_NlSctSz/4) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/2) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL800T):
                  fi->size = 2*A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                   2*(A29_Am29DL800Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2 -
                                      A29_NlSctSz/4) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL400B):
                  fi->size = 2*A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + 2*(A29_NlSctSz/4) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8) ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_NlSctSz/4 + A29_NlSctSz/2 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/8 + A29_NlSctSz/8 +
                                      A29_NlSctSz/2) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               case FlMem2x8 (A29_Am29DL400T):
                  fi->size = 2*A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4) ;
                  fi->spSctAdd[6] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2) ;
                  fi->spSctAdd[5] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8) ;
                  fi->spSctAdd[4] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                   2*(A29_Am29DL400Sz - A29_NlSctSz/4 -
                                      A29_NlSctSz/2 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/8 -
                                      A29_NlSctSz/8 - A29_NlSctSz/2 -
                                      A29_NlSctSz/4) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm2x8Byp ;
                  break ;
               default:
                  *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (A29_RstCmd) ;
                  ErrData = DevID16 ;
                  return (ErrStatus = FlUnkwnDev16Err) ;
            }
            *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (A29_RstCmd) ;
            fi->devsPerWord = 2 ;
            fi->sctErsAlg = (FlMemSctErsAlg) A29_SctErs2x8 ;
            fi->sctErsSusAlg = (FlMemSctErsSusAlg) A29_SctErsSus2x8 ;
            fi->sctErsResAlg = (FlMemSctErsResAlg) A29_SctErsRes2x8 ;
            fi->sctErsStat = (FlMemSctErsStat) A29_SctErsStat2x8 ;
            break ;
         case FlMem1x16 (A29_AMD):
            DevID16 =
               ((volatile Word16 *) FlMemBsAdd) [F29_16Add (A29_DevIDAdd)] ;
            switch (DevID16)
            {
               case FlMem1x16 (A29_Am29LV160B):
                  fi->size = A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz +
                                A29_SpSct2Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29LV160T):
                  fi->size = A29_Am29LV160Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_Am29LV160Sz - A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_Am29LV160Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_Am29LV160Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz ;
                  fi->spSctAdd[0] = FlMemBsAdd + A29_Am29LV160Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz -
                                A29_SpSct3Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29LV800B):
                  fi->size = A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz +
                                A29_SpSct2Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16 ;
                  break ;
               case FlMem1x16 (A29_Am29LV800T):
                  fi->size = A29_Am29LV800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_Am29LV800Sz - A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_Am29LV800Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_Am29LV800Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz ;
                  fi->spSctAdd[0] = FlMemBsAdd + A29_Am29LV800Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz -
                                A29_SpSct3Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16 ;
                  break ;
               case FlMem1x16 (A29_Am29LV400B):
                  fi->size = A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz +
                                A29_SpSct2Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16 ;
                  break ;
               case FlMem1x16 (A29_Am29LV400T):
                  fi->size = A29_Am29LV400Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_Am29LV400Sz - A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_Am29LV400Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_Am29LV400Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz ;
                  fi->spSctAdd[0] = FlMemBsAdd + A29_Am29LV400Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz -
                                A29_SpSct3Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16 ;
                  break ;
               case FlMem1x16 (A29_Am29LV200B):
                  fi->size = A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_SpSct0Sz + A29_SpSct1Sz +
                                A29_SpSct2Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16 ;
                  break ;
               case FlMem1x16 (A29_Am29LV200T):
                  fi->size = A29_Am29LV200Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_Am29LV200Sz - A29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_Am29LV200Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_Am29LV200Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz ;
                  fi->spSctAdd[0] = FlMemBsAdd + A29_Am29LV200Sz -
                                A29_SpSct0Sz - A29_SpSct1Sz - A29_SpSct2Sz -
                                A29_SpSct3Sz ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16 ;
                  break ;
               case FlMem1x16 (A29_Am29DL322B):
                  fi->size = A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL322T):
                  fi->size = A29_Am29DL322Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL322B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       (A29_Am29DL322Sz -
			(fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL323B):
                  fi->size = A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL323T):
                  fi->size = A29_Am29DL323Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL323B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       (A29_Am29DL323Sz -
			(fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL162B):
                  fi->size = A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL162T):
                  fi->size = A29_Am29DL162Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL162B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       (A29_Am29DL162Sz -
			(fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL163B):
                  fi->size = A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B1) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
                                           spSct*(A29_NlSctSz/8) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL163T):
                  fi->size = A29_Am29DL163Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL163B2) ;
                  fi->numSpSct = 8 ;
                  for (spSct = 0 ; spSct < fi->numSpSct ; spSct++)
                     fi->spSctAdd[spSct] = FlMemBsAdd +
		       (A29_Am29DL163Sz -
			(fi->numSpSct - spSct)*(A29_NlSctSz/8)) ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL800B):
                  fi->size = A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_NlSctSz/4 ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 ;
                  fi->spSctAdd[4] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[5] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/8 ;
                  fi->spSctAdd[6] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[7] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/2 ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL800T):
                  fi->size = A29_Am29DL800Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL800B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 ;
                  fi->spSctAdd[6] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 ;
                  fi->spSctAdd[5] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[4] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/2 ;
                  fi->spSctAdd[0] = FlMemBsAdd + A29_Am29DL800Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/2 - A29_NlSctSz/4 ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL400B):
                  fi->size = A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B1) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_NlSctSz/4 ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 ;
                  fi->spSctAdd[4] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[5] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/8 ;
                  fi->spSctAdd[6] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[7] = FlMemBsAdd + A29_NlSctSz/4 +
                                   A29_NlSctSz/2 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/8 +
                                   A29_NlSctSz/8 + A29_NlSctSz/2 ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               case FlMem1x16 (A29_Am29DL400T):
                  fi->size = A29_Am29DL400Sz ;
                  fi->numSubBnk = 2 ;
                  fi->subBnkAdd [1] = (void *) (FlMemBsAdd + A29_Am29DL400B2) ;
                  fi->numSpSct = 8 ;
                  fi->spSctAdd[7] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 ;
                  fi->spSctAdd[6] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 ;
                  fi->spSctAdd[5] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[4] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 ;
                  fi->spSctAdd[3] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 ;
                  fi->spSctAdd[2] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 ;
                  fi->spSctAdd[1] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/2 ;
                  fi->spSctAdd[0] = FlMemBsAdd + A29_Am29DL400Sz -
                                   A29_NlSctSz/4 - A29_NlSctSz/2 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/8 - A29_NlSctSz/8 -
                                   A29_NlSctSz/2 - A29_NlSctSz/4 ;
                  fi->prgmAlg = (FlMemPrgmAlg) A29_Prgm1x16Byp ;
                  break ;
               default:
                  *((volatile Word16 *) FlMemBsAdd) = FlMem1x16 (A29_RstCmd) ;
                  ErrData = DevID16 ;
                  return (ErrStatus = FlUnkwnDev16Err) ;
            }
            *((volatile Word16 *) FlMemBsAdd) = FlMem1x16 (A29_RstCmd) ;
            fi->devsPerWord = 1 ;
            fi->sctErsAlg = (FlMemSctErsAlg) A29_SctErs1x16 ;
            fi->sctErsSusAlg = (FlMemSctErsSusAlg) A29_SctErsSus1x16 ;
            fi->sctErsResAlg = (FlMemSctErsResAlg) A29_SctErsRes1x16 ;
            fi->sctErsStat = (FlMemSctErsStat) A29_SctErsStat1x16 ;
            break ;
         case FlMem2x8 (F29_Hitachi):
            DevID16 =
               ((volatile Word16 *) FlMemBsAdd) [F29_8Add (F29_DevIDAdd)] ;
            if (DevID16 == ManufID16)
            {
               DevID16 =
                  ((volatile Word16 *) FlMemBsAdd) [F29_16Add (F29_DevIDAdd)] ;
               fi->devsPerWord = 1 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm1x16 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs1x16 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus1x16 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes1x16 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat1x16 ;
            }
            else
            {
               fi->devsPerWord = 2 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm2x8 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs2x8 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus2x8 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes2x8 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat2x8 ;
            }
            switch (DevID16)
            {
               case FlMem2x8 (F29_HN29WB800):
                  fi->size = fi->devsPerWord*F29_HN29W800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + fi->devsPerWord*F29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz +
                                        F29_SpSct2Sz) ;
                  break ;
               case FlMem2x8 (F29_HN29WT800):
                  fi->size = fi->devsPerWord*F29_HN29W800Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                fi->devsPerWord*(F29_HN29W800Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz -
                                        F29_SpSct3Sz) ;
                  break ;
               default:
                  *((volatile Word16 *) FlMemBsAdd) =
                     FlMem2x8 (F29_MemRdCmd) ;
                  ErrData = DevID16 ;
                  return (ErrStatus = FlUnkwnDev16Err) ;
            }
            *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (F29_MemRdCmd) ;
            break ;
         case FlMem2x8 (F29_Motorola):
            DevID16 =
               ((volatile Word16 *) FlMemBsAdd) [F29_8Add (F29_DevIDAdd)] ;
            if (DevID16 == ManufID16)
            {
               DevID16 =
                  ((volatile Word16 *) FlMemBsAdd) [F29_16Add (F29_DevIDAdd)] ;
               fi->devsPerWord = 1 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm1x16 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs1x16 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus1x16 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes1x16 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat1x16 ;
            }
            else
            {
               fi->devsPerWord = 2 ;
               fi->prgmAlg = (FlMemPrgmAlg) F29_Prgm2x8 ;
               fi->sctErsAlg = (FlMemSctErsAlg) F29_SctErs2x8 ;
	       fi->sctErsSusAlg = (FlMemSctErsSusAlg) F29_SctErsSus2x8 ;
	       fi->sctErsResAlg = (FlMemSctErsResAlg) F29_SctErsRes2x8 ;
               fi->sctErsStat = (FlMemSctErsStat) F29_SctErsStat2x8 ;
            }
            switch (DevID16)
            {
               case FlMem2x8 (F29_M29F800A3B):
                  fi->size = fi->devsPerWord*F29_M29F800A3Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[0] = FlMemBsAdd ;
                  fi->spSctAdd[1] = FlMemBsAdd + fi->devsPerWord*F29_SpSct0Sz ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz) ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_SpSct0Sz + F29_SpSct1Sz +
                                        F29_SpSct2Sz) ;
                  break ;
               case FlMem2x8 (F29_M29F800A3U):
                  fi->size = fi->devsPerWord*F29_M29F800A3Sz ;
                  fi->numSpSct = 4 ;
                  fi->spSctAdd[3] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz) ;
                  fi->spSctAdd[2] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz) ;
                  fi->spSctAdd[1] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz) ;
                  fi->spSctAdd[0] = FlMemBsAdd +
                                fi->devsPerWord*(F29_M29F800A3Sz - F29_SpSct0Sz -
                                        F29_SpSct1Sz - F29_SpSct2Sz -
                                        F29_SpSct3Sz) ;
                  break ;
               default:
                  *((volatile Word16 *) FlMemBsAdd) =
                     FlMem2x8 (F29_MemRdCmd) ;
                  ErrData = DevID16 ;
                  return (ErrStatus = FlUnkwnDev16Err) ;
            }
            *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (F29_MemRdCmd) ;
            break ;
         default:
            *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (A29_RstCmd) ;
            *((volatile Word16 *) FlMemBsAdd) = FlMem2x8 (F29_MemRdCmd) ;
            ErrData = ManufID16 ;
            return (ErrStatus = FlUnkwnMnf16Err) ;
      }
      if (ManufIDPtr != NULL)
         *ManufIDPtr = (unsigned int) ManufID16 ;
      if (DevIDPtr != NULL)
         *DevIDPtr = (unsigned int) DevID16 ;
   }
   else
      return (ErrStatus = FlStMemIntfErr) ;
   fi->nlSctSz = fi->devsPerWord * F29_NlSctSz;
   for (spSct=0; spSct<fi->numSpSct-1; spSct++) {
      fi->spSctSz[spSct] = fi->spSctAdd[spSct+1] - fi->spSctAdd[spSct];
   }
   fi->spSctSz[spSct] = ((fi->spSctAdd[spSct]) & ~(fi->nlSctSz - 1))
                        + fi->nlSctSz - fi->spSctAdd[spSct];
   return (ErrStatus = Success);
}


/*
 * PUBLIC PROCEDURE: FlMemIdentify
 *
 * Purpose
 *    The procedure "FlMemIdentify" identifies the type of flash memory chips
 *    and their organization in a given static-memory bank, and returns
 *    the information necessary to erase or program them.
 *
 *    The procedure "FlMemIdentify" attempts to recognize if a flash memory
 *    bank is composed of 1 or 2 sub-bank(s), by checking the data in all
 *    locations that are guarantied to return the manufacturer or device
 *    identification codes when the flash memory chips are in auto-select
 *    mode or identification mode (for all supported parts).  In most cases,
 *    the algorithm will succeed after the first iteration or after a small
 *    number of iterations.  The algorithm will require more iterations in a
 *    few pathological cases, where the data programmed in the flash memory
 *    at most of these locations is identical to the manufacturer or device
 *    identification codes.  The algorithm will only fail to recognize that a
 *    flash memory bank is composed of 2 sub-banks in the extremely
 *    pathological case where the data read from all these locations is
 *    identical to the manufacturer or device identification codes in both
 *    sub-banks.
 *
 * Input
 *    StMemBnkNb        Static memory bank number.
 *    FlMemBsAdd	Flash memory virtual base address.
 *    fi        	Pointer to an empty flash memory information
 *              	structure to be filled in.
 *
 * Output
 *    FlMemIdentify	Set to 0 ("Success") if the flash memory was
 *              	successfully identified or to the corresponding error
 *              	number if an error occurred.
 *    fi        	Filled in if success.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variable
 *    "ErrAddress" is updated as well.
 */

#define AutoSelSz	0x00000100	/* Auto-Select Size (max.)         */

unsigned int FlMemIdentify (
   int          	StMemBnkNb,
   unsigned int		FlMemBsAdd,
   FlMemInfo    	*fi)
{
   unsigned int		ManufID, DevID ;
   int          	Divider, DualSubBnk, spSct, subBnk ;
   volatile Word32	*SubBnk1Ptr32, *SubBnk2Ptr32, *AutoSelBsAdd32 ;
   Word32       	Data32_1, Data32_2 ;
   volatile Word16	*SubBnk1Ptr16, *SubBnk2Ptr16, *AutoSelBsAdd16 ;
   Word16       	Data16_1, Data16_2 ;
   FlMemInfo    	SubBnkInfo ;

   /* Stash given info */
   fi->stMemBnkNb = StMemBnkNb;
   if (SubBnkIdentify (StMemBnkNb, FlMemBsAdd, fi, &ManufID, &DevID) !=
       Success)
      return (ErrStatus);
   /* Loop computes: fi->bitsPerDev = fi->bitsPerWord / fi->devsPerWord; */
   fi->bitsPerDev = fi->bitsPerWord;
   for (Divider = fi->devsPerWord ; Divider > 1 ; Divider >>= 1) 
      fi->bitsPerDev >>= 1 ;
   DualSubBnk = FALSE ;
   if (fi->bitsPerWord == 32)
   {
      SubBnk1Ptr32 = (volatile Word32 *) FlMemBsAdd ;
      SubBnk2Ptr32 = (volatile Word32 *) (FlMemBsAdd + fi->size) ;
      if (fi->bitsPerDev == 8)
         while (SubBnk1Ptr32 < (volatile Word32 *) (FlMemBsAdd + fi->size))
         {
            if ((Data32_1 = SubBnk1Ptr32 [F29_8Add (A29_ManufIDAdd)]) !=
                SubBnk2Ptr32 [F29_8Add (A29_ManufIDAdd)] ||
                (Data32_2 = SubBnk1Ptr32 [F29_8Add (A29_DevIDAdd)]) !=
                SubBnk2Ptr32 [F29_8Add (A29_DevIDAdd)])
            {
               DualSubBnk = TRUE ;
               break ;
            }
            if (Data32_1 != (Word32) ManufID || Data32_2 != (Word32) DevID)
            {
               if (fi->numSubBnk > 1 &&
                   SubBnk1Ptr32 >= (volatile Word32 *) fi->subBnkAdd [1])
                  AutoSelBsAdd32 = (volatile Word32 *) fi->subBnkAdd [1] ;
               else
                  AutoSelBsAdd32 = (volatile Word32 *) FlMemBsAdd ;
               AutoSelBsAdd32 [F29_8Add (A29_Unlck1Add)] =
                  FlMem4x8 (A29_Unlck1Cmd) ;
               AutoSelBsAdd32 [F29_8Add (A29_Unlck2Add)] =
                  FlMem4x8 (A29_Unlck2Cmd) ;
               AutoSelBsAdd32 [F29_8Add (A29_AutoSelAdd)] =
                  FlMem4x8 (F29_IDCmd) ;
               if (SubBnk2Ptr32 [F29_8Add (A29_ManufIDAdd)] !=
                   (Word32) ManufID ||
                   SubBnk2Ptr32 [F29_8Add (A29_DevIDAdd)] != (Word32) DevID)
                  DualSubBnk = TRUE ;
               *AutoSelBsAdd32 = FlMem4x8 (A29_RstCmd) ;
               *AutoSelBsAdd32 = FlMem4x8 (F29_MemRdCmd) ;
               break ;
            }
            SubBnk1Ptr32 = &(SubBnk1Ptr32 [AutoSelSz]) ;
            SubBnk2Ptr32 = &(SubBnk2Ptr32 [AutoSelSz]) ;
         }
      else /* if (fi->bitsPerDev == 16) */
         while (SubBnk1Ptr32 < (volatile Word32 *) (FlMemBsAdd + fi->size))
         {
            if ((Data32_1 = SubBnk1Ptr32 [F29_16Add (A29_ManufIDAdd)]) !=
                SubBnk2Ptr32 [F29_16Add (A29_ManufIDAdd)] ||
                (Data32_2 = SubBnk1Ptr32 [F29_16Add (A29_DevIDAdd)]) !=
                SubBnk2Ptr32 [F29_16Add (A29_DevIDAdd)])
            {
               DualSubBnk = TRUE ;
               break ;
            }
            if (Data32_1 != (Word32) ManufID || Data32_2 != (Word32) DevID)
            {
               if (fi->numSubBnk > 1 &&
                   SubBnk1Ptr32 >= (volatile Word32 *) fi->subBnkAdd [1])
                  AutoSelBsAdd32 = (volatile Word32 *) fi->subBnkAdd [1] ;
               else
                  AutoSelBsAdd32 = (volatile Word32 *) FlMemBsAdd ;
               AutoSelBsAdd32 [F29_16Add (A29_Unlck1Add)] =
                  FlMem2x16 (A29_Unlck1Cmd) ;
               AutoSelBsAdd32 [F29_16Add (A29_Unlck2Add)] =
                  FlMem2x16 (A29_Unlck2Cmd) ;
               AutoSelBsAdd32 [F29_16Add (A29_AutoSelAdd)] =
                  FlMem2x16 (F29_IDCmd) ;
               if (SubBnk2Ptr32 [F29_16Add (A29_ManufIDAdd)] !=
                   (Word32) ManufID ||
                   SubBnk2Ptr32 [F29_16Add (A29_DevIDAdd)] != (Word32) DevID)
                  DualSubBnk = TRUE ;
               *AutoSelBsAdd32 = FlMem2x16 (A29_RstCmd) ;
               *AutoSelBsAdd32 = FlMem2x16 (F29_MemRdCmd) ;
               break ;
            }
            SubBnk1Ptr32 = &(SubBnk1Ptr32 [AutoSelSz]) ;
            SubBnk2Ptr32 = &(SubBnk2Ptr32 [AutoSelSz]) ;
         }
   }
   else /* if (fi->bitsPerWord == 16) */
   {
      SubBnk1Ptr16 = (volatile Word16 *) FlMemBsAdd ;
      SubBnk2Ptr16 = (volatile Word16 *) (FlMemBsAdd + fi->size) ;
      if (fi->bitsPerDev == 8)
         while (SubBnk1Ptr16 < (volatile Word16 *) (FlMemBsAdd + fi->size))
         {
            if ((Data16_1 = SubBnk1Ptr16 [F29_8Add (A29_ManufIDAdd)]) !=
                SubBnk2Ptr16 [F29_8Add (A29_ManufIDAdd)] ||
                (Data16_2 = SubBnk1Ptr16 [F29_8Add (A29_DevIDAdd)]) !=
                SubBnk2Ptr16 [F29_8Add (A29_DevIDAdd)])
            {
               DualSubBnk = TRUE ;
               break ;
            }
            if (Data16_1 != (Word16) ManufID || Data16_2 != (Word16) DevID)
            {
               if (fi->numSubBnk > 1 &&
                   SubBnk1Ptr16 >= (volatile Word16 *) fi->subBnkAdd [1])
                  AutoSelBsAdd16 = (volatile Word16 *) fi->subBnkAdd [1] ;
               else
                  AutoSelBsAdd16 = (volatile Word16 *) FlMemBsAdd ;
               AutoSelBsAdd16 [F29_8Add (A29_Unlck1Add)] =
                  FlMem2x8 (A29_Unlck1Cmd) ;
               AutoSelBsAdd16 [F29_8Add (A29_Unlck2Add)] =
                  FlMem2x8 (A29_Unlck2Cmd) ;
               AutoSelBsAdd16 [F29_8Add (A29_AutoSelAdd)] =
                  FlMem2x8 (F29_IDCmd) ;
               if (SubBnk2Ptr16 [F29_8Add (A29_ManufIDAdd)] !=
                   (Word16) ManufID ||
                   SubBnk2Ptr16 [F29_8Add (A29_DevIDAdd)] != (Word16) DevID)
                  DualSubBnk = TRUE ;
               *AutoSelBsAdd16 = FlMem2x8 (A29_RstCmd) ;
               *AutoSelBsAdd16 = FlMem2x8 (F29_MemRdCmd) ;
               break ;
            }
            SubBnk1Ptr16 = &(SubBnk1Ptr16 [AutoSelSz]) ;
            SubBnk2Ptr16 = &(SubBnk2Ptr16 [AutoSelSz]) ;
         }
      else /* if (fi->bitsPerDev == 16) */
         while (SubBnk1Ptr16 < (volatile Word16 *) (FlMemBsAdd + fi->size))
         {
            if ((Data16_1 = SubBnk1Ptr16 [F29_16Add (A29_ManufIDAdd)]) !=
                SubBnk2Ptr16 [F29_16Add (A29_ManufIDAdd)] ||
                (Data16_2 = SubBnk1Ptr16 [F29_16Add (A29_DevIDAdd)]) !=
                SubBnk2Ptr16 [F29_16Add (A29_DevIDAdd)])
            {
               DualSubBnk = TRUE ;
               break ;
            }
            if (Data16_1 != (Word16) ManufID || Data16_2 != (Word16) DevID)
            {
               if (fi->numSubBnk > 1 &&
                   SubBnk1Ptr16 >= (volatile Word16 *) fi->subBnkAdd [1])
                  AutoSelBsAdd16 = (volatile Word16 *) fi->subBnkAdd [1] ;
               else
                  AutoSelBsAdd16 = (volatile Word16 *) FlMemBsAdd ;
               AutoSelBsAdd16 [F29_16Add (A29_Unlck1Add)] =
                  FlMem1x16 (A29_Unlck1Cmd) ;
               AutoSelBsAdd16 [F29_16Add (A29_Unlck2Add)] =
                  FlMem1x16 (A29_Unlck2Cmd) ;
               AutoSelBsAdd16 [F29_16Add (A29_AutoSelAdd)] =
                  FlMem1x16 (F29_IDCmd) ;
               if (SubBnk2Ptr16 [F29_16Add (A29_ManufIDAdd)] !=
                   (Word16) ManufID ||
                   SubBnk2Ptr16 [F29_16Add (A29_DevIDAdd)] != (Word16) DevID)
                  DualSubBnk = TRUE ;
               *AutoSelBsAdd16 = FlMem1x16 (A29_RstCmd) ;
               *AutoSelBsAdd16 = FlMem1x16 (F29_MemRdCmd) ;
               break ;
            }
            SubBnk1Ptr16 = &(SubBnk1Ptr16 [AutoSelSz]) ;
            SubBnk2Ptr16 = &(SubBnk2Ptr16 [AutoSelSz]) ;
         }
   }
   if (DualSubBnk)
   {
      if (SubBnkIdentify (StMemBnkNb, FlMemBsAdd + fi->size, &SubBnkInfo,
                          NULL, NULL) !=
          Success)
         return (ErrStatus);
      if (SubBnkInfo.prgmAlg != fi->prgmAlg ||
          SubBnkInfo.sctErsAlg != fi->sctErsAlg ||
          SubBnkInfo.sctErsStat != fi->sctErsStat ||
          SubBnkInfo.sctErsSusAlg != fi->sctErsSusAlg ||
          SubBnkInfo.sctErsResAlg != fi->sctErsResAlg ||
          SubBnkInfo.devsPerWord != fi->devsPerWord ||
          SubBnkInfo.bitsPerWord != fi->bitsPerWord ||
          SubBnkInfo.dataWdthMsk != fi->dataWdthMsk ||
          SubBnkInfo.nlSctSz != fi->nlSctSz)
         return (ErrStatus = FlIncSubBnkErr) ;
      fi->size += SubBnkInfo.size ;
      for (subBnk = 0 ; subBnk < SubBnkInfo.numSubBnk ; subBnk++)
         fi->subBnkAdd [fi->numSubBnk++] = SubBnkInfo.subBnkAdd [subBnk] ;
      for (spSct = 0 ; spSct < SubBnkInfo.numSpSct ; spSct++)
      {
         fi->spSctAdd [fi->numSpSct] = SubBnkInfo.spSctAdd [spSct] ;
         fi->spSctSz [fi->numSpSct++] = SubBnkInfo.spSctSz [spSct] ;
      }
   }
   return (ErrStatus = Success);
}


#ifndef __linux__  /* No longer used by kernel */
/*
 * PUBLIC PROCEDURE: FlMemPrgm
 *
 * Purpose
 *    The procedure "FlMemPrgm" programs (i.e., erases and re-writes) a given
 *    data buffer into a flash memory. Only the flash-memory sectors storing
 *    the new data are modified.
 *
 * Input
 *    FlMemAdd  	Virtual address, in the flash memory, where to write
 *              	the new data.
 *    DataAdd   	Address of the new data to be written into the flash
 *              	memory.
 *    DataSz    	Size of the new data to be written into the flash
 *              	memory [byte].
 *    Merge     	When set to 0 ("FALSE"), the data previously stored
 *              	in the flash-memory sectors to be modified are
 *              	ignored and un-written locations are left blank
 *              	(i.e., all bits set to 1). Otherwise ("TRUE"), the
 *              	new data is merged with the previously stored data
 *              	before being written into the flash memory.
 *
 * Output
 *    FlMemPrgm 	Set to 0 ("Success") if the programming operation was
 *              	successful, or to the corresponding error number if
 *              	an error occurred.
 *
 * Side-effects
 *    The global variable "ErrStatus" is updated according to the error
 *    status of the procedure call. In case of error, the global variables
 *    "ErrAddress" and "ErrData" are updated as well.
 */

unsigned int FlMemPrgm (
   Address      	FlMemAdd,
   Address      	DataAdd,
   unsigned int 	DataSz,
   int          	Merge)
{
   FlMemInfo            fi;
   unsigned int		FlMemPhAdd,      /* Physical address to write */
                	FlMemBsPhAdd,    /* Physical base address of this bank */
			StMemBnkNb,      /* Number of this bank */
			RemDataSz,
			PrgmSz,
			SctSz,           /* Size of the current sector */
			ErrStat;
   int          	DataCnt,
                	spSct,           /* to iterate over all special sectors */
                	subBnk;          /* to iterate over all sub-banks */
   Address      	FlMemBsAdd,      /* Virtual base address of this bank */
                        FlMemPtr,
			DataPtr,
			SctAdd;
   Word			DataBuf [(4*F29_NlSctSz)/sizeof (Word)];
   /* Temporary variables */
   Word32       	*Word32Ptr1, *Word32Ptr2;
   Word16       	*Word16Ptr1, *Word16Ptr2;
   Byte         	*BytePtr1, *BytePtr2;

#define	FlashPhysAdd(v)	(PhysAdd(v))
#define	FlashVirtAdd(p)	(VirtAdd(p))

   /* Convert write address to physical */
   FlMemPhAdd = (unsigned int) FlashPhysAdd (FlMemAdd) ;

   /* Get physical base address */
   FlMemBsPhAdd = FlMemPhAdd & ~(StMemBnkSp - 1) ;

   /* Range check on physical base address */
   if (!(FlMemBsPhAdd >= _StMemBnk0 && FlMemBsPhAdd <= _StMemBnk3))
   {
      ErrAddress = (Address) FlMemAdd ;
      return (ErrStatus = FlStMemBnkErr) ;
   }

   /* Check that we aren't executing out of the bank to be programmed */
   else if (((unsigned int) FlashPhysAdd (CurrentPC_2 ()) & ~(MemBnkSp - 1)) ==
            FlMemBsPhAdd)
      return (ErrStatus = FlInDevErr) ;

   /* Compute bank number and virtual base address */
   StMemBnkNb = (FlMemBsPhAdd - _StMemBnk0)/StMemBnkSp ;
   FlMemBsAdd = (Address) FlashVirtAdd (FlMemBsPhAdd) ;

   /* Identify kind of chips in use */
   if ((ErrStat = FlMemIdentify(StMemBnkNb, (unsigned int) FlMemBsAdd, &fi)) !=
       Success) {
       return (ErrStat);
   }

   if (FlMemPhAdd + DataSz > FlMemBsPhAdd + fi.size)
      return (ErrStatus = FlSizeErr) ;

   FlMemPtr = FlMemAdd ;
   DataPtr = DataAdd ;

   /* Loop through all sectors being modified */
   for (RemDataSz = DataSz ; RemDataSz > 0 ; RemDataSz -= PrgmSz)
   {
      /* Find size of this sector */
      SctSz = fi.nlSctSz;
      SctAdd = (Address) ((unsigned int) FlMemPtr & ~(SctSz - 1)) ;
      for (spSct = 0 ; spSct < fi.numSpSct ; spSct++)
      {
	 if (FlMemPtr >= (Address) fi.spSctAdd[spSct] &&
             FlMemPtr < (Address) (fi.spSctAdd[spSct] + fi.spSctSz[spSct]))
         {
            SctSz = fi.spSctSz[spSct];
            SctAdd = (Address) fi.spSctAdd[spSct] ;
	    break;
         }
      }
      for (subBnk = 1 ; subBnk < fi.numSubBnk ; subBnk++)
      {
         if (SctAdd < fi.subBnkAdd [subBnk])
            break ;
         FlMemBsAdd = fi.subBnkAdd [subBnk] ;
      }
      if (RemDataSz < SctSz)
         PrgmSz = RemDataSz ;
      else
         PrgmSz = SctSz ;
      if (SctSz - ((unsigned int) FlMemPtr - (unsigned int) SctAdd) < PrgmSz)
         PrgmSz = SctSz - ((unsigned int) FlMemPtr - (unsigned int) SctAdd) ;
      if (Merge &&
          (((unsigned int) DataPtr & fi.dataWdthMsk) != 0 || PrgmSz < SctSz))
      {
	 /* Programming only part of a sector, and Merge is set.  Merge
	    existing flash data and new data into a buffer in RAM */

	 /* Copy entire old sector contents into RAM buffer */
         Word32Ptr1 = (Word32 *) DataBuf ;
         Word32Ptr2 = (Word32 *) SctAdd ;
         for (DataCnt = SctSz ; DataCnt > 0 ; DataCnt -= sizeof (Word32))
            *(Word32Ptr1++) = *(Word32Ptr2++) ;
	 /* Overwrite portion of buffer with new contents */
         BytePtr1 = (Byte *) ((unsigned int) DataBuf +
                              ((unsigned int) FlMemPtr -
                               (unsigned int) SctAdd)) ;
         BytePtr2 = (Byte *) DataPtr ;
         DataCnt = PrgmSz ;
         if ((unsigned int) BytePtr1 % sizeof (Word32) ==
             (unsigned int) BytePtr2 % sizeof (Word32))
         {
	    /* Source and destination have the same alignment (mod 4) */
            if ((unsigned int) BytePtr1 % sizeof (Word16) != 0 &&
                DataCnt >= sizeof (Byte))
            {
               *(BytePtr1++) = *(BytePtr2++) ;
               DataCnt -= sizeof (Byte) ;
            }
            Word16Ptr1 = (Word16 *) BytePtr1 ;
            Word16Ptr2 = (Word16 *) BytePtr2 ;
            if ((unsigned int) Word16Ptr1 % sizeof (Word32) != 0 &&
                DataCnt >= sizeof (Word16))
            {
               *(Word16Ptr1++) = *(Word16Ptr2++) ;
               DataCnt -= sizeof (Word16) ;
            }
            Word32Ptr1 = (Word32 *) Word16Ptr1 ;
            Word32Ptr2 = (Word32 *) Word16Ptr2 ;
            for ( ; DataCnt >= sizeof (Word32) ; DataCnt -= sizeof (Word32))
               *(Word32Ptr1++) = *(Word32Ptr2++) ;
            Word16Ptr1 = (Word16 *) Word32Ptr1 ;
            Word16Ptr2 = (Word16 *) Word32Ptr2 ;
            if (DataCnt >= sizeof (Word16))
            {
               *(Word16Ptr1++) = *(Word16Ptr2++) ;
               DataCnt -= sizeof (Word16) ;
            }
            BytePtr1 = (Byte *) Word16Ptr1 ;
            BytePtr2 = (Byte *) Word16Ptr2 ;
            if (DataCnt > 0)
               *(BytePtr1++) = *(BytePtr2++) ;
         }
         else if ((unsigned int) BytePtr1 % sizeof (Word16) ==
                  (unsigned int) BytePtr2 % sizeof (Word16))
         {
	    /* Source and destination have the same alignment (mod 2) */
            if ((unsigned int) BytePtr1 % sizeof (Word16) != 0 &&
                DataCnt >= sizeof (Byte))
            {
               *(BytePtr1++) = *(BytePtr2++) ;
               DataCnt -= sizeof (Byte) ;
            }
            Word16Ptr1 = (Word16 *) BytePtr1 ;
            Word16Ptr2 = (Word16 *) BytePtr2 ;
            for ( ; DataCnt >= sizeof (Word16) ; DataCnt -= sizeof (Word16))
               *(Word16Ptr1++) = *(Word16Ptr2++) ;
            BytePtr1 = (Byte *) Word16Ptr1 ;
            BytePtr2 = (Byte *) Word16Ptr2 ;
            if (DataCnt > 0)
               *(BytePtr1++) = *(BytePtr2++) ;
         }
         else
	 {
	    /* Source and destination have opposite alignment (mod 2) */
            for ( ; DataCnt > 0 ; DataCnt -= sizeof (Byte))
               *(BytePtr1++) = *(BytePtr2++) ;
	 }
	 /* Now erase and reprogram the sector from the modified buffer */
         ErrStat = (*fi.sctErsAlg) (FlMemBsAdd, SctAdd);
	 if (ErrStat == Success) {
	    while ((ErrStat = (*fi.sctErsStat) (FlMemBsAdd, SctAdd)) == -1) {
	       WAIT_ERASE((void *) SctAdd);
	    }
	 }
         if (ErrStat == Success) {
            ErrStat = (*fi.prgmAlg) (FlMemBsAdd, SctAdd, DataBuf, SctSz);
         }

	if (ErrStat != Success) return (ErrStat);
      }
      else if (((unsigned int) FlMemPtr & fi.dataWdthMsk) != 0 ||
               ((unsigned int) DataPtr & fi.dataWdthMsk) != 0 ||
               (PrgmSz & fi.dataWdthMsk) != 0)
      {
         DataBuf [0] = (Word) -1 ;
         DataBuf [(((unsigned int) FlMemPtr & fi.dataWdthMsk) + PrgmSz - 1) /
                  sizeof (Word)] = (Word) -1 ;
         BytePtr1 = (Byte *) ((unsigned int) DataBuf +
                              ((unsigned int) FlMemPtr & fi.dataWdthMsk)) ;
         BytePtr2 = (Byte *) DataPtr ;
         DataCnt = PrgmSz ;
         if ((unsigned int) BytePtr1 % sizeof (Word32) ==
             (unsigned int) BytePtr2 % sizeof (Word32))
         {
            if ((unsigned int) BytePtr1 % sizeof (Word16) != 0 &&
                DataCnt >= sizeof (Byte))
            {
               *(BytePtr1++) = *(BytePtr2++) ;
               DataCnt -= sizeof (Byte) ;
            }
            Word16Ptr1 = (Word16 *) BytePtr1 ;
            Word16Ptr2 = (Word16 *) BytePtr2 ;
            if ((unsigned int) Word16Ptr1 % sizeof (Word32) != 0 &&
                DataCnt >= sizeof (Word16))
            {
               *(Word16Ptr1++) = *(Word16Ptr2++) ;
               DataCnt -= sizeof (Word16) ;
            }
            Word32Ptr1 = (Word32 *) Word16Ptr1 ;
            Word32Ptr2 = (Word32 *) Word16Ptr2 ;
            for ( ; DataCnt >= sizeof (Word32) ; DataCnt -= sizeof (Word32))
               *(Word32Ptr1++) = *(Word32Ptr2++) ;
            Word16Ptr1 = (Word16 *) Word32Ptr1 ;
            Word16Ptr2 = (Word16 *) Word32Ptr2 ;
            if (DataCnt >= sizeof (Word16))
            {
               *(Word16Ptr1++) = *(Word16Ptr2++) ;
               DataCnt -= sizeof (Word16) ;
            }
            BytePtr1 = (Byte *) Word16Ptr1 ;
            BytePtr2 = (Byte *) Word16Ptr2 ;
            if (DataCnt > 0)
               *(BytePtr1++) = *(BytePtr2++) ;
         }
         else if ((unsigned int) BytePtr1 % sizeof (Word16) ==
                  (unsigned int) BytePtr2 % sizeof (Word16))
         {
            if ((unsigned int) BytePtr1 % sizeof (Word16) != 0 &&
                DataCnt >= sizeof (Byte))
            {
               *(BytePtr1++) = *(BytePtr2++) ;
               DataCnt -= sizeof (Byte) ;
            }
            Word16Ptr1 = (Word16 *) BytePtr1 ;
            Word16Ptr2 = (Word16 *) BytePtr2 ;
            for ( ; DataCnt >= sizeof (Word16) ; DataCnt -= sizeof (Word16))
               *(Word16Ptr1++) = *(Word16Ptr2++) ;
            BytePtr1 = (Byte *) Word16Ptr1 ;
            BytePtr2 = (Byte *) Word16Ptr2 ;
            if (DataCnt > 0)
               *(BytePtr1++) = *(BytePtr2++) ;
         }
         else
            for ( ; DataCnt > 0 ; DataCnt -= sizeof (Byte))
               *(BytePtr1++) = *(BytePtr2++) ;

	 /* Erase and reprogram the sector */
         ErrStat = (*fi.sctErsAlg) (FlMemBsAdd, SctAdd);
	 if (ErrStat == Success) {
	    while ((ErrStat = (*fi.sctErsStat) (FlMemBsAdd, SctAdd)) == -1) {
	       WAIT_ERASE((void *) SctAdd);
	    }
	 }
	 if (ErrStat == Success) {
	    ErrStat = (*fi.prgmAlg) (FlMemBsAdd,
                                    (Address) ((unsigned int) FlMemPtr &
                                               ~fi.dataWdthMsk),
                                    DataBuf,
                                    (((unsigned int) FlMemPtr & fi.dataWdthMsk) +
                                     PrgmSz + fi.dataWdthMsk) &
                                    ~fi.dataWdthMsk);
         }

         if (ErrStat != Success) return (ErrStat) ;
      }
      else
      {
	 /* Erase and reprogram the sector */
         ErrStat = (*fi.sctErsAlg) (FlMemBsAdd, SctAdd);
	 if (ErrStat == Success) {
	    while ((ErrStat = (*fi.sctErsStat) (FlMemBsAdd, SctAdd)) == -1) {
	       WAIT_ERASE((void *) SctAdd);
	    }
	 }
         if (ErrStat == Success) {
            ErrStat = (*fi.prgmAlg) (FlMemBsAdd, FlMemPtr, DataPtr, PrgmSz);
         }
	 if (ErrStat != Success) return (ErrStat) ;
      }
      FlMemPtr = (Address) ((unsigned int) FlMemPtr + PrgmSz) ;
      DataPtr = (Address) ((unsigned int) DataPtr + PrgmSz) ;
   }
   BytePtr1 = (Byte *) FlMemAdd ;
   BytePtr2 = (Byte *) DataAdd ;
   DataCnt = DataSz ;
   if ((unsigned int) BytePtr1 % sizeof (Word32) ==
       (unsigned int) BytePtr2 % sizeof (Word32))
   {
      if ((unsigned int) BytePtr1 % sizeof (Word16) != 0 &&
          DataCnt >= sizeof (Byte))
      {
         if (*(BytePtr1++) != *(BytePtr2++))
         {
            ErrAddress = (Address) &(BytePtr1 [-1]) ;
            ErrData = BytePtr2 [-1] ;
            return (ErrStatus = FlChk8Err) ;
         }
         DataCnt -= sizeof (Byte) ;
      }
      Word16Ptr1 = (Word16 *) BytePtr1 ;
      Word16Ptr2 = (Word16 *) BytePtr2 ;
      if ((unsigned int) Word16Ptr1 % sizeof (Word32) != 0 &&
          DataCnt >= sizeof (Word16))
      {
         if (*(Word16Ptr1++) != *(Word16Ptr2++))
         {
            ErrAddress = (Address) &(Word16Ptr1 [-1]) ;
            ErrData = Word16Ptr2 [-1] ;
            return (ErrStatus = FlChk16Err) ;
         }
         DataCnt -= sizeof (Word16) ;
      }
      Word32Ptr1 = (Word32 *) Word16Ptr1 ;
      Word32Ptr2 = (Word32 *) Word16Ptr2 ;
      for ( ; DataCnt >= sizeof (Word32) ; DataCnt -= sizeof (Word32))
         if (*(Word32Ptr1++) != *(Word32Ptr2++))
         {
            ErrAddress = (Address) &(Word32Ptr1 [-1]) ;
            ErrData = Word32Ptr2 [-1] ;
            return (ErrStatus = FlChk32Err) ;
         }
      Word16Ptr1 = (Word16 *) Word32Ptr1 ;
      Word16Ptr2 = (Word16 *) Word32Ptr2 ;
      if (DataCnt >= sizeof (Word16))
      {
         if (*(Word16Ptr1++) != *(Word16Ptr2++))
         {
            ErrAddress = (Address) &(Word16Ptr1 [-1]) ;
            ErrData = Word16Ptr2 [-1] ;
            return (ErrStatus = FlChk16Err) ;
         }
         DataCnt -= sizeof (Word16) ;
      }
      BytePtr1 = (Byte *) Word16Ptr1 ;
      BytePtr2 = (Byte *) Word16Ptr2 ;
      if (DataCnt > 0 && *(BytePtr1++) != *(BytePtr2++))
      {
         ErrAddress = (Address) &(BytePtr1 [-1]) ;
         ErrData = BytePtr2 [-1] ;
         return (ErrStatus = FlChk8Err) ;
      }
   }
   else if ((unsigned int) BytePtr1 % sizeof (Word16) ==
            (unsigned int) BytePtr2 % sizeof (Word16))
   {
      if ((unsigned int) BytePtr1 % sizeof (Word16) != 0 &&
          DataCnt >= sizeof (Byte))
      {
         if (*(BytePtr1++) != *(BytePtr2++))
         {
            ErrAddress = (Address) &(BytePtr1 [-1]) ;
            ErrData = BytePtr2 [-1] ;
            return (ErrStatus = FlChk8Err) ;
         }
         DataCnt -= sizeof (Byte) ;
      }
      Word16Ptr1 = (Word16 *) BytePtr1 ;
      Word16Ptr2 = (Word16 *) BytePtr2 ;
      for ( ; DataCnt >= sizeof (Word16) ; DataCnt -= sizeof (Word16))
         if (*(Word16Ptr1++) != *(Word16Ptr2++))
         {
            ErrAddress = (Address) &(Word16Ptr1 [-1]) ;
            ErrData = Word16Ptr2 [-1] ;
            return (ErrStatus = FlChk16Err) ;
         }
      BytePtr1 = (Byte *) Word16Ptr1 ;
      BytePtr2 = (Byte *) Word16Ptr2 ;
      if (DataCnt > 0 && *(BytePtr1++) != *(BytePtr2++))
      {
         ErrAddress = (Address) &(BytePtr1 [-1]) ;
         ErrData = BytePtr2 [-1] ;
         return (ErrStatus = FlChk8Err) ;
      }
   }
   else
      for ( ; DataCnt > 0 ; DataCnt -= sizeof (Byte))
         if (*(BytePtr1++) != *(BytePtr2++))
         {
            ErrAddress = (Address) &(BytePtr1 [-1]) ;
            ErrData = BytePtr2 [-1] ;
            return (ErrStatus = FlChk8Err) ;
         }
   return (ErrStatus = Success) ;
}
#endif /* __linux__ */
